/***************************************************************************//**
 * @file Customer.hpp
 * @author Dorian Weber
 * @brief Contains the declaration of the Customer process.
 ******************************************************************************/

#ifndef CUSTOMER_HPP_INCLUDED
#define CUSTOMER_HPP_INCLUDED

#include "Simulation.hpp"

namespace Checkout {
	/**@brief Active class that models customers.
	 */
	class Customer: public odemx::base::Process {
		/**@brief The cart filled with merchandise.
		 */
		std::vector< std::unique_ptr<Merchandise> > cart;

		/**@brief The time needed for this customer to pay.
		 */
		odemx::base::SimTime payTime;

	public:

		/**@brief Constructor for customers.
		 * @param[in] sim      the simulation context
		 * @param[in] payTime  the time needed to pay
		 */
		Customer(Simulation& sim, odemx::base::SimTime payTime)
		:	Process(sim, "Customer"),
			payTime(payTime) {}

		/**@brief Adds another product to the customers cart.
		 * @note This method is used by the CustomerGenerator to
		 * progressively fill the customers cart with merchandise.
		 * @param[in] product  the product to insert
		 */
		void productAdd(Merchandise* product) {
			cart.push_back(std::unique_ptr<Merchandise>(product));
		}

		/**@brief Returns the product count in the customers cart.
		 * @return the product count in the customers cart
		 */
		unsigned int productCount() const {
			return cart.size();
		}

	protected:

		/**@brief Virtual selector function for picking the best lane
		 * if multiple are possible.
		 * @note This can be overwritten in class specializations to
		 * test the effect of different strategies.
		 *
		 * The strategy currently implemented first looks at the number
		 * of items located on each belt and picks the one with the
		 * lowest amount. If multiple belts tie, the number of customers
		 * in considered next. If this is equal as well, the earlier
		 * lane is picked.
		 *
		 * @return register that the customer wants to queue up
		 */
		virtual Register& selectLane();

		/**@brief Transfers the products from the cart onto the belt,
		 * from the tray into the cart and pays the bill.
		 */
		virtual int main();
	};
}

#endif /* CUSTOMER_HPP_INCLUDED */
