/***************************************************************************//**
 * @file CustomerGenerator.hpp
 * @author Dorian Weber
 * @brief Contains the declaration of the CustomerGenerator process.
 ******************************************************************************/

#ifndef CUSTOMERGENERATOR_HPP_INCLUDED
#define CUSTOMERGENERATOR_HPP_INCLUDED

#include <odemx/odemx.h>

namespace Checkout {
	// forward declaration
	struct Simulation;

	/**@brief Active class that generates and configures customer processes.
	 */
	class CustomerGenerator: public odemx::base::Process {
		/**@brief Distribution of arriving customers.
		 *
		 * This is modeled as a negative exponential distribution
		 * because the customers are assumed to arrive independently
		 */
		odemx::random::NegativeExponential arrivalDist;

		/**@brief Distribution of the number of products per customer.
		 *
		 * This is modeled as a poisson distribution because the need
		 * to buy a product is assumed to be motivated by that product
		 * running out at the customers home. This is assumed to be
		 * independent between two products due to different package
		 * sizes and receipes requiring different quantities of each
		 * component.
		 */
		odemx::random::Poisson productDist;

		/**@brief Cost distribution of the merchandise.
		 *
		 * I don't really understand why we model that as negative
		 * exponentially distributed but it is specified in the
		 * assignment.
		 */
		odemx::random::NegativeExponential costDist;

		/**@brief Weight distribution of merchandise.
		 *
		 * The weight is used as an abstraction on how much time the
		 * transfer of a product between two places needs. It is normal
		 * distributed because the transfer time is assumed to be
		 * directly correlated to the handling complexity of the
		 * product. I'm aware that this explanation doesn't really cut
		 * it here but I'm going to leave it at that anyways :).
		 */
		odemx::random::Normal weightDist;

		/**@brief Distribution for the time needed by the customer to
		 * pay.
		 * @note This should probably be correlated to the price.
		 *
		 * This is modeled as a normal distributed quantity because it
		 * directly correlates with the customers ability to find the
		 * correct coins/enter the correct PIN/etc.
		 */
		odemx::random::Normal payTimeDist;

		/**@brief The CustomerGenerator will continue to produce customers
		 * until the simulation duration is exceeded.
		 */
		unsigned int simDuration;

	public:
		/**@brief Statistical quantity for the count of generated
		 * customers.
		 */
		odemx::statistics::Count customers;

		/**@brief Constructor for a single cash register.
		 * @param[in] sim               simulation
		 * @param[in] arrivalRate       expected rate of arrivals
		 * @param[in] productMean       expected number products
		 * @param[in] costMean          expected price per product
		 * @param[in] weightMean        expected weight of a product
		 * @param[in] weightDeviation   deviation of product weights
		 * @param[in] payTimeMean       expected time to finish payment
		 * @param[in] payTimeDeviation  deviation of paytime
		 */
		CustomerGenerator(
			Simulation& sim,
			unsigned int productMean,
			double arrivalRate,
			double costMean,
			double weightMean,
			double weightDeviation,
			double payTimeMean,
			double payTimeDeviation,
			unsigned int simDuration
		);

	protected:
		/**@brief Generates and configures customers and their
		 * merchandise.
		 */
		virtual int main();
	};
}

#endif /* CUSTOMERGENERATOR_HPP_INCLUDED */
