/***************************************************************************//**
 * @file Register.hpp
 * @author Dorian Weber
 * @brief Passive structure for the cash register.
 ******************************************************************************/

#ifndef REGISTER_HPP_INCLUDED
#define REGISTER_HPP_INCLUDED

#include <memory>
#include "Simulation.hpp"
#include "Cashier.hpp"

namespace Checkout {
	/**@brief Passive class modeling the cash register.
	 */
	class Register {
		/**@brief Pointer to the head of the belt for read access.
		 */
		odemx::synchronization::PortHead::Ptr beltHead;

		/**@brief Pointer to the tail of the belt for write access.
		 */
		odemx::synchronization::PortTail::Ptr beltTail;

		/**@brief Mutex controlling write access to the belt.
		 */
		odemx::synchronization::Res beltAccess;

		/**@brief Pointer to the head of the tray for read access.
		 */
		odemx::synchronization::PortHead::Ptr trayHead;

		/**@brief Pointer to the tail of the tray for write access.
		 */
		odemx::synchronization::PortTail::Ptr trayTail;

		/**@brief Mutex controlling read access from the tray.
		 */
		odemx::synchronization::Res trayAccess;

	public:
		/**@brief The cashier process responsible for this register.
		 */
		Cashier cashier;

		/**@brief Constructor for a single cash register.
		 * @param[in] sim      the simulation context
		 * @param[in] beltCap  the maximum number of items on the conveyor belt
		 * @param[in] trayCap  the maximum number of items in the tray
		 */
		Register(Simulation& sim, unsigned int beltCap = -1,
		         unsigned int trayCap = -1)
		:	beltAccess(sim, "BeltAccess", 1, 1),
			trayAccess(sim, "TrayAccess", 1, 1),
			cashier(sim, *this) {
			using namespace odemx::synchronization;

			// create conveyor belt
			beltHead = PortHead::create(
				sim, "Belt", WAITING_MODE, beltCap
			);
			beltTail = beltHead->getTail();

			// create tray deposit
			trayHead = PortHead::create(
				sim, "Tray", WAITING_MODE, trayCap
			);
			trayTail = trayHead->getTail();
		}

		/**@brief Returns the number of customers currently in line for
		 * this register.
		 * @return the number of customers currently in line for this
		 * register
		 */
		unsigned int customerCount() const {
			return beltAccess.getWaitingProcesses().size();
		}

		/**@brief Returns the number of items currently on the belt.
		 * @return the number of items currently on the belt
		 */
		unsigned int itemCount() const {
			return beltHead->count();
		}

		/**@brief Aquire write access to the conveyor belt.
		 */
		void aquireBelt() {
			beltAccess.acquire(1);
		}

		/**@brief Release the conveyor belt for the next customer.
		 */
		void releaseBelt() {
			beltPut(nullptr);
			beltAccess.release(1);
		}

		/**@brief Aquire read access to the tray.
		 */
		void aquireTray() {
			trayAccess.acquire(1);
		}

		/**@brief Release the tray for the next customer.
		 */
		void releaseTray() {
			trayAccess.release(1);
		}

		/**@brief Add another product to the belt.
		 * @attention The customer needs to aquire write access first.
		 * @param[in] product  the product
		 */
		void beltPut(Merchandise* product) {
			beltTail->put(product);
		}

		/**@brief Take a product from the belt.
		 * @return unique_ptr to the product
		 */
		std::unique_ptr<Merchandise> beltGet() {
			return std::unique_ptr<Merchandise>(
				dynamic_cast<Merchandise*>(*beltHead->get())
			);
		}

		/**@brief Put another product into the tray area.
		 * @param[in] product  pointer to the product to put
		 * @note This method takes ownership of the product pointer.
		 */
		void trayPut(Merchandise* product) {
			trayTail->put(product);
		}

		/**@brief Take a product out of the tray.
		 * @attention The customer needs to aquire read access first.
		 * @return unique_ptr to the product
		 */
		std::unique_ptr<Merchandise> trayGet() {
			return std::unique_ptr<Merchandise>(
				dynamic_cast<Merchandise*>(*trayHead->get())
			);
		}
	};
}

#endif /* REGISTER_HPP_INCLUDED */
