/***************************************************************************//**
 * @file Simulation.hpp
 * @author Dorian Weber
 * @brief Contains the custom simulation for the Checkout scenario.
 ******************************************************************************/

#ifndef SIMULATION_HPP_INCLUDED
#define SIMULATION_HPP_INCLUDED

#include <odemx/odemx.h>
#include <vector>
#include <iostream>

#include "CustomerGenerator.hpp"

namespace Checkout {
	// forward declaration
	class Register;

	/**@brief Simulation context for the Checkout scenario.
	 */
	struct Simulation: public odemx::base::Simulation {
		/**@brief Pointer to the generated statistic report.
		 */
		odemx::data::output::OStreamReport report;

		/**@brief Customer generator process responsible for creating
		 * and configuring arriving customers.
		 */
		CustomerGenerator arrivals;

		/**@brief Vector of cash registers.
		 */
		std::vector<Register*> lanes;

		/**@brief Statistics object for recording the time needed per
		 * customer.
		 */
		odemx::statistics::Histogram handlingTime;

		/**@brief Total time spent waiting for customers to arrive on
		 * all lanes.
		 */
		odemx::statistics::Sum idleTime;

		/**@brief Total amount of money exchanged on all lanes.
		 */
		odemx::statistics::Sum salesVol;

		/**@brief Constructor for the simulation.
		 *
		 * @param[in] registerCount     number of cash registers
		 * @param[in] beltCap           belt capacity per register
		 * @param[in] trayCap           tray capacity per register
		 * @param[in] productMean       expected number products
		 * @param[in] arrivalRate       expected rate of arrivals
		 * @param[in] costMean          expected price per product
		 * @param[in] weightMean        expected weight of a product
		 * @param[in] weightDeviation   deviation of product weights
		 * @param[in] paytimeMean       expected time to finish payment
		 * @param[in] paytimeDeviation  deviation of paytime
		 * @param[in] defaultReport     whether to report simulation statistics
		 */
		Simulation(
			unsigned int duration,
			unsigned int registerCount,
			unsigned int beltCap,
			unsigned int trayCap,
			unsigned int productMean,
			double arrivalRate,
			double costMean,
			double weightMean,
			double weightDeviation,
			double paytimeMean,
			double paytimeDeviation,
			bool defaultReport = false
		);

		/**@brief Activates the processes initially.
		 */
		void initSimulation() override;
	};
}

// courtesy includes
#include "Merchandise.hpp"
#include "Cashier.hpp"
#include "Register.hpp"
#include "Customer.hpp"

#endif /* SIMULATION_HPP_INCLUDED */
