#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <cmath>

#include <string>
#include <vector>
#include <utility>
#include <sstream>

#include "main.h"

/* ******************************************************* internal variables */

struct SLXTransferVector* SLXTV;
static std::string Gnuplot;

static const char* const GnuplotRegPath = 
	"SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\App Paths\\gnuplot.exe";

static const char* const Plot2Default =
	"#!%s\n"
	"set terminal Term linewidth 1.5\n"
	"set output File\n"
	"\n"
	"# Line style for axes\n"
	"set style line 80 lt rgb \"#808080\"\n"
	"\n"
	"# Line style for grid, border and tics\n"
	"set style line 81 lt 0\n"
	"set style line 81 lt rgb \"#808080\"\n"
	"set grid back linestyle 81\n"
	"set border 3 back linestyle 80\n"
	"set xtics nomirror\n"
	"set ytics nomirror\n"
	"\n"
	"set key autotitle columnheader\n"
	"set key bottom right box opaque\n"
	"\n"
	"# Line styles: try to pick pleasing colors, rather\n"
	"# than strictly primary colors or hard-to-see colors\n"
	"# like gnuplot's default yellow.\n"
	"set style line 2 lt rgb \"#A00000\" pt 1\n"
	"set style line 3 lt rgb \"#00A000\" pt 6\n"
	"set style line 4 lt rgb \"#5060D0\" pt 2\n"
	"set style line 5 lt rgb \"#F25900\" pt 9\n"
	"set style line 6 lt rgb \"#990070\" pt 3\n"
	"\n"
	"set xlabel 'Modellzeit' offset 0,0.5\n"
	"plot for [i=2:Columns] Trace using 1:(column(i)) with lines ls i\n"
;

/* ******************************************************* internal functions */

static std::string
getStringRegKey(const std::string& path, const std::string& key,
                const std::string& defVal)
{
    HKEY handle;
	CHAR buffer[512];
	DWORD bufferSize = sizeof(buffer);
	LONG rc;
	
	rc = RegOpenKeyEx(HKEY_LOCAL_MACHINE, path.c_str(), 0, KEY_READ, &handle);
	if (rc != ERROR_SUCCESS)
		goto e1;
	
	rc = RegQueryValueEx(handle, key.c_str(), 0, NULL, (LPBYTE) buffer, &bufferSize);
	if (rc != ERROR_SUCCESS)
		goto e2;
	
	RegCloseKey(handle);
	return std::string(buffer);
	
e2:	RegCloseKey(handle);
e1:	return defVal;
}

static std::string
getLastErrorString(DWORD err)
{
	size_t size;
	LPSTR buf = nullptr;
	
	if (err == ERROR_SUCCESS)
		return std::string();

	size = FormatMessage(
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
		NULL, err, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPSTR) &buf, 0, NULL
	);

	std::string message(buf, size);
	LocalFree(buf);
	
	return message;
}

static bool
fileExists(const std::string& path)
{
	return GetFileAttributes(path.c_str()) != INVALID_FILE_ATTRIBUTES;
}

static int
runProgram(const std::string& path, std::string args, std::string& errMsg)
{
	std::string msg;
	STARTUPINFO info = {sizeof(info)};
	PROCESS_INFORMATION process;
	HANDLE readErr = NULL, writeErr = NULL;
	DWORD rc;
	
	/* create a new pipe for stderr to read from */
	{
		SECURITY_ATTRIBUTES sa = {sizeof(sa)};
		sa.bInheritHandle = TRUE;
		CreatePipe(&readErr, &writeErr, &sa, 0);
	}
	
	/* create the child process with redirected stderr */
	info.hStdError = writeErr;
	info.dwFlags |= STARTF_USESTDHANDLES;
	if (!CreateProcess(path.c_str(), &args[0], NULL, NULL, TRUE,
	                   CREATE_NO_WINDOW, NULL, NULL, &info, &process))
	{
		rc = GetLastError();
		errMsg = getLastErrorString(rc);
		CloseHandle(writeErr);
		CloseHandle(readErr);
		return rc;
	}
	
	/* wait until done, get exit code and close write handle */
	WaitForSingleObject(process.hProcess, INFINITE);
	GetExitCodeProcess(process.hProcess, &rc);
	CloseHandle(process.hProcess);
	CloseHandle(writeErr);
	
	/* read error message from redirected stderr */
	if (rc)
	{
		CHAR buffer[1024];
		DWORD count;
		
		while (ReadFile(readErr, buffer, sizeof(buffer), &count, NULL) && count > 0)
			errMsg.append(buffer, count);;
	}
	
	/* close read handle */
	CloseHandle(readErr);
	return rc;
}

/* ******************************************************* exported functions */

EXPORT void
Connect(SLXTransferVector* slx)
{
	SLXTV = slx;
	
	/* check if the gnuplot-program exists */
	if (!fileExists(Gnuplot))
		SLX_Print((char*) "slxplot.dll: Gnuplot could not be located.\n");
}

EXPORT void
PlotToFile(const string_header* fin, int columns, int blocks,
           const string_header* fout, const string_header* fplot, int show)
{
	std::ostringstream cmdLn;
	std::string err;
	const char* term;
	
	if (!fileExists(fplot->string_address))
	{
		FILE* plt = std::fopen(fplot->string_address, "w");
		
		if (plt == NULL)
		{
			SLX_Print(const_cast<char*>("slxplot.dll: "));
			SLX_Print(std::strerror(errno));
			return;
		}
		
		std::fprintf(plt, Plot2Default, Gnuplot.c_str());
		std::fclose(plt);
	}
	
	term = std::strrchr(fout->string_address, '.');
	if (term == NULL)
	{
		SLX_Print(const_cast<char*>("slxplot.dll: '"));
		SLX_Print(fout->string_address);
		SLX_Print(const_cast<char*>("' lacks a file extension determining the output format."));
		return;
	}
	
	cmdLn << " -e \""
	         "Term = '" << (term + 1) << "';"
	         "File = '" << fout->string_address << "';"
	         "Trace = '" << fin->string_address << "';"
	         "Columns = " << columns << ";"
	         "Blocks = " << blocks << ";\" " <<
	         fplot->string_address;
	
	if (runProgram(Gnuplot, cmdLn.str(), err))
	{
		SLX_Print(const_cast<char*>("slxplot.dll: "));
		SLX_Print(const_cast<char*>(err.c_str()));
		return;
	}
	
	if (show)
		ShellExecute(HWND_DESKTOP, "open", fout->string_address, NULL, NULL, SW_SHOW);
}

extern "C" EXPORT
BOOL APIENTRY DllMain(HINSTANCE hinstDLL, DWORD fdwReason, LPVOID lpvReserved)
{
	switch (fdwReason)
	{
	case DLL_PROCESS_ATTACH:
	case DLL_THREAD_ATTACH:
		Gnuplot = getStringRegKey(GnuplotRegPath, "", "gnuplot.exe");
		break;
	}
	
	return TRUE;
}
