//! This example simulates a simple barbershop scenario, demonstrating the use
//! of facilities, random number generation, and statistics collection in the
//! ODEM-rs library.
//!
//! # Overview
//!
//! The simulation models a barbershop with a single barber (`joe`). Customers
//! arrive at random intervals, and each goes through a sequence of actions:
//! arrival, seizing the barber for a haircut, and leaving after the haircut.
//! The simulation also collects statistics on customer waiting times.

use crate::DefaultRng;
use core::pin::pin;
use odem_rs::{prelude::*, sync::Facility, util::random::RngStream};

/// Configuration type.
#[derive(Config, Default)]
struct Barbershop {
	/// The simulation time type.
	#[time]
	time: Time<f64>,
	/// Total duration of the simulation run.
	duration: Time<f64>,
	/// Stream for generating independent random number generators.
	rng_stream: RngStream<DefaultRng>,
	/// The barber facility for haircuts.
	joe_barber: Facility,
	/// Random variable for recording customer wait times.
	wait_time: RandomVariable<Time<f64>>,
}

/// Customer process with access to the barber and a random processing delay.
struct Customer(Time<f64>);

impl Behavior<Barbershop> for Customer {
	type Output = ();

	async fn actions(&self, sim: &Sim<Barbershop>) -> Self::Output {
		// access the barber and record the time for the report
		let arrival_time = sim.now();
		let barber = sim.global().joe_barber.seize().await;
		sim.global().wait_time.tabulate(sim.now() - arrival_time);

		// spend time
		sim.advance(self.0).await;

		// release the barber
		barber.release();
	}
}

/// Orchestrates the barbershop simulation for a given duration.
async fn sim_main(sim: &Sim<Barbershop>) {
	let pool = pool::Pool::dynamic();

	// specify a customer generating fiber
	let customers = pin!(Fiber::new(async {
		let mut rng_a = sim.global().rng_stream.rng();
		let mut rng_s = sim.global().rng_stream.rng();

		loop {
			sim.advance(minute::<f64>(rng_a.random_range(12.0..24.0)))
				.await;
			sim.activate(pool.alloc(Process::new(Customer(minute::<f64>(
				rng_s.random_range(12.0..18.0),
			)))));
		}
	}));

	// activate the fiber and wait for the duration to pass
	let puck = sim.activate(customers);
	sim.advance(sim.global().duration).await;
	puck.abort();

	// finish processing the queue (no more customers arrive)
	let _ = sim.global().joe_barber.seize().await;
}

pub fn barbershop(duration: Time<f64>) -> impl std::fmt::Debug {
	let sim = Simulator::new(Barbershop {
		duration,
		rng_stream: RngStream::from_seed(0),
		..Default::default()
	});

	sim.run(sim_main).unwrap();
	sim.into_inner().wait_time.displayed(uom::si::time::minute)
}
