use crate::DefaultRng;
use rand::Rng;
use simcore_rs::{
	util::{Facility, RandomVariable},
	Process, Sim, Time,
};
use std::{cell::RefCell, fmt::Debug};

/// Globally shared data for the barbershop simulation.
struct Barbershop {
	/// Random number generator for customer arrival times.
	rng_a: RefCell<DefaultRng>,
	/// Random number generator for service times.
	rng_s: RefCell<DefaultRng>,
	/// Facility representing the barber (Joe).
	joe: Facility,
	/// Statistical accumulator for customer wait times.
	wait_time: RandomVariable,
}

/// Represents a customer in the barbershop simulation.
struct Customer;

impl Customer {
	/// Defines the actions performed by a customer in the simulation.
	///
	/// The customer arrives at the barbershop, waits for the barber to be
	/// available, gets a haircut (spends time being serviced), and then leaves.
	pub async fn actions(self, sim: Sim<'_, Barbershop>) {
		// Record the arrival time for wait time calculation.
		let arrival_time = sim.now();

		// Seize the barber (wait if not available).
		sim.global().joe.seize().await;

		// Calculate and record the customer's wait time.
		sim.global().wait_time.tabulate(sim.now() - arrival_time);

		// Simulate the time taken for the haircut.
		let cut = sim.global().rng_s.borrow_mut().random_range(12.0..18.0);
		sim.advance(cut).await;

		// Release the barber for the next customer.
		sim.global().joe.release();
	}
}

/// The main simulation function.
///
/// This function initializes the customer arrival process, runs the simulation
/// for the specified duration, and ensures that all customers are serviced
/// before the simulation ends.
async fn sim_main(sim: Sim<'_, Barbershop>, duration: Time) {
	// Activate a process to generate customers at random intervals.
	sim.activate(async move {
		loop {
			// Wait for a random time before the next customer arrives.
			let wait_time = sim.global().rng_a.borrow_mut().random_range(12.0..24.0);
			sim.advance(wait_time).await;

			// If the simulation time exceeds the duration, stop generating customers.
			if sim.now() >= duration {
				return;
			}

			// Activate a new customer process.
			sim.activate(Customer.actions(sim));
		}
	});

	// Run the simulation until the store closes.
	sim.advance(duration).await;

	// Ensure the barber finishes servicing any remaining customers.
	sim.global().joe.seize().await;
}

/// Entry point for the barbershop simulation.
///
/// Initializes the simulation environment and runs the simulation for a
/// specified duration.
pub fn barbershop(duration: uom::si::f64::Time) -> impl Debug {
	use rand::SeedableRng;

	// Run the simulation and collect the result.
	let mut rng = DefaultRng::seed_from_u64(0);
	let result = simcore_rs::simulation(
		// Initialize the global data for the simulation.
		Barbershop {
			rng_a: RefCell::new(DefaultRng::from_rng(&mut rng)),
			rng_s: RefCell::new(rng),
			joe: Facility::new(),
			wait_time: RandomVariable::new(),
		},
		// Simulation entry point.
		|sim| Process::new(sim, sim_main(sim, duration.get::<uom::si::time::minute>())),
	);

	result.wait_time
}
