use rand::{Rng, SeedableRng};
use rand_distr::{Distribution, Exp, Normal};

use crate::DefaultRng;
use simcore_rs::{
	util::{channel, select, RandomVariable, Receiver, Sender},
	Process, Sim, Time,
};
use std::{cell::RefCell, fmt};

// Constants for simulation parameters.
const HARBOR_DISTANCE: Time = 10.0;
const FERRY_TIMEOUT: Time = 5.0;
const FERRY_CAPACITY: usize = 5;

/// Shared global data for the ferry simulation.
struct CarFerry {
	/// Master random number generator for seeding.
	master_rng: RefCell<DefaultRng>,
	/// Statistical accumulator for ferry cargo lengths.
	ferry_cargo_len: RandomVariable,
	/// Statistical accumulator for ferry load times.
	ferry_load_time: RandomVariable,
	/// Statistical accumulator for car wait times.
	car_wait_time: RandomVariable,
}

impl fmt::Debug for CarFerry {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		f.debug_struct("Ferries")
			.field("ferry_cargo_len", &self.ferry_cargo_len)
			.field("ferry_load_time", &self.ferry_load_time)
			.field("car_wait_time", &self.car_wait_time)
			.finish()
	}
}

/// Represents a car in the simulation.
#[derive(Debug)]
struct Car {
	/// Time when the car arrives at the pier.
	arrival_time: Time,
	/// Duration required to load the car onto the ferry.
	load_duration: Time,
}

/// Represents a pier (harbor) where cars arrive and wait for ferries.
struct Pier {
	/// Channel to send cars to the ferry.
	landing_site: Sender<Car>,
}
impl Pier {
	/// Simulates the actions of a pier in the simulation.
	///
	/// Cars arrive at the pier following an exponential distribution
	/// and are sent to the ferry when it arrives.
	async fn actions(self, sim: Sim<'_, CarFerry>) {
		let mut rng = DefaultRng::from_seed(sim.global().master_rng.borrow_mut().random());

		// Arrival and loading time distributions.
		let arrival_delay = Exp::new(0.1).unwrap();
		let loading_delay = Normal::new(0.5, 0.2).unwrap();

		loop {
			// Wait for the next car to arrive.
			sim.advance(arrival_delay.sample(&mut rng)).await;

			// Create a new car with arrival and load times.
			let car = Car {
				arrival_time: sim.now(),
				load_duration: loading_delay
					.sample_iter(&mut rng)
					.find(|&val| val >= 0.0)
					.unwrap(),
			};
			self.landing_site
				.send(car)
				.await
				.expect("No ferries in the simulation");
		}
	}
}

/// Represents a ferry that transports cars between harbors.
struct Ferry {
	/// Receivers from piers to accept arriving cars.
	piers: Vec<Receiver<Car>>,
}

impl Ferry {
	/// Simulates the actions of a ferry in the simulation.
	///
	/// The ferry travels between harbors, loads cars, waits for a timeout or
	/// until it reaches capacity, and then moves to the next harbor.
	async fn actions(self, sim: Sim<'_, CarFerry>) {
		let mut cargo = Vec::<Car>::with_capacity(FERRY_CAPACITY);

		loop {
			for pier in self.piers.iter() {
				// Unload the cars at the current pier.
				for car in cargo.drain(..) {
					sim.advance(car.load_duration).await;
				}

				let begin_loading = sim.now();

				// Load cars until capacity is reached or timeout occurs.
				for _ in 0..cargo.capacity() {
					match select(sim, pier.recv(), async {
						sim.advance(FERRY_TIMEOUT).await;
						None
					})
					.await
					{
						// A car arrived before timeout.
						Some(car) => {
							sim.global()
								.car_wait_time
								.tabulate(sim.now() - car.arrival_time);
							sim.advance(car.load_duration).await;
							cargo.push(car);
						}
						// Timeout occurred; depart to next harbor.
						None => break,
					}
				}

				// Record ferry loading statistics.
				sim.global()
					.ferry_load_time
					.tabulate(sim.now() - begin_loading);
				sim.global().ferry_cargo_len.tabulate(cargo.len() as f64);

				// Travel to the next harbor.
				sim.advance(HARBOR_DISTANCE).await;
			}
		}
	}
}

/// The main simulation function.
///
/// Sets up the piers and ferries, activates their processes, and runs the simulation
/// for the specified duration.
async fn sim_main(sim: Sim<'_, CarFerry>, duration: Time, ferries: usize, harbors: usize) {
	let mut ports = Vec::with_capacity(harbors);

	// Create all the harbors (piers).
	for _ in 0..harbors {
		let (sx, rx) = channel();
		let harbor = Pier { landing_site: sx };
		sim.activate(harbor.actions(sim));
		ports.push(rx);
	}

	// Create all the ferries.
	for i in 0..ferries {
		let ferry = Ferry {
			piers: ports
				.iter()
				.skip(i)
				.chain(ports.iter().take(i))
				.cloned()
				.collect(),
		};
		sim.activate(ferry.actions(sim));
	}

	// Run the simulation for the specified duration.
	sim.advance(duration).await;

	// Handle any cars that weren't picked up by a ferry.
	for port in ports {
		for _ in 0..port.len() {
			let car = port.recv().await.unwrap();
			sim.global()
				.car_wait_time
				.tabulate(sim.now() - car.arrival_time);
		}
	}
}

/// Entry point for the ferry simulation.
///
/// Initializes the simulation environment and runs the simulation for one week.
pub fn ferry(duration: uom::si::f64::Time, ferries: usize, harbors: usize) -> impl fmt::Debug {
	simcore_rs::simulation(
		// Global shared data.
		CarFerry {
			master_rng: RefCell::new(DefaultRng::seed_from_u64(0)),
			ferry_cargo_len: RandomVariable::new(),
			ferry_load_time: RandomVariable::new(),
			car_wait_time: RandomVariable::new(),
		},
		// Simulation entry point.
		|sim| {
			Process::new(
				sim,
				sim_main(
					sim,
					duration.get::<uom::si::time::minute>(),
					ferries,
					harbors,
				),
			)
		},
	)
}
