//! This crate provides a simulation of the classic Dining Philosophers problem
//! using the ODEM-rs simulation library.
//!
//! The Dining Philosophers problem is a classic synchronization and concurrency
//! problem where a number of philosophers sit at a dining table, and each
//! philosopher alternates between thinking and eating. The challenge is usually
//! to design a solution that avoids deadlock and ensures proper resource
//! sharing (forks in this case). This simulation model attempts to determine
//! statistics about the amount of time it would take for a deadlock to occur if
//! the deadlock is not prevented.

use std::{fmt, rc::Rc};

use rand_distr::{Distribution, Exp, Normal, Uniform};
use tracing::debug;
use uom::si::time;

use odem_rs::prelude::*;
use rayon::prelude::*;

// delay related constants
const THINKING: f64 = 1.0;
const EATING: (f64, f64) = (0.5, 0.2);
const DELAY: (f64, f64) = (0.2, 0.4);

/// Configuration type for the Dining Philosophers simulation.
#[derive(Config, Default)]
struct DiningPhilosophers {
	/// The type and initial value of the model time.
	#[time]
	time: Time<f64>,
	/// The number of dining philosophers.
	count: usize,
	/// A stream of independent random-number generators.
	rng_stream: RngStream,
}

/// Represents the dining table with forks in the Dining Philosophers scenario.
struct Table {
	/// Forks on the table, represented as binary [Control] variables
	/// (true if taken).
	forks: Vec<Control<bool>>,
	/// Number of forks currently held by philosophers.
	forks_held: Control<usize>,
	/// Number of philosophers waiting for forks.
	forks_awaited: Control<usize>,
}

impl Table {
	/// Creates a new dining table with a specified number of forks.
	fn new(forks: usize) -> Self {
		Table {
			forks: (0..forks).map(|_| Control::new(false)).collect(),
			forks_held: Control::new(0),
			forks_awaited: Control::new(0),
		}
	}

	/// Blocks until the requested fork is available and acquires it.
	async fn acquire_fork(&self, i: usize) {
		let num = i % self.forks.len();
		self.forks_awaited.set(self.forks_awaited.get() + 1);

		until!(!self.forks[num]).await;

		self.forks[num].set(true);
		self.forks_awaited.set(self.forks_awaited.get() - 1);
		self.forks_held.set(self.forks_held.get() + 1);
	}

	/// Releases the requested fork back to the table.
	fn release_fork(&self, i: usize) {
		self.forks[i % self.forks.len()].set(false);
		self.forks_held.set(self.forks_held.get() - 1);
	}
}

/// Represents a philosopher in the Dining Philosophers scenario.
struct Philosopher {
	table: Rc<Table>,
	seat: usize,
}

impl Behavior<DiningPhilosophers> for Philosopher {
	type Output = ();

	/// Defines the actions a philosopher goes through in the Dining
	/// Philosophers scenario.
	async fn actions(&self, sim: &Sim<DiningPhilosophers>) -> Self::Output {
		let mut rng = sim.global().rng_stream.rng();
		let thinking_duration = Exp::new(THINKING).unwrap();
		let eating_duration = Normal::new(EATING.0, EATING.1).unwrap();
		let artificial_delay = Uniform::new(DELAY.0, DELAY.1).unwrap();

		loop {
			debug!("pondering the nature of things");

			// spend some time pondering the nature of things
			sim.advance(minute::<f64>(thinking_duration.sample(&mut rng)))
				.await;

			debug!("getting hungry, time to eat");

			// acquire the first fork
			self.table.acquire_fork(self.seat).await;

			debug!("picked up the left fork");

			// introduce an artificial delay to leave room for deadlocks
			sim.advance(minute::<f64>(artificial_delay.sample(&mut rng)))
				.await;

			// acquire the second fork
			self.table.acquire_fork(self.seat + 1).await;

			debug!("picked up the right fork");

			// spend some time eating
			sim.advance(minute::<f64>(
				eating_duration
					.sample_iter(&mut rng)
					.find(|&val| val >= 0.0)
					.unwrap(),
			))
			.await;

			debug!("finished eating, returning the forks");

			// release the forks
			self.table.release_fork(self.seat + 1);
			self.table.release_fork(self.seat);
		}
	}
}

/// Runs the Dining Philosophers simulation once and returns the model time
/// until the deadlock.
async fn sim_main(sim: &Sim<DiningPhilosophers>) -> Time<f64> {
	let count = sim.global().count;
	let table = Rc::new(Table::new(count));
	let arena = pool!(count);

	// create and activate the philosopher-processes
	for i in 0..count {
		sim.activate(arena.alloc(Process::new(Philosopher {
			table: table.clone(),
			seat: i,
		})));
	}

	// wait for the precise configuration indicating a deadlock
	// (we technically don't have to do this, because deadlocks imply that no
	//  processes can advance, causing the simulation to end;
	//  SLX doesn't allow us to record the time of that event, so we can't use
	//  it here either)
	until!(table.forks_held == count && table.forks_awaited == count).await;

	// return the current system time
	sim.now()
}

/// Executes the Dining Philosophers simulation with multithreading and returns
/// model time statistics related to the deadlock.
pub fn philosophers(count: usize, reruns: usize) -> impl fmt::Debug {
	// use thread-based parallelism to concurrently run simulation models
	(0..reruns)
		.into_par_iter()
		.map(|seed| {
			Simulator::new(DiningPhilosophers {
				time: second(0.0),
				count,
				rng_stream: RngStream::from_seed(seed as u64),
			})
			.run(sim_main)
			.unwrap()
		})
		.fold(RandomVariable::new::<2>, |var, duration| {
			var.tabulate(duration);
			var
		})
		.reduce(RandomVariable::default, RandomVariable::join)
		.displayed(time::minute)
}

/// Executes the Dining Philosophers simulation with single-threading and
/// returns the average time until the deadlock.
pub fn philosophers_st(count: usize, reruns: usize) -> impl fmt::Debug {
	// run the simulations sequentially
	let random_var = RandomVariable::new::<2>();
	let mut simulator = Simulator::new(DiningPhilosophers {
		count,
		..DiningPhilosophers::default()
	});

	for seed in 0..reruns {
		simulator.inner_mut().rng_stream = RngStream::from_seed(seed as u64);
		random_var.tabulate(simulator.run(sim_main).unwrap());
	}

	random_var.displayed(time::minute)
}
