//! A discrete event simulation of the Dining Philosophers problem using
//! `simcore_rs`, the simulator core developed as part of my (Dorian Weber)
//! dissertation.

use rand::{Rng, SeedableRng};
use rand_distr::{Distribution, Exp, Normal, Uniform};
use rayon::prelude::*;
use simcore_rs::{
	util::{until, Control, RandomVariable},
	Process, Sim, Time,
};

use crate::DefaultRng;

use std::{
	cell::{Cell, RefCell},
	rc::Rc,
};

// delay related constants
const THINKING: f64 = 1.0;
const EATING: (f64, f64) = (0.5, 0.2);
const DELAY: (f64, f64) = (0.2, 0.4);

/// Shared global data for the simulation.
struct Philosophers {
	/// Master random number generator for seeding.
	master_rng: RefCell<DefaultRng>,
	/// Cell to store the simulation duration until deadlock.
	sim_duration: Cell<Time>,
}

/// Represents the shared table with forks for the philosophers.
struct Table {
	/// Controls for each fork, indicating whether it is held.
	forks: Vec<Control<bool>>,
	/// Counter for the number of forks currently held.
	forks_held: Control<usize>,
	/// Counter for the number of forks currently awaited.
	forks_awaited: Control<usize>,
}

impl Table {
	/// Creates a new table with a specified number of forks.
	fn new(forks: usize) -> Self {
		Table {
			forks: (0..forks).map(|_| Control::new(false)).collect(),
			forks_held: Control::default(),
			forks_awaited: Control::default(),
		}
	}

	/// Acquires a fork at a given position, blocking until it is available.
	async fn acquire_fork(&self, i: usize) {
		let num = i % self.forks.len();
		self.forks_awaited.set(self.forks_awaited.get() + 1);
		until(&self.forks[num], |fork| !fork.get()).await;
		self.forks[num].set(true);
		self.forks_awaited.set(self.forks_awaited.get() - 1);
		self.forks_held.set(self.forks_held.get() + 1);
	}

	/// Releases a fork at a given position back to the table.
	fn release_fork(&self, i: usize) {
		self.forks[i % self.forks.len()].set(false);
		self.forks_held.set(self.forks_held.get() - 1);
	}
}

/// Represents a philosopher in the simulation.
struct Philosopher {
	/// Reference to the shared table.
	table: Rc<Table>,
	/// The seat index of the philosopher at the table.
	seat: usize,
	/// Random number generator for this philosopher.
	rng: DefaultRng,
}

impl Philosopher {
	/// Simulates the actions of a philosopher.
	///
	/// The philosopher alternates between thinking and eating, and tries to
	/// acquire forks.
	async fn actions(mut self, sim: Sim<'_, Philosophers>) {
		let thinking_duration = Exp::new(THINKING).unwrap();
		let eating_duration = Normal::new(EATING.0, EATING.1).unwrap();
		let artificial_delay = Uniform::new(DELAY.0, DELAY.1).unwrap();

		loop {
			// Spend some time pondering the nature of things.
			sim.advance(thinking_duration.sample(&mut self.rng)).await;

			// Acquire the first fork.
			self.table.acquire_fork(self.seat).await;

			// Introduce an artificial delay to leave room for deadlocks.
			sim.advance(artificial_delay.sample(&mut self.rng)).await;

			// Acquire the second fork.
			self.table.acquire_fork(self.seat + 1).await;

			// Spend some time eating.
			sim.advance(
				eating_duration
					.sample_iter(&mut self.rng)
					.find(|&val| val >= 0.0)
					.unwrap(),
			)
			.await;

			// Release the forks.
			self.table.release_fork(self.seat + 1);
			self.table.release_fork(self.seat);
		}
	}
}

/// Runs a single simulation instance of the Dining Philosophers problem.
///
/// Initializes the table and philosophers, and waits until a deadlock occurs.
async fn sim_main(sim: Sim<'_, Philosophers>, count: usize) {
	let table = Rc::new(Table::new(count));

	// Create the philosopher processes and seat them.
	for i in 0..count {
		sim.activate(
			Philosopher {
				table: table.clone(),
				seat: i,
				rng: DefaultRng::from_seed(sim.global().master_rng.borrow_mut().random()),
			}
			.actions(sim),
		);
	}

	// Wait for the precise configuration indicating a deadlock.
	until(
		(&table.forks_held, &table.forks_awaited),
		|(held, awaited)| held.get() == count && awaited.get() == count,
	)
	.await;

	// Record the current simulation time.
	sim.global().sim_duration.set(sim.now());
}

/// Runs multiple simulation instances in parallel and collects statistics.
///
/// # Arguments
///
/// * `count` - The number of philosophers (and forks) in each simulation.
/// * `reruns` - The number of simulation runs to perform.
///
/// # Returns
///
/// A `RandomVar` containing statistical data of simulation durations until
/// deadlock.
pub fn philosophers(count: usize, reruns: usize) -> RandomVariable {
	// Use thread-based parallelism to concurrently run simulation models.
	(1..=reruns)
		.into_par_iter()
		.map(|i| {
			simcore_rs::simulation(
				// Global data.
				Philosophers {
					master_rng: RefCell::new(DefaultRng::seed_from_u64(i as u64)),
					sim_duration: Cell::default(),
				},
				// Simulation entry point.
				|sim| Process::new(sim, sim_main(sim, count)),
			)
			.sim_duration
			.get()
		})
		.fold(RandomVariable::new, |var, duration| {
			var.tabulate(duration);
			var
		})
		.reduce(RandomVariable::new, |var_a, var_b| {
			var_a.merge(&var_b);
			var_a
		})
}
