//! This module provides a set of types and implementations to simulate and
//! manage the flow of visitors in a museum environment. It includes various
//! [Station] specializations like `Switch`, `Exponat`, `WaitingArea`,
//! `SeminarRoom`, and `Desk`, each with specific roles and behaviors in the
//! visitor management system.
//!
//! ## Overview
//! - [`Switch`]: A station type without a specified visitor stay time,
//!   primarily used for distributing visitors to other stations.
//! - [`Exponat`]: Represents exhibit areas where visitors actively stay;
//!   includes time and group capacity management.
//! - [`WaitingArea`]: A station for waiting areas with an unknown stay time for
//!   visitors, managing people capacity.
//! - [`SeminarRoom`]: A station for presentations or movies with known stay
//!   time and visitor capacity.
//! - [`Desk`]: Represents registers or counters for handling payments and
//!   tickets, with processing time management.
//!
//! Each station type implements the `Publisher` trait for subscription
//! management and the `Timesink` trait for asynchronous interaction with the
//! `StasiMuseum` simulation environment. These traits allow for dynamic visitor
//! flow control, including taking in, hosting, and releasing visitor groups,
//! as well as providing visual representations through icons.
//! 
//! [Station]: museum::Station

use async_trait::async_trait;
use rand::{Rng, RngCore};
use rand_distr::Normal;

use museum::Timesink;
use odem_rs::{
	prelude::*,
	sync::{Facility, Publisher},
	util::uom::*
};

use super::{Group, minute, StasiMuseum};

/* ************************************************************ Station Types */

/// Station-type that doesn't have a specified time for visitors to stay.
///
/// It's typically used to distribute visitors to other stations.
#[derive(Publisher)]
pub struct Switch {
	/// Contains the number of people at the switch.
	#[subscribe]
	people: Control<usize>,
	/// The total capacity of the switch for visitors.
	capacity: usize,
}

impl Switch {
	/// Creates a new [Switch] with a specific capacity for visitors.
	pub fn new(capacity: usize) -> Self {
		Self { people: Control::new(0), capacity }
	}
	
	/// Checks if the [Switch] is currently empty (no visitors).
	#[inline]
	pub fn is_empty(&self) -> bool {
		self.people.get() == 0
	}
	
	/// Determines if there is enough space for a specified number of people.
	#[inline]
	pub fn has_space(&self, people: usize) -> bool {
		self.capacity - self.people.get() >= people
	}
	
	/// Gets the current number of people at the [Switch].
	#[inline]
	pub fn people(&self) -> usize {
		self.people.get()
	}
}

#[async_trait(?Send)]
impl Timesink<StasiMuseum> for Switch {
	async fn take(&self, visitor: &Group) {
		until!(self.capacity - self.people >= visitor.members).await;
		self.people.update(|people| people + visitor.members);
	}
	
	async fn give(&self, visitor: &Group) {
		self.people.update(|people| people - visitor.members);
	}
	
	fn icon(&self) -> Option<&'static str> {
		Some(r#"image="picto/Verteiler.svg""#)
	}
}

/// Station-type that is actively stayed at by visitors.
///
/// Typically that includes the actual attractions of the museum.
#[derive(Publisher)]
pub struct Exponat {
	/// Contains the number of groups at the exponat.
	#[subscribe]
	groups: Control<usize>,
	capacity: usize,
	visit_time: Normal<f64>,
}

impl Exponat {
	/// Creates a new [Exponat] with a specific normal-distributed distribution
	/// of time to stay and a maximal number of groups to visit at a time.
	pub fn new(expected: Time<f64>, deviation: Time<f64>, capacity: usize) -> Self {
		Self {
			groups: Control::new(0),
			capacity,
			visit_time: Normal::new(
				expected.get::<time::minute>(),
				deviation.get::<time::minute>()
			).unwrap(),
		}
	}
	
	/// Checks if the [Exponat] is currently empty (no visitors).
	#[inline]
	pub fn is_empty(&self) -> bool {
		self.groups.get() == 0
	}
	
	/// Determines if there is enough space for a specified number of groups.
	#[inline]
	pub fn has_space(&self, groups: usize) -> bool {
		self.capacity - self.groups.get() >= groups
	}
	
	/// Gets the current number of groups at the [Exponat].
	#[inline]
	pub fn groups(&self) -> usize {
		self.groups.get()
	}
}

#[async_trait(?Send)]
impl Timesink<StasiMuseum> for Exponat {
	async fn take(&self, _visitor: &Group) {
		until!(self.groups < self.capacity).await;
		self.groups.update(|count| count + 1);
	}
	
	async fn stay(&self, sim: &Sim<StasiMuseum>, rng: &mut dyn RngCore) {
		sim.advance(minute(rng.sample_iter(&self.visit_time).find(|&dt| dt > 0.0).unwrap())).await;
	}
	
	async fn give(&self, _visitor: &Group) {
		self.groups.update(|count| count - 1);
	}
	
	fn icon(&self) -> Option<&'static str> {
		Some(r#"image="picto/Exponat.svg""#)
	}
}

/// Station-type for waiting areas with an unknown time for visitors to stay.
#[derive(Publisher)]
pub struct WaitingArea {
	/// Contains the number of people at the waiting area.
	#[subscribe]
	people: Control<usize>,
	capacity: usize
}

impl WaitingArea {
	/// Creates a new [WaitingArea] with a specific capacity for visitors.
	pub fn new(capacity: usize) -> Self {
		Self { people: Control::new(0), capacity }
	}
	
	/// Checks if the [WaitingArea] is currently empty (no visitors).
	#[inline]
	pub fn is_empty(&self) -> bool {
		self.people.get() == 0
	}
	
	/// Determines if there is enough space for a specified number of people.
	#[inline]
	pub fn has_space(&self, people: usize) -> bool {
		self.capacity - self.people.get() >= people
	}
	
	/// Gets the current number of people at the [WaitingArea].
	#[inline]
	pub fn people(&self) -> usize {
		self.people.get()
	}
}

#[async_trait(?Send)]
impl Timesink<StasiMuseum> for WaitingArea {
	async fn take(&self, visitor: &Group) {
		until!(self.capacity - self.people >= visitor.members).await;
		self.people.update(|people| people + visitor.members);
	}
	
	async fn give(&self, visitor: &Group) {
		self.people.update(|people| people - visitor.members);
	}
	
	fn icon(&self) -> Option<&'static str> {
		Some(r#"image="picto/Wartebereich.svg""#)
	}
}

/// Station-type for presentations or movies.
///
/// Has a known stay time and a limit for visitors (not groups).
#[derive(Publisher)]
pub struct SeminarRoom {
	/// Contains the number of people in the seminar room.
	#[subscribe]
	people: Control<usize>,
	capacity: usize,
	visit_time: Normal<f64>,
}

impl SeminarRoom {
	/// Creates a new [SeminarRoom] with a normal-distributed distribution of
	/// staying time and a capacity for visitors.
	pub fn new(expected: Time<f64>, deviation: Time<f64>, capacity: usize) -> Self {
		Self {
			people: Control::new(0), capacity,
			visit_time: Normal::new(expected.get::<time::minute>(), deviation.get::<time::minute>()).unwrap(),
		}
	}
	
	/// Checks if the [SeminarRoom] is currently empty (no visitors).
	#[inline]
	pub fn is_empty(&self) -> bool {
		self.people.get() == 0
	}
	
	/// Determines if there is enough space for a specified number of people.
	#[inline]
	pub fn has_space(&self, people: usize) -> bool {
		self.capacity - self.people.get() >= people
	}
	
	/// Gets the current number of people in the [SeminarRoom].
	#[inline]
	pub fn people(&self) -> usize {
		self.people.get()
	}
}

#[async_trait(?Send)]
impl Timesink<StasiMuseum> for SeminarRoom {
	async fn take(&self, visitor: &Group) {
		until!(self.capacity - self.people >= visitor.members).await;
		self.people.update(|people| people + visitor.members);
	}
	
	async fn stay(&self, sim: &Sim<StasiMuseum>, rng: &mut dyn RngCore) {
		sim.advance(minute(rng.sample_iter(&self.visit_time).find(|&dt| dt > 0.0).unwrap())).await;
	}
	
	async fn give(&self, visitor: &Group) {
		self.people.update(|people| people - visitor.members);
	}
	
	fn icon(&self) -> Option<&'static str> {
		Some(r#"image="picto/Seminarraum.svg""#)
	}
}

/// Station-type for registers, typically for handling payments and tickets.
pub struct Desk {
	in_use: Facility,
	visit_time: Normal<f64>
}

impl Desk {
	/// Creates a new [Desk] with a specific normal-distributed distribution
	/// for the time necessary for processing each visitor.
	pub fn new(expected: Time<f64>, deviation: Time<f64>) -> Self {
		Self {
			in_use: Facility::new(),
			visit_time: Normal::new(
				expected.get::<time::minute>(),
				deviation.get::<time::minute>()
			).unwrap()
		}
	}
}

#[async_trait(?Send)]
impl Timesink<StasiMuseum> for Desk {
	async fn stay(&self, sim: &Sim<StasiMuseum>, rng: &mut dyn RngCore) {
		let desk = self.in_use.seize().await;
		sim.advance(minute(rng.sample_iter(&self.visit_time).find(|&dt| dt > 0.0).unwrap())).await;
		desk.release();
	}
	
	fn icon(&self) -> Option<&'static str> {
		Some(r#"image="picto/Schalter.svg""#)
	}
}
