//! The `calendar` module provides a flexible event scheduling mechanism for
//! managing tasks in a simulation.
//!
//! It introduces the `Scheduler` trait, a generic interface for unifying access
//! to an event calendar, along with the associated `PlanState` trait for
//! managing the state of scheduled tasks. The module also includes the
//! `Calendar` structure, implementing a task calendar based on a user-provided
//! `Scheduler`.
//!
//! ## Scheduler Trait
//!
//! The [`Scheduler`] trait defines methods for scheduling, deferring,
//! activating, purging, and extracting tasks within a simulation. It offers a
//! unified interface for interacting with the event calendar, allowing
//! customization through configuration-specific implementations.
//!
//! ## PlanState Trait
//!
//! The [`PlanState`] trait is implemented by the calendar-specific event state
//! for managed tasks. It provides methods to extract the model-time and rank of
//! the next activation if in the corresponding state.
//!
//! ## Calendar Structure
//!
//! The [`Calendar`] structure acts as a container for managing scheduled tasks
//! based on the provided `Scheduler`. It offers methods for scheduling,
//! deferring, activating, purging, and extracting tasks. The calendar is
//! resettable, allowing for the reuse of the same instance in subsequent
//! simulation runs.

use core::{
	cell::{Cell, RefCell},
	fmt::Debug,
};
pub use v3::DefaultPlan;

use crate::{
	config::{Config, Rank, Time},
	fsm::*,
	ptr::Irc,
	task::{token, Task},
};

pub mod ordering;
mod v3;

/* ********************************************************** Scheduler Trait */

/// A trait for unifying access to the event calendar.
pub trait Scheduler {
	/// The type of the associated [configuration](Config) for this scheduler.
	type Config: ?Sized + Config;

	/// The type of the state payload when managed by the calendar.
	type State: PlanState<
		Time = <Self::Config as Config>::Time,
		Rank = <Self::Config as Config>::Rank,
	>;

	/// Creates a new scheduler from a configuration.
	fn new(config: &Self::Config) -> Self;

	/// Schedules a task for a later (or the current) model-time.
	///
	/// Calling this method with a model-time of [now] is semantically
	/// equivalent to calling [activate].
	///
	/// [activate]: Self::activate
	/// [now]: Self::now
	fn schedule<'brand>(
		&mut self,
		task: Irc<Task<'brand, Self::Config>>,
		idle: token::Idle<'brand>,
		time: <Self::Config as Config>::Time,
	) -> token::Next<'brand>;

	/// Defers a task to be executed after all of the other tasks in this
	/// time-slice have completed.
	fn defer<'brand>(
		&mut self,
		task: Irc<Task<'brand, Self::Config>>,
		busy: token::Busy<'brand>,
	) -> token::Next<'brand>;

	/// Schedules an owned task for immediate execution.
	///
	/// This is semantically equivalent to calling [schedule] with the
	/// model-time returned by [now].
	///
	/// [schedule]: Self::schedule
	/// [now]: Self::now
	fn activate<'brand>(
		&mut self,
		task: Irc<Task<'brand, Self::Config>>,
		idle: token::Idle<'brand>,
	) -> token::Next<'brand>;

	/// Purges a task from the calendar.
	fn purge<'brand>(
		&mut self,
		task: &Task<'brand, Self::Config>,
		next: token::Next<'brand>,
	) -> token::Idle<'brand>;

	/// Extracts the next puck from the calendar. Returns `None` if no
	/// puck has been scheduled.
	fn extract(&mut self) -> Option<Irc<Task<'static, Self::Config>>>;

	/// Returns `true` if the calendar is empty and `false` if at least one
	/// [Task] is scheduled.
	fn is_empty(&self) -> bool;

	/// Returns the current model time.
	fn now(&self) -> <Self::Config as Config>::Time;
}

/* ***************************************************** Calendar Event State */

/// A trait implemented by the calendar-specific event state for managed tasks.
pub trait PlanState: Debug {
	/// The [Time] type of this state.
	type Time: Time;

	/// The [Rank] type of this state.
	type Rank: Rank;

	/// Extracts the model-time of the next activation if in the corresponding
	/// state.
	fn time(&self) -> Option<Self::Time>;

	/// Extracts the local rank if in the corresponding state.
	fn rank(&self) -> Option<Self::Rank>;
}

/* ***************************************************************** Calendar */

/// Structure that implements a [Task] calendar based on a user-provided
/// [Scheduler].
pub struct Calendar<C: ?Sized + Config> {
	plan: RefCell<C::Plan>,
	now: Cell<C::Time>,
}

impl<C: ?Sized + Config> Calendar<C> {
	/// Creates a new [Task] calendar from a [Scheduler].
	#[inline]
	pub(crate) fn new(config: &C) -> Self {
		Calendar {
			plan: RefCell::new(C::Plan::new(config)),
			now: Cell::new(config.default_time()),
		}
	}

	/// Schedules a task for a later (or the current) model-time. Returns with
	/// a [Causality]-error if the task was scheduled for a past model-time.
	#[inline]
	pub fn schedule<'brand>(
		&self,
		task: Irc<Task<'brand, C>>,
		idle: token::Idle<'brand>,
		time: C::Time,
	) -> token::Next<'brand> {
		let _span = task.branded_share(&idle).enter_span();
		self.plan.borrow_mut().schedule(task.clone(), idle, time)
	}

	/// Defers a task to be executed after all of the other tasks in this
	/// time-slice have completed.
	#[inline]
	pub fn defer<'brand>(
		&self,
		task: Irc<Task<'brand, C>>,
		busy: token::Busy<'brand>,
	) -> token::Next<'brand> {
		self.plan.borrow_mut().defer(task.clone(), busy)
	}

	/// Schedules an owned task for immediate execution.
	///
	/// The difference between calling this method and calling [schedule] with
	/// the model-time returned by [now] is that the activated task becomes
	/// part of the current batch of active tasks, potentially displacing other
	/// tasks with lower priority whereas the call to [schedule] will finish
	/// executing the current batch and only then activate the provided task.
	///
	/// [schedule]: Self::schedule
	/// [now]: Self::now
	#[inline]
	pub fn activate<'brand>(
		&self,
		task: Irc<Task<'brand, C>>,
		idle: token::Idle<'brand>,
	) -> token::Next<'brand> {
		let _span = task.branded_share(&idle).enter_span();
		self.plan.borrow_mut().activate(task.clone(), idle)
	}

	/// Purges a task from the calendar.
	#[inline]
	pub fn purge<'brand>(
		&self,
		task: &Task<'brand, C>,
		next: token::Next<'brand>,
	) -> token::Idle<'brand> {
		let _span = task.branded_share(&next).enter_span();
		self.plan.borrow_mut().purge(task, next)
	}

	/// Extracts the next puck from the calendar and runs a custom closure,
	/// returning both it and the result of the closure. Returns `None` if no
	/// puck has been scheduled.
	#[inline]
	pub fn extract(&self) -> Option<Irc<Task<'static, C>>> {
		let mut plan = self.plan.borrow_mut();
		let next = plan.extract();
		self.now.set(plan.now());
		next
	}

	/// Returns `true` if the calendar is empty and `false` if at least one
	/// [Task] is scheduled.
	#[inline]
	pub fn is_empty(&self) -> bool {
		self.plan.borrow().is_empty()
	}

	/// Returns the current model-time.
	#[inline]
	pub fn now(&self) -> C::Time {
		self.now.get()
	}
}

/* *************************************************** Guarded Task Reference */

/// Newtype wrapper that transitions tasks away on drop.
///
/// This simplifies clearing of the [DefaultPlan] since the
/// destructor of the owned pointer type also transitions the [Task] away from
/// the [TaskState] indicating where in the calendar it is located, preventing
/// hard-to-track bugs when the calendar is cleared before the task drops and
/// the task's state indicates that it is still scheduled, moving or active.
///
/// [TaskState]: crate::erased::TaskState
#[repr(transparent)]
pub(crate) struct IdleOnDrop<C: ?Sized + Config>(pub Irc<Task<'static, C>>);

impl<C: ?Sized + Config> IdleOnDrop<C> {
	/// Creates a new [IdleOnDrop]-wrapper that accepts [Irc] of [Task] in
	/// state [Next](token::Next).
	///
	/// This constructor freezes the state of the underlying task until either
	/// its `Drop` is executed or the task is extracted using [Self::into_inner].
	#[inline]
	pub fn new<'brand>(task: Irc<Task<'brand, C>>, next: &token::Next<'brand>) -> Self {
		// ignore the witness
		let _ = next;

		// freeze the state
		unsafe {
			task.enter();
		}

		// construct the instance
		Self(Irc::map(task, Task::detach))
	}

	/// Extracts the inner [Irc] without invoking the destructor.
	#[inline]
	pub fn into_inner(self) -> Irc<Task<'static, C>> {
		// unfreeze the state
		unsafe {
			self.0.leave();
		}

		// extract the inner `Irc` without running drop
		unsafe { core::mem::transmute(self) }
	}
}

impl<C: ?Sized + Config> From<Irc<Task<'_, C>>> for IdleOnDrop<C> {
	#[inline]
	fn from(task: Irc<Task<'_, C>>) -> Self {
		task.brand(move |task, once| {
			let next = task.token(once).into_next().unwrap();
			Self::new(task.clone(), &next)
		})
	}
}

impl<C: ?Sized + Config> From<IdleOnDrop<C>> for Irc<Task<'static, C>> {
	#[inline]
	fn from(guard: IdleOnDrop<C>) -> Self {
		guard.into_inner()
	}
}

impl<C: ?Sized + Config> Drop for IdleOnDrop<C> {
	fn drop(&mut self) {
		// unfreeze the state
		unsafe {
			self.0.leave();
		}

		// transition into the 'Idle' state
		(*self.0).brand(move |task, once| {
			let next = task.token(once).into_next().unwrap();
			let _: token::Idle<'_> = task.state().transition(next, ());
		});
	}
}
