#[allow(unused_imports)]
use core::{
	cmp::Ordering,
	fmt,
	num::Wrapping,
	ops::{Add, AddAssign, Range, RangeFrom, RangeFull, RangeTo, Sub, SubAssign},
};

/* ********************************************************************* Mark */

/// A tag-type used to preserve the order of inserted tasks into the calendar.
///
/// It consists of two integers: the first serves as a generation counter that
/// is increased before starting another time slot, the second ensures that
/// processes, i.e. [tasks] with shared data, are executed in batches in
/// insertion order.
///
/// The purpose of this construction is to sort tasks according to the order
/// their underlying processes were inserted and to be able to invalidate the
/// previous order by incrementing the generation counter.
///
/// [tasks]: super::Task
#[derive(Copy, Clone, Debug, Default, Ord, PartialOrd, Eq, PartialEq)]
pub struct EpochCounter<T> {
	/// Generation counter used to distinguish different epochs in the calendar.
	epoch: T,
	/// Order attribute used for sorting tasks in the same epoch.
	order: T,
}

/// New-type-wrapper for cyclic-ordered unsigned integers.
///
/// Cyclic ordering refers to the property that the elements are arranged in a
/// ring and an element `a` is considered 'lesser' than an element `b` if
/// the distance between `b` and `a` is lesser going backwards than it is going
/// forwards.
///
/// This technically violates the transitive property of the order-relation
/// since we can easily find a sequence `x < y < z < x` that spans the whole
/// range of the integer type. However, every *upper* and *lower set* spanning
/// less than half the integer range is totally ordered.
#[derive(Copy, Clone, Debug, Default, PartialEq, Eq)]
#[cfg_attr(doc, derive(PartialOrd, Ord))]
pub struct Cyclic<T>(pub Wrapping<T>);

macro_rules! impl_mark {
	(@impl $UID:ident) => {
		#[allow(dead_code)]
		impl EpochCounter<$UID> {
			/// Constructs a new (potentially non-stale) mark.
			#[inline]
			pub const fn new() -> Self {
				Self { epoch: 1, order: 0 }
			}

			/// Constructs a new mark that is guaranteed to be stale.
			#[inline]
			pub const fn stale() -> Self {
				Self { epoch: 0, order: 0 }
			}

			/// Tests whether two marks are of the same generation.
			#[inline]
			pub const fn is_stale(&self, master: &Self) -> bool {
				self.epoch != master.epoch
			}

			/// Returns the new mark with increased order.
			#[inline]
			pub const fn next_gen(self) -> Self {
				Self {
					epoch: self.epoch,
					order: self.order.wrapping_add(1)
				}
			}

			/// Returns the new mark with increased epoch.
			///
			/// The generation counter will wrap around at the numeric boundaries but
			/// avoids generation 0 since that is also the generation of a newly
			/// created mark.
			#[inline]
			pub const fn next_epoch(self) -> Self {
				Self {
					epoch: if let Some(gen) = self.epoch.checked_add(1) { gen } else { 1 },
					order: 0
				}
			}

			/// Returns the order-attribute with cyclic ordering.
			#[inline]
			pub const fn order(&self) -> Cyclic<$UID> {
				Cyclic(Wrapping(self.order))
			}
		}
	};

	(@doc) => {
		impl<T> EpochCounter<T> {
			/// Constructs a new (potentially non-stale) mark.
			#[inline]
			pub const fn new() -> Self { unimplemented!() }

			/// Constructs a new mark that is guaranteed to be stale.
			#[inline]
			pub const fn stale() -> Self { unimplemented!() }

			/// Tests whether two marks are of the same generation.
			#[inline]
			pub const fn is_stale(&self, master: &Self) -> bool { unimplemented!() }

			/// Returns the new mark with increased order.
			#[inline]
			pub const fn next_gen(self) -> Self { unimplemented!() }

			/// Returns the new mark with increased epoch.
			///
			/// The generation counter will wrap around at the numeric boundaries but
			/// avoids generation 0 since that is also the generation of a newly
			/// created mark.
			#[inline]
			pub const fn next_epoch(self) -> Self { unimplemented!() }

			/// Returns the order-attribute with cyclic ordering.
			#[inline]
			pub const fn order(&self) -> Cyclic<T> { unimplemented!() }
		}
	};

	($($UID:ident),* $(,)?) => {
		$(impl_mark!(@impl $UID);)*
	};
}

macro_rules! impl_cyclic {
	(@impl $UID:ident -> $SID:ident) => {
		impl PartialOrd for Cyclic<$UID> {
			#[inline]
			fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
				Some(self.cmp(other))
			}
		}

		impl Ord for Cyclic<$UID> {
			#[inline]
			fn cmp(&self, other: &Self) -> Ordering {
				match (self.0.0.wrapping_sub(other.0.0) as $SID).signum() {
					0 => Ordering::Equal,
					1 => Ordering::Greater,
					_ => Ordering::Less
				}
			}
		}
	};

	($($UID:ident -> $SID:ident),* $(,)?) => {
		$(impl_cyclic!(@impl $UID -> $SID);)*
	};
}

#[cfg(not(doc))]
impl_mark!(u8, u16, u32, u64, usize);

#[cfg(doc)]
impl_mark!(@doc);

#[cfg(not(doc))]
impl_cyclic!(u8 -> i8, u16 -> i16, u32 -> i32, u64 -> i64, usize -> isize);

// forwarding implementation for Cyclic<T> to Wrapping<T>
impl<T: fmt::Display> fmt::Display for Cyclic<T> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		self.0.fmt(f)
	}
}

impl<T> Add for Cyclic<T>
where
	Wrapping<T>: Add<Output = Wrapping<T>>,
{
	type Output = Self;

	#[inline]
	fn add(self, other: Self) -> Self {
		Cyclic(self.0 + other.0)
	}
}

impl<T> AddAssign for Cyclic<T>
where
	Wrapping<T>: AddAssign,
{
	#[inline]
	fn add_assign(&mut self, other: Self) {
		self.0 += other.0;
	}
}

impl<T> AddAssign<T> for Cyclic<T>
where
	Wrapping<T>: AddAssign<T>,
{
	#[inline]
	fn add_assign(&mut self, other: T) {
		self.0 += other;
	}
}

impl<T> Sub for Cyclic<T>
where
	Wrapping<T>: Sub<Output = Wrapping<T>>,
{
	type Output = Self;

	#[inline]
	fn sub(self, other: Self) -> Self {
		Cyclic(self.0 - other.0)
	}
}

impl<T> SubAssign for Cyclic<T>
where
	Wrapping<T>: SubAssign,
{
	#[inline]
	fn sub_assign(&mut self, other: Self) {
		self.0 -= other.0;
	}
}

impl<T> SubAssign<T> for Cyclic<T>
where
	Wrapping<T>: SubAssign<T>,
{
	#[inline]
	fn sub_assign(&mut self, other: T) {
		self.0 -= other;
	}
}

/* *************************************************************** Partitions */

/// A range-type used to encode fully-ordered, non-overlapping intervals of
/// contiguous integer numbers for use as a precedence type within the
/// simulation library.
///
/// The intervals are fully ordered in depth-first fashion, i.e. super
/// intervals sort before their sub intervals and neighboring intervals are
/// sorted in ascending order. This is unambiguous because all the generated
/// intervals are non-overlapping, hence the name of this type. An exciting
/// property is that two intervals may be compared by simply comparing
/// their inner integer type, making this kind of hierarchical comparison
/// very efficient.
///
/// # Layout
/// The internal representation uses only one unsigned integer type instead
/// of two to store the interval. The memory layout is as follows:
///
/// |   Offset   |  Nesting |
/// | :--------: | :------: |
/// | `n-1 .. e` | `e .. 0` |
///
/// where `n` is the size in bits and `e` is `log_2(n)`. This reserves
/// `p` bits to express a range that is encoded as a power of two.
///
/// ## Example
///
/// For `Partition<u8>` the 3 least significant bits are reserved to encode
/// the greatest possible distance of 256 from any origin.
/// The layout for that case is:
///
/// |  Offset  | Nesting  |
/// | :------: | :------: |
/// | `7 .. 3` | `3 .. 0` |
///
/// The instance `Partition(0u8)` represents the interval `0..32`, since the
/// offset is `0` and the Partition has not been nested.
///
/// The instance `Partition(1u8)` represents the interval `0..16`, since the
/// offset is still `0` and the Partition has a nesting of one, reducing the
/// total range in half.
///
/// The instance `Partition(9u8)` represents the interval `1..16`, since the
/// offset is now `1` and the Partition has a nesting of one.
#[derive(Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Default)]
pub struct Partition<T>(pub T);

/// An iterator over sub-intervals that are generated by [splitting] the
/// combined interval.
///
/// [splitting]: Partition::split
struct Iter<T> {
	next: Partition<T>,
	size: T,
}

macro_rules! impl_prec {
	(@impl $UID:ident $($GEN:ident)?) => {
		impl<$($GEN)*> Partition<$UID> {
			/// The number of bits used to store the nesting level. This
			/// constant is restricted by the total number of bits available to
			/// the underlying integer type since it represents the log2-size
			/// of the interval.
			const EXP_BITS: u32 = $UID::BITS.ilog2();

			/// The number of bits used to store the start-point of the
			/// partition, i.e. the *precision*.
			const NUM_BITS: u32 = $UID::BITS - Self::EXP_BITS;

			/// A bit-mask for all of the bits used to store the nesting.
			const EXP_MASK: $UID = !(!0 << Self::EXP_BITS);

			/// Creates a new partition representing the universal range.
			#[inline]
			pub const fn new() -> Self {
				Self(0)
			}

			/// Splits the current interval into `n` sub-intervals and returns
			/// `Some` iterator to them if the precision supports the split or
			/// `None` if there aren't enough bits left to encode all of the
			/// generated intervals.
			#[inline]
			pub const fn split(self, n: $UID) -> Option<impl ExactSizeIterator<Item = Self>> {
				let bits = n.next_power_of_two().trailing_zeros();

				if bits <= self.capacity() {
					Some(Iter {
						next: Self(self.0 + bits as $UID),
						size: n
					})
				} else {
					None
				}
			}

			/// Selects the next interval at the current level of nesting.
			///
			/// This produces unexpected results if the number of intervals that
			/// this partition has been split into previously is smaller
			/// than the number of the next interval.
			#[inline]
			const fn advance(self) -> Self {
				self.advance_by(1)
			}

			/// Selects the next specified interval at the current level of
			/// nesting.
			///
			/// The same caveats as in [advance] apply here as well.
			///
			/// [advance]: Self::advance
			#[inline]
			const fn advance_by(self, count: u32) -> Self {
				Self(
					self.0.wrapping_add(
						(count as $UID).wrapping_shl(self.0.wrapping_neg() as u32)
					)
				)
			}

			/// Returns the number of (still) possible splits as a power of two.
			#[inline]
			pub const fn capacity(self) -> u32 {
				Self::NUM_BITS - self.nesting() as u32
			}

			/// Returns the starting point of the interval that is represented
			/// by the partition.
			#[inline]
			const fn offset(self) -> $UID {
				self.0 >> Self::EXP_BITS
			}

			/// Returns the level of nesting. This is equal to the power-of-two
			/// of the number of intervals that a newly generated partition
			/// has been split into and is initially equal to 0.
			#[inline]
			const fn nesting(self) -> $UID {
				self.0 & Self::EXP_MASK
			}

			/// Calculates the distance encoded in the exponent bits of the
			/// range.
			const fn delta(self) -> $UID {
				(!0) as $UID >> Self::EXP_BITS >> self.nesting()
			}

			/// Returns the half-open interval [start, end) represented by this
			/// partition.
			#[inline]
			pub const fn range(self) -> core::ops::Range<$UID> {
				let lo = self.offset();
				let hi = lo + self.delta();
				// let hi = ((self.0 | (!0 >> self.nesting())) >> Self::EXP_BITS);
				lo..hi+1
			}

			/// Returns the half-open interval [start,end) represented by this
			/// partition and normalized as floating-point numbers in [0,1).
			#[inline]
			pub fn float_range(self) -> core::ops::Range<f64> {
				let max = (1 as $UID) << Self::NUM_BITS;
				let range = self.range();

				(range.start as f64) / (max as f64)
				.. (range.end as f64) / (max as f64)
			}
		}

		impl<$($GEN)*> From<$UID> for Partition<$UID> {
			#[inline]
			fn from(value: $UID) -> Self {
				Self((value << Self::EXP_BITS) | (Self::NUM_BITS as $UID))
			}
		}

		impl<$($GEN)*> fmt::Display for Partition<$UID> {
			fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
				f.debug_tuple("Partition")
				 .field(&self.float_range())
				 .finish()
			}
		}

		impl<$($GEN)*> fmt::Debug for Partition<$UID> {
			fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
				f.debug_struct("Partition")
				 .field("irange", &self.range())
				 .field("frange", &self.float_range())
				 .field("capacity", &self.capacity())
				 .finish()
			}
		}

		impl<$($GEN)*> fmt::Binary for Partition<$UID> {
			fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
				write!(
					f, "Partition({:0num$b}_{:0exp$b})",
					self.offset(), self.nesting(),
					exp = Self::EXP_BITS as usize,
					num = Self::NUM_BITS as usize
				)
			}
		}

		impl<$($GEN)*> Iterator for Iter<$UID> {
			type Item = Partition<$UID>;

			#[inline]
			fn next(&mut self) -> Option<Self::Item> {
				if self.size > 0 {
					let next = self.next;
					self.size -= 1;
					self.next = next.advance();
					Some(next)
				} else {
					None
				}
			}

			#[inline]
			fn size_hint(&self) -> (usize, Option<usize>) {
				(self.len(), Some(self.len()))
			}

			#[inline]
			fn count(self) -> usize {
				self.len()
			}
		}

		impl<$($GEN)*> ExactSizeIterator for Iter<$UID> {
			#[inline]
			fn len(&self) -> usize {
				self.size as usize
			}
		}
	};

	(@doc $UID:ident) => {
		impl_prec!(@impl $UID $UID);
	};

	($($UID:ident),* $(,)?) => {
		$(impl_prec!(@impl $UID);)*
	};
}

#[cfg(not(doc))]
impl_prec!(u8, u16, u32, u64, u128, usize);

#[cfg(doc)]
impl_prec!(@doc T);

#[cfg(test)]
mod tests {
	use super::*;

	#[test]
	fn partition_capacity() {
		assert_eq!(Partition::<u8>::new().capacity(), 8 - 3);
		assert_eq!(Partition::<u16>::new().capacity(), 16 - 4);
		assert_eq!(Partition::<u32>::new().capacity(), 32 - 5);
		assert_eq!(Partition::<u64>::new().capacity(), 64 - 6);
		assert_eq!(Partition::<u128>::new().capacity(), 128 - 7);
	}

	#[test]
	fn partition_deep_nesting() {
		let p1 = Partition::<u8>::new();

		for p2 in p1.split(4).expect("remaining capacity is 5") {
			for p3 in p2.split(4).expect("remaining capacity is 3") {
				for p4 in p3.split(2).expect("remaining capacity is 1") {
					assert_eq!(p4.capacity(), 0);
				}
			}
		}
	}

	#[test]
	fn partition_shallow_nesting() {
		let p1 = Partition::<u8>::new();

		for p2 in p1.split(32).expect("remaining capacity is 5") {
			assert_eq!(p2.capacity(), 0);
		}
	}
}
