use core::cmp::{Ordering, Reverse};

use intrusive_collections::{Adapter, KeyAdapter, LinkOps, PointerOps};

use crate::{
	calendar::ordering::Cyclic,
	calendar::{IdleOnDrop, Scheduler},
	config::Config,
	fsm::*,
	simulator::Prec,
	task,
};

use super::State;

/// Helper structure for keys that only implement [PartialOrd], e.g. [f32] and
/// [f64], that are expected to totally order at runtime.
#[repr(transparent)]
pub(crate) struct ForceOrd<T>(T);

impl<T: PartialEq> PartialEq for ForceOrd<T> {
	#[inline]
	fn eq(&self, other: &Self) -> bool {
		self.0.eq(&other.0)
	}
}

impl<T: PartialOrd> Eq for ForceOrd<T> {}

impl<T: PartialOrd> PartialOrd for ForceOrd<T> {
	#[inline]
	fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
		Some(self.cmp(other))
	}
}

impl<T: PartialOrd> Ord for ForceOrd<T> {
	#[inline]
	fn cmp(&self, other: &Self) -> Ordering {
		self.0.partial_cmp(&other.0).unwrap_or(Ordering::Equal)
	}
}

/* ************************************************************* Key Adapters */

/// Helper macro to generate an implementation for wrapper [Adapter].
macro_rules! intrusive_adapter_wrapper {
	() => {};

	($name:ident) => {
		unsafe impl<C: ?Sized + Config> Adapter for $name<C> {
			type LinkOps = <task::Adapter<C, IdleOnDrop<C>> as Adapter>::LinkOps;
			type PointerOps = <task::Adapter<C, IdleOnDrop<C>> as Adapter>::PointerOps;

			#[inline]
			unsafe fn get_value(&self, link: <Self::LinkOps as LinkOps>::LinkPtr) -> *const <Self::PointerOps as PointerOps>::Value {
				self.0.get_value(link)
			}

			#[inline]
			unsafe fn get_link(&self, value: *const <Self::PointerOps as PointerOps>::Value) -> <Self::LinkOps as LinkOps>::LinkPtr {
				self.0.get_link(value)
			}

			#[inline]
			fn link_ops(&self) -> &Self::LinkOps {
				self.0.link_ops()
			}

			#[inline]
			fn link_ops_mut(&mut self) -> &mut Self::LinkOps {
				self.0.link_ops_mut()
			}

			#[inline]
			fn pointer_ops(&self) -> &Self::PointerOps {
				self.0.pointer_ops()
			}
		}
	};

	($name:ident, $($tail:ident),*) => {
		intrusive_adapter_wrapper!($name);
		intrusive_adapter_wrapper!($($tail),*);
	};

	($($name:ident,)+) => {
		intrusive_adapter_wrapper!($($name),*);
	};
}

/// Modified [Adapter] that extracts the next model time from a [Task] when
/// asked for a key.
pub(crate) struct TimeKey<C: ?Sized + Config>(pub task::Adapter<C, IdleOnDrop<C>>);

/// Modified [Adapter] that extracts the process rank from a [Task] when
/// asked for a key.
pub(crate) struct RankKey<C: ?Sized + Config>(pub task::Adapter<C, IdleOnDrop<C>>);

/// Modified [Adapter] that extracts the precedence from a [Task] when
/// asked for a key.
pub(crate) struct PrecKey<C: ?Sized + Config>(pub task::Adapter<C, IdleOnDrop<C>>);

// derive the non-key-specific functionality from the wrapped adapter
intrusive_adapter_wrapper!(TimeKey, RankKey, PrecKey);

impl<'a, C: ?Sized + Config> KeyAdapter<'a> for TimeKey<C>
where
	C::Plan: Scheduler<State = State<C>>,
{
	type Key = (ForceOrd<C::Time>, Cyclic<u32>);

	fn get_key(&self, task: &'a task::Task<'static, C>) -> Self::Key {
		use super::State::Time;

		let state = task.state().borrow();

		let state = match &*state {
			task::State::Next(state) => state,
			_ => unreachable!(),
		};

		match state {
			Time(count, time) => (ForceOrd(*time), *count),
			state => unreachable!(
				"task in Time layer is in state `{:?}` instead of `Time`",
				state
			),
		}
	}
}

impl<'a, C: ?Sized + Config> KeyAdapter<'a> for RankKey<C>
where
	C::Plan: Scheduler<State = State<C>>,
{
	type Key = (Reverse<C::Rank>, Cyclic<u32>, Cyclic<u32>);

	fn get_key(&self, task: &'a task::Task<'static, C>) -> Self::Key {
		use super::State::Rank;

		task.brand(|task, once| {
			let state = task.state().borrow();

			let state = match &*state {
				task::State::Next(state) => state,
				_ => unreachable!(),
			};

			match state {
				Rank(count, rank) => (
					Reverse(rank.clone()),
					task.mark(&task.token(once).into_init().unwrap())
						.get()
						.order(),
					*count,
				),
				state => unreachable!(
					"task in Rank layer is in state `{:?}` instead of `Rank`",
					state
				),
			}
		})
	}
}

impl<'a, C: ?Sized + Config> KeyAdapter<'a> for PrecKey<C>
where
	C::Plan: Scheduler<State = State<C>>,
{
	type Key = (Prec, Cyclic<u32>);

	fn get_key(&self, task: &'a task::Task<'static, C>) -> Self::Key {
		use super::State::Prec;
		let state = task.state().borrow();

		let state = match &*state {
			task::State::Next(state) => state,
			_ => unreachable!(),
		};

		match state {
			Prec(count) => (task.prec(), *count),
			state => unreachable!(
				"task in Prec layer is in state `{:?}` instead of `Prec`",
				state
			),
		}
	}
}
