//! This module provides an event calendar implementation for scheduling and
//! managing tasks in a discrete event simulation system. The calendar organizes
//! tasks based on their activation time, rank, and precedence. It uses a
//! layered structure to efficiently handle scheduling and activation of tasks
//! in a simulation.
//!
//! ## Key Components
//!
//! The module includes the following key components:
//!
//! - [`LayerStack<C>`]: The main structure representing the event calendar.
//!   It contains layers for sorting tasks based on activation time, rank, and
//!   precedence. The layers ensure efficient retrieval and activation of tasks
//!   according to their scheduling attributes.
//!
//! - [`State<C>`]: An enumeration representing the possible states of a task in
//!   the event calendar. States include time-scheduled, deferred,
//!   rank-scheduled, and precedence-scheduled states.
//!
//! - Various helper functions: The module includes functions for adding tasks
//!   to different layers, removing tasks, and extracting tasks based on their
//!   scheduling attributes.
//!
//! The event calendar is designed to be generic and can be configured with a
//! user-defined configuration type implementing [`Config`] that implements the
//! necessary traits for time, rank, and task handling.

use intrusive_collections::{LinkedList, RBTree};

use adapter::*;

use crate::{
	config::{Config, Time},
	fmt,
	fsm::*,
	ptr::Irc,
	simulator::Mark,
	task::{self, token, Adapter, Task},
};

use super::{ordering::Cyclic, IdleOnDrop, PlanState, Scheduler};

mod adapter;

/* ************************************************************** Layer Stack */

/// The structure of the [Calendar] used to sort Tasks according to activation
/// time, rank, and precedence.
pub struct LayerStack<C: ?Sized + Config> {
	/// A generator for distinct process markings which is used to sort tasks
	/// assigned to the same process together.
	pid: Mark,

	/// A generator for distinct task markings which is used to ensure stability
	/// in the ordering of tasks with identical time, mark and precedence keys.
	ord: Cyclic<u32>,

	/// A pair of current model time and mark of the active process.
	top: (C::Time, Mark),

	/// The time layer, sorting tasks according to their model time.
	time: RBTree<TimeKey<C>>,

	/// The hold layer, containing deferred tasks.
	hold: LinkedList<Adapter<C, IdleOnDrop<C>>>,

	/// The rank layer, sorting tasks with identical model time according to
	/// rank and process mark.
	rank: RBTree<RankKey<C>>,

	/// The precedence layer, sorting tasks with identical model time, rank and
	/// process mark according to their individual precedence.
	prec: RBTree<PrecKey<C>>,
}

impl<C: ?Sized + Config<Plan = Self>> LayerStack<C> {
	/// Creates a new, empty calendar layer stack with a initial model time.
	#[inline]
	fn new(start: C::Time) -> Self {
		LayerStack {
			pid: Mark::new(),
			ord: Cyclic::default(),
			top: (start, Mark::stale()),
			time: RBTree::new(TimeKey(Adapter::NEW)),
			hold: LinkedList::new(Adapter::NEW),
			rank: RBTree::new(RankKey(Adapter::NEW)),
			prec: RBTree::new(PrecKey(Adapter::NEW)),
		}
	}

	/// Adds a [Task] to the time-layer, transitioning it into the `Next` state.
	#[inline]
	fn push_time<'brand>(
		&mut self,
		task: Irc<Task<'brand, C>>,
		idle: token::Idle<'brand>,
		time: C::Time,
	) -> token::Next<'brand> {
		// increment the cyclic order for stability
		self.ord += 1;

		// transition into state `Next`
		let next: token::Next<'_> = task.state().transition(idle, State::Time(self.ord, time));

		// insert the task into the time layer
		self.time.insert(IdleOnDrop::new(task, &next));

		// return the token signifying the `Next` state
		next
	}

	/// Adds a [Task] to the hold-layer, transitioning it into the `Next` state
	/// and reactivating it once all other non-deferred Tasks at the current
	/// model time have been activated.
	#[inline]
	fn push_hold<'brand>(
		&mut self,
		task: Irc<Task<'brand, C>>,
		busy: token::Busy<'brand>,
	) -> token::Next<'brand> {
		let sim = task.branded_share(&busy).sim();
		let idle = sim.unslot(&task, busy);

		// transition into state `Next`
		let next: token::Next<'_> = task.state().transition(idle, State::Hold);

		// append the task to the hold list
		self.hold.push_back(IdleOnDrop::new(task, &next));

		// return the token signifying the `Next` state
		next
	}

	/// Adds a [Task] to the rank- or prec-layer, transitioning it into the
	/// `Next` state.
	///
	/// Which layer it is added to depends on whether another Task of the same
	/// Process is currently active or not. If it is, the Task is added directly
	/// to the prec-layer, bypassing the rank-layer. This is independent of
	/// whether the Process rank has changed or not to better model the Process
	/// as being selected in the scheduler, not any individual Fiber.
	fn push_rank<'brand>(
		&mut self,
		task: Irc<Task<'brand, C>>,
		idle: token::Idle<'brand>,
		rank: C::Rank,
	) -> token::Next<'brand> {
		// retrieve the process mark
		let mark = task.mark(idle.as_init());

		// increment the cyclic order for stability
		self.ord += 1;

		// test if a task belonging to the same process has been inserted yet
		if mark.get().is_stale(&self.pid) {
			// no: update the Mark of the corresponding process
			let next = self.pid;
			self.pid = self.pid.next_gen();
			mark.set(next);
		} else if mark.get() == self.mark() {
			// yes and it's the currently active process: bypass the Rank layer
			let next: token::Next<'_> = task.state().transition(idle, State::Prec(self.ord));
			self.prec.insert(IdleOnDrop::new(task, &next));
			return next;
		}

		// insert the task into the Rank layer
		let next: token::Next<'_> = task.state().transition(idle, State::Rank(self.ord, rank));
		self.rank.insert(IdleOnDrop::new(task, &next));
		next
	}

	/// Removes a [Task] from the layer stack, transitioning it into state
	/// `Idle`.
	fn remove<'brand>(
		&mut self,
		task: &Task<'brand, C>,
		next: token::Next<'brand>,
	) -> token::Idle<'brand> {
		use State::*;

		// remove the task from the calendar and drop it
		match &*task.state().borrow() {
			task::State::Next(state) => match state {
				Time(_, _) => unsafe { self.time.cursor_mut_from_ptr(task.detach()) }.remove(),
				Hold => unsafe { self.hold.cursor_mut_from_ptr(task.detach()) }.remove(),
				Rank(_, _) => unsafe { self.rank.cursor_mut_from_ptr(task.detach()) }.remove(),
				Prec(_) => unsafe { self.prec.cursor_mut_from_ptr(task.detach()) }.remove(),
			},
			_ => unreachable!(),
		}
		.unwrap()
		.into_inner();

		// transition into state `Idle`
		task.state().transition(next, ())
	}

	/// Extracts a [Task] from the calendar, reorganizing the layers and
	/// advancing model-time if necessary.
	#[inline]
	fn pop(&mut self) -> Option<IdleOnDrop<C>> {
		self.prec
			.front_mut()
			.remove()
			.or_else(|| {
				if self.pull_rank() || self.pull_hold() || self.pull_time() {
					self.prec.front_mut().remove()
				} else {
					None
				}
			})
	}

	/// Returns whether the calendar is empty.
	#[inline]
	fn is_empty(&self) -> bool {
		self.prec.is_empty() && self.rank.is_empty() && self.hold.is_empty() && self.time.is_empty()
	}

	/// Returns the [Mark] of the currently active Process.
	#[inline]
	fn mark(&self) -> Mark {
		self.top.1
	}

	/// Attempts to extract the [Tasks] with the smallest model time,
	/// redistributing them over the layer stack. Returns `false` if the
	/// time-layer is empty.
	///
	/// [Tasks]: Task
	fn pull_time(&mut self) -> bool {
		// select and remove the first element of the Time tree
		let mut cur = self.time.front_mut();
		let mut guard = match cur.remove() {
			Some(guard) => guard,
			None => return false,
		};

		// update the current model time
		self.top.0 = (*guard.0).brand(|task, once| {
			task.next_time(&task.token(once).into_next().unwrap())
				.unwrap()
		});

		// invalidate all earlier marks
		self.pid = self.pid.next_epoch();

		// insert all the tasks into the rank tree
		loop {
			(*guard.0).brand(|task, once| {
				// extract the mark
				let next = task.token(once).into_next().unwrap();
				let mark = task.mark(&next);

				// check if the current mark of the task is valid
				if mark.get().is_stale(&self.pid) {
					// generate a new mark
					let next = self.pid;
					self.pid = self.pid.next_gen();
					mark.set(next);
				}

				// transition into the state corresponding to the Rank layer
				self.ord += 1;
				let rank = task.branded_share(&next).rank();
				task.state().update(next, State::Rank(self.ord, rank));
			});

			self.rank.insert(guard);

			// check if the model-time of the next task is the current one
			if let Some(task) = cur.get() {
				let next_time = task.brand(|task, once| {
					// we know statically that all the tasks in the calendar are
					// in state `Next(Time)`
					task.next_time(task.token(once).as_next().unwrap()).unwrap()
				});

				// continue this loop if the next task's model time is identical
				if next_time == self.top.0 {
					guard = cur.remove().unwrap();
					continue;
				}
			}

			// otherwise advance to extracting the tasks of the first process
			break self.pull_rank();
		}
	}

	/// Attempts to extract all of the deferred [Tasks], redistributing them
	/// into the lower layers. Returns `false` if there are no deferred Tasks.
	///
	/// [Tasks]: Task
	fn pull_hold(&mut self) -> bool {
		// exit early if the list is empty
		if self.hold.is_empty() {
			return false;
		}

		// invalidate all earlier marks
		self.pid = self.pid.next_epoch();

		// extract all the deferred tasks and insert them into the Rank tree
		for guard in self.hold.take() {
			(*guard.0).brand(|task, once| {
				let next = task.token(once).into_next().unwrap();
				let mark = task.mark(&next);

				if mark.get().is_stale(&self.pid) {
					// generate a new mark
					let next = self.pid;
					self.pid = self.pid.next_gen();
					mark.set(next);
				}

				// transition into the state corresponding to the Rank layer
				self.ord += 1;
				let rank = task.branded_share(&next).rank();
				task.state().update(next, State::Rank(self.ord, rank));
			});

			self.rank.insert(guard);
		}

		// advance to extracting the tasks of the first process
		self.pull_rank()
	}

	/// Attempts to extract the next batch of [Tasks] with the highest Rank from
	/// the Rank layer, redistributing them into the Prec layer. Returns `false`
	/// if the Rank layer is empty.
	///
	/// [Tasks]: Task
	fn pull_rank(&mut self) -> bool {
		// select and remove the first element from the Rank tree
		let mut cur = self.rank.front_mut();
		let mut guard = match cur.remove() {
			Some(guard) => guard,
			None => return false,
		};

		// update the current mark
		self.top.1 =
			(*guard.0).brand(|task, once| task.mark(task.token(once).as_init().unwrap()).get());

		// insert all the tasks into the Precedence tree
		loop {
			self.ord += 1;

			(*guard.0).brand(|task, once| {
				let next = task.token(once).into_next().unwrap();
				task.state().update(next, State::Prec(self.ord));
			});

			self.prec.insert(guard);

			// test if the mark of the next task exists and is identical
			if let Some(task) = cur.get() {
				let next_order = (*task).brand(|task, once| {
					task.mark(&task.token(once).into_init().unwrap())
						.get()
						.order()
				});

				if next_order == self.top.1.order() {
					guard = cur.remove().unwrap();
					continue;
				}
			}

			// otherwise exit the loop
			break true;
		}
	}
}

/* ****************************************************** Calendar Trait Impl */

/// Type-alias for the default [Scheduler].
pub type DefaultPlan<C> = LayerStack<C>;

impl<C: ?Sized + Config<Plan = Self>> Scheduler for LayerStack<C> {
	type Config = C;
	type State = State<C>;

	#[inline]
	fn new(config: &C) -> Self {
		LayerStack::new(config.default_time())
	}

	#[inline]
	fn schedule<'brand>(
		&mut self,
		task: Irc<Task<'brand, C>>,
		idle: token::Idle<'brand>,
		time: C::Time,
	) -> token::Next<'brand> {
		self.push_time(task, idle, time)
	}

	#[inline]
	fn defer<'brand>(
		&mut self,
		task: Irc<Task<'brand, C>>,
		busy: token::Busy<'brand>,
	) -> token::Next<'brand> {
		self.push_hold(task, busy)
	}

	#[inline]
	fn activate<'brand>(
		&mut self,
		task: Irc<Task<'brand, C>>,
		idle: token::Idle<'brand>,
	) -> token::Next<'brand> {
		let rank = task.branded_share(&idle).rank();
		self.push_rank(task, idle, rank)
	}

	#[inline]
	fn purge<'brand>(
		&mut self,
		task: &Task<'brand, C>,
		next: token::Next<'brand>,
	) -> token::Idle<'brand> {
		self.remove(task, next)
	}

	#[inline]
	fn extract(&mut self) -> Option<Irc<Task<'static, Self::Config>>> {
		Some(self.pop()?.into_inner())
	}

	#[inline]
	fn is_empty(&self) -> bool {
		self.is_empty()
	}

	#[inline]
	fn now(&self) -> C::Time {
		self.top.0
	}
}

/* ************************************************************** Event State */

/// Enumeration over possible payloads of the [Next]-state of a [Task].
///
/// [Next]: state::Next
pub enum State<C: ?Sized + Config> {
	/// State of a task that is scheduled for a later model time.
	Time(Cyclic<u32>, C::Time),
	/// State of a deferred task.
	Hold,
	/// State of a task that is scheduled for the current model time and sorted
	/// according to priority with respect to other tasks scheduled for the
	/// current model time.
	Rank(Cyclic<u32>, C::Rank),
	/// State of a task that is scheduled and associated with the currently
	/// executed process
	Prec(Cyclic<u32>),
}

impl<C: ?Sized + Config> fmt::Debug for State<C> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		match self {
			Self::Time(_, time) => write!(f, "Time({})", time.display()),
			Self::Hold => write!(f, "Hold"),
			Self::Rank(_, rank) => write!(f, "Rank({rank:?})"),
			Self::Prec(_) => write!(f, "Prec"),
		}
	}
}

impl<C: ?Sized + Config> fmt::Display for State<C> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		use crate::config::Time;

		match self {
			Self::Time(_, time) => write!(f, "Time({})", time.display()),
			Self::Hold => write!(f, "Hold"),
			Self::Rank(_, rank) => write!(f, "Rank({rank:?})"),
			Self::Prec(_) => write!(f, "Prec"),
		}
	}
}

/* *************************************************** Event State Trait Impl */

impl<C: ?Sized + Config> PlanState for State<C> {
	type Time = C::Time;
	type Rank = C::Rank;

	#[inline]
	fn time(&self) -> Option<C::Time> {
		match self {
			Self::Time(_, time) => Some(*time),
			_ => None,
		}
	}

	#[inline]
	fn rank(&self) -> Option<C::Rank> {
		match self {
			Self::Rank(_, rank) => Some(rank.clone()),
			_ => None,
		}
	}
}
