//! Provides type-erased and object-safe traits that can be used to dynamically
//! dispatch over [Simulators], regardless of their [configuration].
//!
//! [Simulators]: crate::Simulator
//! [configuration]: Config

use super::{Config, Prec};
use core::{any::Any, fmt, marker::PhantomData, ptr::NonNull, task::RawWaker};

use crate::{
	simulator::SimBox, task, task::erased::State as TaskState, IntrusivelyCounted, Irc, Label,
};

/// Type-erased trait for [Simulation contexts](crate::Sim).
pub trait Sim: Any + IntrusivelyCounted<Inner = SimBox> {
	/// Returns the currently active and config-erased version of a [TaskPuck].
	///
	/// [TaskPuck]: crate::Puck
	fn active(&self) -> Option<TaskPuck>;

	/// Returns the current time as a displayable object.
	fn now(&self) -> &dyn fmt::Display;

	/// Returns the global (shared) data of the simulator.
	fn global(&self) -> &dyn Any;

	/// Constructs a [RawWaker] for the currently active process.
	fn waker(&self) -> RawWaker;

	/// Defers the currently active process.
	fn defer(&self);
}

/// Type-erased trait for [Tasks](task::Task).
pub trait Task: Any + IntrusivelyCounted + fmt::Debug {
	/// Returns a type-erased reference to the shared process data.
	fn subject(&self) -> &dyn Any;

	/// Returns the task's [Label].
	fn label(&self) -> Label;

	/// Returns the task's [Precedence].
	///
	/// [Precedence]: Prec
	fn prec(&self) -> Prec;

	/// Returns the task's [TaskState].
	fn state(&self) -> TaskState;
}

type DynTask = dyn Task<Inner = task::TaskBox>;

/// [Config]-erased type for [TaskPucks](task::Puck).
pub struct TaskPuck(Irc<DynTask>);

impl TaskPuck {
	/// Converts a [Config]-erased [TaskPuck] into a Config-dependent
	/// [TaskPuck](task::Puck) or returns the original object if the
	/// configuration didn't agree with the actual type.
	#[inline]
	pub fn downcast<C: Config>(self) -> Result<task::Puck<C>, Self> {
		match self.0.downcast::<task::Task<'static, C>>() {
			Ok(task) => Ok(task::Puck(task)),
			Err(old) => Err(TaskPuck(old)),
		}
	}

	/// Returns a reference to the instance shared among all the [Fiber]
	/// associated with the same process as this one.
	///
	/// [Fiber]: crate::Fiber
	#[inline]
	pub fn shared(&self) -> &dyn Any {
		self.0.subject()
	}

	/// Returns the name of the underlying [Process](crate::Process).
	#[inline]
	pub fn label(&self) -> Label {
		self.0.label()
	}

	/// Returns the current [Precedence] of the underlying [Fiber].
	///
	/// [Precedence]: Prec
	/// [Fiber]: crate::Fiber
	#[inline]
	pub fn prec(&self) -> Prec {
		self.0.prec()
	}

	/// Returns a copy of this task's [state](TaskState).
	#[inline]
	pub fn state(&self) -> TaskState {
		self.0.state()
	}
}

impl<C: Config> From<task::Puck<C>> for TaskPuck {
	#[inline]
	fn from(task: task::Puck<C>) -> Self {
		TaskPuck(unsafe { Irc::from_raw(Irc::into_raw(task.into()) as NonNull<DynTask>) })
	}
}

impl Clone for TaskPuck {
	#[inline]
	fn clone(&self) -> Self {
		TaskPuck(self.0.clone())
	}
}

impl fmt::Debug for TaskPuck {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		self.0.fmt(f)
	}
}

/// Currently active thread-local and type-erased [Simulator].
///
/// [Simulator]: Sim
use shared::SIMULATOR;

/// Sets the passed [Simulator](crate::base::Simulator)-reference as the new
/// active simulator, returning a [Guard] that resets the active simulator
/// on drop.
#[must_use = "dropping this guard restores the previous thread-local simulator"]
#[inline]
pub(crate) fn hook(sim: &impl Sim) -> impl Drop + '_ {
	Guard::new(sim)
}

/// Executes a closure receiving a type-erased [simulation context].
/// Returns `None` if no simulator is currently active.
///
/// [simulation context]: crate::Sim
#[inline]
pub fn with<R>(f: impl FnOnce(&dyn Sim) -> R) -> Option<R> {
	SIMULATOR.with(|inner| inner.get().map(|sim| f(unsafe { sim.as_ref() })))
}

/// Returns a [TaskPuck] for the currently active task, if this thread is
/// currently running a simulation.
#[inline]
pub fn active() -> Option<TaskPuck> {
	SIMULATOR.with(|inner| inner.get().and_then(|sim| unsafe { sim.as_ref() }.active()))
}

/// Guards the currently active [Simulator] in this thread, restoring the
/// previous state when this guard is dropped.
///
/// Guard-instances form an implicit chain over the stack, allowing nested
/// simulation runs within the same thread.
///
/// [Simulator]: crate::base::Simulator
struct Guard<'s> {
	/// The previous type-erased simulator.
	prev: Option<NonNull<dyn Sim>>,
	/// A marker for the lifetime.
	mark: PhantomData<&'s ()>,
}

impl<'s> Guard<'s> {
	/// Creates a new Guard, pushing it on top of the implicit simulator-stack.
	#[inline]
	fn new(sim: &'s impl Sim) -> Self {
		Guard {
			prev: SIMULATOR.replace(Some(NonNull::from(sim))),
			mark: PhantomData,
		}
	}
}

impl Drop for Guard<'_> {
	#[inline]
	fn drop(&mut self) {
		SIMULATOR.replace(self.prev);
	}
}

/* *************************************************************** Model Time */

/// Helper that allows displaying the model time of the currently active
/// simulator in human-readable format.
pub struct ModelTime<W>(pub W)
where
	W: Fn(&mut dyn fmt::Write, &dyn Sim) -> fmt::Result;

impl<W> fmt::Display for ModelTime<W>
where
	W: Fn(&mut dyn fmt::Write, &dyn Sim) -> fmt::Result,
{
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		with(move |sim| self.0(f, sim)).unwrap_or(Ok(()))
	}
}

/// Constructs a constant [ModelTime] instance that may be used to return the
/// model time of the currently active simulator for this thread.
///
/// The macro accepts a format string and optional, additional arguments as
/// specified for the [format_args]-macro. The format string may reference the
/// `time`-variable.
///
/// This is really meant in combination with the `tracing`-crate, but it is
/// possible to be used independently through the [Display]-trait.
///
/// # Example
/// ```
/// # use odem_rs_core::{Sim, simulation, model_time};
/// # async fn sim_main(sim: &Sim) {}
/// fn main() {
///     tracing_subscriber::fmt()
///         .with_timer(model_time!("[{time:#}]"))
///         .init();
///
///     simulation(Default::default(), sim_main).unwrap();
///}
/// ```
///
/// [Display]: fmt::Display
#[macro_export]
macro_rules! model_time {
	() => { $crate::model_time!("[{time}]") };

	($($arg:tt)*) => {
		$crate::erased::ModelTime(
			move |w,s| ::core::fmt::write(w, ::core::format_args!($($arg)*, time = s.now()))
		)
	};
}

// make ModelTime compatible to the FormatTime-trait from tracing-subscriber
#[cfg(feature = "tracing-subscriber")]
impl<W> tracing_subscriber::fmt::time::FormatTime for ModelTime<W>
where
	W: Fn(&mut dyn fmt::Write, &dyn Sim) -> fmt::Result,
{
	#[inline]
	fn format_time(&self, w: &mut tracing_subscriber::fmt::format::Writer<'_>) -> fmt::Result {
		with(move |sim| self.0(w, sim)).unwrap_or(Ok(()))
	}
}

/* **************************************************** Normalized Model Time */

/// Helper for displaying normalized model-time in digital-clock-format.
#[derive(Copy, Clone)]
pub struct ClockTime<U> {
	/// The value-part of the time.
	value: isize,
	/// Indicator of the unit of time.
	_unit: PhantomData<U>,
}

impl ClockTime<()> {
	/// Constructs a [displayable](fmt::Display) clock time in seconds.
	#[inline]
	pub const fn seconds(value: isize) -> impl fmt::Display {
		ClockTime::<Seconds> {
			value,
			_unit: PhantomData,
		}
	}

	/// Constructs a [displayable](fmt::Display) clock time in minutes.
	#[inline]
	pub const fn minutes(value: isize) -> impl fmt::Display {
		ClockTime::<Minutes> {
			value,
			_unit: PhantomData,
		}
	}

	/// Constructs a [displayable](fmt::Display) clock time in hours.
	#[inline]
	pub const fn hours(value: isize) -> impl fmt::Display {
		ClockTime::<Hours> {
			value,
			_unit: PhantomData,
		}
	}

	/// Constructs a [displayable](fmt::Display) clock time in days.
	#[inline]
	pub const fn days(value: isize) -> impl fmt::Display {
		ClockTime::<Days> {
			value,
			_unit: PhantomData,
		}
	}
}

impl<U> ClockTime<U> {
	/// Private method that prints the time's sign.
	#[inline]
	fn write_sign(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		if self.value < 0 {
			write!(f, "-")
		} else if f.sign_plus() {
			write!(f, "+")
		} else {
			Ok(())
		}
	}
}

impl fmt::Display for ClockTime<Days> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		self.write_sign(f)?;
		Days(self.value.unsigned_abs()).fmt(f)
	}
}

impl fmt::Display for ClockTime<Hours> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		self.write_sign(f)?;
		Hours(self.value.unsigned_abs()).fmt(f)
	}
}

impl fmt::Display for ClockTime<Minutes> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		self.write_sign(f)?;
		Minutes(self.value.unsigned_abs()).fmt(f)
	}
}

impl fmt::Display for ClockTime<Seconds> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		self.write_sign(f)?;
		Seconds(self.value.unsigned_abs()).fmt(f)
	}
}

struct Seconds(usize);
struct Minutes(usize);
struct Hours(usize);
struct Days(usize);

impl fmt::Display for Days {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		self.0.fmt(f).and_then(|_| f.write_str("d"))
	}
}

impl fmt::Display for Hours {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		let mut quantity = self.0;

		if quantity >= 24 || f.precision().map(|w| w > 3).unwrap_or(false) {
			Days(quantity / 24).fmt(f)?;
			quantity %= 24;
			f.write_str(" ")?;
		}

		write!(f, "{:02}h", quantity)
	}
}

impl fmt::Display for Minutes {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		let mut quantity = self.0;

		if quantity >= 60 || f.precision().map(|w| w > 2).unwrap_or(false) {
			Hours(quantity / 60).fmt(f)?;
			quantity %= 60;
			f.write_str(" ")?;
		}

		write!(f, "{:02}m", quantity)
	}
}

impl fmt::Display for Seconds {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		let mut quantity = self.0;

		if quantity >= 60 || f.precision().map(|w| w > 1).unwrap_or(false) {
			Minutes(quantity / 60).fmt(f)?;
			quantity %= 60;
			f.write_str(" ")?;
		}

		write!(f, "{:02}s", quantity)
	}
}

/* ****************************************** Shared Simulator Helper Modules */

#[cfg(feature = "std")]
mod shared {
	use super::Sim;
	use std::{cell::Cell, ptr::NonNull};

	thread_local! {
		pub static SIMULATOR: Cell<Option<NonNull<dyn Sim>>> = const { Cell::new(None) };
	}
}

#[cfg(not(feature = "std"))]
mod shared {
	use super::Sim;
	use core::{
		cell::Cell,
		ptr::NonNull,
		sync::atomic::{AtomicBool, Ordering},
	};

	pub static SIMULATOR: LocalKey = LocalKey::new();

	pub struct LocalKey {
		sim: Cell<Option<NonNull<dyn Sim>>>,
		lock: AtomicBool,
	}

	impl LocalKey {
		const fn new() -> Self {
			Self {
				sim: Cell::new(None),
				lock: AtomicBool::new(true),
			}
		}

		pub fn with<R>(&self, f: impl FnOnce(&Cell<Option<NonNull<dyn Sim>>>) -> R) -> R {
			assert!(
				self.lock.swap(false, Ordering::SeqCst),
				"detected multiple threads running simulators"
			);

			let res = f(&self.sim);

			self.lock.store(true, Ordering::SeqCst);
			return res;
		}

		pub fn replace(&self, sim: Option<NonNull<dyn Sim>>) -> Option<NonNull<dyn Sim>> {
			assert!(
				self.lock.swap(false, Ordering::SeqCst),
				"detected multiple threads running simulators"
			);

			let res = self.sim.replace(sim);

			self.lock.store(true, Ordering::SeqCst);
			return res;
		}
	}

	unsafe impl Sync for LocalKey {}
}
