//! A module for core-specific error types.

pub use crate::task::error::*;

/* ************************************************************** Error Types */

/// Enumeration of errors for the functions and structures contained in this
/// module of the simulation library.
#[derive(Copy, Clone, Debug, thiserror::Error)]
pub enum Error {
	/// The simulation stopped when the simulator detected a deadlock.
	#[error(transparent)]
	Deadlock(#[from] Deadlock),
	/// The simulation was stopped early, before completion.
	#[error(transparent)]
	Incomplete(#[from] Incomplete),
	/// Task is already bound.
	#[error(transparent)]
	NotBorn(#[from] NotBorn),
	/// Task is not idle.
	#[error(transparent)]
	NotIdle(#[from] NotIdle),
	/// Task is not active.
	#[error(transparent)]
	NotBusy(#[from] NotBusy),
	/// Task is not scheduled.
	#[error(transparent)]
	NotNext(#[from] NotNext),
	/// Task hasn't completed yet.
	#[error(transparent)]
	NotDone(#[from] NotDone),
	/// Task hasn't been released yet.
	#[error(transparent)]
	NotGone(#[from] NotGone),
	/// Task hasn't been bound yet.
	#[error(transparent)]
	NotInit(#[from] NotInit),
	/// Task hasn't terminated.
	#[error(transparent)]
	NotTerm(#[from] NotTerm),
	/// Signals an attempt to schedule a task at an earlier model-time.
	#[error("causality error")]
	Causality,
}

/* ******************************************************** Minor Error Types */

/// Signals the detection of a deadlock during a simulation run.
#[derive(Copy, Clone, Debug, thiserror::Error)]
#[error("deadlock detected")]
pub struct Deadlock;

/// Indicates that a simulation run was stopped early, before completion or
/// running into a [Deadlock].
#[derive(Copy, Clone, Debug, thiserror::Error)]
#[error("incomplete simulation run")]
pub struct Incomplete;

/// Signals that an event at the current model time caused a change in that
/// event's past, violating causality.
#[derive(Copy, Clone, Debug, thiserror::Error)]
#[error(
	"task was scheduled at model-time {cause:?} (cause) for model-time \
         {effect:?} (effect), violating causality"
)]
pub struct Causality<T> {
	pub cause: T,
	pub effect: T,
}

impl<T> From<Causality<T>> for Error {
	#[inline]
	fn from(_: Causality<T>) -> Self {
		Error::Causality
	}
}
