//! The `fiber` module supplies integrated threads of execution, known as
//! [`Fiber`], within the context of the simulation library.
//!
//! # Overview
//!
//! Fibers represent the combination of a [Task] and a [Future], allowing them
//! to share state with other threads in the same [Process]. These fibers can
//! reference shared state and each other, provided that the shared state
//! outlives them. Unlike tasks, which only support simple *success*/*failure*
//! return values, fibers record their actual return value. This enables their
//! creator to retrieve it upon termination.
//!
//! Fibers may also be terminated prematurely, removing them from the simulation
//! calendar, releasing resources, and recording their result as a *failure* for
//! any interested parties.
//!
//! ## Key Components
//!
//! - **Fiber:** The main struct representing a fiber, combining a type-erased
//!   handle to a future (`Task`) and a return-type-erased future instance
//!   (`FiberState`). It supports operations like activation, termination, and
//!   result extraction.
//!
//! - **FiberState:** Internal helper type erasing the return type of future.
//!   It contains a union of future and result instances, allowing for the safe
//!   erasure of the return type while preserving the ability to extract the
//!   result upon termination.
//!
//! - **Builder:** A builder type aiding in the configuration of fibers with
//!   detailed specifications, such as the future lifecycle, precedence, source
//!   code location, and a finalization routine.
//!
//! # Examples
//!
//! Creating and activating a simple fiber in a simulation context:
//!
//! ```
//! # use odem_rs_core::{Sim, Fiber, unique};
//! # use core::pin::pin;
//! async fn sim_main(sim: &Sim) {
//!     // create and initialize a local variable
//!     let c = 3;
//!
//!     // create a fiber referring to it and the simulation context
//!     let fib = unique!(Fiber::new(async {
//!         for i in 0..c {
//!             println!("Hello {i}");
//!             sim.advance(1.0).await;
//!         }
//!     }));
//!
//!     // activate the fiber
//!     sim.activate(fib);
//!
//!     // pass time
//!     sim.advance((c + 1) as f64).await;
//! }
//! ```
//!
//! # Memory-Safety and Drop
//! Some memory-safety guarantees related to accessing process-shared data
//! within the lifecycle of a Fiber hinge on the assumption that a Fiber's
//! [Drop] is called before the creator-[Process] terminates.
//!
//! However, Rust demands that memory-safety not be violated even if no
//! destructors run at all. The reasoning is, that it is virtually impossible
//! to statically prevent memory leaks from occuring even in purely safe code,
//! e.g. by constructing circular dependencies using [Rc].
//!
//! The library has been hardened towards this type of bug, mainly by inserting
//! additional checks and checksums. We distinguish two different types of leak:
//! 1. [forgetting] to call [drop], or
//! 2. [leaking] Fiber from within the Process using the heap.
//!
//! ## [`forget`]-type leaks
//! Forgetting to call a Fiber's destructor crucially requires *moving* it into
//! the forget-function.
//! We divide the situations for potential memory-unsafety into the following
//! categories:
//! 1. after the construction of the fiber, before its activation,
//! 2. after its activation, before its termination,
//! 3. after its termination.
//!
//! Possible dangling pointers that lead to memory unsafety upon dereferencing
//! are found in the Calendar if the Fiber is scheduled or active, and in the
//! [Waker] and [TaskPuck]-objects distributed to other parts of the simulation.
//!
//! ### Case 1
//! The first case is harmless because the Fiber isn't scheduled yet and no
//! TaskPuck exist, since these are derived from [Puck]-handles that can only
//! be created as the result of Task activation.
//!
//! ### Case 2
//! Activating a Fiber requires constructing a [FiberPuck] that requires a
//! commitment to [pinning] the Fiber which prevents *Moving* it afterward.
//! This guarantee is independent of the lifetime of the Puck.
//!
//! In case a FiberPuck is forgotten, no memory unsafety may occur because
//! the Pin-guarantees prevent the referenced Fiber from being moved during the
//! entirety of its life which includes moving it into the forget-function.
//!
//! We can therefore conclude that the destructor of the Fiber runs, which will
//! clean up all the dangling references or panic if that isn't feasible,
//! leading only to well-defined program behavior.
//!
//! ### Case 3
//! In this case, no dangling pointers can be found in the Calendar since the
//! Fiber is no longer scheduled. Furthermore, [pinning] guarantees prevent
//! moving the Fiber even after its termination and before the destructor has
//! completed.
//!
//! ## [`Box::leak`]-type leaks
//! The problem here may occur if the Fiber outlives the Process that spawned
//! it, because the Fiber contains references to data owned by the Process.
//! Memory-safety is secured by reference-counting the number of active
//! references to the Process and is checked in the drop-implementation of
//! the Process. Detecting active Fiber during the drop of the root-Fiber leads
//! to a panic and could only lead to memory unsafety if the Process was never
//! destructed, which is prevented in the Process implementation.
//!
//! [forgetting]: core::mem::forget
//! [`forget`]: core::mem::forget
//! [leaking]: Box::leak
//! [pinning]: core::pin
//! [drop]: Drop::drop
//! [Process]: crate::Process
//! [Waker]: core::task::Waker
//! [FiberPuck]: Puck
//! [Puck]: crate::Puck
//! [Rc]: std::rc::Rc

use core::{
	any::Any,
	cell::RefCell,
	fmt,
	future::{Future, IntoFuture, Pending},
	marker::PhantomData,
	mem::ManuallyDrop,
	panic::Location,
	pin::Pin,
	ptr::NonNull,
	task::{Context, Poll},
};

use crate::{
	config::DefaultConfig,
	error,
	fsm::*,
	ptr::{IntrusivelyCounted, Irc, IrcBox, IrcBoxed},
	task::{erased::State as TaskState, token, Share, Task},
	Active, Config, Dispatch, ExitStatus, Label, Prec, Sim, Unique,
};

/* ******************************************************************** Fiber */

/// Light-weight thread of execution that shares state with other threads
/// within the same [Process].
///
/// Fibers are fundamentally ([Task], [Future]) pairs that can reference shared
/// state and each other, as long as that state outlives them. Since tasks don't
/// support return values other than a simple *success*/*failure*, fiber erase
/// their actual return value but record it, allowing their parent to retrieve
/// it upon their termination.
///
/// Fiber may also be terminated before completion, which will remove them from
/// the calendar, release all of their resources and record their result as
/// *failure* to any interested party.
///
/// # Examples
/// As stated, Fiber are light-weight and can easily refer to the variables
/// in their lexical scope.
/// ```
/// # use odem_rs_core::{Sim, Fiber, unique};
/// async fn sim_main(sim: &Sim) {
///     // create and initialize a local variable
///     let c = 3;
///     
///     // create a fiber referring to it and the simulation context
///     let fib = unique!(Fiber::new(async {
///         for i in 0..c {
///             println!("Hello {i}");
///             sim.advance(1.0).await;
///         }
///     }));
///     
///     // activate the fiber
///     sim.activate(fib);
///     
///     // pass time
///     sim.advance((c + 1) as f64).await;
/// }
/// ```
///
/// [Process]: crate::Process
#[pin_project::pin_project(PinnedDrop, !Unpin)]
pub struct Fiber<'brand, C: ?Sized + Config, F: Future, T = Unchecked> {
	/// A type-erased handle to the future contained within this fiber.
	#[pin]
	task: Task<'brand, C>,
	/// The return-type-erased future instance representing the inner state
	/// of this fiber.
	#[pin]
	state: RefCell<FiberState<F, T>>,
}

impl<C: ?Sized + Config, F: Future> Fiber<'static, C, F> {
	/// Creates a new fiber with a sequence of actions, expressed through a
	/// future.
	#[track_caller]
	pub fn new(actions: F) -> Self {
		Fiber::build().with_actions(actions).finish()
	}
}

impl Fiber<'static, DefaultConfig, Pending<()>> {
	/// Creates and returns a [Builder] for fibers, that is useful if complex
	/// initialization is desired.
	pub const fn build() -> Builder {
		Builder::new()
	}
}

impl<'brand, C: ?Sized + Config, F: Future, T> Fiber<'brand, C, F, T> {
	/// Sets the precedence of the fiber.
	#[inline]
	pub fn set_prec(&mut self, prec: Prec) {
		self.task.set_prec(prec);
	}

	/// Returns a copy of the fiber's [task state](TaskState).
	#[inline]
	pub fn state(&self) -> TaskState {
		self.task.state().borrow().erased()
	}

	/// Extracts the return value if the fiber is terminated.
	#[inline]
	pub fn result(&self) -> Result<F::Output, error::NotDone> {
		self.brand(|fiber, once| Ok(fiber.inner_result(fiber.token(once).into_done()?).1))
	}

	/// Aborts the execution of this fiber, removing it from the calendar,
	/// dropping the inner future and the return value if applicable.
	#[inline]
	pub fn abort(&self) {
		self.brand(|fiber, once| {
			fiber.inner_abort(fiber.token(once));
		});
	}

	/// Converts the specific brand into a generic brand, breaking the
	/// connection with the equally branded token.
	#[inline]
	pub fn detach(&self) -> &Fiber<'static, C, F, T> {
		unsafe { core::mem::transmute(self) }
	}

	/// Returns an enumeration copy of the internal [State](token::State).
	#[inline]
	pub fn token(&self, once: Once<'brand>) -> token::State<'brand> {
		self.task.token(once)
	}

	/// Returns an exclusive reference to the finalizer if in state `Born`.
	#[inline]
	pub fn finalizer(&mut self, born: &token::Born<'brand>) -> &mut T {
		let _ = born;

		unsafe { &mut self.state.get_mut().future.1 }
	}

	/// Aborts the execution of this fiber, removing it from the calendar,
	/// dropping the inner future and the return value if applicable.
	fn inner_abort(&self, state: token::State<'brand>) -> token::Gone<'brand> {
		use token::State::*;

		// drop impls should not be able to perform further task transitions
		// but we better check to be sure
		let task = &self.task;

		// set the correct span
		let _span = state
			.as_init()
			.map(|init| task.branded_share(init).enter_span());

		// analyze the state
		let once: Once<'_> = match state {
			// the future has already been dropped -> quit
			Gone(gone) => return gone,
			// the return value is pending -> drop it and quit
			Done(done) => return self.inner_result(done).0,
			// the fiber is scheduled -> remove it from the calendar
			Next(next) => task.deschedule(next).into(),
			// the fiber is active -> remove it from the active slot
			Busy(busy) => task.deactivate(busy).into(),
			// the fiber is born or idle -> drop the future and transition
			state @ (Born(_) | Idle(_)) => state.into(),
		};

		// drop the future before performing extra transitions to prevent
		// undefined behavior in user-defined drop impls that rely on the
		// current task state
		let (once, ()) = self.debrand(once, |fiber| unsafe {
			fiber.state.borrow_mut().unchecked_drop();
		});

		// perform the final transition into the `Gone`-state
		let gone = match task.token(once) {
			Born(born) => task.state().transition(born, Err(crate::Failure)),
			Idle(idle) => task.state().transition(idle, Err(crate::Failure)),
			state => unreachable!(
				"expected task to be 'Born' or 'Idle' but it is \
				 in state '{:?}' instead",
				state.erased()
			),
		};

		// now reactivate pending tasks
		task.wake_pending();

		// return the corresponding token to the caller
		gone
	}

	/// Extracts the return value of a terminated fiber.
	#[inline]
	fn inner_result(&self, done: token::Done<'brand>) -> (token::Gone<'brand>, F::Output) {
		let task = &self.task;
		let rc = task.branded_result(done.as_term());

		// set the correct span
		let _span = task.branded_share(&done).enter_span();

		// SAFETY: we know that the result has been written and not been read
		// in state 'Done'
		unsafe {
			(
				task.state().transition(done, rc),
				self.state
					.borrow_mut()
					.unchecked_result()
					.unwrap_unchecked(),
			)
		}
	}

	/// Binds shared data to an unbound fiber.
	///
	/// # Safety
	/// The caller has to ensure that the shared-data-reference outlives the
	/// (active) part of the tasks' life.
	pub(crate) unsafe fn bind<'a>(
		self: Pin<&'a mut Self>,
		born: token::Born<'brand>,
		vtab: NonNull<dyn Dispatch + 'a>,
		share: NonNull<Share<C>>,
	) -> token::Idle<'brand> {
		self.project().task.bind(born, vtab, share)
	}
}

impl<'p, C: ?Sized + Config, F: Future, T> Unique<'p, Fiber<'_, C, F, T>> {
	/// Used during startup of a simulation run to initialize the root fiber.
	#[inline]
	pub(crate) fn boot(self, share: Pin<&'p Share<C>>) -> Puck<'p, C, F, T>
	where
		T: Settle<F::Output>,
	{
		self.brand(move |fiber, once| {
			let born = fiber.token(once).into_born().unwrap();
			let hook = fiber.as_ptr();
			let share = NonNull::from(share.get_ref());

			// SAFETY: The tagged fiber was originally unique, pinned, and in
			// state 'Born'. The shared data lives at least `'p` long, so
			// creating the puck is safe.
			unsafe { Puck::new(fiber, born, hook, share) }
		})
	}
}

impl<'b, C: ?Sized + Config, F: Future, T> Stateful for Fiber<'b, C, F, T> {
	type Brand = &'b ();

	#[inline]
	unsafe fn enter(&self) {
		self.task.enter();
	}

	#[inline]
	unsafe fn leave(&self) {
		self.task.leave();
	}
}

impl<'b, C: ?Sized + Config, F: Future, T> Rebrand<'b> for Fiber<'b, C, F, T> {
	type Kind<'a> = Fiber<'a, C, F, T>;
}

impl<C: ?Sized + Config, F: Future, T> fmt::Debug for Fiber<'_, C, F, T> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		let mut s = f.debug_struct("Fiber");
		s.field("task", &self.task)
			.field("future", &core::any::type_name::<F>())
			.finish()
	}
}

impl<C: ?Sized + Config, F: Future, T: Settle<F::Output>> Dispatch for Fiber<'_, C, F, T> {
	fn poll(self: Pin<&Self>, cx: &mut Context<'_>) -> Poll<ExitStatus> {
		assert!(self.state().is_busy(), "only active tasks can be polled");

		let mut state = self.state.borrow_mut();

		// SAFETY: self is pinned and access to future is defined in state Busy
		unsafe { Pin::new_unchecked(&mut *state.future.0) }
			.poll(cx)
			.map(move |rv| {
				// SAFETY: the future just terminated
				unsafe { state.unchecked_terminate(rv) }
			})
	}
}

impl<C: ?Sized + Config, F: Future, T: Settle<F::Output>> Active<C> for Fiber<'_, C, F, T> {
	type Puck<'p>
		= Puck<'p, C, F, T>
	where
		Self: 'p;

	#[inline]
	unsafe fn bind<'p>(
		this: Unique<'p, Self>,
		vtab: NonNull<dyn Dispatch + 'p>,
		sctx: NonNull<Share<C>>,
	) -> Result<Self::Puck<'p>, error::NotBorn> {
		this.brand(move |fiber, once| {
			// ensure that the fiber hasn't been bound already
			let born = fiber.token(once).into_born()?;

			// SAFETY: the shared data of the root-fiber outlives the
			// newly bound fiber, because fibers have to terminate
			// before their process does, which is enforced by the
			// borrow-checker
			Ok(unsafe { Puck::new(fiber, born, vtab, sctx) })
		})
	}
}

unsafe impl<'brand, C: ?Sized + Config, F: Future, T> IntrusivelyCounted
	for Fiber<'brand, C, F, T>
{
	type Inner = <Task<'brand, C> as IntrusivelyCounted>::Inner;

	#[inline]
	fn irc_box(&self) -> &IrcBox<Self::Inner> {
		self.task.irc_box()
	}
}

impl<'brand, C: ?Sized + Config, F: Future, T> AsRef<Task<'brand, C>> for Fiber<'brand, C, F, T> {
	#[inline]
	fn as_ref(&self) -> &Task<'brand, C> {
		&self.task
	}
}

#[pin_project::pinned_drop]
impl<C: ?Sized + Config, F: Future, T> PinnedDrop for Fiber<'_, C, F, T> {
	fn drop(self: Pin<&mut Self>) {
		// acquire ownership to the fiber in order to prevent reclaim from
		// being called
		unsafe {
			self.irc_box().inner.acquire();
		}

		// abort the execution
		self.abort();

		// relinquish ownership without running reclaim
		unsafe {
			self.irc_box().inner.release();
		}
	}
}

/* *************************************************************** Fiber Puck */

/// An exclusive reference to a pinned [Fiber].
///
/// The underlying [Task] does not expire at the end of its scope but is
/// continued until it either terminates normally or its [Drop] destructor is
/// executed. This is safe because pinning guarantees that the fiber will
/// not be moved before it is dropped.
pub struct Puck<'p, C: ?Sized + Config, F: Future, T = Unchecked>(
	Irc<Fiber<'static, C, F, T>>,
	PhantomDrop<Pin<&'p mut Fiber<'static, C, F, T>>>,
);

impl<'p, C: ?Sized + Config, F: Future, T> Puck<'p, C, F, T> {
	/// Creates a new [Puck] from an exclusively referenced and pinned [Fiber]
	/// in state [Born](TaskState::Born).
	///
	/// # Safety
	/// The caller has to ensure that both the virtual [Dispatch] table and
	/// the reference to the [Share] stay valid for the duration of the fiber's
	/// life.
	#[inline]
	unsafe fn new<'b>(
		mut fiber: Guard<Unique<'p, Fiber<'b, C, F, T>>>,
		born: token::Born<'b>,
		vtab: NonNull<dyn Dispatch + 'p>,
		share: NonNull<Share<C>>,
	) -> Self {
		// bind the shared process data to it
		let idle = fiber.as_mut().bind(born, vtab, share);

		// initialize the pinned fiber reference
		Puck(
			Irc::map(Irc::from(Guard::into_inner(fiber, idle)), Fiber::detach),
			PhantomData,
		)
	}
	
	/// Returns a reference to the [shared context](Share) bound to this fiber.
	pub fn share(&self) -> &Share<C> {
		// SAFETY: Pucks can only be created during initialization, after
		// which they have shared data bound to them
		unsafe {
			self.0.task.share().unwrap_unchecked()
		}
	}
	
	/// Sets the rank of this fiber.
	///
	/// Note that the rank only becomes effective during the next selected
	/// model time. You can force a rescheduling by [deferring] the
	/// corresponding fiber.
	///
	/// [deferring]: crate::ops::defer
	#[inline]
	pub fn set_rank(&self, rank: C::Rank) {
		self.0.task.share().unwrap().set_rank(rank);
	}

	/// Terminates the referenced fiber prematurely.
	///
	/// This method consumes the reference to prevent it from being called twice
	/// or used to extract the return value later.
	#[inline]
	pub fn abort(self) {
		self.0.abort();
	}
}

impl<C: ?Sized + Config, F: Future, T: Settle<F::Output>> fmt::Debug for Puck<'_, C, F, T> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		self.0.fmt(f)
	}
}

impl<C: ?Sized + Config, F: Future, T: Settle<F::Output>> crate::Puck<C> for Puck<'_, C, F, T> {
	fn result(&mut self) -> Option<Self::Output> {
		self.0.result().ok()
	}

	fn wake(&mut self) -> Result<(), error::NotIdle> {
		Task::wake(Irc::map(self.0.clone(), |inner| &inner.task))
	}

	fn subject(&self) -> &dyn Any {
		self.share().item()
	}

	fn sim(&self) -> &Sim<C> {
		self.share().sim()
	}

	fn label(&self) -> Label {
		self.share().label()
	}

	fn time(&self) -> Option<C::Time> {
		self.0.task.time()
	}

	fn rank(&self) -> C::Rank {
		self.share().rank()
	}

	fn prec(&self) -> Prec {
		self.0.task.prec()
	}

	fn state(&self) -> TaskState {
		self.0.task.state().borrow().erased()
	}
	
	fn own(&self) -> Irc<Task<'static, C>> {
		Irc::map(self.0.clone(), |inner| &inner.task)
	}
}

impl<C: ?Sized + Config, F: Future, T: Settle<F::Output>> From<Puck<'_, C, F, T>>
	for Irc<Fiber<'static, C, F, T>>
{
	fn from(value: Puck<'_, C, F, T>) -> Self {
		value.0
	}
}

impl<C: ?Sized + Config, F: Future, T: Settle<F::Output>> AsRef<Task<'static, C>>
	for Puck<'_, C, F, T>
{
	fn as_ref(&self) -> &Task<'static, C> {
		&self.0.task
	}
}

impl<C: ?Sized + Config, F: Future, T: Settle<F::Output>> IntoFuture for Puck<'_, C, F, T> {
	type Output = F::Output;
	type IntoFuture = crate::ops::Join<C, Self>;

	#[inline]
	fn into_future(self) -> Self::IntoFuture {
		crate::ops::join(self)
	}
}

/* **************************************** Return-Type-Erased Future Adapter */

/// Internal helper type erasing the return type of future.
///
/// This generic type is used to implement the [Future] trait with an
/// [Output](Future::Output) of `Result<(),()>` but allow extraction of the
/// result using methods.
///
/// The type is implemented as a union to potentially save some space, because
/// either the future is still pending (and there is no result yet) or it is
/// ready (and there is no future anymore).
union FiberState<F: Future, T> {
	/// A future and finalization routine.
	///
	/// The future still contains a user-defined return type but the
	/// [Settle] routine knows how to convert it into a result of type
	/// [ExitStatus]. The pair is used to erase the return-type of the future.
	future: (ManuallyDrop<F>, ManuallyDrop<T>),
	/// The result of the future stored after its termination.
	result: ManuallyDrop<Option<F::Output>>,
}

impl<F: Future, T> FiberState<F, T> {
	/// Initializes a new return-type-erased instance from a future returning
	/// a `Result`.
	#[inline]
	const fn new(actions: F, settle: T) -> Self {
		FiberState {
			future: (ManuallyDrop::new(actions), ManuallyDrop::new(settle)),
		}
	}

	/// Unconditionally (and unsafely) drops the inner future.
	///
	/// # Safety
	/// The caller is responsible for calling this method not more than once
	/// and only if the future hasn't terminated.
	#[inline]
	unsafe fn unchecked_drop(&mut self) {
		ManuallyDrop::drop(&mut self.future.0);
		ManuallyDrop::drop(&mut self.future.1);
	}

	/// Unconditionally (and unsafely) returns the result of the inner future.
	///
	/// # Safety
	/// The caller is responsible for calling this method only after the future
	/// has terminated, but it is allowed to be called multiple times.
	#[inline]
	unsafe fn unchecked_result(&mut self) -> Option<F::Output> {
		self.result.take()
	}

	/// Unconditionally (and unsafely) drops the inner future and writes out
	/// the result.
	///
	/// # Safety
	/// The caller is responsible for only calling this method once at the end
	/// of the future's life.
	#[inline]
	unsafe fn unchecked_terminate(&mut self, result: F::Output) -> ExitStatus
	where
		T: Settle<F::Output>,
	{
		// drop-guard to protect against inconsistent states due to panic-caused
		// stack-unwinding
		struct DropGuard<R> {
			slot: *mut ManuallyDrop<Option<R>>,
			result: Option<R>,
		}

		impl<R> Drop for DropGuard<R> {
			#[inline]
			fn drop(&mut self) {
				// perform the assignment as promised
				// SAFETY: the raw pointer is valid per construction
				unsafe {
					self.slot.write(ManuallyDrop::new(self.result.take()));
				}
			}
		}

		// ensure that the result is assigned before this function is left,
		// even if drop or finalize panic
		let guarded_result = DropGuard {
			// SAFETY: self.result is uninitialized, so avoid creating a mutable
			// reference to it
			slot: core::ptr::addr_of_mut!(self.result),
			result: Some(result),
		};

		// move the finalizer out of the state before calling any functions
		// that may panic in order to prevent resource leaks
		let term = ManuallyDrop::take(&mut self.future.1);

		// carefully drop the pinned future without moving it
		ManuallyDrop::drop(&mut self.future.0);

		// use the finalizer to convert the result
		term.settle(guarded_result.result.as_ref().unwrap_unchecked())
	}
}

/* ************************************************ Canonical Future Adapters */

/// A trait for converting the result of a [Fiber] into an [ExitStatus].
///
/// The trait has a blanket implementation for [`FnOnce`]-closures but can
/// also be further specialized for user-defined types if needed.
pub trait Settle<R> {
	/// Method hook that is called once a fiber completes with a return value.
	/// The return value of this method indicates whether the result is counted
	/// as success (`Ok`) or failure (`Err`).
	fn settle(self, result: &R) -> ExitStatus;
}

// blanket implementation for closures
impl<R, F> Settle<R> for F
where
	F: FnOnce(&R) -> ExitStatus,
{
	#[inline(always)]
	fn settle(self, result: &R) -> ExitStatus {
		self(result)
	}
}

/// Finalization adapter that converts `Poll::Ready` values into the `Ok`
/// variant of a `Result<Success,Failure>`.
pub struct Unchecked;

impl<R> Settle<R> for Unchecked {
	#[inline(always)]
	fn settle(self, _result: &R) -> ExitStatus {
		Ok(crate::Success)
	}
}

/// Finalization adapter that converts `Poll::Ready` values into an
/// [`ExitStatus`] for `bool`, [`Option`], and [`Result`], using the
/// `Ok`-variant for `true`, `Some`, and `Ok` respectively, and the `Err`
/// variant otherwise.
pub struct Checked;

impl Settle<bool> for Checked {
	#[inline(always)]
	fn settle(self, result: &bool) -> ExitStatus {
		if *result {
			Ok(crate::Success)
		} else {
			Err(crate::Failure)
		}
	}
}

impl<R, T> Settle<Result<R, T>> for Checked {
	#[inline(always)]
	fn settle(self, result: &Result<R, T>) -> ExitStatus {
		if result.is_ok() {
			Ok(crate::Success)
		} else {
			Err(crate::Failure)
		}
	}
}

impl<T> Settle<Option<T>> for Checked {
	#[inline(always)]
	fn settle(self, result: &Option<T>) -> ExitStatus {
		if result.is_some() {
			Ok(crate::Success)
		} else {
			Err(crate::Failure)
		}
	}
}

/* ************************************************************ Fiber Builder */

/// Builder-type helping to configure [Fiber] in more detail.
///
/// # Examples
/// ```
/// # use odem_rs_core::{Sim, Prec, Fiber, unique};
/// # async fn sim_main(sim: &Sim) {
///     # let fiber =
/// Fiber::build()                     // create builder
///     .with_actions(async { true })  // specify lifecycle
///     .with_precedence(Prec::new())  // initialize precedence
///     .checked()                     // store result as exit code
///     .finish()                      // instantiate fiber
///     # ;
///     # sim.activate(unique!(fiber));
/// # }
/// ```
pub struct Builder<F = (), S = Unchecked> {
	future: F,
	location: &'static Location<'static>,
	finalizer: S,
	precedence: Prec,
}

impl Builder {
	/// Constructs a new (default) Builder.
	const fn new() -> Self {
		Builder {
			future: (),
			location: Location::caller(),
			finalizer: Unchecked,
			precedence: Prec::new(),
		}
	}
}

impl<F, S> Builder<F, S> {
	/// Sets a future for the lifecycle of the [Fiber].
	#[track_caller]
	pub fn with_actions<X: IntoFuture>(self, actions: X) -> Builder<X, S> {
		Builder {
			future: actions,
			location: Location::caller(),
			finalizer: self.finalizer,
			precedence: self.precedence,
		}
	}

	/// Sets the [Precedence](Prec) of the [Fiber].
	pub fn with_precedence(self, precedence: impl Into<Prec>) -> Self {
		Builder {
			precedence: precedence.into(),
			..self
		}
	}

	/// Sets an explicit source-code [Location] for the [Fiber].
	pub fn with_location(self, location: &'static Location<'static>) -> Self {
		Builder { location, ..self }
	}

	/// Sets a user-defined [Settle] that is run once the lifecycle of the
	/// [Fiber] terminates.
	pub fn with_finalizer<X>(self, finalizer: X) -> Builder<F, X> {
		Builder {
			future: self.future,
			location: self.location,
			finalizer,
			precedence: self.precedence,
		}
	}

	/// Sets a specific [Settle] routine that is used to convert a `bool`,
	/// [`Option`], or [`Result`] return type into an [ExitStatus] that may
	/// be observed by other processes or fiber.
	pub fn checked(self) -> Builder<F, Checked> {
		self.with_finalizer(Checked)
	}
}

impl<F: IntoFuture, S: Settle<F::Output>> Builder<F, S> {
	/// Finalizes the construction of the [Fiber] and returns it.
	pub fn finish<C: ?Sized + Config>(self) -> Fiber<'static, C, F::IntoFuture, S> {
		Fiber {
			task: Task::new(self.precedence, self.location),
			state: RefCell::new(FiberState::new(self.future.into_future(), self.finalizer)),
		}
	}
}

/* ******************************************************* PhantomDrop helper */

/// This type-alias combines [PhantomData] with a Drop-check, ensuring that
/// even phantom types that don't implement the Drop traits themselves present
/// like they would. This is necessary to prevent code like in the following
/// example to compile which would lead to a runtime error due to increased
/// reference counts during drop.
///
/// ```compile_fail,E0716
/// use odem_rs_core::*;
///
/// async fn sim_main(sim: &Sim) {
///     let puck = sim.activate(unique!(Fiber::new(async {})));
/// }
/// ```
type PhantomDrop<T> = PhantomData<PhantomDropInner<T>>;

/// Newtype wrapping a `T` and explicitly implementing the [Drop]-trait.
struct PhantomDropInner<T: ?Sized>(T);

impl<T: ?Sized> Drop for PhantomDropInner<T> {
	#[inline]
	fn drop(&mut self) {}
}
