//! Module for basic scheduling operations that can be called as free functions.
//!
//! These operations are useful for the definition of synchronization
//! structures.

use crate::{
	config::Config,
	ptr::Irc,
	simulator::Sim,
	task::Task,
	Puck,
};

use core::{
	future::{poll_fn, Future},
	pin::Pin,
	task::{Context, Poll},
};

/// Returns a [Future] that can be awaited in order to reconstruct the currently
/// active [simulation context] from the [Context] provided implicitly by the
/// future.
///
/// This function has to be awaited for technical reasons but doesn't actually
/// cause any scheduling operations to occur.
/// 
/// [simulation context]: Sim
pub fn sim<C: Config>() -> impl Future<Output = Option<Irc<Sim<C>>>> {
	poll_fn(|cx| Poll::Ready(Sim::from_context(cx)))
}

/// Suspends the currently active puck until rescheduled by some other puck.
pub fn sleep() -> impl Future<Output = ()> {
	let mut ready = false;
	poll_fn(move |_cx| {
		if ready {
			Poll::Ready(())
		} else {
			ready = true;
			Poll::Pending
		}
	})
}

/// Cedes control to the other pucks scheduled at the current model time,
/// independent of the priority of the active puck.
///
/// The active puck regains control after all of the other pucks have completed
/// their transition. The model time is not advanced for the active puck.
pub fn defer() -> impl Future<Output = ()> {
	let mut ready = false;
	poll_fn(move |cx| {
		if ready {
			Poll::Ready(())
		} else {
			ready = true;
			cx.waker().wake_by_ref();
			Poll::Pending
		}
	})
}

/// Returns a [Waker] for the currently active puck that may be stored for later
/// activation.
///
/// This function has to be awaited for technical reasons but doesn't actually
/// cause any scheduling operations to occur.
///
/// [Waker]: core::task::Waker
pub fn waker() -> impl Future<Output = core::task::Waker> {
	poll_fn(|cx| Poll::Ready(cx.waker().clone()))
}

/// Returns a [Future] that can be awaited in order to advance to the
/// termination or abortion of a [Task] within the simulation.
#[inline]
pub const fn join<C, P>(puck: P) -> Join<C, P>
where
	C: ?Sized + Config,
	P: Puck<C> + AsRef<Task<'static, C>> + Unpin,
{
	Join {
		publisher: puck,
		subscriber: None,
	}
}

/// Future that suspends the currently active [Task] until another task
/// indicates completion or abortion, returning its result.
pub struct Join<C: ?Sized + Config, P: Puck<C> + AsRef<Task<'static, C>>> {
	/// The task being awaited.
	publisher: P,
	/// The reactivated task.
	///
	/// Is `None` before suspending and after reactivation, contains the
	/// task to reactivate in-between.
	subscriber: Option<Irc<Task<'static, C>>>,
}

impl<C, P> Drop for Join<C, P>
where
	C: ?Sized + Config,
	P: Puck<C> + AsRef<Task<'static, C>>,
{
	fn drop(&mut self) {
		// clean up if we're being aborted
		if let Some(active) = self.subscriber.take() {
			unsafe {
				self.publisher.as_ref().remove_pending(&active);
			}
		}
	}
}

impl<C, P> Future for Join<C, P>
where
	C: ?Sized + Config,
	P: Puck<C> + AsRef<Task<'static, C>>,
{
	type Output = P::Output;

	fn poll(mut self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<Self::Output> {
		// check if we're already subscribed
		match self.subscriber.take() {
			None => {
				// no: determine the active task and subscribe
				let active = Irc::from(self.publisher.sim().active().expect("active task"));
				self.subscriber = Some(active.clone());
				self.publisher.as_ref().insert_pending(active);

				// wait for reactivation
				Poll::Pending
			}
			Some(_) => {
				// yes: we've been reactivated => the result is available
				Poll::Ready(self.publisher.result().expect("result available"))
			}
		}
	}
}
