//! This module provides the `Process` structure, which represents an isolated
//! thread of execution with its own copy of externally visible shared data.
//!
//! # Overview
//!
//! Each process can be thought of as the root of a tree made up of fibers,
//! each having its unique copy of shared data accessible from the outside.
//!
//! The lifecycle of processes prevents leakage of internal state into the
//! surrounding lexical context and prevents the surrounding context from
//! influencing the process execution via side effects.
//!
//! Process isolation is achieved through the `Actions` trait, which
//! parameterizes the `Process` structure. The `Actions` trait allows users to
//! define processes without explicitly naming the asynchronous lifecycle,
//! providing fine-grained control over the interface visibility.
//!
//! ## Key Components
//!
//!  - **Process:** The [`Process`] structure represents an isolated thread of
//!    execution with its own shared data and lifecycle. It automatically
//!    dereferences to the shared data it has been initialized with, and mutable
//!    references to the shared data may only be obtained before a process is
//!    activated, which is checked at compile-time.
//!
//!  - **Actions:** The [`Actions`] trait is used as the single parameter to the
//!    `Process` structure, aiding in the specification of user-defined
//!    behavior. It allows users to specify the simulation configuration, the
//!    type of shared data accepted, the resulting type of futures, and the
//!    [Settle] action taken once the process terminates normally.
//!
//!  - **Lifecycle:** The [`Lifecycle`] trait allows implementors to be used
//!    like a function that can be called once with a shared reference to some
//!    data and a simulation context, returning a `Future` that can reference
//!    these arguments. It is used in the `Actions` trait in an all-quantified
//!    (HRTB) form over the lifetime of its arguments and return value,
//!    enabling compiler-enforced process isolation.
//!
//!  - **Builder:** A builder type aiding in the configuration of processes with
//!   detailed specifications, such as a unique name, lifecycle, rank, and
//!   source code location.
//!
//! ## Examples
//!
//! Simple process with a lifecycle and no return value:
//!
//! ```
//! # use odem_rs_core::{process::{Actions, Process}, Sim, config::DefaultConfig, unique};
//! struct Foo(usize);
//!
//! impl Foo {
//!     fn new(n: usize) -> Process<impl Actions<Subject = Self, Config = DefaultConfig>> {
//!         Process::new(Foo(n), Foo::actions)
//!     }
//!
//!     async fn actions(&self, sim: &Sim) {}
//! }
//!
//! async fn sim_main(sim: &Sim) {
//!     let foo = unique!(Foo::new(3));
//!     let h = sim.activate(foo);
//!     sim.advance(10.0).await;
//! }
//! ```
//!
//! Simple process with lifecycle returning a value:
//!
//! ```
//! # use odem_rs_core::{ops::defer, Puck, process::{Actions, Process}, Sim, config::DefaultConfig, unique};
//! struct Foo(usize);
//!
//! impl Foo {
//!     fn new(n: usize) -> Process<impl Actions<Subject = Self, Output = f32, Config = DefaultConfig>> {
//!         Process::new(Foo(n), Foo::actions)
//!     }
//!
//!     async fn actions(&self, sim: &Sim) -> f32 {
//!         self.0 as f32
//!     }
//! }
//!
//! async fn sim_main(sim: &Sim) {
//!     let foo = unique!(Foo::new(3));
//!     let x: f32 = sim.activate(foo).await;
//! }
//! ```
//!
//! # Memory-Safety and Drop
//! As is the case for [Fiber], certain memory-safety guarantees for Processes
//! hinge on the execution of their [Drop] implementation.
//!
//! However, Rust asserts that memory-safety guarantees must be upheld even in
//! the complete absence of destructor calls with the reasoning - as far as the
//! library author understands it - that it is impossible to statically prevent
//! memory leaks from occuring even in purely safe code, e.g. by constructing
//! circular dependencies using [Rc].
//!
//! The library has been hardened towards preventing this type of bug, mainly
//! by adding additional checks and checksums. We distinguish two different
//! types of leak: (1) purely [forgetting] to call [drop], or (2) [leaking]
//! Processes via the heap.
//!
//! ## [`forget`]-type leaks
//! Forgetting to call a Process' destructor crucially requires *taking
//! ownership* of the Process and *moving* it into the forget-function.
//! We distinguish three cases where this could potentially happen:
//! 1. right after the construction of the process, before its activation,
//! 2. after its activation, before its termination,
//! 3. after its termination.
//!
//! Possible dangling pointers that lead to memory unsafety upon dereferencing
//! are found in the [Fiber]-instances of the fibers that are still active when
//! the process-memory is reclaimed.
//!
//! The first case is harmless because no Fibers have been instantiated yet and
//! so there is nothing to clean up.
//!
//! In the second case, the process must have been pinned to be activated,
//! precluding it from being moved at all.
//!
//! Since pinning provides guarantees for the entirety of an object's life,
//! the process can also not be moved after termination of its root fiber.
//!
//! ## [`Box::leak`]-type leaks
//! Since processes are isolated from each other, leaking a process doesn't
//! cause memory unsafety: it will simply be descheduled at the end of the
//! simulation run and then continue to exist until the end of the program.
//!
//! [forgetting]: core::mem::forget
//! [`forget`]: core::mem::forget
//! [leaking]: Box::leak
//! [drop]: Drop::drop
//! [Rc]: std::rc::Rc

use core::{
	any::{type_name, Any},
	future::{pending, Future, IntoFuture, Pending},
	hint::unreachable_unchecked,
	marker::PhantomData,
	ops::{Deref, DerefMut},
	panic::Location,
	pin::Pin,
	ptr::NonNull,
	task::{Context, Poll},
};

use crate::{
	config::{Config, DefaultConfig},
	error,
	fiber::{Builder as FiberBuilder, Checked, Fiber, Puck as FiberPuck, Settle, Unchecked},
	simulator::Sim,
	task::{erased::State as TaskState, Share, Task},
	Active, Dispatch, ExitStatus, Irc, Label, Prec, Unique,
};

/* *************************************************** Internal Actions-Trait */

/// This trait is used as the single parameter to the [Process] structure and
/// is intended to be helpful for the construction of user-defined processes.
///
/// The main problem that is solved by the trait concerns the fact that the
/// asynchronous [Lifecycle] of a process cannot be named since its definition
/// is generated by the compiler. By generating a [Process] whose only
/// type-argument `impl`s the Actions-trait, a user can get around naming
/// the lifecycle-part of the process.
///
/// Note that it is possible to underspecify the result type, e.g. by omitting
/// information about the associated [Object](Actions::Subject)-type. This
/// effectively hides those parts of the interface that rely on this
/// information, giving a very fine-grained control about what is possible to
/// do by the caller and what isn't.
pub trait Actions {
	/// Specifies the [simulation configuration](Config) used by the
	/// [Action](Self::Action) type.
	type Config: Config;
	/// Specifies the type of shared data accepted by the [Action](Self::Action)
	/// type.
	type Subject: Any;
	/// Specifies the resulting type of all the potential futures produced
	/// by the [Action](Self::Action) type.
	type Output;
	/// Specifies the [Settle]-action taken once the process terminates
	/// normally.
	type Settle: Settle<Self::Output>;

	/// Specifies a set of methods all-quantified over lifetime that can be used
	/// to derive a [Future] given a reference to [shared] data and a
	/// [simulation context](Sim).
	///
	/// This construction is used to enforce process-isolation statically
	/// because it prevents the produced future from explicitly referencing
	/// non-static context outside the provided simulation context and the
	/// reference to the shared data. Consequently, this reduces the
	/// surface-area of side effects regarding the inner process state to
	/// the well-defined interface provided by these two references, analogous
	/// to the way operating systems are used to orchestrate interactions
	/// between system processes through a well-defined interface.
	///
	/// Crucially, it is enough to demand the HRTB without actually having to
	/// run it for different lifetimes in order to trigger the corresponding
	/// checks in the compiler.
	///
	/// See the [Lifecycle]-trait for more information.
	///
	/// [shared]: Self::Subject
	type Action: for<'p> Lifecycle<'p, Self::Config, Self::Subject, Output = Self::Output>;
}

/// This trait allows its implementors to be used like a function that can be
/// called once with a shared reference to some data and a
/// [simulation context](Sim) and return a [Future] that can reference
/// these arguments.
///
/// It is used in the [Actions]-trait in an all-quantified (HRTB) form over the
/// lifetime of its arguments and return value, enabling compiler-enforced
/// [Process]-isolation.
pub trait Lifecycle<'p, C: Config, I: Any + ?Sized> {
	/// Specifies the [Output](Future::Output)-type of the [Future].
	type Output;
	/// The resulting [Future] of the [init](Self::init)-method.
	type Future: Future<Output = Self::Output> + 'p;

	/// Trait-method that uses the implementor-instance to construct the
	/// [Future](Self::Future) given the shared reference to some data and
	/// the [simulation context](Sim).
	fn init(self, item: &'p I, sim: &'p Sim<C>) -> Self::Future;
}

/* ****************************************************************** Process */

/// Isolated thread of execution that has its own copy of externally visible
/// shared data.
///
/// Each [Process] can be thought of the root of a tree made up of [Fiber] that
/// has its own unique copy of data that is shared amongst the fiber in the
/// tree and is accessible from the outside.
///
/// The [Lifecycle] of processes prevents leakage of internal state into the
/// surrounding (lexical) context as well as preventing the surrounding context
/// from influencing the execution of the process via side effects. This is in
/// contrast to the way [Fiber] allow the free sharing of lexically visible
/// state between each other and their creators.
///
/// Process isolation is achieved through the [Actions]-trait parameterizing
/// the structure. More concretely, the lifecycle of a process is prevented from
/// sharing data with its surrounding context by asking for a higher-ranked
/// trait-bound on the lifetime of arguments and return types, preventing
/// non-`'static` data, i.e. locally visible identifiers, from being referenced
/// by the closure.
///
/// Note that processes automatically dereference to the shared data they have
/// been initialized with as a matter of convenience. Mutable references to the
/// shared data may only be obtained before a Process is [activated] which is
/// checked at compile-time.
///
/// # Examples
/// Simple process with a lifecycle and no return value:
/// ```
/// # use odem_rs_core::{process::{Actions, Process}, Sim, config::DefaultConfig, Puck, unique};
/// struct Foo(usize);
///
/// impl Foo {
///     fn new(n: usize) -> Process<impl Actions<
///         Subject = Self,
///         Config = DefaultConfig
///     >> { Process::new(Foo(n), Foo::actions) }
///
///     async fn actions(&self, sim: &Sim) {}
/// }
///
/// async fn sim_main(sim: &Sim) {
///     let foo = unique!(Foo::new(3));
///     let h = sim.activate(foo);
///     sim.advance(10.0).await;
/// }
/// ```
///
/// Simple process with lifecycle returning a value:
/// ```
/// # use odem_rs_core::{ops::defer, process::{Actions, Process}, Sim, config::DefaultConfig, Puck, unique};
/// struct Foo(usize);
///
/// impl Foo {
///     fn new(n: usize) -> Process<impl Actions<
///         Subject = Self,
///         Output = f32,
///         Config = DefaultConfig
///     >> { Process::new(Foo(n), Foo::actions) }
///
///     async fn actions(&self, sim: &Sim) -> f32 {
///         self.0 as f32
///     }
/// }
///
/// async fn sim_main(sim: &Sim) {
///     let foo = unique!(Foo::new(3));
///     let x: f32 = sim.activate(foo).await;
/// }
/// ```
///
/// [activated]: Sim::activate
#[pin_project::pin_project]
pub struct Process<A: Actions> {
	/// An enumeration capturing the runtime state.
	#[pin]
	inner: Inner<A>,

	/// The shared process data.
	#[pin]
	item: A::Subject,
}

/// Inner enumeration for the runtime states of the `Process`: `Born`, `Bust`,
/// and `Live`.
///
/// The `Born`-state records the initialization of the process before its
/// activation. The `Bust`-state is an intermediate state during activation
/// and can only be observed if the constructor-function of the future panics.
/// The `Live`-state is the state after initialization has occurred.
#[pin_project::pin_project(
	project = StateProject,
	project_ref = StateProjectRef,
	project_replace = StateOwn
)]
enum Inner<A: Actions> {
	/// Pre-activation state containing an inert future, the process name, its
	/// initial rank, and a fiber builder missing the future.
	Born {
		/// Contains a callable function constructing a future from shared data
		/// and a simulation context.
		action: A::Action,
		/// The base name of the process.
		name: &'static str,
		/// An optional initial rank.
		rank: Option<<A::Config as Config>::Rank>,
		/// A [FiberBuilder] in intermediate state without the lifecycle future.
		builder: FiberBuilder<(), A::Settle>,
	},
	/// Intermediate state during activation of a process. Should only be
	/// observable if `A::Action` panicked.
	Bust,
	/// Post-activation state containing a [Fiber] and the [process data](Share)
	/// shared between all fiber contained in the process.
	Live {
		/// Post-activation state containing a [Fiber] constructed by the
		/// action function.
		#[pin]
		fiber: RootFiber<'static, A>,
		/// Data shared between all the [Fiber] associated with this process.
		///
		/// This data is accessible from outside given any type of [Puck] by calling
		/// the [Puck::subject]-method and downcasting the resulting [Any] into the
		/// concrete type.
		#[pin]
		share: Share<A::Config>,
	},
}

/// Type-alias for the [Future]-type constructed by the [Lifecycle]-trait.
type RootFuture<'l, A> = <<A as Actions>::Action as Lifecycle<
	'l,
	<A as Actions>::Config,
	<A as Actions>::Subject,
>>::Future;

/// Type-alias for the [Fiber]-type constructed by the [Lifecycle]-trait.
type RootFiber<'l, A> =
	Fiber<'static, <A as Actions>::Config, RootFuture<'l, A>, <A as Actions>::Settle>;

/// Type-alias for the [FiberPuck]-type constructed by the [Lifecycle]-trait.
type RootPuck<'l, A> =
	FiberPuck<'l, <A as Actions>::Config, RootFuture<'l, A>, <A as Actions>::Settle>;

impl Process<()> {
	/// Creates a new process with a lifecycle that produces futures that
	/// cannot fail.
	///
	/// The inability to fail refers to the property that - as far as the
	/// underlying [Task] is concerned - the process can only either be
	/// cancelled, panic, or terminate successfully, i.e. with a result of `Ok`.
	///
	/// It doesn't mean that the return type of the process cannot be a
	/// `Result`, just that the `Err` variant is not recorded as a failure but
	/// as an acceptable (`Ok`) result of the execution.
	///
	/// If a process with a publicly observable failure is required, use the
	/// [Builder] to configure it instead.
	#[track_caller]
	pub fn new<C, I, R, A>(subject: I, actions: A) -> Process<Action<C, I, A, Unchecked>>
	where
		C: Config,
		I: Any,
		A: for<'l> Lifecycle<'l, C, I, Output = R>,
	{
		Self::build()
			.with_subject(subject)
			.with_actions(actions)
			.finish()
	}

	/// Creates a [Builder] that can be used to configure the process
	/// before instantiating it.
	pub const fn build<C: Config>() -> Builder<C> {
		Builder::new()
	}
}

impl<A: Actions> Deref for Process<A> {
	type Target = A::Subject;

	#[inline]
	fn deref(&self) -> &Self::Target {
		&self.item
	}
}

impl<A: Actions> DerefMut for Process<A> {
	#[inline]
	fn deref_mut(&mut self) -> &mut Self::Target {
		&mut self.item
	}
}

impl<A: Actions> Dispatch for Process<A> {
	fn poll(self: Pin<&Self>, cx: &mut Context<'_>) -> Poll<ExitStatus> {
		match self.project_ref().inner.project_ref() {
			StateProjectRef::Live { fiber, .. } => fiber.poll(cx),
			_ => unreachable!("the process should always be active"),
		}
	}
}

impl<A: Actions> Active<A::Config> for Process<A> {
	type Puck<'p>
		= Puck<'p, A>
	where
		Self: 'p;

	unsafe fn bind<'p>(
		this: Unique<'p, Self>,
		vtab: NonNull<dyn Dispatch + 'p>,
		sctx: NonNull<Share<A::Config>>,
	) -> Result<Self::Puck<'p>, error::NotBorn> {
		let mut this = this.into_inner().project();

		// check the current process state and bail early if it isn't `Born`
		match this.inner.as_ref().project_ref() {
			StateProjectRef::Born { .. } => {}
			StateProjectRef::Bust => return Err(error::NotBorn(TaskState::Gone)),
			StateProjectRef::Live { fiber, .. } => return Err(error::NotBorn(fiber.state())),
		}

		// replace the `Born` state by `Bust` for the transition
		match this.inner.as_mut().project_replace(Inner::Bust) {
			StateOwn::Born {
				action,
				name,
				rank,
				builder,
			} => {
				use core::mem::transmute;

				// extract the simulation context from the shared context
				let item = this.item.into_ref().get_ref();
				let sim = unsafe { sctx.as_ref() }.sim();
				
				// construct the fiber by executing the action;
				// this may panic, leaving the process in state `Bust`
				let live = builder.with_actions(action.init(item, sim)).finish();

				// overwrite the previous state, settling the created fiber in
				// its final memory location
				this.inner.set(Inner::Live {
					fiber: transmute::<RootFiber<'p, A>, RootFiber<'static, A>>(live),
					share: Share::new(
						sim.clone(),
						item,
						rank.unwrap_or(sim.config().default_rank()),
						name,
					),
				});

				match this.inner.project() {
					StateProject::Live { fiber, share } => {
						let live = Unique::new_unchecked(transmute::<
							Pin<&'p mut RootFiber<'static, A>>,
							Pin<&'p mut RootFiber<'p, A>>,
						>(fiber));

						Ok(Puck(Active::bind(
							live,
							vtab,
							NonNull::from(share.into_ref().get_ref()),
						)?))
					}
					_ => unreachable_unchecked(),
				}
			}
			// SAFETY: we know that the process is in state `Born`
			_ => unreachable_unchecked(),
		}
	}
}

/* ************************************************************* Process Puck */

/// The root [Puck](FiberPuck) of the [Process].
pub struct Puck<'p, A: Actions>(RootPuck<'p, A>);

impl<C: Config, A: Actions<Config = C>> Puck<'_, A> {
	/// Sets the rank of the root fiber.
	///
	/// Note that the rank only becomes effective during the next selected
	/// model time. You can force a rescheduling by [deferring] the
	/// corresponding fiber.
	///
	/// [deferring]: crate::ops::defer
	#[inline]
	pub fn set_rank(&self, rank: C::Rank) {
		self.0.set_rank(rank);
	}

	/// Returns a reference to the subject.
	#[inline]
	pub fn subject(&self) -> &A::Subject {
		use crate::Puck;

		self.0.subject().downcast_ref::<A::Subject>().unwrap()
	}

	/// Terminates the root fiber prematurely.
	///
	/// This method consumes the reference to prevent it from being called twice
	/// or used to extract the return value later.
	pub fn abort(self) {
		self.0.abort();
	}
}

impl<C: Config, A: Actions<Config = C>> crate::Puck<C> for Puck<'_, A> {
	fn result(&mut self) -> Option<Self::Output> {
		self.0.result()
	}

	fn wake(&mut self) -> Result<(), error::NotIdle> {
		self.0.wake()
	}

	fn subject(&self) -> &dyn Any {
		self.0.subject()
	}

	fn sim(&self) -> &Sim<C> {
		self.0.sim()
	}

	fn label(&self) -> Label {
		self.0.label()
	}

	fn time(&self) -> Option<C::Time> {
		self.0.time()
	}

	fn rank(&self) -> C::Rank {
		self.0.rank()
	}

	fn prec(&self) -> Prec {
		self.0.prec()
	}

	fn state(&self) -> TaskState {
		self.0.state()
	}

	fn own(&self) -> Irc<Task<'static, C>> {
		self.0.own()
	}
}

impl<A: Actions> AsRef<Task<'static, A::Config>> for Puck<'_, A> {
	fn as_ref(&self) -> &Task<'static, A::Config> {
		self.0.as_ref()
	}
}

impl<A: Actions> IntoFuture for Puck<'_, A> {
	type Output = A::Output;
	type IntoFuture = crate::ops::Join<A::Config, Self>;

	#[inline]
	fn into_future(self) -> Self::IntoFuture {
		crate::ops::join(self)
	}
}

/* ********************************************************** Process Builder */

/// Used for configuring [Process] instances before they are instantiated.
///
/// Instantiating processes this way allows to set a custom name and initial
/// rank.
pub struct Builder<C: Config, I = (), A = (), S = Unchecked> {
	/// The shared item.
	subject: I,
	/// The lifecycle of the process.
	actions: A,
	/// The custom name or the name derived from the type name of `I`.
	name: Option<&'static str>,
	/// The initial rank of the created process.
	rank: Option<C::Rank>,
	/// The [Settle]-routine called at the end of the process' life.
	settle: S,
	/// The [Location] of the lifecycle in the source code.
	location: &'static Location<'static>,
	/// Marker for the [Config].
	_config: PhantomData<C>,
}

impl<C: Config> Builder<C, (), ()> {
	/// Creates a new, empty process builder.
	#[inline(always)]
	const fn new() -> Self {
		Builder {
			subject: (),
			actions: (),
			name: None,
			rank: None,
			settle: Unchecked,
			location: Location::caller(),
			_config: PhantomData,
		}
	}
}

impl<C: Config, I: Any, A, S> Builder<C, I, A, S> {
	/// Sets the shared item (the subject) that the created process operates on.
	pub fn with_subject<X: Any>(self, object: X) -> Builder<C, X, A, S> {
		Builder {
			subject: object,
			actions: self.actions,
			name: self.name,
			rank: self.rank,
			settle: self.settle,
			location: self.location,
			_config: self._config,
		}
	}

	/// Sets the lifecycle for this process builder with a future that cannot
	/// fail.
	///
	/// The inability to fail refers to the property that - as far as the
	/// underlying [Task] is concerned - the process can only either be
	/// cancelled, panic, or terminate successfully, i.e. with a result of `Ok`.
	///
	/// It doesn't mean that the return type of the process cannot be a
	/// `Result`, just that the `Err` variant is not recorded as a failure but
	/// as an acceptable (`Ok`) result of the execution.
	#[track_caller]
	#[inline(always)]
	pub fn with_actions<X>(self, actions: X) -> Builder<C, I, X, S>
	where
		X: for<'l> Lifecycle<'l, C, I>,
	{
		Builder {
			subject: self.subject,
			actions,
			name: self.name,
			rank: self.rank,
			settle: self.settle,
			location: Location::caller(),
			_config: self._config,
		}
	}

	/// Sets an explicit source-code [Location] for the [Process].
	///
	/// The location will otherwise default to the location of the lifecycle.
	#[inline(always)]
	pub fn with_location(self, location: &'static Location<'static>) -> Self {
		Builder { location, ..self }
	}

	/// Sets a custom name for the instantiated process.
	///
	/// The name will default to the type name of the shared item.
	#[inline(always)]
	pub fn with_name(self, name: &'static str) -> Self {
		Builder {
			name: Some(name),
			..self
		}
	}

	/// Sets a custom starting rank for the instantiated process.
	///
	/// The default rank is specified in the [configuration](Config).
	#[inline(always)]
	pub fn with_rank(self, rank: C::Rank) -> Self {
		Builder {
			rank: Some(rank),
			..self
		}
	}

	/// Sets a user-defined [Settle] that is run once the lifecycle of the
	/// [Process] terminates.
	#[inline(always)]
	pub fn with_finalizer<X>(self, finalizer: X) -> Builder<C, I, A, X> {
		Builder {
			subject: self.subject,
			actions: self.actions,
			name: self.name,
			rank: self.rank,
			settle: finalizer,
			location: self.location,
			_config: self._config,
		}
	}

	/// Sets a specific [Settle] routine that is used to convert a `bool`,
	/// [`Option`], or [`Result`] return type into an [ExitStatus] that may
	/// be observed by other processes or fiber.
	#[inline(always)]
	pub fn checked(self) -> Builder<C, I, A, Checked> {
		self.with_finalizer(Checked)
	}
}

impl<C: Config, I: Any, R, A, S> Builder<C, I, A, S>
where
	A: for<'l> Lifecycle<'l, C, I, Output = R>,
	S: Settle<R>,
{
	/// Instantiates the configured process from the builder.
	#[inline(always)]
	pub fn finish(self) -> Process<Action<C, I, A, S>> {
		Process {
			item: self.subject,
			inner: Inner::Born {
				action: self.actions,
				name: if let Some(name) = self.name {
					name
				} else {
					type_name::<I>()
				},
				rank: self.rank,
				builder: Fiber::build()
					.with_finalizer(self.settle)
					.with_location(self.location),
			},
		}
	}
}

/* **************************************************** Trait Implementations */

/// Marker-type that is used to determine the configuration-, instance-,
/// and action-types of the process.
pub struct Action<C, I, A, T>(PhantomData<(C, I, A, T)>);

impl<C: Config, I: Any, R, A, T: Settle<R>> Actions for Action<C, I, A, T>
where
	A: for<'l> Lifecycle<'l, C, I, Output = R>,
{
	type Config = C;
	type Subject = I;
	type Output = R;
	type Settle = T;
	type Action = A;
}

impl Actions for () {
	type Config = DefaultConfig;
	type Subject = ();
	type Output = ();
	type Settle = Unchecked;
	type Action = Pending<()>;
}

impl<'p, C, I> Lifecycle<'p, C, I> for Pending<()>
where
	C: Config,
	I: Any,
{
	type Output = ();
	type Future = Pending<()>;

	#[inline]
	fn init(self, _item: &'p I, _sim: &'p Sim<C>) -> Self::Future {
		pending()
	}
}

impl<'p, C, I, A, F> Lifecycle<'p, C, I> for A
where
	C: Config,
	I: Any,
	A: FnOnce(&'p I, &'p Sim<C>) -> F,
	F: Future + 'p,
{
	type Output = F::Output;
	type Future = F;

	#[inline]
	fn init(self, item: &'p I, sim: &'p Sim<C>) -> Self::Future {
		self(item, sim)
	}
}
