//! The `irc` module provides an intrusive reference-counting smart pointer,
//! `Irc`, along with traits for implementing intrusively counted types.

use core::{
	any::{Any, TypeId},
	borrow::Borrow,
	fmt,
	marker::PhantomData,
	ops::{Deref, DerefMut},
	panic::{Location, RefUnwindSafe, UnwindSafe},
	pin::Pin,
	ptr::NonNull,
};

/* ******************************************************************* Traits */

/// A trait that can be implemented by types that track references using an
/// internal reference counter.
///
/// # Safety
/// The implementation must guarantee that the `IrcBox` is owned by the
/// implementor, i.e. will be dropped together with `Self`. This usually means
/// that the `IrcBox` is embedded in the type implementing this trait.
pub unsafe trait IntrusivelyCounted {
	/// The intrusive type containing the reference count and additional
	/// meta-data.
	type Inner: IrcBoxed;

	/// Returns a reference to the embedded object implementing [IrcBoxed].
	/// It is used by [Irc] to opaquely manipulate the reference counter.
	fn irc_box(&self) -> &IrcBox<Self::Inner>;
}

/// A trait allowing an [`Irc`] to increment and decrement the reference count
/// as well as recycling an unreachable object.
///
/// # Safety
/// The implementor has to ensure that the `acquire` and `release` methods
/// provided by the trait correctly keep track of the inner reference counts
/// in order for the `Irc` derived from them to be sound.
pub unsafe trait IrcBoxed {
	/// Increments the reference counter, returning its updated value.
	///
	/// # Safety
	/// The caller of this method is responsible to call it only in order to
	/// register an additional reference to the object containing this
	/// [`IrcBox`]. Calling it too often will lead to a program abort once the
	/// `IrcBox` is dropped.
	unsafe fn acquire(&self) -> usize;

	/// Decrements the internal reference counter, returning a function pointer
	/// responsible for recycling if it reaches zero.
	///
	/// The reason for this convoluted mechanism is that the implementing
	/// [`IrcBox`] may recover the original type and reclaim it using an
	/// exclusive reference. This cannot be done in the function directly
	/// because a shared reference to the `IrcBox` exists, precluding the
	/// existence of mutable references to objects that include the `IrcBox`
	/// itself.
	///
	/// # Safety
	/// The caller of this method is only allowed to call it once a reference to
	/// the object containing this `IrcBox` is destroyed. Calling it in other
	/// cases can lead to dangling pointers and is therefore not memory safe.
	unsafe fn release(&self) -> Option<fn(NonNull<Self>)>;
}

/* ************************************************************ Irc Structure */

/// The intrusive version of a [Rc] without support for [Weak] pointer.
///
/// [Rc]: std::rc::Rc
/// [Weak]: std::rc::Weak
#[repr(transparent)]
pub struct Irc<T: ?Sized + IntrusivelyCounted>(NonNull<T>);

impl<T: ?Sized + IntrusivelyCounted> Irc<T> {
	/// Creates a new intrusively counted pointer from an [`Invariant`] value.
	///
	/// Because the value is borrowed exclusively, we can be sure that it
	/// currently is the only reference to the value. Because of the invariant
	/// lifetime `'p` associated with the `Invariant`, we can be sure that it
	/// can *never* be borrowed again for the remainder of its lifetime.
	/// Finally, because the value is [pinned] and `T` implements
	/// [`IntrusivelyCounted`], we can be sure that its inner [`IrcBox`] will
	/// either be dropped or remain valid indefinitely.
	///
	/// [pinned]: core::pin
	pub fn new<'p>(value: Pin<&'p mut Invariant<'p, T>>) -> Self {
		// increase the reference count to one
		unsafe {
			value.irc_box().inner.acquire();
		}
	
		// strip away the lifetime requirements
		Irc(NonNull::from(&mut **unsafe {
			Pin::into_inner_unchecked(value)
		}))
	
		// dangling pointers are prevented by the Drop impl of `IrcBox`
	}
	
	// pub fn new(value: Pin<&mut T>) -> Option<Self> {
	// 	let irc_box = value.irc_box();
	// 	
	// 	match unsafe { irc_box.inner.acquire() } {
	// 		0 => unreachable!("fewer than 0 references to this Irc"),
	// 		1 => Some(Irc(NonNull::from(&mut *unsafe {
	// 			Pin::into_inner_unchecked(value)
	// 		}))),
	// 		_ => unsafe {
	// 			irc_box.inner.release();
	// 			None
	// 		}
	// 	}
	// }

	/// Strips the outer structure from the `Irc`, revealing the inner
	/// unprotected reference. This operation does not decrease the reference
	/// counter and should be used in tandem with [`from_raw`] to restore the
	/// smart pointer at a later time.
	///
	/// [`from_raw`]: Self::from_raw
	pub const fn into_raw(this: Self) -> NonNull<T> {
		// copy the inner pointer
		let inner = this.0;

		// forget calling drop
		core::mem::forget(this);

		// return the pointer
		inner
	}

	/// Reconstructs the `Irc` from an unprotected reference.
	///
	/// # Safety
	/// The associated function is marked as unsafe because it is the caller's
	/// responsibility to ensure that this reference has originally been the result
	/// of a call to [`into_raw`].
	///
	/// [`into_raw`]: Self::into_raw
	pub unsafe fn from_raw(inner: NonNull<T>) -> Self {
		Irc(inner)
	}

	/// Allows the (limited) projection of a composite [Irc] into an Irc of one
	/// of its member variables, provided that member variable also intrusively
	/// contains a reference counter.
	pub fn map<F, R>(this: Self, f: F) -> Irc<R>
	where
		F: FnOnce(&T) -> &R,
		R: ?Sized + IntrusivelyCounted,
	{
		// store the pointer to the inner IrcBox
		let inner = this.irc_box();

		// perform the conversion
		let value = f(&*this);

		// assert that the IrcBox hasn't changed due to the projection;
		// this should be optimized away by the compiler
		assert!(
			core::ptr::addr_eq(inner, value.irc_box()),
			"expected the mapping to yield an Irc with the same IrcBox"
		);

		// construct the new Irc, taking the provenance from `this`
		let res = Irc(with_provenance(NonNull::from(value), this.0));

		// forget the original Irc
		core::mem::forget(this);

		res
	}

	/// Converts the `Irc` into type `V` if it is of this type, or returns the
	/// old `Irc` if it isn't.
	pub fn downcast<V>(self) -> Result<Irc<V>, Irc<T>>
	where
		T: Any,
		V: 'static + IntrusivelyCounted,
	{
		if (*self).type_id() == TypeId::of::<V>() {
			let res = Irc(self.0.cast::<V>());
			core::mem::forget(self);
			Ok(res)
		} else {
			Err(self)
		}
	}
}

impl<T: ?Sized + IntrusivelyCounted> From<super::Unique<'_, T>> for Irc<T> {
	fn from(value: super::Unique<'_, T>) -> Self {
		unsafe {
			value.irc_box().inner.acquire();
		}
		Irc(NonNull::from(unsafe {
			Pin::into_inner_unchecked(value.into_inner())
		}))
	}
}

impl<T: ?Sized + IntrusivelyCounted> Deref for Irc<T> {
	type Target = T;

	fn deref(&self) -> &Self::Target {
		unsafe { self.0.as_ref() }
	}
}

impl<T: ?Sized + IntrusivelyCounted> Clone for Irc<T> {
	fn clone(&self) -> Self {
		// increment the internal reference counter and copy the inner pointer
		unsafe {
			self.irc_box().inner.acquire();
		}
		Irc(self.0)
	}
}

impl<T: ?Sized + IntrusivelyCounted> Borrow<T> for Irc<T> {
	fn borrow(&self) -> &T {
		self
	}
}

impl<T: ?Sized + IntrusivelyCounted> Drop for Irc<T> {
	fn drop(&mut self) {
		if let Some(recycle) = unsafe { self.irc_box().inner.release() } {
			recycle(NonNull::from(&self.irc_box().inner));
		}
	}
}

impl<T: ?Sized + IntrusivelyCounted> Unpin for Irc<T> {}

impl<T: ?Sized + IntrusivelyCounted + fmt::Debug> fmt::Debug for Irc<T> {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		unsafe { self.0.as_ref() }.fmt(f)
	}
}

impl<T: ?Sized + IntrusivelyCounted + fmt::Display> fmt::Display for Irc<T> {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		unsafe { self.0.as_ref() }.fmt(f)
	}
}

// T: Sync is enough for Irc<T> to be Send + Sync since we're moving references
unsafe impl<T: Sync + ?Sized + IntrusivelyCounted> Send for Irc<T> {}

unsafe impl<T: Sync + ?Sized + IntrusivelyCounted> Sync for Irc<T> {}

impl<T: RefUnwindSafe + ?Sized + IntrusivelyCounted> UnwindSafe for Irc<T> {}

/* ****************************************************** Provenance Transfer */

/// Transfers the provenance from one non-null pointer to another.
///
/// This function assigns the provenance of the `provenance` pointer to the
/// `target` pointer. Provenance refers to the origin or ownership context of a
/// pointer, which is crucial for maintaining memory safety and correctness in
/// low-level operations. See [here] for Rust's strict provenance.
///
/// # Safety
///
/// This function performs raw pointer manipulation and assumes that the layout
/// of fat pointers with metadata remains consistent.
/// 
/// [here]: https://doc.rust-lang.org/core/ptr/index.html#strict-provenance
#[inline(always)]
fn with_provenance<S, T>(mut target: NonNull<T>, provenance: NonNull<S>) -> NonNull<T>
where
	S: ?Sized,
	T: ?Sized,
{
	// Create a thin pointer by casting `provenance` to `u8` and setting its
	// address to that of `target`. This combines the address of `target` with
	// the provenance of `provenance`.
	let target_thin_ptr = provenance.cast::<u8>().with_addr(target.addr());
	
	// Obtain a mutable reference to the `target` pointer and cast it to a
	// pointer to `NonNull<u8>`. This allows direct manipulation of the thin
	// pointer portion of the potentially fat `NonNull<T>`.
	let ptr_to_fat_ptr = NonNull::from(&mut target).cast::<NonNull<u8>>();
	
	// Overwrite the thin pointer part of the fat `target` pointer with the new
	// thin pointer that carries the desired provenance. This operation
	// preserves the original address while updating its provenance.
	//
	// SAFETY: This is safe provided that the layout of fat pointers with
	// metadata does not change.
	// TODO: use with_metadata_of() once stable
	unsafe {
		ptr_to_fat_ptr.write(target_thin_ptr);
	}
	
	// Return the updated `target` pointer with the new provenance.
	target
}

/* *********************************************** Invariant Lifetime Wrapper */

/// A weightless wrapper type that tags a value of type `T` with an invariant
/// lifetime `'p`.
///
/// The purpose of this type is to enable the following pattern:
///
/// ```
/// # use odem_rs_core::ptr::Invariant;
/// fn indefinitely_borrow<'p, T>(value: &'p mut Invariant<'p, T>) {
///     // ...
/// }
///
/// let mut value = Invariant::new(42);
/// indefinitely_borrow(&mut value);
/// // indefinitely_borrow(&mut value); // error: value is mutably borrowed
/// // std::mem::forget(value);         // error: value is still borrowed
///```
#[derive(Debug, Clone, Copy, Default)]
pub struct Invariant<'p, T: ?Sized> {
	_mark: PhantomData<fn(&'p T) -> &'p T>,
	value: T,
}

impl<T> Invariant<'_, T> {
	/// Creates a new `Invariant` wrapping the given value.
	pub const fn new(value: T) -> Self {
		Self {
			_mark: PhantomData,
			value,
		}
	}

	/// Consumes the `Invariant`, returning the wrapped value.
	pub fn into_inner(self) -> T {
		self.value
	}
}

impl<T: ?Sized> Deref for Invariant<'_, T> {
	type Target = T;

	fn deref(&self) -> &Self::Target {
		&self.value
	}
}

impl<T: ?Sized> DerefMut for Invariant<'_, T> {
	fn deref_mut(&mut self) -> &mut Self::Target {
		&mut self.value
	}
}

/* ********************************************************* IrcBox Structure */

/// Type that has to be stored inside a structure in order to enable creating
/// [`Irc`] to it.
#[derive(Debug)]
pub struct IrcBox<T: ?Sized + IrcBoxed> {
	/// The location in the code.
	pub loc: &'static Location<'static>,
	/// The inner value containing the reference counters.
	pub inner: T,
}

impl<T: IrcBoxed> IrcBox<T> {
	/// Creates a new `IrcBox` with the inner value.
	#[track_caller]
	pub const fn new(inner: T) -> Self {
		IrcBox::with_location(inner, Location::caller())
	}

	/// Creates a new `IrcBox` with a specific [Location].
	pub const fn with_location(inner: T, loc: &'static Location<'static>) -> Self {
		IrcBox { loc, inner }
	}
}

impl<T: IrcBoxed + Default> Default for IrcBox<T> {
	#[track_caller]
	fn default() -> Self {
		Self::new(T::default())
	}
}

impl<T: ?Sized + IrcBoxed> Drop for IrcBox<T> {
	fn drop(&mut self) {
		// prime an abort guard in case we have to panic
		let abort_guard = scopeguard::guard((), |()| {
			panic!("aborting due to dangling Irc");
		});

		// ensure that no references point to this instance
		match unsafe { self.inner.acquire() } {
			// this should not be possible
			0 => unreachable!("fewer than zero references to this Irc"),
			// abort the abort
			1 => core::mem::forget(abort_guard),
			// this panic leads to an immediate abort which is necessary
			// because running all the destructors in the stack frames above
			// will perform a ton of illegal memory accesses
			n => panic!(
				"dropping the value created at '{}' leaves {} reference{} dangling",
				self.loc,
				n - 1,
				if n > 2 { "s" } else { "" }
			),
		}
	}
}
