//! This module defines the `Unique<P>` struct, a type representing an owning
//! *unique* pointer.
//!
//! It's designed to provide safe, exclusive ownership of the data it points to,
//! ensuring that there is no other reference to the same data during its
//! lifetime.
//!
//! ## Features
//!
//! - `Unique<T>`: A struct that encapsulates an exclusive, pinned pointer to
//!   data. It guarantees that the data referred to is not aliased elsewhere for
//!   as long as the `Unique` exists.
//!
//! - Safe creation through the `unique!` macro: This macro provides a
//!   convenient and safe way to create `Unique` pointers, ensuring that the
//!   data is pinned and owned uniquely by the pointer.
//!
//! - Utility methods: The module provides essential methods for working with
//!   `Unique` pointers, including creating new unique pointers
//!   (`new_unchecked`), mapping to another unique pointer (`map_unchecked`),
//!   and converting to and from raw pointers (`from_raw` and `into_raw`).

use core::{
	ops::{Deref, DerefMut},
	pin::Pin,
	ptr::NonNull,
};

/// An owning *unique* pointer to some data.
///
/// This pointer-type behaves as if it owned the referenced memory by
/// disallowing (safe) duplication of its contents for as long as it lives.
/// It is possible to access the contained pointer through mutable or shared
/// referencing, but clone or copy is implemented and the pointer cannot be
/// moved out of the type without its destruction.
///
/// The simulation framework uses this type to prevent accidental mis-use of
/// active objects by requiring that the activation routines, e.g.
/// [Sim::activate](crate::Sim::activate) be given a reference that uniquely
/// refers to the object being scheduled. That prevents double-activations
/// as well as ensures that the initial [Irc](crate::Irc) track their reference
/// counts correctly.
///
/// # Examples
/// Correct example:
/// ```
/// # use odem_rs_core::*;
/// async fn sim_main(sim: &Sim) {
///     let fiber = unique!(Fiber::new(async {}));
///     sim.activate(fiber);
/// }
/// ```
/// Prevented mis-use:
/// ```compile_fail,E0382
/// # use odem_rs_core::*;
/// async fn sim_main(sim: &Sim) {
///     let fiber = unique!(Fiber::new(async {}));
///     sim.activate(fiber);
///     // later...
///     sim.activate(fiber);
/// }
/// ```
#[repr(transparent)]
pub struct Unique<'p, T: ?Sized> {
	#[doc(hidden)]
	pub pointer: Pin<&'p mut T>,
}

impl<'p, T: ?Sized> Unique<'p, T> {
	/// Creates a new, unique pointer.
	///
	/// # Safety
	/// The caller has to ensure that the exclusive `T` is the *only* pointer
	/// available for the whole lifetime of the referent.
	///
	/// Local `Unique` pointers can be safely created using the [unique!]-macro.
	///
	/// [unique!]: crate::unique
	#[inline]
	pub unsafe fn new_unchecked(pointer: Pin<&'p mut T>) -> Self {
		Unique { pointer }
	}

	/// Returns the inner pointer.
	///
	/// Calling this method revokes the guarantees of uniqueness, since pinned
	/// pointers may be duplicated, e.g. using [Pin::as_mut].
	#[inline]
	pub fn into_inner(self) -> Pin<&'p mut T> {
		self.pointer
	}

	/// Maps the unique, pinned inner pointer to another unique, pinned pointer
	/// using a closure.
	///
	/// # Safety
	/// The supplied closure has to return a pin-projected result of the same
	/// allocation that it's called on.
	pub unsafe fn map_unchecked<F, R>(self, f: F) -> Unique<'p, R>
	where
		F: FnOnce(Pin<&'p mut T>) -> Pin<&'p mut R>,
	{
		// SAFETY: projecting to a member of a unique pointer also yields
		// a unique pointer, since uniqueness (like ownership) is transitive
		Unique::new_unchecked(f(self.into_inner()))
	}

	/// Converts a [NonNull] raw-pointer into a `Unique` pinned pointer.
	///
	/// # Safety
	/// The user is responsible to only call this function with the result
	/// obtained by [Self::into_raw] and only at most once.
	#[inline]
	pub unsafe fn from_raw(mut pointer: NonNull<T>) -> Self {
		Unique {
			pointer: Pin::new_unchecked(pointer.as_mut()),
		}
	}

	/// Converts the `Unique` pointer into a [NonNull] raw-pointer.
	///
	/// This method is safe because the uniqueness of the pointer is only
	/// assured for as long as the `Unique` lives, and it gets moved into this
	/// function.
	#[inline]
	pub fn into_raw(self) -> NonNull<T> {
		NonNull::from(unsafe { self.pointer.get_unchecked_mut() })
	}

	/// Creates a raw pointer to the exclusive reference contained inside the
	/// `Unique`.
	///
	/// This is safe but will invalidate the mutable reference and every
	/// pointer derived from it once it is being dereferenced. Only use with
	/// extreme care!
	#[inline]
	pub fn as_ptr(&self) -> NonNull<T> {
		// SAFETY: the unique is a pinned, exclusive reference to a T, so to
		// transmute it into a NonNull<T> is possible
		unsafe { core::ptr::read(self as *const Self as *const NonNull<T>) }
	}
}

impl<'p, T: ?Sized> Deref for Unique<'p, T> {
	type Target = Pin<&'p mut T>;

	#[inline]
	fn deref(&self) -> &Self::Target {
		&self.pointer
	}
}

impl<T: ?Sized> DerefMut for Unique<'_, T> {
	#[inline]
	fn deref_mut(&mut self) -> &mut Self::Target {
		&mut self.pointer
	}
}

/// Safely creates a local [Unique] pointer that owns the pinned data it
/// contains.
#[macro_export]
macro_rules! unique {
	($value:expr $(,)?) => {
		$crate::ptr::Unique {
			pointer: core::pin::pin!($value),
		}
	};
}
