//! This module contains the heart of the library: the [Simulator] and the
//! [simulation context].
//!
//! Combined, these two structures implement *executor* and *spawner* for the
//! *future*-based implementation of coroutines. Conceptually, the executor is
//! responsible for selecting the next future to be executed and the spawner
//! allows new futures to be injected into the executor.
//!
//! In our implementation, the executor is realized by the [Simulator], the
//! spawner is realized by the [simulation context] and futures are realized
//! through [Tasks](Task).
//!
//! [simulation context]: Sim

use core::{
	any::Any,
	cell::Cell,
	fmt,
	future::Future,
	num::NonZero,
	ops::Add,
	pin::pin,
	ptr::NonNull,
	task::{Context, RawWaker, RawWakerVTable, Waker},
};

use crate::{
	calendar::{ordering, Calendar},
	config::{Config, DefaultConfig, Time},
	erased,
	error::{Causality, Deadlock, Error, Incomplete, NotBorn},
	fiber::Fiber,
	fsm::*,
	ptr::{IntrusivelyCounted, Irc, IrcBox, IrcBoxed},
	task::{token, Puck as TaskPuck, Share, Task},
	unique, Active, Puck, Unique,
};

mod ops;

/// The type used to mark the processes of incoming tasks.
///
/// This is used to prevent interleaved execution of tasks from different
/// processes by enforcing the following property: all tasks belonging to some
/// process `A` are sorted before any tasks belonging to process  `B` iff any
/// task of process `A` is scheduled before all the tasks of process `B`.
pub type Mark = ordering::EpochCounter<u32>;

/// The precedence type for ordering the tasks that are executed as part of the
/// same process.
///
/// The way this type works is by partitioning an integer range (say of `usize`)
/// into a set of non-overlapping intervals, each represented by the first
/// integer that is part of the respective interval. Initially, the interval
/// contains the whole range (`..`) but can be refined at every level into
/// smaller sub-intervals that are sorted according to their first integer and
/// nesting level. This has the effect of allowing sub-tasks to inherit the
/// precedence of their parent and to recursively allow the same for their
/// children. Since the number of bits is finite, this is bound to fail at some
/// level of recursion, but it should not place undue limitations on the user.
pub type Prec = ordering::Partition<u32>;

/* ****************************************************** Simulation Function */

/// Simplified function that creates a new [default] configuration, runs the
/// provided `async` closure to completion and returns it.
///
/// Use it in situations when both input and output of the simulation is
/// communicated through a configuration initialized to a default value.
///
/// [default]: Default
pub fn simulation<C, F>(main: F) -> Result<C, Deadlock>
where
	C: Config + Default,
	F: for<'s> SimMainClosure<'s, C, Output = ()>,
{
	let sim = Simulator::default();
	sim.run(main).map(move |()| sim.into_inner())
}

/* **************************************************************** Simulator */

/// This trait abstracts over any callable entity (closures or functions) that
/// can be invoked with a reference to a [simulation context] and returns a
/// future that may reference it.
///
/// It is needed until [async closures] become stable which should happen on the
/// 20th of february 2025 and makes this trait obsolete.
///
/// [simulation context]: Sim
/// [async closures]: https://rust-lang.github.io/rfcs/3668-async-closures.html
pub trait SimMainClosure<'s, C: Config> {
	/// The future returned by the implementor. It must implement [`Future`]
	/// can borrow the simulation context.
	type Fut: Future<Output = Self::Output> + 's;

	/// The output type produced by the future `Fut`.
	type Output;

	/// Invokes the closure with the given simulation context.
	fn call(self, sim: &'s Sim<C>) -> Self::Fut;
}

// blanket implementation for callables and closures
impl<'s, C, F, R> SimMainClosure<'s, C> for F
where
	C: Config,
	F: FnOnce(&'s Sim<C>) -> R,
	R: Future + 's,
{
	type Fut = R;
	type Output = R::Output;

	#[inline]
	fn call(self, sim: &'s Sim<C>) -> Self::Fut {
		self(sim)
	}
}

/// The principal driver of a simulation run which can be initialized using a
/// [configuration].
///
/// [configuration]: Config
#[derive(Copy, Clone, Default)]
pub struct Simulator<C>(C);

impl<C> Simulator<C> {
	/// Creates a new simulator from a configuration.
	pub const fn new(config: C) -> Self {
		Self(config)
	}

	/// Converts the simulator back into a configuration.
	pub fn into_inner(self) -> C {
		self.0
	}
	
	/// Grants shared access to the inner configuration.
	pub fn inner(&self) -> &C {
		&self.0
	}
	
	/// Grants mutable access to the inner configuration.
	pub fn inner_mut(&mut self) -> &mut C {
		&mut self.0
	}
}

impl<C: Config> Simulator<C> {
	/// Executes the simulation model provided in the closure and returns its
	/// result in the `Ok`-variant of a [Result].
	///
	/// The `Err`-variant is reserved for errors that occur during the
	/// event-loop. Only a [Deadlock] can be reported.
	#[track_caller]
	pub fn run<F, R>(&self, main: F) -> Result<R, Deadlock>
	where
		// TODO: replace with `F: AsyncFnOnce(&Sim<C>) -> R` once stable
		F: for<'s> SimMainClosure<'s, C, Output = R>,
	{
		// Create a fresh simulation context.
		// SAFETY: the reference to `self` outlives the simulation context.
		// (done to prevent mentioning the lifetime in the simulation context)
		let sim = unique!(unsafe { Sim::new(&self.0) });

		// Wrap the simulation context into a reference-counted `Irc`.
		let sim = Irc::from(sim);

		// Register the simulation context thread-locally to support tracing.
		// The hook will take care of de-registration in its destructor.
		let _hook = erased::hook(&*sim);

		// Bootstrap the root process.
		let share = pin!(Share::root(sim.clone()));
		let share = share.into_ref();
		let span = share.enter_span();
		let root = unique!(Fiber::new(main.call(&sim)));
		let mut puck = root.boot(share);
		drop(span);
		let task = puck.own();

		// Schedule the root fiber at the current model-time.
		task.brand(|task, once| {
			let idle = task.token(once).into_idle().unwrap();
			sim.calendar().activate(task.clone(), idle);
		});

		// Run the simulator until the root fiber terminates.
		{
			let waker = ShallowWaker::new(sim.clone());
			// SAFETY: the `ShallowWaker` is valid for the lifetime of the waker
			let waker = unsafe { waker.as_waker() };
			let mut cx = Context::from_waker(&waker);
			let root = puck.as_ref();

			// Ensure that the active slot is cleared on exit.
			scopeguard::defer! { sim.active.set(None); }

			loop {
				if let Some(task) = sim.calendar.extract() {
					let res = task.brand(|task, once| {
						let next = task.token(once).into_next().unwrap();
						let _span = task.branded_share(&next).enter_span();
						let busy: token::Busy<'_> = task.state().transition(next, ());

						// set the active task
						sim.active
							.set(Some(TaskPuck::new(task.clone(), busy.as_init())));

						// poll the task and check whether the root task terminated
						task.poll(busy, &mut cx).is_ready() && core::ptr::eq(task.detach(), root)
					});

					if res {
						break;
					}
				} else {
					return Err(Deadlock);
				}
			}
		}

		// Extract and return the result.
		Ok(puck.result().unwrap())
	}

	/// Executes the simulation model provided in the closure, that has to
	/// return a result of type [Result], and returns its result.
	///
	/// The result of simulation run can also be a [Deadlock] if the scheduler
	/// detects that it is not possible to schedule any more tasks after some
	/// model-time has passed.
	///
	/// [Deadlock]: Error::Deadlock
	#[track_caller]
	pub fn with_result<F, T, E>(&self, main: F) -> Result<T, E>
	where
		// TODO: replace with `F: AsyncFnOnce(&Sim<C>) -> R` once stable
		F: for<'s> SimMainClosure<'s, C, Output = Result<T, E>>,
		E: From<Deadlock>,
	{
		self.run(main).unwrap_or_else(|err| Err(E::from(err)))
	}

	// #[track_caller]
	// pub fn steps<F,R>(&self, main: F) -> Steps<'_, C, R>
	// where
	// 	F: for<'s> SimMainClosure<'s, C, Output = R>,
	// {
	//
	// }
}

impl<C> From<C> for Simulator<C> {
	fn from(config: C) -> Self {
		Self::new(config)
	}
}

/* ******************************************************* Simulation Context */

/// Simulation context that can be used to advance the [model-time], spawn
/// additional [Processes] or [Fibers], access [global data], the [current]
/// model-time and the [active] Fiber, among other things.
///
/// It can only be accessed during a simulation run.
///
/// [Processes]: crate::Process
/// [Fibers]: Fiber
/// [global data]: Config::Data
/// [model-time]: Config::Time
/// [current]: Sim::now
/// [active]: Sim::active
pub struct Sim<C: ?Sized + Config = DefaultConfig> {
	/// The event calendar.
	calendar: Calendar<C>,
	/// The currently active task.
	active: Cell<Option<TaskPuck<C>>>,
	/// The inner box to enable intrusive reference counting.
	sim_box: IrcBox<SimBox>,
	/// A global counter for unique process-IDs.
	#[cfg(not(feature = "alloc"))]
	pid_gen: Cell<NonZero<usize>>,
	/// A per-process-type counter for unique process-IDs.
	#[cfg(feature = "alloc")]
	pid_gen: core::cell::RefCell<hashbrown::HashMap<core::any::TypeId, NonZero<usize>>>,
	/// The thread id of the thread currently executing this simulation.
	///
	/// This is used to perform a runtime check on whether a `Waker` has been
	/// sent to another thread and attempted to wake from there. It is necessary
	/// because `Waker`s are `Send + Sync` but our executor doesn't support it.
	///
	/// Will be made obsolete by [local wakers].
	///
	/// [local wakers]: https://github.com/rust-lang/rust/issues/118959
	#[cfg(feature = "std")]
	thread_id: std::thread::ThreadId,
	/// Pointer to the user-supplied configuration.
	config: NonNull<C>,
}

impl<C: Config> Sim<C> {
	/// Creates a new simulation context from a [configuration](Config).
	///
	/// # Safety
	/// The caller has to ensure that the reference to the configuration
	/// outlives the created simulation context.
	///
	/// This could also be solved via the borrow checker but would require the
	/// introduction of an additional lifetime which makes referring to the
	/// simulation context in function signatures quite a bit less ergonomic.
	#[inline]
	unsafe fn new(config: &C) -> Self {
		#[cfg(not(feature = "alloc"))]
		{
			Sim {
				calendar: Calendar::new(config),
				active: Cell::new(None),
				sim_box: IrcBox::new(SimBox::new()),
				pid_gen: Cell::new(NonZero::<usize>::MIN),
				#[cfg(feature = "std")]
				thread_id: std::thread::current().id(),
				config: NonNull::from(config),
			}
		}

		#[cfg(feature = "alloc")]
		{
			Sim {
				calendar: Calendar::new(config),
				active: Cell::new(None),
				sim_box: IrcBox::new(SimBox::new()),
				pid_gen: core::cell::RefCell::new(hashbrown::HashMap::new()),
				#[cfg(feature = "std")]
				thread_id: std::thread::current().id(),
				config: NonNull::from(config),
			}
		}
	}

	/// Gets a [simulation context] from a waker.
	///
	/// This can be used to recover the current simulation context from the
	/// context argument if the configuration is known, making it possible to
	/// distribute the simulation context by piggybacking on Rust's implicit
	/// [Context].
	///
	/// [simulation context]: Sim
	pub fn from_context(context: &Context<'_>) -> Option<Irc<Self>> {
		let waker = context.waker();

		// comparing for pointer equality is sufficient because there is only
		// one vtable for all simulators
		if core::ptr::eq(waker.vtable(), &SIM_VTABLE) {
			// SAFETY: it's a shallow waker, so it is safe to reconstruct the
			// simulation context
			let waker = unsafe { &*(waker.data() as *const ShallowWaker) };
			waker.0.clone().downcast::<Self>().ok()
		} else {
			// none of our wakers
			None
		}
	}
}

impl<C: ?Sized + Config> Sim<C> {
	/// Removes the currently active task.
	#[inline]
	pub(crate) fn unslot<'brand>(
		&self,
		task: &Task<'brand, C>,
		busy: token::Busy<'brand>,
	) -> token::Idle<'brand> {
		let _puck = self.active.take();

		// sanity check in debug mode
		debug_assert!(
			_puck.is_some_and(|puck| puck.is_same(task)),
			"unslotted task in state `Busy` was not in the active slot"
		);

		task.state().transition(busy, ())
	}

	/// Returns the thread-id of the thread running the simulation.
	#[cfg(feature = "std")]
	#[inline]
	pub(crate) fn get_thread_id(&self) -> std::thread::ThreadId {
		self.thread_id
	}

	/// Returns a shared reference to the configuration.
	#[inline]
	pub fn config(&self) -> &C {
		unsafe { self.config.as_ref() }
	}

	/// Returns a reference to the shared data.
	#[inline]
	pub fn global(&self) -> &C::Data {
		self.config().global_data()
	}

	/// Returns the [Puck] of the currently active task.
	#[inline]
	pub fn active(&self) -> Option<TaskPuck<C>> {
		let active = self.active.take();
		let result = active.clone();
		self.active.set(active);
		result
	}

	/// Returns the current model time.
	#[inline]
	pub fn now(&self) -> C::Time {
		self.calendar().now()
	}

	/// Schedules the currently active task at a later model time.
	///
	/// Panics if no task is currently active or if the proposed model time has
	/// already passed. For non-panicking version of this method, see
	/// [`try_advance`].
	///
	/// [`try_advance`]: Self::try_advance
	#[inline]
	pub fn advance(&self, dt: impl Into<C::Time>) -> impl Future<Output = ()> + '_
	where
		C::Time: Add<Output = C::Time>,
	{
		ops::TryAdvance::new(self, dt.into()).unwrap()
	}

	/// Schedules the currently active task at a later model time.
	///
	/// This method has the same function as [`advance`] but takes an absolute
	/// time-point for the proposed reactivation time rather than a relative
	/// one.
	///
	/// For a non-panicking alternative, see [`try_advance_to`].
	///
	/// [`advance`]: Self::advance
	/// [`try_advance_to`]: Self::try_advance_to
	#[inline]
	pub fn advance_to(&self, time: impl Into<C::Time>) -> impl Future<Output = ()> + '_ {
		ops::TryAdvanceTo::new(self, time.into()).unwrap()
	}

	/// Attempts to advance the currently active task to a later model time.
	///
	/// Fails with an `Err`-result if either no task is active or if the
	/// proposed model time has already passed.
	#[inline]
	pub fn try_advance(
		&self,
		dt: impl Into<C::Time>,
	) -> impl Future<Output = Result<(), Causality<C::Time>>> + '_
	where
		C::Time: Add<Output = C::Time>,
	{
		ops::TryAdvance::new(self, dt.into())
	}

	/// Attempts to advance the currently active task to a later model time.
	///
	/// This method has the same function as [`try_advance`] but takes an
	/// absolute time-point for the proposed reactivation time rather than a
	/// relative one.
	///
	/// [`try_advance`]: Self::try_advance
	#[inline]
	pub fn try_advance_to(
		&self,
		time: impl Into<C::Time>,
	) -> impl Future<Output = Result<(), Causality<C::Time>>> + '_ {
		ops::TryAdvanceTo::new(self, time.into())
	}

	/// Activates a new task by binding it to the simulation context,
	/// scheduling it at the current model time, and returning a [Puck] to it.
	///
	/// Panics if the task has already been activated. For a non-panicking
	/// version of this method, see [`try_activate`].
	///
	/// [`try_activate`]: Self::try_activate
	#[track_caller]
	#[inline]
	pub fn activate<'a, A>(&'a self, actions: Unique<'a, A>) -> A::Puck<'a>
	where
		A: Active<C>,
	{
		self.try_activate(actions).unwrap_or_else(|err| {
			#[cfg(feature = "tracing")]
			tracing::error!(%err);
			panic!("{}", err);
		})
	}

	/// Activates a new task by binding it to the simulation context,
	/// scheduling it at a future model time, and returning a [Puck] to it.
	///
	/// Panics if the task has already been activated or if the proposed model
	/// time for activation has passed. For a non-panicking version of this
	/// method, see [`try_schedule`].
	///
	/// [`try_schedule`]: Self::try_schedule
	#[track_caller]
	#[inline]
	pub fn schedule<'a, A>(&'a self, actions: Unique<'a, A>, dt: impl Into<C::Time>) -> A::Puck<'a>
	where
		A: Active<C>,
		C::Time: Add<Output = C::Time>,
	{
		self.try_schedule(actions, dt).unwrap_or_else(|err| {
			#[cfg(feature = "tracing")]
			tracing::error!(%err);
			panic!("{}", err);
		})
	}

	/// Attempts to activate a new task by binding it to the simulation context,
	/// scheduling it at the current model time, and returning a [Puck] to it.
	///
	/// Doesn't panic if the task has already been activated before, but returns
	/// an `Err`-result instead.
	pub fn try_activate<'a, A>(&'a self, actions: Unique<'a, A>) -> Result<A::Puck<'a>, Error>
	where
		A: Active<C>,
	{
		// bind the simulation context
		let puck = unsafe { self.bind(actions)? };

		// activate the puck
		puck.own().brand(|task, once| {
			// successfully bound tasks are in state 'Idle'
			let idle = task.token(once).into_idle()?;
			
			// set the correct span
			let _span = task.branded_share(&idle).enter_span();
			
			// schedule the puck
			self.calendar().activate(task.clone(), idle);
			
			Ok(puck)
		})
	}

	/// Attempts to activate a new task by binding it to the simulation context,
	/// scheduling it at a later model time, and returning a [Puck] to it.
	///
	/// Doesn't panic if the task has already been activated before or if the
	/// proposed model time has already passed, but returns an `Err`-result
	/// instead.
	pub fn try_schedule<'a, A>(
		&'a self,
		actions: Unique<'a, A>,
		dt: impl Into<C::Time>,
	) -> Result<A::Puck<'a>, Error>
	where
		A: Active<C>,
		C::Time: Add<Output = C::Time>,
	{
		use core::cmp::Ordering::*;
		let time: C::Time = self.now() + dt.into();

		match time.partial_cmp(&self.now()) {
			None | Some(Less) => Err(Error::from(Causality {
				cause: self.now(),
				effect: time,
			})),
			Some(order) => {
				// bind the simulation context
				let puck = unsafe { self.bind(actions)? };

				// schedule or activate the puck
				puck.own().brand(|task, once| {
					// successfully bound tasks are in state 'Idle'
					let idle = task.token(once).into_idle()?;

					// set the correct span
					let _span = task.branded_share(&idle).enter_span();

					// schedule the puck
					let calendar = self.calendar();

					let irc = task.clone();
					if order == Greater {
						calendar.schedule(irc, idle, time);
					} else {
						calendar.activate(irc, idle);
					}

					Ok(puck)
				})
			}
		}
	}

	/// Returns an awaitable [Future] that will suspend the currently executed
	/// task and reschedule it after every other task that is scheduled at the
	/// current model time had a chance to run, independent of its [Rank].
	///
	/// [Rank]: crate::config::Rank
	#[inline]
	pub fn defer(&self) -> impl Future<Output = ()> + '_ {
		ops::Defer::new(self)
	}

	/// Generates a new, simulation-run-unique process id.
	#[inline]
	pub(crate) fn pid_gen<I: 'static>(&self) -> NonZero<usize> {
		#[cfg(not(feature = "alloc"))]
		// simply increment the global process counter to obtain a new pid
		{
			let next = self.pid_gen.get();
			self.pid_gen
				.set(next.checked_add(1).unwrap_or(NonZero::<usize>::MIN));
			next
		}

		#[cfg(feature = "alloc")]
		// look-up the process-type-specific pid to increment
		{
			let mut pid_gen = self.pid_gen.borrow_mut();
			let counter = pid_gen
				.entry(core::any::TypeId::of::<I>())
				.or_insert(NonZero::<usize>::MIN);
			let next = *counter;
			*counter = next.checked_add(1).unwrap_or(NonZero::<usize>::MIN);
			next
		}
	}

	/// Returns a reference to the [event calendar](DefaultPlan).
	#[inline]
	pub(crate) fn calendar(&self) -> &Calendar<C> {
		&self.calendar
	}

	/// Activates the active object, returning a [Puck] on success and an
	/// [Error] if the operation failed.
	///
	/// This is slightly more complicated than simply calling the [bind]-method
	/// on the [Active]-trait because we also need to store the *actual* type
	/// of the instance being activated in order to allow dynamic dispatching
	/// to it later.
	///
	/// # Safety
	/// The caller has to ensure that the simulation context remains valid
	/// at least until the active object expires.
	///
	/// [bind]: Active::bind
	#[inline]
	unsafe fn bind<'a, A>(&'a self, actions: Unique<'a, A>) -> Result<A::Puck<'a>, NotBorn>
	where
		A: Active<C>,
	{
		// create a raw reference to the pinned task;
		// this will be the most specialized version, enabling dynamic dispatch
		let hook = actions.as_ptr();

		// create a reference to the context of the currently active object
		let active = unsafe { &*self.active.as_ptr() }.as_ref().unwrap();
		let share = NonNull::from(active.share());

		// derive the puck from it
		Active::bind(actions, hook, share)
	}
}

impl<C: Config> fmt::Debug for Sim<C> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		f.debug_tuple("Sim").field(&self.now()).finish()
	}
}

impl<C: Config> erased::Sim for Sim<C> {
	fn active(&self) -> Option<erased::TaskPuck> {
		self.active().map(Into::into)
	}

	fn now(&self) -> &dyn fmt::Display {
		// SAFETY: repr(transparent) guarantees that the two structures are laid
		// out identically in memory so that this transmute is safe.
		unsafe { core::mem::transmute::<&Sim<C>, &DisplayHelper<C>>(self) }
	}

	fn global(&self) -> &dyn Any {
		self.global()
	}

	fn waker(&self) -> RawWaker {
		self.active().expect("active task").into_waker()
	}

	fn defer(&self) {
		if let Some(active) = self.active() {
			active.0.brand(|active, once| {
				let busy = active
					.token(once)
					.into_busy()
					.expect("active task is 'Busy'");
				self.calendar().defer(active.clone(), busy);
			});
		}
	}
}

/// *Newtype* adapter that implements [Display](fmt::Display) for a simulation
/// context by displaying the current model-time.
#[repr(transparent)]
struct DisplayHelper<C: Config>(Sim<C>);

impl<C: Config> fmt::Display for DisplayHelper<C> {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		self.0.now().format(f)
	}
}

unsafe impl<C: ?Sized + Config> IntrusivelyCounted for Sim<C> {
	type Inner = SimBox;

	fn irc_box(&self) -> &IrcBox<Self::Inner> {
		&self.sim_box
	}
}

/* ************************************************************** Sim Irc Box */

pub struct SimBox {
	/// Contains a counter for the number of references to the outer [Sim].
	refs: Cell<usize>,
}

impl SimBox {
	const fn new() -> Self {
		Self { refs: Cell::new(0) }
	}
}

unsafe impl IrcBoxed for SimBox {
	unsafe fn acquire(&self) -> usize {
		let refs = self.refs.get() + 1;
		self.refs.set(refs);
		refs
	}

	unsafe fn release(&self) -> Option<fn(NonNull<Self>)> {
		self.refs.set(self.refs.get() - 1);
		None
	}
}

/* ****************************************************** Simulation Iterator */

// /// An iterator-like adapter to allow [Simulators](Simulator) to be executed
// /// step-wise.
// #[pin_project::pin_project]
// pub struct Steps<'s, C: ?Sized + Config, F: Future> {
// 	/// A reference to the simulator.
// 	sim: &'s Sim<C>,
// 	/// The root fiber for the simulation run.
// 	#[pin]
// 	root: Fiber<C,F>,
// 	/// Associated shared data of the root fiber.
// 	share: Share<C,()>,
// }
//
// impl<'s, C: Config, F: Future + 's> Steps<'s,C,F> {
// 	/// Creates a new simulation iterator from a simulator and a root fiber.
// 	#[inline]
// 	fn new(sim: &'s Sim<C>, root: Fiber<C,F>) -> Self {
// 		// reset the active task and the simulator
// 		sim.active.set(None);
// 		sim.reset();
//
// 		// construct the iterator
// 		Steps { sim, root, share: unsafe { Share::root(sim) } }
// 	}
//
// 	#[inline]
// 	pub fn iter(self: Pin<&mut Self>) -> Iter<C> {
// 		self.into_iter()
// 	}
//
// 	#[inline]
// 	pub fn result(&self) -> Result<F::Output, Error> {
// 		// extract the result (or handle the error)
// 		self.root.result().map_err(|_|
// 			if self.sim.calendar().is_empty() && self.sim.active().is_none() {
// 				Deadlock.into()
// 			} else {
// 				Incomplete.into()
// 			}
// 		)
// 	}
// }
//
// impl<'s, C: Config, F: Future> IntoIterator for Pin<&'s mut Steps<'_,C,F>> {
// 	type Item = TaskPuck<C>;
// 	type IntoIter = Iter<'s,C>;
//
// 	#[inline]
// 	fn into_iter(self) -> Self::IntoIter {
// 		let projection = self.project();
//
// 		// activate the root process
// 		let root = unsafe {
// 			Unique::new_unchecked(projection.root)
// 		}.brand(|fiber, once| {
// 			let born = fiber.proj_ref().proj_ref().state(once).to_born().expect("not in state `Born`");
// 			fiber.boot(born, projection.share)
// 		}).as_task_puck();
//
// 		root.0.brand(|task, once| {
// 			let idle = unsafe { token::Idle::new(once) };
// 			projection.sim.calendar().activate(task.clone(), idle);
// 		});
//
// 		// construct the actual iterator
// 		Iter { sim: projection.sim, root }
// 	}
// }
//
// /// An iterator for a single simulation-run.
// pub struct Iter<'s, C: ?Sized + Config> {
// 	/// A reference to the underlying simulator.
// 	sim: &'s Sim<C>,
// 	/// The [Puck] of the root fiber.
// 	root: TaskPuck<C>,
// }
//
// impl<'s, C: Config> Iterator for Iter<'s,C> {
// 	type Item = TaskPuck<C>;
//
// 	fn next(&mut self) -> Option<Self::Item> {
// 		// check if there is a task active currently
// 		if let Some(puck) = self.sim.active() {
// 			// initialize the future context
// 			let waker = ShallowWaker::waker(self.sim);
// 			let mut cx = Context::from_waker(&waker);
//
// 			// set the global simulator
// 			let _guard = erased::hook(self.sim);
// 			let is_root = puck == self.root;
//
// 			let ready = puck.0.brand(move |task, once| {
// 				// we know that the task in the `active` slot is in state `Busy`
// 				let busy = unsafe { token::Busy::new(once) };
//
// 				// poll the task and return its `ready` state
// 				task.proj_ref().poll(busy, &mut cx).is_ready()
// 			});
//
// 			// clear the active task
// 			self.sim.active.set(None);
//
// 			// if the task is ready, and it's the root, exit the loop
// 			if ready && is_root { return None; }
// 		}
//
// 		// set and return the next active task if there is one
// 		self.sim.calendar().extract(|task, busy| {
// 			// load the task into the active slot
// 			self.sim.active.set(Some(TaskPuck::new(task.clone(), busy.as_init())));
// 			// initialize the task puck
// 			TaskPuck::new(task, busy.as_init())
// 		}).map(|(_, puck)| puck)
// 	}
// }

/* ************************************************************ Shallow Waker */

/// Internal waker that can only be used to create a copy of the currently
/// active task or to defer it.
struct ShallowWaker(Irc<dyn erased::Sim>);

/// Virtual function table for the [simulation executor](Simulator), independent
/// of the [configuration](Config).
///
/// This construction ensures that there is a globally unique address for the
/// simulators, allowing us to restore the [simulation context](Sim) from the
/// context and the configuration.
static SIM_VTABLE: RawWakerVTable = RawWakerVTable::new(
	ShallowWaker::clone,
	ShallowWaker::nop,
	ShallowWaker::wake_by_ref,
	ShallowWaker::nop,
);

impl ShallowWaker {
	/// Creates a [Waker] for a simulator that can be cloned into [Puck]
	/// wakers. Waking this waker has the effect of deferring the currently
	/// active task, i.e. the task is reawakened at the current model time
	/// after all the other tasks have run.
	fn new<C: Config>(sim: Irc<Sim<C>>) -> Self {
		// erase the concrete configuration
		Self(Irc::map(sim, |inner| inner as &dyn erased::Sim))
	}

	/// Constructs a [Waker] from the reference of a shallow waker.
	///
	/// # Safety
	/// The caller is responsible to ensure that the reference to the shallow
	/// waker is valid for the lifetime of the waker.
	unsafe fn as_waker(&self) -> Waker {
		let waker = RawWaker::new(self as *const Self as *const (), &SIM_VTABLE);
		Waker::from_raw(waker)
	}

	/// No-op waker-function.
	fn nop(_: *const ()) {}

	/// Constructs a [RawWaker] for the active process.
	unsafe fn clone(this: *const ()) -> RawWaker {
		// perform a dynamic dispatch to recover the configuration
		(*(this as *const Self)).0.waker()
	}

	/// Defers the currently active process.
	unsafe fn wake_by_ref(this: *const ()) {
		// perform a dynamic dispatch to recover the configuration
		(*(this as *const Self)).0.defer();
	}
}
