//! Module for basic scheduling operations implemented in a space-optimized way.

use core::{
	fmt::Display,
	future::Future,
	ops::Add,
	pin::Pin,
	task::{Context, Poll},
};

use crate::{config::Config, error::Causality, fsm::*};

use super::Sim;

/* *********************************************************** Try Advance To */

/// Structure that serves as a future that attempts to schedule the calling
/// task at a later, fixed model-time.
pub(super) struct TryAdvanceTo<'s, C: ?Sized + Config> {
	/// The simulation context. Doubles as a state since it is extracted after
	/// the first call.
	sim: Option<&'s Sim<C>>,
	/// The fixed model-time to schedule the calling task for.
	when: C::Time,
}

impl<'s, C: ?Sized + Config> TryAdvanceTo<'s, C> {
	/// Instantiates the structure.
	#[inline]
	pub(super) const fn new(sim: &'s Sim<C>, when: C::Time) -> Self {
		TryAdvanceTo {
			sim: Some(sim),
			when,
		}
	}

	/// Unwraps the future, panicking and printing an error message if
	/// unsuccessful and returning the unwrapped result otherwise.
	#[inline]
	pub(super) const fn unwrap(self) -> Unwrap<Self> {
		Unwrap(self)
	}
}

impl<C: ?Sized + Config> Future for TryAdvanceTo<'_, C> {
	type Output = Result<(), Causality<C::Time>>;

	fn poll(mut self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<Self::Output> {
		use core::cmp::Ordering::*;

		if let Some(sim) = self.sim.take() {
			sim.active()
				.map(|active| {
					active.0.brand(|task, once| {
						let busy = task
							.token(once)
							.into_busy()
							.expect("active task should be in state 'Busy'");
						let idle = sim.unslot(&task, busy);
						let now = sim.now();
						let calendar = sim.calendar();
						let when = self.when;

						match when.partial_cmp(&now) {
							None | Some(Less) => Poll::Ready(Err(Causality {
								cause: now,
								effect: when,
							})),
							Some(Greater) => {
								calendar.schedule(task.clone(), idle, when);
								Poll::Pending
							}
							Some(Equal) => {
								calendar.activate(task.clone(), idle);
								Poll::Pending
							}
						}
					})
				})
				.expect("active task")
		} else {
			Poll::Ready(Ok(()))
		}
	}
}

/* ************************************************************** Try Advance */

/// Structure that serves as a future that attempts to schedule the calling
/// task at a later, relative model time.
pub(super) struct TryAdvance<'s, C: ?Sized + Config> {
	/// The simulation context. Doubles as a state since it is extracted after
	/// the first call.
	sim: Option<&'s Sim<C>>,
	/// A model-time relative to the binding time to schedule the calling task
	/// for.
	dt: C::Time,
}

impl<'s, C: ?Sized + Config> TryAdvance<'s, C> {
	/// Instantiates the structure.
	#[inline]
	pub(super) const fn new(sim: &'s Sim<C>, dt: C::Time) -> Self {
		TryAdvance { sim: Some(sim), dt }
	}

	/// Unwraps the future, panicking and printing an error message if
	/// unsuccessful and returning the unwrapped result otherwise.
	#[inline]
	pub(super) const fn unwrap(self) -> Unwrap<Self> {
		Unwrap(self)
	}
}

impl<C: ?Sized + Config> Future for TryAdvance<'_, C>
where
	C::Time: Add<Output = C::Time>,
{
	type Output = Result<(), Causality<C::Time>>;

	#[inline]
	fn poll(mut self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<Self::Output> {
		use core::cmp::Ordering::*;

		if let Some(sim) = self.sim.take() {
			sim.active()
				.map(|active| {
					active.0.brand(|task, once| {
						let busy = task
							.token(once)
							.into_busy()
							.expect("active task should be in state 'Busy'");
						let idle = sim.unslot(&task, busy);
						let now = sim.now();
						let calendar = sim.calendar();
						let when: C::Time = now + self.dt;

						match when.partial_cmp(&now) {
							None | Some(Less) => Poll::Ready(Err(Causality {
								cause: now,
								effect: when,
							})),
							Some(Greater) => {
								calendar.schedule(task.clone(), idle, when);
								Poll::Pending
							}
							Some(Equal) => {
								calendar.activate(task.clone(), idle);
								Poll::Pending
							}
						}
					})
				})
				.expect("no active task")
		} else {
			Poll::Ready(Ok(()))
		}
	}
}

/* ******************************************************************** Defer */

/// Structure that serves as a future that can be awaited in order to defer
/// execution for the currently active task.
pub(super) struct Defer<'s, C: ?Sized + Config> {
	/// A simulation context before polling or `None` after the first poll.
	sim: Option<&'s Sim<C>>,
}

impl<'s, C: ?Sized + Config> Defer<'s, C> {
	/// Initializes a new defer-future.
	#[inline]
	pub(super) const fn new(sim: &'s Sim<C>) -> Self {
		Defer { sim: Some(sim) }
	}
}

impl<'s, C: ?Sized + Config> Future for Defer<'s, C> {
	type Output = ();

	#[inline]
	fn poll(mut self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<Self::Output> {
		match self.sim.take() {
			Some(sim) => {
				if let Some(active) = sim.active() {
					active.0.brand(|active, once| {
						let busy = active
							.token(once)
							.into_busy()
							.expect("active task is 'Busy'");
						sim.calendar().defer(active.clone(), busy);
					});
				}
				Poll::Pending
			}
			None => Poll::Ready(()),
		}
	}
}

/* ***************************************************************** Adapters */

/// Future adapter that attempts to unwrap the inner future, panicking if it
/// resolves into an `Err`-result.
#[must_use = "Futures must be awaited in order to execute them"]
#[pin_project::pin_project]
pub(super) struct Unwrap<F: ?Sized>(#[pin] F);

impl<T, E: Display, F: ?Sized + Future<Output = Result<T, E>>> Future for Unwrap<F> {
	type Output = T;

	#[inline]
	fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
		match self.project().0.poll(cx) {
			Poll::Ready(Ok(res)) => Poll::Ready(res),
			Poll::Ready(Err(err)) => panic!("{}", err),
			Poll::Pending => Poll::Pending,
		}
	}
}
