use core::{marker::PhantomData, mem::offset_of, ptr::NonNull};

use intrusive_collections::{DefaultLinkOps, LinkOps};

use super::{Link, Task};
use crate::{
	config::Config,
	ptr::{IntrusivelyCounted, Irc},
};

/* ******************************************************** Intrusive Adapter */

/// Adapter for [Tasks] to be inserted into intrusive data structures.
///
/// [Tasks]: Task
pub struct Adapter<C: ?Sized + Config, P = Irc<Task<'static, C>>> {
	link_ops: <Link as DefaultLinkOps>::Ops,
	pointer_ops: PointerOps<Task<'static, C>, P>,
}

impl<C: ?Sized + Config, P> Adapter<C, P> {
	pub const NEW: Self = Adapter {
		link_ops: <Link as DefaultLinkOps>::NEW,
		pointer_ops: PointerOps(PhantomData),
	};
}

impl<C: ?Sized + Config, P> Default for Adapter<C, P> {
	#[inline]
	fn default() -> Self {
		Self::NEW
	}
}

impl<C: ?Sized + Config, P> Clone for Adapter<C, P> {
	#[inline]
	fn clone(&self) -> Self {
		*self
	}
}

impl<C: ?Sized + Config, P> Copy for Adapter<C, P> {}

unsafe impl<C, P> intrusive_collections::Adapter for Adapter<C, P>
where
	C: ?Sized + Config,
	P: From<Irc<Task<'static, C>>> + Into<Irc<Task<'static, C>>>,
{
	type LinkOps = <Link as DefaultLinkOps>::Ops;
	type PointerOps = PointerOps<Task<'static, C>, P>;

	#[inline]
	unsafe fn get_value(
		&self,
		link: <Self::LinkOps as LinkOps>::LinkPtr,
	) -> *const Task<'static, C> {
		intrusive_collections::container_of!(link.as_ptr(), Task<'static, C>, hook)
	}

	#[inline]
	unsafe fn get_link(
		&self,
		value: *const Task<'static, C>,
	) -> <Self::LinkOps as LinkOps>::LinkPtr {
		let ptr = (value as *const u8).add(offset_of!(Task<'static, C>, hook));
		NonNull::new_unchecked(ptr as *mut _)
	}

	#[inline]
	fn link_ops(&self) -> &Self::LinkOps {
		&self.link_ops
	}

	#[inline]
	fn link_ops_mut(&mut self) -> &mut Self::LinkOps {
		&mut self.link_ops
	}

	#[inline]
	fn pointer_ops(&self) -> &Self::PointerOps {
		&self.pointer_ops
	}
}

/// Another helper structure that helps with the conversion between raw
/// [Task] pointers and owned ones.
pub struct PointerOps<V, P>(PhantomData<(V, P)>);

impl<V, P> PointerOps<V, P> {
	pub const NEW: Self = PointerOps(PhantomData);
}

impl<V, P> Clone for PointerOps<V, P> {
	#[inline]
	fn clone(&self) -> Self {
		*self
	}
}

impl<V, P> Copy for PointerOps<V, P> {}

unsafe impl<V, P> intrusive_collections::PointerOps for PointerOps<V, P>
where
	V: IntrusivelyCounted,
	P: From<Irc<V>> + Into<Irc<V>>,
{
	type Value = V;
	type Pointer = P;

	#[inline]
	unsafe fn from_raw(&self, value: *const V) -> P {
		P::from(Irc::from_raw(NonNull::new_unchecked(value as *mut V)))
	}

	#[inline]
	fn into_raw(&self, ptr: P) -> *const V {
		Irc::into_raw(P::into(ptr)).as_ptr() as *const V
	}
}
