//! The `task` module provides the implementation of a type-erased, intrusively
//! linked futures abstraction used in the scheduler of the simulation library.
//! Tasks are associated with a calendar, allowing them to be scheduled and
//! managed based on model time.
//!
//! # Overview
//!
//! The [`Task`] struct represents a task that can be scheduled and executed
//! within the simulation library. It is designed to handle type-erased futures
//! and supports features such as branding to symbolically tie instances to
//! specific lifetimes, allowing for statically guaranteed safety regarding
//! state transitions.
//!
//! ## Key Concepts
//!
//!  - **Task Lifecycle:** Tasks transition through states such as [`Born`],
//!    [`Idle`], [`Next`], [`Busy`], [`Done`], and [`Gone`]. The module provides
//!    methods to create, schedule, activate, and handle the termination of
//!    tasks.
//!
//!  - **Type Erasure and Branding:** Tasks are type-erased, allowing them to
//!    accommodate different future types. They support branding to symbolically
//!    tie instances to specific lifetimes, enabling static safety guarantees
//!    during composition and chaining of method calls.
//!
//!  - **Calendar Integration:** Tasks are associated with a calendar, enabling
//!    scheduling based on model time. The module provides methods to interact
//!    with the calendar, including insertion, removal, and awakening of tasks.
//!
//!  - **Shared Data Binding:** Tasks can be bound to shared data, enabling
//!    collaboration between tasks and allowing them to access shared data
//!    throughout their lifecycle.
//!
//!  - **Precendence and Prioritization:** Tasks have a precedence value that
//!    determines their priority relative to other tasks from the same process.
//!    The module offers methods to retrieve and set the precedence of a task.
//!
//! [`Born`]: State::Born
//! [`Idle`]: State::Idle
//! [`Next`]: State::Next
//! [`Busy`]: State::Busy
//! [`Done`]: State::Done
//! [`Gone`]: State::Gone

use core::{
	any::Any,
	cell::{Cell, OnceCell},
	fmt,
	hint::unreachable_unchecked,
	panic::Location,
	pin::Pin,
	ptr::NonNull,
	task::{Context, Poll},
};

use intrusive_collections::{LinkedList, RBTreeLink};

pub use adapter::{Adapter, PointerOps};
pub use puck::Puck;
pub use share::{Label, Share};

use crate::{
	calendar::Scheduler,
	error::NotIdle,
	fsm::*,
	ptr::{IntrusivelyCounted, Irc, IrcBox, IrcBoxed},
	simulator::Mark,
	Config, Dispatch, ExitStatus, PlanState, Prec,
};

mod adapter;
mod puck;
mod share;

/* ********************************************************************* Task */

/// The type of intrusive link stored in every [Task].
pub type Link = RBTreeLink;

/// A structure of type-erased, intrusively linked [futures].
///
/// Pointers to instances of this type are used in the scheduler of the
/// simulation library. They are oblivious to lifetime-restrictions and
/// erase the concrete types of the future and the process instance.
/// The latter types are recovered during dynamic lookup but the
/// lifetime-restrictions have to be observed by the abstractions provided
/// by the simulation library.
///
/// Tasks may be *branded* using an additional lifetime that symbolically
/// ties [Token] to specific instances. This allows composition and chaining of
/// method calls that depend on specific [Task-States] without having to
/// continuously test the current states within the method implementations,
/// allowing for statically guaranteed safety.
///
/// [futures]: core::future::Future
/// [Token]: token
/// [Task-States]: State
pub struct Task<'brand, C: ?Sized + Config> {
	/// An intrusive link for insertion of a task into the calendar.
	hook: Link,
	/// Intrusive reference counter and raw pointer to the most specialized
	/// version of the type containing this task.
	///
	/// This is needed in order to provide run-time polymorphism over the type
	/// of future being executed as well as what code is run on drop. See
	/// [Dispatch] for further information.
	task_box: IrcBox<TaskBox>,
	/// Pointer to data shared across different tasks.
	share: OnceCell<NonNull<Share<C>>>,
	/// A list of tasks waiting for this task to terminate.
	pending: Cell<LinkedList<Adapter<C>>>,
	/// Precedence of this task relative to other tasks from the same process.
	prec: Prec,
	/// Current state of this task instance.
	state: StateMachine<'brand, State<C>>,
}

impl<C: ?Sized + Config> Task<'static, C> {
	/// Creates a new task initialized in [State::Born] with specific location
	/// information attached.
	///
	/// [State::Born]: erased::State::Born
	pub(crate) fn new(prec: Prec, location: &'static Location<'static>) -> Self {
		Self {
			hook: Link::new(),
			task_box: IrcBox::with_location(TaskBox::new(), location),
			share: OnceCell::new(),
			pending: Cell::new(LinkedList::new(Adapter::NEW)),
			prec,
			state: StateMachine::default(),
		}
	}

	pub(crate) fn result(&self) -> Option<ExitStatus> {
		self.brand(|task, once| {
			task.token(once)
				.into_term()
				.map(|state| task.branded_result(&state))
				.ok()
		})
	}

	/// Reactivates the task at the current model-time.
	pub(crate) fn wake(this: Irc<Self>) -> Result<(), NotIdle> {
		this.clone().brand(|task, once| {
			let idle = task.token(once).into_idle()?;

			task.branded_share(&idle)
				.sim()
				.calendar()
				.activate(task.clone(), idle);

			Ok(())
		})
	}

	pub(crate) fn time(&self) -> Option<C::Time> {
		self.brand(|task, once| {
			task.token(once)
				.into_next()
				.ok()
				.and_then(|next| task.next_time(&next))
		})
	}
}

impl<'brand, C: ?Sized + Config> Task<'brand, C> {
	/// Returns the number of references to this task.
	#[inline]
	pub(crate) fn use_count(&self) -> usize {
		self.task_box.inner.refs.get()
	}

	/// Returns a copy of the internal [State](erased::State).
	#[inline]
	pub(crate) fn state(&self) -> &StateMachine<'brand, State<C>> {
		&self.state
	}

	/// Returns the current [Prec] of this task.
	#[inline]
	pub(crate) fn prec(&self) -> Prec {
		self.prec
	}

	/// Sets the new [Prec] of this task.
	#[inline]
	pub(crate) fn set_prec(&mut self, prec: Prec) {
		self.prec = prec;
	}

	/// Adds another task to the list of tasks to be awakened upon terminating.
	#[inline]
	pub(crate) fn insert_pending(&self, other: Irc<Task<'static, C>>) {
		let mut list = self.pending.take();
		list.push_back(other);
		self.pending.set(list);
	}

	/// Removes a previously added task from the list of pending tasks.
	///
	/// # Safety
	/// It is the caller's responsibility to ensure that the task had been
	/// added previously via [Self::insert_pending].
	#[inline]
	pub(crate) unsafe fn remove_pending(&self, other: &Task<'static, C>) {
		let mut list = self.pending.take();
		list.cursor_mut_from_ptr(other.detach()).remove();
		self.pending.set(list);
	}

	/// Awakens all pending tasks and clears the list.
	#[inline]
	pub(crate) fn wake_pending(&self) {
		for task in self.pending.take() {
			Task::wake(task).ok();
		}
	}

	/// Converts the specific brand into a generic brand, breaking the
	/// connection with the equally branded token.
	#[inline]
	pub(crate) fn detach(&self) -> &Task<'static, C> {
		unsafe { core::mem::transmute(self) }
	}

	/// Performs a runtime-check if this task has been dereferenced on the
	/// same thread as the one the executor is running and panicks if that is
	/// not the case.
	///
	/// This function doesn't (yet) work without feature "std".
	#[inline]
	fn is_same_thread<I>(&self, _init: &I) -> bool
	where
		I: Into<token::Init<'brand>>,
	{
		#[cfg(feature = "std")]
		{
			// compare thread ids during creation and currently in order to
			// runtime-fail if the task has been moved across trait-boundaries
			let thread_id = self.branded_share(_init).sim().get_thread_id();
			thread_id == std::thread::current().id()
		}

		#[cfg(not(feature = "std"))]
		{
			// TODO: fix unsoundness associated with Waker being Send
			// fingers crossed
			true
		}
	}

	/// Returns an enumeration copy of the internal [State](token::State).
	pub(crate) fn token(&self, once: Once<'brand>) -> token::State<'brand> {
		self.state.token(once)
	}

	/// Binds a task to [shared data].
	///
	/// # Safety
	/// The caller is responsible to ensure that the shared-data-reference
	/// outlives the (active) part of the task's life.
	///
	/// [shared data]: Share
	pub(crate) unsafe fn bind<'a>(
		mut self: Pin<&'a mut Self>,
		born: token::Born<'brand>,
		vtab: NonNull<dyn Dispatch + 'a>,
		share: NonNull<Share<C>>,
	) -> token::Idle<'brand> {
		// assign the reference to the shared data
		self.share.set(share).unwrap();

		// set the pointer to the virtual dispatch table
		self.task_box.inner.vptr = Some(core::mem::transmute::<
			NonNull<dyn Dispatch + 'a>,
			NonNull<dyn Dispatch + 'static>,
		>(vtab));

		// transition into state `Idle`
		self.state.transition(born, ())
	}

	/// Purges the task from the calendar.
	#[inline]
	pub(crate) fn deschedule(&self, next: token::Next<'brand>) -> token::Idle<'brand> {
		let share = self.branded_share(&next);

		// remove the task from the calendar
		share.sim().calendar().purge(self, next)
	}

	/// Removes the active state from the task.
	#[inline]
	pub(crate) fn deactivate(&self, busy: token::Busy<'brand>) -> token::Idle<'brand> {
		let share = self.branded_share(&busy);

		// de-register the task from the active cell
		share.sim().unslot(self, busy)
	}

	/// Polls the underlying future of this task, implemented as a private
	/// method to protect users from awaiting non-activated fiber.
	pub(crate) fn poll(&self, busy: token::Busy<'brand>, cx: &mut Context<'_>) -> Poll<()> {
		// read the virtual function from the table

		// SAFETY: the busy-token testifies that the task has been bound, which
		// ensures that the vptr to the virtual function table has been set;
		// pinning is ensured by the binding routine requiring it
		let vtab = unsafe { Pin::new_unchecked(self.task_box.inner.vptr.unwrap().as_ref()) };

		// temporarily escape the branding to allow Future::poll() to rebrand
		// without accidentally creating two token for the same instance
		let (once, res) = self.state.debrand(busy, move |_| vtab.poll(cx));

		// analyze the resulting state
		match self.token(once).into_busy() {
			Ok(busy) => match res {
				Poll::Ready(result) => {
					// busy -> done
					let _: token::Done<'_> = self.state.transition(busy, result);

					// reactivate pending tasks on completion
					self.wake_pending();

					Poll::Ready(())
				}
				Poll::Pending => {
					// busy -> idle
					let _: token::Idle<'_> = self.state.transition(busy, ());
					Poll::Pending
				}
			},
			Err(_err) => {
				debug_assert!(
					res.is_pending(),
					"task in state `{:?}` should not have been able to terminate",
					_err.0
				);
				Poll::Pending
			}
		}
	}

	/// Returns a reference to the shared data for this task.
	#[inline]
	pub(crate) fn branded_share<'s, I>(&'s self, init: &I) -> &'s Share<C>
	where
		I: Into<token::Init<'brand>>,
	{
		let _ = init;

		// SAFETY: the token witness testifies that the shared data is
		// initialized because `bind` has been called
		unsafe { self.share.get().unwrap_unchecked().as_ref() }
	}

	pub(crate) fn share(&self) -> Option<&Share<C>> {
		// SAFETY: once bound, the pointer stays valid
		self.share.get().map(|inner| unsafe { inner.as_ref() })
	}

	/// Returns the model time that the task will be activated next if the task
	/// is in state [`Next(Time)`](EventState::Time).
	#[inline]
	pub(crate) fn next_time(&self, next: &token::Next<'brand>) -> Option<C::Time> {
		let _ = next;

		match &*self.state.borrow() {
			State::Next(state) => state.time(),
			_ => unsafe { unreachable_unchecked() },
		}
	}

	/// Returns the local rank of the task if the task is in state
	/// [`Next(Rank)`](EventState::Rank).
	#[inline]
	pub(crate) fn local_rank(&self, next: &token::Next<'brand>) -> Option<C::Rank> {
		let _ = next;

		match &*self.state.borrow() {
			State::Next(state) => state.rank(),
			_ => unsafe { unreachable_unchecked() },
		}
	}

	/// Returns the [Cell] containing the current mark of the shared data
	/// associated with this task.
	///
	/// This value is used to organize different tasks with identical shared
	/// data in the calendar such that they are executed in a contiguous batch.
	#[inline]
	pub(crate) fn mark<'s, I>(&'s self, init: &I) -> &'s Cell<Mark>
	where
		I: Into<token::Init<'brand>>,
	{
		self.branded_share(init).mark()
	}

	/// Returns the exit code of this task.
	#[inline]
	pub(crate) fn branded_result<T>(&self, _: &T) -> ExitStatus
	where
		T: Into<token::Term<'brand>>,
	{
		// SAFETY: the task is in state `Done` or `Gone` per the token witness
		match &*self.state.borrow() {
			State::Done(result) => *result,
			State::Gone(result) => *result,
			_ => unsafe { unreachable_unchecked() },
		}
	}
}

// Tasks offer unique methods for Branded variants
impl<'b, C: ?Sized + Config> Stateful for Task<'b, C> {
	type Brand = &'b ();

	#[inline]
	unsafe fn enter(&self) {
		self.state.enter();
	}

	#[inline]
	unsafe fn leave(&self) {
		self.state.leave();
	}
}

impl<'b, C: ?Sized + Config> Rebrand<'b> for Task<'b, C> {
	type Kind<'a> = Task<'a, C>;
}

// Tasks only contain pointers to pinned data and are not pinned themselves
impl<C: ?Sized + Config> Unpin for Task<'_, C> {}

impl<C: ?Sized + Config> fmt::Debug for Task<'_, C> {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		use crate::config::Time;
		let mut s = f.debug_struct("Task");

		s.field("state", &self.state.borrow().erased())
			.field("location", self.task_box.loc);

		self.brand(|state, once| {
			let token = state.token(once);
			let init = token.as_init()?;

			s.field("label", &state.branded_share(init).label());
			let next = token.as_next()?;

			if let Some(time) = state.next_time(next) {
				if size_of::<C::Time>() > 0 {
					s.field("time", &time.display());
				}
			} else if let Some(rank) = state.local_rank(next) {
				if size_of::<C::Rank>() > 0 {
					s.field("rank", &rank);
				}
				s.field("mark", &state.mark(init).get());
			}

			None::<()>
		});

		s.field("prec", &self.prec())
			.field("refs", &self.use_count())
			.finish()
	}
}

impl<C: Config> crate::erased::Task for Task<'static, C> {
	fn subject(&self) -> &dyn Any {
		self.share().map_or(&(), |shared| shared.item())
	}

	fn label(&self) -> Label {
		self.share()
			.map_or(Label::default(), |shared| shared.label())
	}

	fn prec(&self) -> Prec {
		self.prec()
	}

	fn state(&self) -> erased::State {
		self.state().borrow().erased()
	}
}

unsafe impl<C: ?Sized + Config> IntrusivelyCounted for Task<'_, C> {
	type Inner = TaskBox;

	#[inline]
	fn irc_box(&self) -> &IrcBox<Self::Inner> {
		&self.task_box
	}
}

/* ************************************************************** Irc Support */

/// Container type used to support [intrusive reference counting](Irc).
pub struct TaskBox {
	/// Contains a counter for the number of references to the outer [Task].
	refs: Cell<usize>,
	/// Pointer to the most-specialized version of the instance.
	/// Used to recover type information during dynamic dispatch.
	vptr: Option<NonNull<dyn Dispatch>>,
}

impl TaskBox {
	/// Creates a new irc box from a location with zero references and no
	/// dispatcher.
	const fn new() -> Self {
		TaskBox {
			refs: Cell::new(0),
			vptr: None,
		}
	}
}

unsafe impl IrcBoxed for TaskBox {
	unsafe fn acquire(&self) -> usize {
		let refs = self.refs.get() + 1;
		self.refs.set(refs);
		refs
	}

	unsafe fn release(&self) -> Option<fn(NonNull<Self>)> {
		self.refs.set(self.refs.get() - 1);

		(self.refs.get() == 0 && self.vptr.is_some()).then_some(
			// return a function reclaiming the outer type if no references point
			// to the Task anymore; this separation is necessary to prevent
			// overlapping references to this Task from self and (indirectly) from
			// the inner dyn object
			|this| unsafe {
				// restore the outer type and copy the pointer to the virtual table
				let mut vptr = this.as_ref().vptr.unwrap_unchecked();

				// call the reclaim-method on an exclusive reference to the dyn
				// object; the temporary reference to the task has been dropped
				// previously and no other active references exist to the task
				// which allows us to take this exclusive reference
				Pin::new_unchecked(vptr.as_mut()).reclaim();
			},
		)
	}
}

/* ************************************************************** Task States */

crate::state_machine! {
	/// Represents the state of a [Task] instance throughout its lifecycle.
	///
	/// A Task starts in the [`Born`] state without references to shared data or
	/// its future, as provided by its owning [Fiber] or [Process]. After
	/// binding pinned references to it, the Task moves to the [`Idle`] state,
	/// indicating it is unscheduled.
	///
	/// From the [`Idle`] state, the Task can be scheduled by inserting it into
	/// the event calendar at a specific model time, transitioning it to the
	/// [`Next`] state. When the model time reaches this point and the Task is
	/// activated, it enters the [`Busy`] state.
	///
	/// Upon completion, the Task transitions to the [`Done`] state, indicating
	/// that a return value is available for extraction. At any point before
	/// normal termination, the Task can be canceled, moving it to the [`Gone`]
	/// state and dropping its bound future. The [`Gone`] state is the final
	/// state in a Task's lifecycle.
	///
	/// [Fiber]: crate::Fiber
	/// [Process]: crate::Process
	/// [`Born`]: State::Born
	/// [`Idle`]: State::Idle
	/// [`Next`]: State::Next
	/// [`Busy`]: State::Busy
	/// [`Done`]: State::Done
	/// [`Gone`]: State::Gone
	#[derive(Default)]
	pub enum State<C: Config> {
		/// State of a task signifying a non-bound future and shared data.
		#[default]
		Born -> {Idle, Gone},
		/// State of a task that is waiting for external reactivation.
		Idle -> {Next, Gone},
		/// State of the task that is currently active. Only one task may be in
		/// this state during one simulation run at any one time.
		Busy -> {Idle, Done},
		/// State of a task that is managed by the calendar.
		Next(<C::Plan as Scheduler>::State) -> {Idle, Busy},
		/// State of a completed task with a result available for extraction.
		Done(ExitStatus) -> {Gone},
		/// State of a task that cannot be scheduled anymore.
		Gone(ExitStatus) -> {}
	}

	/// Meta state for all non-[`Born`] states.
	///
	/// [`Born`]: State::Born
	pub Init = {Idle, Busy, Next, Done, Gone};

	/// Meta state for [`Done`] and [`Gone`] states and a subset of the [`Init`]
	/// meta state.
	///
	/// [`Done`]: State::Done
	/// [`Gone`]: State::Gone
	/// [`Init`]: token::Init
	pub Term: Init = {Done, Gone};
}
