use core::{
	any::Any,
	fmt,
	future::IntoFuture,
	ptr::NonNull,
	task::{RawWaker, RawWakerVTable},
	mem::ManuallyDrop,
};

use crate::{error::NotIdle, fsm::*, task::{erased::State as TaskState, token, Label, Task}, Config, ExitStatus, IntrusivelyCounted, Irc, IrcBoxed, Prec, Share, Sim};

/* *********************************************************** Task Reference */

/// Reference-counted wrapper around a pinned task.
///
/// Instances of this type may be cloned and moved around the simulation model.
/// They allow access to the shared process state and implement the abstract
/// [Waker](core::task::Waker) interface.
///
/// The reference counter does not prolong the lifetime of a [Task] but instead
/// serves as an indicator that dropping the referenced task would lead to
/// dangling references if its reference counter is greater than zero.
/// Thus, the actual mechanics of this type are more akin to a checksum than a
/// reference-counted type since it falls upon the user to drop their `Puck`s
/// before the referenced Task is dropped.
///
/// This design preserves memory-safety for tasks with erased lifetimes at the
/// cost of reduced user-friendliness since bugs only surface during destruction
/// of tasks, making them potentially hard to track.
/// However, it also allows for `no_std` environments without dynamic memory
/// allocation to run simulations as it is not necessary to allocate tasks
/// on the heap.
pub struct Puck<C: ?Sized + Config>(pub(crate) Irc<Task<'static, C>>);

impl<C: ?Sized + Config> Puck<C> {
	/// Creates a new [Task] [Puck] from a task that has been initialized
	/// through binding.
	pub fn new<'brand>(task: Irc<Task<'brand, C>>, _: &token::Init<'brand>) -> Self {
		Puck(Irc::map(task, |inner| inner.detach()))
	}
	
	/// Returns a reference to the [shared context](Share) bound to this task.
	pub fn share(&self) -> &Share<C> {
		// SAFETY: Pucks can only be created during initialization, after
		// which they have shared data bound to them
		unsafe {
			self.0.share().unwrap_unchecked()
		}
	}
	
	/// Returns whether this task pin points to a given [Task].
	pub fn is_same(&self, other: &Task<'_, C>) -> bool {
		core::ptr::eq(&*self.0, other.detach())
	}
}

// implement the waker interface for task pins
impl<C: Config> Puck<C> {
	pub(crate) const VTABLE: RawWakerVTable = RawWakerVTable::new(
		Self::waker_clone,
		Self::waker_wake,
		Self::waker_wake_by_ref,
		Self::waker_drop,
	);

	/// Converts a task pin into a [RawWaker] without decreasing the reference
	/// count of the inner task.
	#[inline]
	pub(crate) fn into_waker(self) -> RawWaker {
		// initialize the raw waker
		RawWaker::new(
			Irc::into_raw(self.0).cast().as_ptr() as *const (),
			&Self::VTABLE,
		)
	}

	/// Reconstructs the original [Puck] from the raw pointer.
	///
	/// This function also ensures that the call didn't come from a different
	/// thread than the one that the executor runs in.
	///
	/// # Safety
	/// The caller has to ensure that the provided raw pointer was the one
	/// originally used in the call to [Self::into_waker].
	pub(crate) unsafe fn restore(task: *const ()) -> Self {
		// reconstruct the pointer to the task
		let task = &*(task as *const Task<'static, C>);

		// determine whether this thread is different from the one that the
		// executor is running in
		let same_thread = task.brand(|task, once| {
			// SAFETY: we originally used a puck to create the waker
			let init = task.token(once).into_init().unwrap();
			task.is_same_thread(&init)
		});

		// assert that the waker hasn't been called from a different thread
		assert!(
			same_thread,
			"simulation-associated waker may not be moved across thread boundaries"
		);

		// now it's safe to reconstruct the puck
		Puck(Irc::from_raw(NonNull::from(task)))
	}

	unsafe fn waker_clone(task: *const ()) -> RawWaker {
		// restore the puck from the raw pointer
		let puck = Self::restore(task);

		// increase the reference counter
		puck.0.irc_box().inner.acquire();

		// create another (identical) raw waker
		puck.into_waker()
	}

	unsafe fn waker_wake(task: *const ()) {
		// restore the puck and reactivate the underlying task
		Task::wake(Self::restore(task).0).ok();
	}

	unsafe fn waker_wake_by_ref(task: *const ()) {
		// restore the puck without dropping it later
		let puck = ManuallyDrop::new(Self::restore(task));

		// reactivate the underlying task
		Task::wake(puck.0.clone()).ok();
	}

	unsafe fn waker_drop(task: *const ()) {
		// restore the puck and drop it immediately
		Self::restore(task);
	}
}

impl<C: ?Sized + Config> From<Puck<C>> for Irc<Task<'static, C>> {
	fn from(value: Puck<C>) -> Self {
		value.0
	}
}

impl<C: ?Sized + Config> crate::Puck<C> for Puck<C> {
	fn result(&mut self) -> Option<ExitStatus> {
		self.0.result()
	}

	fn wake(&mut self) -> Result<(), NotIdle> {
		Task::wake(self.0.clone())
	}

	fn subject(&self) -> &dyn Any {
		self.share().item()
	}

	fn sim(&self) -> &Sim<C> {
		self.share().sim()
	}

	fn label(&self) -> Label {
		self.share().label()
	}

	fn time(&self) -> Option<C::Time> {
		self.0.time()
	}
	
	fn rank(&self) -> C::Rank {
		self.share().rank()
	}
	
	fn prec(&self) -> Prec {
		self.0.prec()
	}

	fn state(&self) -> TaskState {
		self.0.state().borrow().erased()
	}
	
	fn own(&self) -> Irc<Task<'static, C>> {
		self.0.clone()
	}
}

impl<C: ?Sized + Config> AsRef<Task<'static, C>> for Puck<C> {
	fn as_ref(&self) -> &Task<'static, C> {
		&self.0
	}
}

impl<C: ?Sized + Config> IntoFuture for Puck<C> {
	type Output = ExitStatus;
	type IntoFuture = crate::ops::Join<C, Self>;

	#[inline]
	fn into_future(self) -> Self::IntoFuture {
		crate::ops::join(self)
	}
}

impl<C: ?Sized + Config> AsRef<Irc<Task<'static, C>>> for Puck<C> {
	fn as_ref(&self) -> &Irc<Task<'static, C>> {
		&self.0
	}
}

impl<C: ?Sized + Config> fmt::Debug for Puck<C> {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		self.0.fmt(f)
	}
}

impl<C: ?Sized + Config> Clone for Puck<C> {
	fn clone(&self) -> Self {
		Puck(self.0.clone())
	}
}

impl<C: ?Sized + Config> PartialEq for Puck<C> {
	fn eq(&self, other: &Self) -> bool {
		self.is_same(&other.0)
	}
}

impl<C: ?Sized + Config> Eq for Puck<C> {}
