use core::{any::Any, cell::Cell, fmt, num::NonZero, ptr::NonNull};

#[cfg(feature = "tracing")]
use tracing::{error_span, span::Entered, Span};

use crate::{
	config::Config,
	simulator::{Mark, Sim},
	ptr::Irc
};

/* ************************************************************** Shared Data */

/// Structure for shared data between tasks associated with the same process.
pub struct Share<C: ?Sized + Config> {
	/// Pointer to the simulator.
	sim: Irc<Sim<C>>,
	/// Holds the [Span] of the process.
	#[cfg(feature = "tracing")]
	span: Span,
	/// Contains the type name of the process.
	name: &'static str,
	/// Contains the process-unique ID for this specific instance.
	pid: Option<NonZero<usize>>,
	/// Marks the priority of tasks with shared data.
	rank: Cell<C::Rank>,
	/// Marks the order in which tasks associated with the same process have
	/// been inserted.
	mark: Cell<Mark>,
	/// A reference to the shared instance over which tasks with shared data
	/// operate over.
	item: NonNull<dyn Any>,
}

impl<C: ?Sized + Config> Share<C> {
	/// Initialize the shared data for the root fiber of a simulation run.
	pub(crate) fn root(sim: Irc<Sim<C>>) -> Self {
		let default_rank = sim.config().default_rank();
		Self {
			sim,
			#[cfg(feature = "tracing")]
			span: error_span!(parent: None, "SimMain"),
			name: "SimMain",
			pid: None,
			rank: Cell::new(default_rank),
			mark: Cell::new(Mark::stale()),
			item: NonNull::from(&()),
		}
	}
	
	/// Initialize the shared data for a new process.
	///
	/// # Safety
	/// The caller is responsible that `item` outlives the shared data. 
	pub(crate) unsafe fn new<I: Any>(sim: Irc<Sim<C>>, item: &I, rank: C::Rank, name: &'static str) -> Share<C> {
		// Generate a fresh process id.
		let pid = Some(sim.pid_gen::<I>());
		
		// Initialize the shared process context.
		Share {
			sim,
			#[cfg(feature = "tracing")]
			span: error_span!(
				parent: None, "Process",
				label = %Label { name, pid }
			),
			name,
			pid,
			rank: Cell::new(rank),
			mark: Cell::new(Mark::stale()),
			item: NonNull::from(item),
		}
	}

	/// Returns the preset shared priority.
	pub fn rank(&self) -> C::Rank {
		self.rank.get()
	}

	/// Returns the chosen identifier for the type of the shared instance.
	pub fn name(&self) -> &'static str {
		self.name
	}

	/// Returns the ID of the process associated with the shared instance if
	/// one has been set or `None` if it hasn't.
	pub fn pid(&self) -> Option<NonZero<usize>> {
		self.pid
	}

	/// Returns a reference of the contained instance.
	pub fn item(&self) -> &dyn Any {
		unsafe { self.item.as_ref() }
	}

	/// Returns the process-unique [Label] for this instance.
	pub fn label(&self) -> Label {
		Label {
			name: self.name(),
			pid: self.pid(),
		}
	}

	/// Returns a [Cell] with the currently stored [Mark].
	pub fn mark(&self) -> &Cell<Mark> {
		&self.mark
	}

	/// Sets the shared priority.
	pub fn set_rank(&self, rank: C::Rank) {
		self.rank.set(rank);
	}

	/// Enters the [Span] associated with the process owning the shared data.
	#[cfg(feature = "tracing")]
	pub fn enter_span(&self) -> Entered<'_> {
		self.span.enter()
	}

	#[cfg(not(feature = "tracing"))]
	pub const fn enter_span(&self) {}

	/// Returns a reference to the simulation context.
	pub(crate) fn sim(&self) -> &Irc<Sim<C>> {
		&self.sim
	}
}

// there is nothing inside the shared context that needs pinning
impl<C: ?Sized + Config> Unpin for Share<C> {}

impl<C: Config> fmt::Debug for Share<C>
where
	C::Rank: fmt::Debug,
{
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		f.debug_struct("Share")
			.field("name", &self.name())
			.field("pid", &self.pid())
			.field("mark", &self.mark.get())
			.field("rank", &self.rank())
			.finish()
	}
}

/* ************************************************************ Process Label */

/// Process-specific identifier that uniquely identifies an instance over the
/// course of a simulation run.
#[derive(Default, Debug, Copy, Clone, PartialEq, Eq)]
pub struct Label {
	/// The process name.
	pub name: &'static str,
	/// The process-name-specific ID.
	pub pid: Option<NonZero<usize>>,
}

impl Label {
	/// Returns an iterator of string slices into the process name that
	/// corresponds to a pretty-printed version of the type path.
	pub fn pretty_name(&self) -> impl Iterator<Item = &str> {
		self.name
			.split_inclusive(&['<', '>', '(', ')', '[', ']', ',', ';'])
			.filter_map(|part| part.rsplit(':').next())
	}
}

impl fmt::Display for Label {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		if !f.alternate() {
			// use the pretty-printed format
			for ident in self.pretty_name() {
				f.write_str(ident)?;
			}
		} else {
			// print the whole name
			f.write_str(self.name)?;
		}

		// only print the pid if it exists
		if let Some(pid) = self.pid {
			write!(f, "#{pid}")
		} else {
			Ok(())
		}
	}
}
