use proc_macro2::{Span, TokenStream};
use quote::{quote, ToTokens};
use syn::{Data, DeriveInput, GenericParam};

pub(crate) fn config_derive_impl(mut config: DeriveInput) -> TokenStream {
	// set the default time-type and value
	let mut time_type = quote!(f64);
	let mut time_val = quote!(0.0);

	// set the default rank-type and value
	let mut rank_type = quote!(());
	let mut rank_val = quote!();

	// the scheduler cannot be overwritten using this macro
	let plan_type = quote!(::odem_rs::core::calendar::DefaultPlan<Self>);
	let data_type = &config.ident;

	// add the `'static` bound to the generic parameters
	let static_lt = syn::Lifetime::new("'static", Span::call_site());
	for param in config.generics.params.iter_mut() {
		match param {
			GenericParam::Lifetime(syn::LifetimeParam { bounds, .. }) => {
				bounds.push(static_lt.clone());
			}
			GenericParam::Type(syn::TypeParam { bounds, .. }) => {
				bounds.push(syn::TypeParamBound::Lifetime(static_lt.clone()));
			}
			GenericParam::Const(_) => {}
		}
	}

	// extract the names of the generic parameters
	let (impl_generics, ty_generics, where_clause) = config.generics.split_for_impl();

	// check the type of data passed into the macro
	match &config.data {
		// return an error for enumerations
		Data::Enum(_) => {
			return quote!(compile_error!(
				"derive(Config) is not supported for enumerations"
			))
		}

		// return an error for unions
		Data::Union(_) => {
			return quote!(compile_error!("derive(Config) is not supported for unions"))
		}

		// analyze structures for "time" and "rank" attributes overwriting the
		// default configuration
		Data::Struct(syn::DataStruct { fields, .. }) => {
			for (field, id) in fields.iter().zip((0..).map(syn::Index::from)) {
				for attr in field.attrs.iter() {
					if let Some(ident) = attr.meta.path().get_ident() {
						let ty = &field.ty;
						let access = field
							.ident
							.as_ref()
							.map_or_else(|| id.to_token_stream(), ToTokens::to_token_stream);

						if ident == "time" {
							time_type = quote!(#ty);
							time_val = quote!(self.#access);
						} else if ident == "rank" {
							rank_type = quote!(#ty);
							rank_val = quote!(self.#access.clone());
						}
					}
				}
			}
		}
	}

	// generate the trait implementation
	quote! {
		impl #impl_generics ::odem_rs::core::config::Config for #data_type #ty_generics #where_clause {
			type Time = #time_type;
			type Rank = #rank_type;
			type Data = Self;
			type Plan = #plan_type;

			#[inline]
			fn default_time(&self) -> Self::Time { #time_val }

			#[inline]
			fn default_rank(&self) -> Self::Rank { #rank_val }

			#[inline]
			fn global_data(&self) -> &Self { self }
		}
	}
}
