use proc_macro2::{Span, TokenStream};
use quote::{quote, quote_spanned};
use std::collections::{HashMap, HashSet};
use syn::{parse2, spanned::Spanned, visit::Visit, visit_mut::VisitMut, Expr, Local, PatIdent};

pub(crate) fn control_expr_impl(mut expr: Expr) -> TokenStream {
	// create an identifier for the parameter of the generated function;
	// mixed_site ensures that this identifier doesn't collide with user ids
	let args = syn::Ident::new("args", Span::mixed_site());

	// rewrite the expression and collect (potential) control variables
	let mut extract_vars = ExtractVars::new(&args);
	extract_vars.visit_expr_mut(&mut expr);

	// collect the set of control variables
	let mut ctrl_set = Vec::from_iter(extract_vars.registry);

	// ensure that no more than 12 are referenced
	if ctrl_set.len() > 12 {
		let msg = format!(
			"{} expressions referenced, but only up to 12 allowed",
			ctrl_set.len()
		);
		return quote!(compile_error!(#msg));
	}

	// sort the control set ascending by the index
	ctrl_set.sort_unstable_by_key(|(_, v)| *v);

	// select only the expressions
	let ctrl = ctrl_set.iter().map(|(k, _)| k);

	// generate the control expression
	quote!(ControlExpr::new(
		(#((&#ctrl).deref().expr_term().pub_term().expr_or_publisher(),)*),
		|&#args| #expr
	))
}

/* ************************************************* Expression-Tree Rewriter */

/// Implements the visitor-pattern to extract possible control variables from
/// the abstract syntax tree with an expression root and rewrites it to refer
/// to the extracted and resolved control variables instead.
struct ExtractVars<'x> {
	/// Maps known possible control variables to their index in the argument
	/// tuple. Making this a [HashMap] automatically contracts identical control
	/// variables into a single member of the argument tuple.
	registry: HashMap<Expr, usize>,
	/// Layered symbol table containing detected local bindings that are not to
	/// be classified as potential control variables and thereby rewritten.
	symtab: LayeredTable,
	/// Name of the argument to the generated closure.
	arg_name: &'x syn::Ident,
}

impl<'x> ExtractVars<'x> {
	/// Creates a new visitor that replaces potential control variables with
	/// references to the specified argument.
	#[inline]
	fn new(arg_name: &'x syn::Ident) -> Self {
		ExtractVars {
			registry: HashMap::new(),
			symtab: LayeredTable::new(),
			arg_name,
		}
	}

	/// Internal method that checks whether an expression in the syntax tree
	/// could be a control variable.
	///
	/// Control variables are restricted to combinations of path, field access,
	/// and index, and cannot contain references to local variables.
	#[inline]
	fn is_control_var(&self, expr: &Expr) -> bool {
		let mut visitor = IsControlVar::new(&self.symtab);
		visitor.visit_expr(expr);
		visitor.result
	}
}

impl VisitMut for ExtractVars<'_> {
	fn visit_expr_mut(&mut self, i: &mut Expr) {
		use Expr::*;

		match i {
			Async(expr) => {
				self.symtab.push_layer();
				self.visit_block_mut(&mut expr.block);
				self.symtab.pop_layer();
			}
			Block(expr) => {
				self.symtab.push_layer();
				self.visit_block_mut(&mut expr.block);
				self.symtab.pop_layer();
			}
			Closure(expr) => {
				// keep the closure's local variables in a layer of their own
				self.symtab.push_layer();

				for var in expr.inputs.iter_mut() {
					self.visit_pat_mut(var);
				}

				self.visit_expr_mut(&mut expr.body);
				self.symtab.pop_layer();
			}
			Field(_) | Index(_) | Path(_) => {
				// test if the expression could refer to a control variable
				if self.is_control_var(i) {
					// rewrite the expression and assign a field to it
					let span = i.span();
					let arg = self.arg_name;
					let id = self.registry.len();
					let id = self.registry.entry(i.clone()).or_insert(id);
					let id = syn::Index::from(*id);

					*i = parse2(quote_spanned!(span => #arg.#id.get())).unwrap();
				} else {
					// otherwise recurse into the expression as usual
					syn::visit_mut::visit_expr_mut(self, i);
				}
			}
			ForLoop(expr) => {
				// parse the expression first in order not to miss any external
				// references
				self.visit_expr_mut(&mut expr.expr);

				// open a new block in the symbol table
				self.symtab.push_layer();
				self.visit_pat_mut(&mut expr.pat);
				self.visit_block_mut(&mut expr.body);
				self.symtab.pop_layer();
			}
			If(expr) => {
				// open a new block in the symbol table for the variables
				// declared in let-bindings
				self.symtab.push_layer();

				// visit the condition before visiting the then-block
				self.visit_expr_mut(&mut expr.cond);
				self.visit_block_mut(&mut expr.then_branch);

				// if the statement has an else-part, clear declared variables
				// from the then-block but keep the layer open
				if let Some((_, alt)) = expr.else_branch.as_mut() {
					self.symtab.clear_layer();
					self.visit_expr_mut(alt);
				}

				self.symtab.pop_layer();
			}
			Let(expr) => {
				// visit the right-hand-side first before visiting the pattern
				// used for the bound identifiers in order to ensure replacing
				// non-local variables in the expression even if identically
				// named local variables come into scope afterwards
				self.visit_expr_mut(&mut expr.expr);
				self.visit_pat_mut(&mut expr.pat);
			}
			Loop(expr) => {
				self.symtab.push_layer();
				self.visit_block_mut(&mut expr.body);
				self.symtab.pop_layer();
			}
			Match(expr) => {
				// visit the top expression first
				self.visit_expr_mut(&mut expr.expr);

				// visit match-arms sequentially and put them in their own block
				self.symtab.push_layer();
				for arm in expr.arms.iter_mut() {
					self.symtab.clear_layer();
					self.visit_pat_mut(&mut arm.pat);

					if let Some((_, guard)) = arm.guard.as_mut() {
						self.visit_expr_mut(guard);
					}

					self.visit_expr_mut(&mut arm.body);
				}
				self.symtab.pop_layer();
			}
			_ => syn::visit_mut::visit_expr_mut(self, i),
		}
	}

	fn visit_local_mut(&mut self, i: &mut Local) {
		// visit the initializer first in order to delay the variable decls
		if let Some(init) = i.init.as_mut() {
			self.visit_expr_mut(&mut init.expr);
		}

		// visit the pattern next
		self.visit_pat_mut(&mut i.pat);
	}

	fn visit_pat_ident_mut(&mut self, i: &mut PatIdent) {
		// insert the identifier into the symbol table
		self.symtab.insert(i.ident.clone());
	}
}

/* ********************************** Control Variable Identification Visitor */

/// Implements the visitor pattern to detect whether an expression in the syntax
/// tree qualifies as a potential control variable.
///
/// Control variables may be constructed from combinations of Path,
/// Field-Access, Index-Operations, and Method-Calls and may not contain
/// references to local variables.
struct IsControlVar<'v> {
	table: &'v LayeredTable,
	result: bool,
}

impl<'v> IsControlVar<'v> {
	/// Creates the visitor and provides it with a table of the local variables.
	#[inline]
	const fn new(symtab: &'v LayeredTable) -> Self {
		IsControlVar {
			table: symtab,
			result: true,
		}
	}

	/// Evaluates whether an expression references a local variable, returning
	/// `true` if that is the case and `false` if it isn't.
	#[inline]
	fn ref_local_var(&self, expr: &Expr) -> bool {
		let mut visitor = RefLocalVar::new(self.table);
		visitor.visit_expr(expr);
		visitor.result
	}
}

impl<'ast> Visit<'ast> for IsControlVar<'_> {
	fn visit_expr(&mut self, i: &'ast Expr) {
		use Expr::*;

		match i {
			Field(expr) => {
				self.visit_expr(&expr.base);
			}
			Index(expr) => {
				if self.ref_local_var(&expr.index) {
					self.result = false;
				} else {
					self.visit_expr(&expr.expr);
				}
			}
			MethodCall(expr) => {
				for arg in expr.args.iter() {
					if self.ref_local_var(arg) {
						self.result = false;
						return;
					}
				}

				self.visit_expr(&expr.receiver);
			}
			Path(expr) => {
				if let Some(id) = expr.path.get_ident() {
					if self.table.contains(id) {
						self.result = false;
					}
				}
			}
			_ => {
				self.result = false;
			}
		}
	}
}

/// Implements the visitor pattern to detect whether an expression references
/// a variable located in a [LayeredTable] containing local variables.
///
/// This is used in order to prevent mis-identifying expressions referencing
/// local variables e.g. in their index. Function-calls and arithmetic
/// operations not referencing them are allowed.
struct RefLocalVar<'v> {
	table: &'v LayeredTable,
	result: bool,
}

impl<'v> RefLocalVar<'v> {
	/// Creates the visitor and provides it with a table of the local variables.
	#[inline]
	const fn new(symtab: &'v LayeredTable) -> Self {
		RefLocalVar {
			table: symtab,
			result: false,
		}
	}
}

impl<'ast> Visit<'ast> for RefLocalVar<'_> {
	#[inline]
	fn visit_ident(&mut self, i: &'ast syn::Ident) {
		if self.table.contains(i) {
			self.result = true;
		}
	}
}

/* ***************************************************** Layered Symbol Table */

/// Block-structured symbol table to capture the identities of local
/// Rust-identifiers.
struct LayeredTable(Vec<HashSet<syn::Ident>>);

impl LayeredTable {
	/// Creates a new, empty symbol table.
	#[inline]
	const fn new() -> Self {
		LayeredTable(Vec::new())
	}

	/// Adds a new layer, allowing shadowed from lower layers to be restored
	/// after the removal of this layer.
	#[inline]
	fn push_layer(&mut self) {
		self.0.push(HashSet::new());
	}

	/// Empties the current top layer.
	#[inline]
	fn clear_layer(&mut self) {
		if let Some(layer) = self.0.last_mut() {
			layer.clear();
		}
	}

	/// Removes the current top layer, restoring the symbol table to its
	/// previous state.
	#[inline]
	fn pop_layer(&mut self) {
		self.0.pop();
	}

	/// Inserts a new symbol into the current top layer.
	fn insert(&mut self, id: syn::Ident) -> bool {
		if let Some(set) = self.0.last_mut() {
			set.insert(id)
		} else {
			false
		}
	}

	/// Searches the symbol table for the existence of an identifier.
	fn contains(&self, id: &syn::Ident) -> bool {
		for layer in self.0.iter().rev() {
			if layer.contains(id) {
				return true;
			}
		}

		false
	}
}
