use proc_macro2::TokenStream;
use quote::{quote, ToTokens};
use syn::{Data, DeriveInput};

pub(crate) fn publisher_derive_impl(publisher: DeriveInput) -> TokenStream {
	// extract the names of the generic parameters
	let mut types = Vec::new();
	let mut vars = Vec::new();
	let mut indices = Vec::new();

	// check the type of data passed into the macro
	match &publisher.data {
		// return an error for enumerations
		Data::Enum(_) => {
			return quote!(compile_error!(
				"derive(Publisher) is not supported for enumerations"
			))
		}

		// return an error for unions
		Data::Union(_) => {
			return quote!(compile_error!(
				"derive(Publisher) is not supported for unions"
			))
		}

		// search structures for "subscribe" attributes
		Data::Struct(syn::DataStruct { fields, .. }) => {
			let mut index = 0;
			for (field, id) in fields.iter().zip((0..).map(syn::Index::from)) {
				for attr in field.attrs.iter() {
					if let Some(ident) = attr.meta.path().get_ident() {
						if ident == "subscribe" {
							let ty = &field.ty;
							let access = field
								.ident
								.as_ref()
								.map_or_else(|| id.to_token_stream(), ToTokens::to_token_stream);

							types.push(quote!(#ty));
							vars.push(quote!(self.#access));
							indices.push(syn::Index::from(index));
							index += 1;
						}
					}
				}
			}
		}
	}

	let (impl_generics, ty_generics, where_clause) = publisher.generics.split_for_impl();
	let data_type = &publisher.ident;

	// generate the trait implementation
	quote! {
		impl #impl_generics ::odem_rs::sync::Publisher for #data_type #ty_generics #where_clause {
			type Link = ::odem_rs::sync::Tuple<(#(<#types as ::odem_rs::sync::Publisher>::Link,)*)>;

			#[inline]
			unsafe fn subscribe(&self, link: core::pin::Pin<&Self::Link>) {
				#(#vars.subscribe(link.map_unchecked(|inner| &inner.0.#indices));)*
			}

			#[inline]
			unsafe fn unsubscribe(&self, link: core::pin::Pin<&Self::Link>) {
				#(#vars.unsubscribe(link.map_unchecked(|inner| &inner.0.#indices));)*
			}
		}
	}
}
