//! A module for stat-specific error types.

/* ************************************************************** Error Types */

/// Enumeration of errors for the functions and structures contained in this
/// module of the simulation library.
#[derive(Copy, Clone, Debug, thiserror::Error)]
pub enum Error {
	/// Signals an attempt to tabulate a value at a time point that is in the
	/// past of the previous tabulation.
	#[error("attempted tabulation of a value in the past of the previous tabulation")]
	Causality,
	/// Signals an attempt to tabulate a value with negative weight.
	#[error("attempted tabulation of a value with negative weight")]
	NegativeWeight,
}

/* ******************************************************** Minor Error Types */

/// Indicates that an attempt to tabulate a value at a time point that is in the
/// past of the previous tabulation has been made.
#[derive(Debug, thiserror::Error)]
#[error("tabulated value at time {next:?} happens before the previous value at {curr:?}")]
pub struct Causality<T> {
	pub curr: T,
	pub next: T,
}

impl<T> From<Causality<T>> for Error {
	#[inline]
	fn from(_: Causality<T>) -> Self {
		Error::Causality
	}
}

/// Indicates that an attempt to tabulate a value with negative weight has been
/// made.
#[derive(Debug, thiserror::Error)]
#[error("tabulated value has negative weight {weight:?}")]
pub struct NegativeWeight<W> {
	pub weight: W,
}

impl<W> From<NegativeWeight<W>> for Error {
	#[inline]
	fn from(_: NegativeWeight<W>) -> Self {
		Error::NegativeWeight
	}
}
