//! Contains the additional functionality enabled by crate-feature "uom" by
//! making the si-part of [uom] compatible with the statistical collections.
//!  
//! [uom]: https://crates.io/crates/uom

use uom::{
	fmt::DisplayStyle,
	num_traits::Num,
	si::{self, fmt::QuantityArguments, Dimension, Quantity, Unit, Units},
	Conversion,
};

use crate::random_variable::{RandomVariable, Sample};

/// Type of displayable objects returned by [RandomVariable]'s formatting
/// methods.
pub struct RandomQuantity<'s, D, U, V, N, const M: usize>
where
	D: Dimension + ?Sized,
	U: Units<V> + ?Sized,
	V: Num + Conversion<V>,
	Quantity<D, U, V>: Sample,
{
	/// Reference to the random variable.
	rv: &'s RandomVariable<Quantity<D, U, V>, M>,
	/// The [DisplayStyle] with which to format the unit.
	style: DisplayStyle,
	/// The unit-type.
	unit: N,
}

/// A trait to indicate which [Dimensions](Dimension) a [Unit] is compatible
/// with.
///
/// Unfortunately, the uom-crate doesn't ship with a meta-function like this,
/// so we have to improvise.
pub trait Compatible<D: Dimension + ?Sized>: Unit {
	/// Creates a statically type-checked and displayable object that can be
	/// used for formatting [Quantities](Quantity).
	fn format_args<U: Units<V> + ?Sized, V: Num + Conversion<V>>(
		self,
		value: Quantity<D, U, V>,
		style: DisplayStyle,
	) -> QuantityArguments<D, U, V, Self>;
}

// implement additional traits for built-in storage types
storage_types! {
	types: PrimInt, Float;

	use uom::{
		Conversion,
		fmt::DisplayStyle,
		si::{Dimension, Units, Quantity}
	};

	use core::{fmt, marker::PhantomData};

	use crate::random_variable::{RandomVariable, Sample};

	use super::{RandomQuantity, Compatible};

	// allow SI quantities to be used in RandomVariables
	impl<D,U> Sample for Quantity<D,U,V>
	where D: Dimension + ?Sized, U: Units<V> + ?Sized {
		#[inline]
		fn quantify(&self) -> f64 {
			#![allow(clippy::unnecessary_cast)]
			self.value as f64
		}

		#[inline]
		fn qualify(val: f64) -> Self {
			Quantity {
				dimension: PhantomData,
				units: PhantomData,
				value: val as V
			}
		}
	}

	// extend RandomVariables with display-methods for compatible SI quantities
	impl<D,U,const M: usize> RandomVariable<Quantity<D,U,V>,M>
	where D: Dimension + ?Sized,
		  U: Units<V> + ?Sized,
	{
		/// Returns a [displayable](fmt::Display) object that can be used as a
		/// format argument and displays the computed statistical moments with
		/// units of measurement in [abbreviated] form.
		///
		/// [abbreviated]: DisplayStyle::Abbreviation
		#[inline]
		pub fn display<N>(&self, unit: N) -> RandomQuantity<D,U,V,N,M>
			where N: Compatible<D> + Conversion<V, T = <V as Conversion<V>>::T> {
			RandomQuantity { rv: self, style: DisplayStyle::Abbreviation, unit }
		}

		/// Returns a [displayable](fmt::Display) object that can be used as a
		/// format argument and displays the computed statistical moments with
		/// units of measurement in [descriptive] form.
		///
		/// [descriptive]: DisplayStyle::Description
		#[inline]
		pub fn display_full<N>(&self, unit: N) -> RandomQuantity<D,U,V,N,M>
			where N: Compatible<D> + Conversion<V, T = <V as Conversion<V>>::T> {
			RandomQuantity { rv: self, style: DisplayStyle::Description, unit }
		}
	}

	// display (dimension-) compatible RandomQuantities
	impl<D,U,N> fmt::Debug for RandomQuantity<'_,D,U,V,N,0>
	where D: Dimension + ?Sized,
		  U: Units<V> + ?Sized,
		  N: Compatible<D> + Conversion<V, T = <V as Conversion<V>>::T>,
	{
		fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
			let mut dbs = f.debug_struct("RandomVariable");
			let len = self.rv.count();

			dbs.field("n", &len);

			if len > 0 {
				dbs.field("min", &self.unit.format_args(
					self.rv.min().unwrap(), self.style
				)).field("max", &self.unit.format_args(
					self.rv.max().unwrap(), self.style
				));
			}

			dbs.finish()
		}
	}

	impl<D,U,N> fmt::Debug for RandomQuantity<'_,D,U,V,N,1>
	where D: Dimension + ?Sized,
		  U: Units<V> + ?Sized,
		  N: Compatible<D> + Conversion<V, T = <V as Conversion<V>>::T>,
	{
		fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
			let mut dbs = f.debug_struct("RandomVariable");
			let len = self.rv.count();

			dbs.field("n", &len);

			if len > 0 {
				dbs.field("min", &self.unit.format_args(
					self.rv.min().unwrap(), self.style
				)).field("max", &self.unit.format_args(
					self.rv.max().unwrap(), self.style
				)).field("mean", &self.unit.format_args(
					Quantity::<D,U,V>::qualify(self.rv.mean()), self.style
				));
			}

			dbs.finish()
		}
	}

	impl<D,U,N> fmt::Debug for RandomQuantity<'_,D,U,V,N,2>
	where D: Dimension + ?Sized,
		  U: Units<V> + ?Sized,
		  N: Compatible<D> + Conversion<V, T = <V as Conversion<V>>::T>,
	{
		fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
			let mut dbs = f.debug_struct("RandomVariable");
			let len = self.rv.count();

			dbs.field("n", &len);

			if len > 0 {
				dbs.field("min", &self.unit.format_args(
					self.rv.min().unwrap(), self.style
				)).field("max", &self.unit.format_args(
					self.rv.max().unwrap(), self.style
				)).field("mean", &self.unit.format_args(
					Quantity::<D,U,V>::qualify(self.rv.mean()), self.style
				)).field("sdev", &self.unit.format_args(
					Quantity::<D,U,V>::qualify(self.rv.std_dev()), self.style
				));
			}

			dbs.finish()
		}
	}

	impl<D,U,N> fmt::Debug for RandomQuantity<'_,D,U,V,N,3>
	where D: Dimension + ?Sized,
		  U: Units<V> + ?Sized,
		  N: Compatible<D> + Conversion<V, T = <V as Conversion<V>>::T>,
	{
		fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
			let mut dbs = f.debug_struct("RandomVariable");
			let len = self.rv.count();

			dbs.field("n", &len);

			if len > 0 {
				dbs.field("min", &self.unit.format_args(
					self.rv.min().unwrap(), self.style
				)).field("max", &self.unit.format_args(
					self.rv.max().unwrap(), self.style
				)).field("mean", &self.unit.format_args(
					Quantity::<D,U,V>::qualify(self.rv.mean()), self.style
				)).field("sdev", &self.unit.format_args(
					Quantity::<D,U,V>::qualify(self.rv.std_dev()), self.style
				)).field("skew", &self.rv.skew());
			}

			dbs.finish()
		}
	}

	impl<D,U,N> fmt::Debug for RandomQuantity<'_,D,U,V,N,4>
	where D: Dimension + ?Sized,
		  U: Units<V> + ?Sized,
		  N: Compatible<D> + Conversion<V, T = <V as Conversion<V>>::T>,
	{
		fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
			let mut dbs = f.debug_struct("RandomVariable");
			let len = self.rv.count();

			dbs.field("n", &len);

			if len > 0 {
				dbs.field("min", &self.unit.format_args(
					self.rv.min().unwrap(), self.style
				)).field("max", &self.unit.format_args(
					self.rv.max().unwrap(), self.style
				)).field("mean", &self.unit.format_args(
					Quantity::<D,U,V>::qualify(self.rv.mean()), self.style
				)).field("sdev", &self.unit.format_args(
					Quantity::<D,U,V>::qualify(self.rv.std_dev()), self.style
				)).field("skew", &self.rv.skew())
				  .field("kurt", &self.rv.kurtosis());
			}

			dbs.finish()
		}
	}
}

/// Macro to add compatibility to a uom-[Quantity] to the simulation library.
#[macro_export]
macro_rules! uom_compatible {
	($($($P:ident)::+),* $(,)?) => {$(
		impl<T> $crate::uom_si::Compatible<$($P)::*::Dimension> for T
		where T: $($P)::*::Unit {
			#[inline(always)]
			fn format_args<U: ::uom::si::Units<V> + ?Sized, V: ::uom::num_traits::Num + ::uom::Conversion<V>>(
				self, value: ::uom::si::Quantity<$($P)::*::Dimension,U,V>, style: ::uom::fmt::DisplayStyle
			) -> uom::si::fmt::QuantityArguments<$($P)::*::Dimension,U,V,Self> {
				value.into_format_args(self, style)
			}
		}
	)*}
}

uom_compatible!(
	si::absement,
	si::acceleration,
	si::action,
	si::amount_of_substance,
	si::angle,
	si::angular_acceleration,
	si::angular_jerk,
	si::angular_velocity,
	si::area,
	si::areal_density_of_states,
	si::areal_mass_density,
	si::areal_number_density,
	si::areal_number_rate,
	si::available_energy,
	si::capacitance,
	si::catalytic_activity,
	si::catalytic_activity_concentration,
	si::curvature,
	si::diffusion_coefficient,
	si::dynamic_viscosity,
	si::electric_charge,
	si::electric_charge_areal_density,
	si::electric_charge_linear_density,
	si::electric_charge_volumetric_density,
	si::electric_current,
	si::electric_current_density,
	si::electric_dipole_moment,
	si::electric_displacement_field,
	si::electric_field,
	si::electric_flux,
	si::electric_permittivity,
	si::electric_potential,
	si::electric_quadrupole_moment,
	si::electrical_conductance,
	si::electrical_conductivity,
	si::electrical_mobility,
	si::electrical_resistance,
	si::electrical_resistivity,
	si::energy,
	si::force,
	si::frequency,
	si::frequency_drift,
	si::heat_capacity,
	si::heat_flux_density,
	si::heat_transfer,
	si::inductance,
	si::information,
	si::information_rate,
	si::jerk,
	si::length,
	si::linear_density_of_states,
	si::linear_mass_density,
	si::linear_number_density,
	si::linear_number_rate,
	si::linear_power_density,
	si::luminance,
	si::luminous_intensity,
	si::magnetic_field_strength,
	si::magnetic_flux,
	si::magnetic_flux_density,
	si::magnetic_moment,
	si::magnetic_permeability,
	si::mass,
	si::mass_concentration,
	si::mass_density,
	si::mass_flux,
	si::mass_rate,
	si::molality,
	si::molar_concentration,
	si::molar_energy,
	si::molar_flux,
	si::molar_heat_capacity,
	si::molar_mass,
	si::molar_radioactivity,
	si::molar_volume,
	si::moment_of_inertia,
	si::momentum,
	si::power,
	si::pressure,
	si::radiant_exposure,
	si::radioactivity,
	si::ratio,
	si::reciprocal_length,
	si::solid_angle,
	si::specific_area,
	si::specific_heat_capacity,
	si::specific_power,
	si::specific_radioactivity,
	si::specific_volume,
	si::surface_electric_current_density,
	si::temperature_coefficient,
	si::temperature_gradient,
	si::temperature_interval,
	si::thermal_conductance,
	si::thermal_conductivity,
	si::thermodynamic_temperature,
	si::time,
	si::torque,
	si::velocity,
	si::volume,
	si::volume_rate,
	si::volumetric_density_of_states,
	si::volumetric_heat_capacity,
	si::volumetric_number_density,
	si::volumetric_number_rate,
	si::volumetric_power_density,
);
