//! A module containing a data structure for intrusively linked lists of
//! [notifyable] objects that is used to construct [Chains] of [Tasks] on the
//! stack.
//!
//! [Chains]: Chain
//! [notifyable]: Subscriber
//! [Tasks]: odem_rs_core::task::Task

use core::{cell::Cell, fmt, marker::PhantomData, task::Waker};
use std::pin::Pin;

use intrusive_collections::{intrusive_adapter, LinkedList, LinkedListLink, UnsafeRef};

use crate::{Publisher, Subscriber};

/* ******************************************************************** Chain */

/// A shared list of intrusively [linked] objects that can be [notified].
///
/// [linked]: Link
/// [notified]: Subscriber
pub struct Chain<N> {
	inner: Cell<LinkedList<LinkAdapter<N>>>,
}

impl<N> Chain<N> {
	/// Creates a new, empty chain.
	#[inline]
	pub fn new() -> Self {
		Chain {
			inner: Cell::new(LinkedList::new(LinkAdapter::NEW)),
		}
	}

	/// Removes all of the queued-up objects without notifying them.
	#[inline]
	pub fn clear(&mut self) {
		self.inner.get_mut().clear();
	}

	/// Constructs a Notifier for the first linked object.
	///
	/// At most one object is notified once the [`go`]-method is called and the
	/// chain isn't empty, but the caller may execute additional methods that
	/// change which object is notified and what happens in that case.
	///
	/// The notified object (if any) is automatically removed from the chain.
	///
	/// [`go`]: NotifyOne::go
	#[inline]
	pub fn notify_one(&self) -> NotifyOne<N>
	where
		N: Subscriber,
	{
		NotifyOne::new(self)
	}

	/// Constructs a Notifier for all of the linked objects.
	///
	/// Potentially all of the objects in the chain are notified when the [`go`]
	/// method is called, but the caller may execute additional methods that
	/// change which objects are notified and what happens in that case.
	///
	/// All notified objects are automatically removed from the chain.
	///
	/// [`go`]: NotifyAll::go
	#[inline]
	pub fn notify_all(&self) -> NotifyAll<N>
	where
		N: Subscriber,
	{
		NotifyAll::new(self)
	}

	/// Removes the first element of the chain.
	///
	/// Returns `None` if the chain is empty.
	#[inline]
	fn pop_front(&self) -> Option<UnsafeRef<Link<N>>> {
		let mut inner = self.inner.take();
		let res = inner.pop_front();
		self.inner.set(inner);
		res
	}
}

impl<N: Subscriber> Subscriber for Chain<N> {
	/// Notifies all of the linked objects, removing them all from the chain.
	#[inline]
	fn notify(&self) {
		self.notify_all().go();
	}
}

impl<N: Subscriber> Publisher for Chain<N> {
	type Link = Link<N>;

	#[inline]
	unsafe fn subscribe(&self, link: Pin<&Self::Link>) {
		let mut inner = self.inner.take();
		inner.push_back(UnsafeRef::from_raw(&*link));
		self.inner.set(inner);
	}

	#[inline]
	unsafe fn unsubscribe(&self, link: Pin<&Self::Link>) {
		if link.hook.is_linked() {
			let mut inner = self.inner.take();
			inner.cursor_mut_from_ptr(&*link).remove();
			self.inner.set(inner);
		}
	}
}

impl<N: fmt::Debug> fmt::Debug for Chain<N> {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		let inner = self.inner.take();
		let res = inner.fmt(f);
		self.inner.set(inner);
		res
	}
}

impl<N> Default for Chain<N> {
	#[inline]
	fn default() -> Self {
		Self::new()
	}
}

impl<N> Drop for Chain<N> {
	#[inline]
	fn drop(&mut self) {
		self.clear();
	}
}

/* *********************************************************** Intrusive Link */

/// A wrapper type to augment objects that may be notified into intrusive
/// links that can be stored in intrusive collections.
#[pin_project::pin_project]
pub struct Link<N: ?Sized> {
	/// Marks the structure as non-Send and non-Sync.
	_mark: PhantomData<*const ()>,
	/// Contains the intrusive link.
	#[pin]
	hook: LinkedListLink,
	/// The object to be notified.
	pub note: N,
}

impl<N> Link<N> {
	/// Initializes a new link with the notifiable object.
	#[inline]
	pub const fn new(waker: N) -> Self {
		Link {
			hook: LinkedListLink::new(),
			note: waker,
			_mark: PhantomData,
		}
	}

	/// Returns whether this link is in use or not.
	#[inline]
	pub fn is_linked(&self) -> bool {
		self.hook.is_linked()
	}

	/// Strips the outer link and returns the inner `N`.
	#[inline]
	pub fn into_inner(self) -> N {
		self.note
	}
}

impl<N> From<N> for Link<N> {
	#[inline]
	fn from(value: N) -> Self {
		Self::new(value)
	}
}

impl<N: ?Sized + Subscriber> Subscriber for Link<N> {
	#[inline]
	fn notify(&self) {
		self.note.notify();
	}
}

impl<N: fmt::Debug> fmt::Debug for Link<N> {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		f.debug_struct("Link")
			.field("hook", &self.hook)
			.field("note", &self.note)
			.finish()
	}
}

impl fmt::Debug for Link<dyn Subscriber> {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		f.debug_struct("Link").field("hook", &self.hook).finish()
	}
}

intrusive_adapter! {
	/// An adapter for the [Link]-structure, allowing it to be inserted into an
	/// intrusive [LinkedList].
	pub LinkAdapter<N> = UnsafeRef<Link<N>>: Link<N> {
		hook: LinkedListLink
	}
}

/* ***************************************************** Notification Helpers */

/// Trait used to trigger the notification process after the
/// notify-configuration is complete.
pub trait Trigger {
	/// The result-type of the triggered notification operation.
	type Output;

	/// The notification operation itself.
	fn trigger(self) -> Self::Output;
}

/* *************************************************************** Notify One */

/// Notification builder that seeks, notifies and removes the first fitting
/// object in a [Chain].
#[must_use = "value must be consumed with go() in order to run"]
pub struct NotifyOne<'c, N, P = (), F = ()> {
	/// A reference to the [Chain].
	chain: &'c Chain<N>,
	/// A boolean function that is used to skip elements in the chain that are
	/// not ready to be notified.
	filter: P,
	/// A function that is called at most once for the selected chain element.
	///
	/// This function is called before the actual notification is submitted.
	action: F,
}

impl<'c, N: Subscriber> NotifyOne<'c, N, (), ()> {
	/// Creates a notification configuration that notifies the first object in
	/// the chain.
	const fn new(chain: &'c Chain<N>) -> Self {
		NotifyOne {
			chain,
			filter: (),
			action: (),
		}
	}
}

impl<'c, N: Subscriber, P, F> NotifyOne<'c, N, P, F> {
	/// Configures the notification configuration with a boolean predicate that
	/// allows skipping the first sequence of objects that aren't ready to be
	/// notified.
	#[inline]
	pub fn filter<X: Fn(&N) -> bool>(self, filter: X) -> NotifyOne<'c, N, X, F> {
		NotifyOne {
			chain: self.chain,
			filter,
			action: self.action,
		}
	}

	/// Configures what happens after a suitable object has been found for
	/// notification but before it is actually notified.
	#[inline]
	pub fn then<X: FnOnce(&N) -> R, R>(self, action: X) -> NotifyOne<'c, N, P, X> {
		NotifyOne {
			chain: self.chain,
			filter: self.filter,
			action,
		}
	}

	/// Triggers the notification.
	#[inline]
	pub fn go(self) -> <Self as Trigger>::Output
	where
		Self: Trigger,
	{
		Trigger::trigger(self)
	}
}

impl<N, P, F, R> Trigger for NotifyOne<'_, N, P, F>
where
	N: Subscriber,
	P: Fn(&N) -> bool,
	F: FnOnce(&N) -> R,
{
	type Output = Option<R>;

	fn trigger(self) -> Self::Output {
		// take a local copy of the chain and replace it with an empty list
		let mut chain = self.chain.inner.replace(LinkedList::new(LinkAdapter::NEW));

		// get a mutable cursor for iteration and removal
		let mut cur = chain.front_mut();
		let mut out = None;

		// iterate over the elements
		while let Some(link) = cur.get() {
			// check the predicate
			if (self.filter)(&link.note) {
				// found the element: call the function, notify and remove
				let res = (self.action)(&link.note);
				link.notify();
				cur.remove();

				// store the result and break off the loop
				out = Some(res);
				break;
			}
		}

		// restore the chain
		self.chain.inner.replace(chain);

		out
	}
}

impl<N: Subscriber, P: Fn(&N) -> bool> Trigger for NotifyOne<'_, N, P, ()> {
	type Output = ();

	fn trigger(self) -> Self::Output {
		// take a local copy of the chain and replace it with an empty list
		let mut chain = self.chain.inner.replace(LinkedList::new(LinkAdapter::NEW));

		// get a mutable cursor for iteration and removal
		let mut cur = chain.front_mut();

		// iterate over the elements
		while let Some(link) = cur.get() {
			// check the predicate
			if (self.filter)(&link.note) {
				// found the element: notify and remove
				link.notify();
				cur.remove();
				break;
			}
		}

		// restore the chain
		self.chain.inner.replace(chain);
	}
}

impl<N: Subscriber, F: FnOnce(&N) -> R, R> Trigger for NotifyOne<'_, N, (), F> {
	type Output = Option<R>;

	fn trigger(self) -> Self::Output {
		// immediately select the first link to notifiy
		self.chain.pop_front().map(move |link| {
			// run the provided closure before notifying
			let res = (self.action)(&link.note);
			link.notify();
			res
		})
	}
}

impl<N: Subscriber> Trigger for NotifyOne<'_, N, (), ()> {
	type Output = ();

	#[inline]
	fn trigger(self) -> Self::Output {
		// immediately select the first link to notify
		if let Some(link) = self.chain.pop_front() {
			link.notify();
		}
	}
}

/* *************************************************************** Notify All */

/// Notification builder that locates, notifies and removes all fitting objects
/// in a [Chain].
#[must_use = "value must be consumed with go() in order to run"]
pub struct NotifyAll<'c, N, P = (), F = ()> {
	/// A reference to the [Chain].
	chain: &'c Chain<N>,
	/// A boolean function that is used to skip elements that are not ready to
	/// be notified.
	filter: P,
	/// A function that is called for every fitting element immediately prior
	/// to its notification.
	///
	/// This function may be used to aggregate data over the sequence of fitting
	/// objects.
	action: F,
}

impl<'c, N: Subscriber> NotifyAll<'c, N, (), ()> {
	/// Creates a notification configuration that notifies all of the objects
	/// contained in the chain.
	///
	/// All successfully notified objects are removed from the chain.
	#[inline]
	const fn new(chain: &'c Chain<N>) -> Self {
		NotifyAll {
			chain,
			filter: (),
			action: (),
		}
	}
}

impl<'c, N: Subscriber, P, F> NotifyAll<'c, N, P, F> {
	/// Configures the notification configuration with a boolean predicate that
	/// allows skipping the objects that aren't ready to be notified.
	#[inline]
	pub fn filter<X: Fn(&N) -> bool>(self, filter: X) -> NotifyAll<'c, N, X, F> {
		NotifyAll {
			chain: self.chain,
			filter,
			action: self.action,
		}
	}

	/// Configures what happens for every suitable object that has been found
	/// for notification right before it is actually notified.
	#[inline]
	pub fn then<X: FnMut(&N)>(self, action: X) -> NotifyAll<'c, N, P, X> {
		NotifyAll {
			chain: self.chain,
			filter: self.filter,
			action,
		}
	}

	/// Triggers the notification.
	#[inline]
	pub fn go(self) -> <Self as Trigger>::Output
	where
		Self: Trigger,
	{
		Trigger::trigger(self)
	}
}

impl<N, P, F> Trigger for NotifyAll<'_, N, P, F>
where
	N: Subscriber,
	P: Fn(&N) -> bool,
	F: FnMut(&N),
{
	type Output = ();

	fn trigger(mut self) -> Self::Output {
		// take a local copy of the chain and replace it with an empty list
		let mut chain = self.chain.inner.replace(LinkedList::new(LinkAdapter::NEW));

		// get a mutable cursor for iteration and removal
		let mut cur = chain.front_mut();

		// iterate over the elements
		while let Some(link) = cur.get() {
			// check the predicate
			if (self.filter)(&link.note) {
				// located an element: call the function, notify and remove
				(self.action)(&link.note);
				link.notify();
				cur.remove();
			} else {
				cur.move_next();
			}
		}

		// restore the chain
		self.chain.inner.replace(chain);
	}
}

impl<N: Subscriber, P: Fn(&N) -> bool> Trigger for NotifyAll<'_, N, P, ()> {
	type Output = ();

	fn trigger(self) -> Self::Output {
		// take a local copy of the chain and replace it with an empty list
		let mut chain = self.chain.inner.replace(LinkedList::new(LinkAdapter::NEW));

		// get a mutable cursor for iteration and removal
		let mut cur = chain.front_mut();

		// iterate over the elements
		while let Some(link) = cur.get() {
			// check the predicate
			if (self.filter)(&link.note) {
				// located an element: remove and notify
				cur.remove().unwrap().notify();
			} else {
				cur.move_next();
			}
		}

		// restore the chain
		self.chain.inner.replace(chain);
	}
}

impl<N: Subscriber, F: FnMut(&N)> Trigger for NotifyAll<'_, N, (), F> {
	type Output = ();

	fn trigger(mut self) -> Self::Output {
		while let Some(link) = self.chain.pop_front() {
			(self.action)(&link.note);
			link.notify();
		}
	}
}

impl<N: Subscriber> Trigger for NotifyAll<'_, N, (), ()> {
	type Output = ();

	#[inline]
	fn trigger(self) -> Self::Output {
		while let Some(link) = self.chain.pop_front() {
			link.notify();
		}
	}
}

/* *********************************************************** Waker Notifier */

impl Subscriber for Waker {
	#[inline]
	fn notify(&self) {
		self.wake_by_ref();
	}
}
