//! This module contains support for MPMC-channels which synchronize sender and
//! receiver tasks over a message queue.

use core::{
	cell::{Cell, UnsafeCell},
	fmt,
	future::Future,
	hint::unreachable_unchecked,
	marker::PhantomData,
	mem::MaybeUninit,
	pin::Pin,
	task::{Context, Poll, Waker},
};
use std::ptr::NonNull;

use odem_rs_core::ptr::{IntrusivelyCounted, Irc, IrcBox, IrcBoxed};

use crate::{
	chain::{Chain, Link},
	error::{RecvError, SendError, TryRecvError, TrySendError},
	Publisher, Subscriber,
};

#[cfg(feature = "alloc")]
pub mod shared;

/* ********************************************************** Generic Channel */

/// Alias for a runtime-type-generic [RingBuf] to be used when abstracting from
/// the underlying buffer of a [Channel].
pub type AnyBuf<T> = dyn RingBuf<Item = T>;

/// A channel which can be used to exchange values between concurrent [Tasks] of
/// the simulation library.
///
/// This channel uses a generic backing store to buffer sent but not yet
/// received messages. The backing store must implement the [RingBuf] trait and
/// depending on its type may use heap or stack memory to store its items.
///
/// The channel itself may be kept on the stack.
///
/// [Tasks]: odem_rs_core::task::Task
pub struct Channel<R: ?Sized + RingBuf>(Inner<R::Item, R>);

/// Helper structure to allow the unsizing coercion to function in stable Rust.
///
/// The reason this indirection is needed is that unsizing doesn't work if the
/// unsized template argument is referred to more than once in the struct.
/// Using this structure allows us to separate the associated type of the
/// [`RingBuf`]-trait from the potentially unsized ring buffer.
struct Inner<T, R: ?Sized + RingBuf<Item = T>> {
	/// A chain of tasks waiting for reactivation as soon as there is a message
	/// to send or to receive.
	chain: Chain<OneShot<T, Waker>>,
	/// The number of tasks waiting to receive or to send a message. Negative
	/// values indicate tasks waiting to receive a message, positive numbers
	/// indicate tasks waiting to send a message and 0 indicates that no tasks
	/// are waiting to send or receive messages.
	pending: Cell<isize>,
	irc_box: IrcBox<ChannelBox>,
	/// A ring buffer containing messages to be received. May have capacity `0`,
	/// in which case the corresponding channel is a rendezvous channel.
	buffer: UnsafeCell<R>,
}

pub struct ChannelBox {
	ref_count: Cell<usize>,
}

impl<R: RingBuf> Channel<R> {
	/// Creates a new channel that uses the backing store passed as an argument.
	#[inline]
	#[track_caller]
	pub fn buffered(buffer: R) -> Self {
		Channel(Inner {
			chain: Chain::new(),
			pending: Cell::new(0),
			irc_box: IrcBox::new(ChannelBox {
				ref_count: Cell::new(0),
			}),
			buffer: UnsafeCell::new(buffer),
		})
	}
}

impl<R: RingBuf + ?Sized> Channel<R> {
	/// Closes the channel causing any waiting sender or receiver to return
	/// immediately with an error indicating a disconnect. Any additional send
	/// operation after the channel is closed fails immediately. Receive
	/// operations fail once the message buffer is empty.
	#[inline]
	pub fn close(&self) {
		self.0.pending.set(isize::MIN);
		self.0.chain.notify_all().go();
	}

	/// Returns whether this channel is closed or not.
	#[inline]
	pub fn is_closed(&self) -> bool {
		self.0.pending.get() == isize::MIN
	}

	/// Resets the channel by clearing the message buffer and re-opening it if
	/// it has been closed before. Panics if there are still [send](Self::send)
	/// or [recv](Self::recv) operations pending.
	#[inline]
	pub fn reset(&self) {
		// reset the number of pending sender/receiver
		match self.0.pending.get() {
			0 | isize::MIN => self.0.pending.set(0),
			n => panic!(
				"resetting the channel failed because there are {} pending \
				 {}-operations",
				n.abs(),
				if n > 0 { "send" } else { "recv" }
			),
		}

		// clear the message buffer
		unsafe { &mut *self.0.buffer.get() }.clear();
	}

	/// Attempt to send the value without waiting.
	///
	/// This can fail for two reasons:
	/// 1. the channel is closed, i.e. [`close`] has been called,
	/// 2. the message buffer is full.
	///
	/// The `Err`-result contains the initially provided item in either case.
	///
	/// [`close`]: Channel::close
	pub fn try_send(&self, item: R::Item) -> Result<(), TrySendError<R::Item>> {
		let pending = self.0.pending.get();

		// test if there are receivers waiting for a message
		if pending < 0 {
			// test if the channel is still open
			if pending > isize::MIN {
				// decrease the number of pending messages
				self.0.pending.set(pending + 1);

				// transmit the item to the first receiver
				self.0
					.chain
					.notify_one()
					.then(|c| c.give(item))
					.go()
					.expect("receiver waiting for message");

				Ok(())
			} else {
				// closed: don't accept any more messages
				Err(TrySendError::Disconnected(item))
			}
		} else {
			// insert the message into the buffer
			self.give(item).map_err(TrySendError::Full)
		}
	}

	/// Returns a future that gets fulfilled when the value has been written to
	/// the channel. If the channel gets closed while sending is in progress,
	/// it will fail, and the future will deliver the value back.
	#[inline]
	pub const fn send(&self, item: R::Item) -> SendFuture<R> {
		SendFuture::new(self, item)
	}

	/// Attempts to receive a value of the channel without waiting.
	///
	/// This can fail for two reasons:
	/// 1. the channel is closed, i.e. [`close`] has been called,
	/// 2. the message buffer is empty.
	///
	/// [`close`]: Channel::close
	pub fn try_recv(&self) -> Result<R::Item, TryRecvError> {
		// define a lambda extracting an item from the buffer or
		// returning the corresponding error if that is not possible
		let pending = match self.0.pending.get() {
			pending if pending > 0 => Ok({
				// receive a message from the first sender
				let item = unsafe {
					self.0
						.chain
						.notify_one()
						.then(OneShot::take)
						.go()
						.unwrap_unchecked()
						.unwrap_unchecked()
				};

				self.0.pending.set(pending - 1);
				item
			}),
			isize::MIN => Err(TryRecvError::Disconnected),
			_ => Err(TryRecvError::Empty),
		};

		// attempt to extract a message from the underlying buffer
		if let Some(msg) = self.take() {
			// test if there are senders waiting to transmit a message
			if let Ok(item) = pending {
				self.give(item)
					.ok()
					.expect("buffer should not be full after an element has been extracted");
			}

			Ok(msg)
		} else {
			pending
		}
	}

	/// Returns a future that gets fulfilled when a value is written to the
	/// channel. If the channels gets closed, the future will resolve to an
	/// `Err` variant.
	#[inline]
	pub const fn recv(&self) -> RecvFuture<R> {
		RecvFuture::new(self)
	}

	/// Returns the number of elements currently queued to be received.
	#[inline]
	pub fn count(&self) -> isize {
		unsafe { &*self.0.buffer.get() }.len() as isize + self.0.pending.get()
	}

	/// Places an item into the internal message buffer. If the buffer is
	/// already full, the item is returned to the caller through the `Err`
	/// variant.
	#[inline]
	fn give(&self, item: R::Item) -> Result<(), R::Item> {
		unsafe { &mut *self.0.buffer.get() }.give(item)
	}

	/// Takes the first item out of the message buffer. Returns `None` if the
	/// buffer is empty.
	#[inline]
	fn take(&self) -> Option<R::Item> {
		unsafe { &mut *self.0.buffer.get() }.take()
	}
}

impl<T> Channel<Rendezvous<T>> {
	/// Creates a type of channel that doesn't use a backing store at all,
	/// instead directly synchronizing senders and receivers, passing messages.
	///
	/// This means that every [`send`] blocks unless and until there is a
	/// receiver to accept the message.
	///
	/// [`send`]: Self::send
	#[inline]
	pub fn unbuffered() -> Self {
		Channel::buffered(Rendezvous::new())
	}
}

impl<T> Channel<ArrayBuf<0, T>> {
	/// Creates a channel with a *fixed* size backing store, allowing sender and
	/// receiver to decouple if the buffer isn't full.
	///
	/// This means that [`send`] blocks only if the buffer is full.
	///
	/// [`send`]: Self::send
	#[inline]
	pub fn array<const N: usize>() -> Channel<ArrayBuf<N, T>> {
		Channel::buffered(ArrayBuf::new())
	}
}

#[cfg(feature = "alloc")]
impl<T> Channel<alloc::collections::VecDeque<T>> {
	/// Creates a channel with a *variably* sized backing store, allowing sender
	/// and receiver to decouple completely.
	///
	/// This means that [`send`] never blocks and [`try_send`] only fails for
	/// closed channels.
	///
	/// [`send`]: Self::send
	/// [`try_send`]: Self::try_send
	#[inline]
	pub fn deque() -> Self {
		Channel::buffered(Default::default())
	}
}

impl<R: RingBuf + Default> Default for Channel<R> {
	#[inline]
	fn default() -> Self {
		Self::buffered(R::default())
	}
}

unsafe impl<R: ?Sized + RingBuf> IntrusivelyCounted for Channel<R> {
	type Inner = ChannelBox;

	fn irc_box(&self) -> &IrcBox<Self::Inner> {
		&self.0.irc_box
	}
}

unsafe impl IrcBoxed for ChannelBox {
	#[inline]
	unsafe fn acquire(&self) -> usize {
		let refs = self.ref_count.get() + 1;
		self.ref_count.set(refs);
		refs
	}

	#[inline]
	unsafe fn release(&self) -> Option<fn(NonNull<Self>)> {
		self.ref_count.set(self.ref_count.get() - 1);
		None
	}
}

/// Creates a new channel, returning the sender/receiver halves.
///
/// All data sent on the [Sender] half will become available on the [Receiver]
/// in the same order as it was sent. If the underlying [RingBuf] is dynamically
/// sized, no [send] will block. [recv] will block until a message is available
/// while there is at least one Sender alive (including clones).
///
/// Both Sender and Receiver may be cloned, allowing receiving and sending
/// messages over the same channel from different parts of the program.
///
/// The underlying [Channel] is closed automatically when the last Sender
/// or Receiver drops.
///
/// [send]: Sender::send
/// [recv]: Receiver::recv
#[cfg(feature = "alloc")]
#[inline]
pub fn channel<R>(buffer: R) -> (shared::Sender<R>, shared::Receiver<R>)
where
	R: RingBuf,
{
	use alloc::rc::Rc;
	let channel = Rc::new(Channel::buffered(buffer));
	(
		shared::Sender(Rc::downgrade(&channel)),
		shared::Receiver(channel),
	)
}

// #[inline]
// pub fn irc_channel<R>(chan: Pinnable<Channel<R>>) -> (Sender<R>, Receiver<R>)
// where R: RingBuf {
// 	let channel = Irc::from(chan);
// 	(Sender(channel.clone()), Receiver(channel))
// }

/* **************************************************** Sender/Receiver Parts */

pub struct Sender<R: ?Sized + RingBuf>(Irc<Channel<R>>);

impl<'r, R: RingBuf + 'r> Sender<R> {
	/// Returns a `Sender` that is type-erased regarding the buffer underlying
	/// the [Channel].
	#[inline]
	pub fn erase(self) -> Sender<dyn RingBuf<Item = R::Item> + 'r> {
		Sender(Irc::map(self.0.clone(), |inner| inner as &Channel<_>))
	}
}

impl<R: ?Sized + RingBuf> Sender<R> {
	/// Attempt to send the value without waiting.
	///
	/// This can fail for two reasons:
	/// 1. the channel is closed, i.e. [`close`] has been called,
	/// 2. the message buffer is full.
	///
	/// The `Err`-result contains the initially provided item in either case.
	///
	/// [`close`]: Channel::close
	#[inline]
	pub fn try_send(&self, item: R::Item) -> Result<(), TrySendError<R::Item>> {
		self.0.try_send(item)
	}

	/// Returns a future that gets fulfilled when the value has been written to
	/// the channel. If the channel gets closed while the send is in progress,
	/// sending the value will fail, and the future will deliver the value back.
	#[inline]
	pub fn send(&self, item: R::Item) -> SendFuture<R> {
		SendFuture::new(&self.0, item)
	}

	/// Returns the number of elements currently inside the channel of this
	/// sender.
	#[inline]
	pub fn count(&self) -> isize {
		self.0.count()
	}

	/// Closes the channel from the sending side. Already sent values may
	/// still be received.
	#[inline]
	pub fn close(&self) {
		self.0.close();
	}
}

impl<R: RingBuf + ?Sized> Clone for Sender<R> {
	#[inline]
	fn clone(&self) -> Self {
		Sender(self.0.clone())
	}
}

pub struct Receiver<R: ?Sized + RingBuf>(Irc<Channel<R>>);

impl<'r, R: RingBuf + 'r> Receiver<R> {
	/// Returns a `Sender` that is type-erased regarding the buffer underlying
	/// the [Channel].
	#[inline]
	pub fn erase(self) -> Receiver<dyn RingBuf<Item = R::Item> + 'r> {
		Receiver(Irc::map(self.0.clone(), |inner| inner as &Channel<_>))
	}
}

impl<R: ?Sized + RingBuf> Receiver<R> {
	/// Attempts to receive a value of the channel without waiting.
	///
	/// This can fail for two reasons:
	/// 1. the channel is closed, i.e. [`close`] has been called,
	/// 2. the message buffer is empty.
	///
	/// [`close`]: Channel::close
	#[inline]
	pub fn try_recv(&self) -> Result<R::Item, TryRecvError> {
		self.0.try_recv()
	}

	/// Returns a future that gets fulfilled when a value is written to the
	/// channel. If the channels gets closed, the future will resolve to an
	/// `Err` variant.
	#[inline]
	pub fn recv(&self) -> RecvFuture<R> {
		self.0.recv()
	}

	/// Returns the number of elements currently inside the channel of this
	/// receiver.
	#[inline]
	pub fn count(&self) -> isize {
		self.0.count()
	}

	/// Closes the channel from the receiving side. Already sent values may
	/// still be received.
	#[inline]
	pub fn close(&self) {
		self.0.close();
	}
}

impl<R: ?Sized + RingBuf> Clone for Receiver<R> {
	#[inline]
	fn clone(&self) -> Self {
		Receiver(self.0.clone())
	}
}

/* ***************************************************** Send/Receive Futures */

/// Future for the [send](Channel::send) operation.
#[pin_project::pin_project(PinnedDrop)]
pub struct SendFuture<'s, R: RingBuf + ?Sized> {
	/// The channel used to send the message.
	chan: &'s Channel<R>,
	/// Either the message or the link into the pending chain.
	#[pin]
	state: SendFutureState<R::Item>,
}

impl<'s, R: RingBuf + ?Sized> SendFuture<'s, R> {
	/// Creates a [send](Channel::send)-future and primes it with a message.
	///
	/// The send-operation is attempted once this future is resolved.
	#[inline]
	const fn new(chan: &'s Channel<R>, msg: R::Item) -> Self {
		SendFuture {
			chan,
			state: SendFutureState::Primed(Some(msg)),
		}
	}
}

impl<R: RingBuf + ?Sized> Future for SendFuture<'_, R> {
	type Output = Result<(), SendError<R::Item>>;

	fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
		let mut projection = self.project();

		match projection.state.as_mut().project() {
			SendFutureStateProject::Primed(msg) => {
				// extract the message
				let msg = msg.take().unwrap();

				// attempt to send it
				match projection.chan.try_send(msg) {
					Err(TrySendError::Full(msg)) => {
						// the message queue is already full
						// create a link and subscribe to the pending chain
						unsafe {
							projection.state.link(
								&projection.chan.0.chain,
								cx.waker().clone(),
								msg,
							);
						}
						// increase the number of pending senders
						projection
							.chan
							.0
							.pending
							.set(projection.chan.0.pending.get() + 1);
						// defer
						Poll::Pending
					}
					Err(TrySendError::Disconnected(msg)) => {
						// the channel has been closed
						// return an error
						Poll::Ready(Err(SendError(msg)))
					}
					Ok(()) => {
						// the message was sent successfully
						// return with success
						Poll::Ready(Ok(()))
					}
				}
			}
			SendFutureStateProject::Linked(link) => {
				// this future has been reactivated after subscribing
				// to the pending chain

				// either the message has been received (success)
				// or the channel has been closed (failure)
				Poll::Ready(if let Some(msg) = link.note.take() {
					Err(SendError(msg))
				} else {
					Ok(())
				})
			}
		}
	}
}

#[pin_project::pinned_drop]
impl<R: RingBuf + ?Sized> PinnedDrop for SendFuture<'_, R> {
	#[inline]
	fn drop(self: Pin<&mut Self>) {
		let projection = self.project();

		// make sure to unlink the soon-to-be-dangling link if the future has
		// been dropped after subscribing but before reactivation
		if unsafe { projection.state.as_ref().unlink(&projection.chan.0.chain) } {
			projection
				.chan
				.0
				.pending
				.set(projection.chan.0.pending.get() - 1);
		}
	}
}

/// Future for the [recv](Channel::recv) operation.
#[pin_project::pin_project(PinnedDrop)]
pub struct RecvFuture<'s, R: RingBuf + ?Sized> {
	/// The channel used to receive the message.
	chan: &'s Channel<R>,
	/// Either `None` if no linking has been attempted or `Some` link into the
	/// pending chain.
	#[pin]
	link: Option<Link<OneShot<R::Item, Waker>>>,
}

impl<'s, R: RingBuf + ?Sized> RecvFuture<'s, R> {
	/// Creates a new [recv](Channel::recv)-future without a link.
	#[inline]
	const fn new(chan: &'s Channel<R>) -> Self {
		RecvFuture { chan, link: None }
	}
}

impl<R: RingBuf + ?Sized> Future for RecvFuture<'_, R> {
	type Output = Result<R::Item, RecvError>;

	fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
		let mut inner = self.project();

		match inner.link.as_ref().as_pin_ref() {
			None => match inner.chan.try_recv() {
				Err(TryRecvError::Empty) => {
					// receiving a message has failed because of an empty queue

					// create the link
					inner
						.link
						.set(Some(Link::new(OneShot::new(cx.waker().clone(), None))));

					// subscribe to the channel's chain
					unsafe {
						inner
							.chan
							.0
							.chain
							.subscribe(inner.link.as_ref().as_pin_ref().unwrap_unchecked());
					}

					// increase the number of pending receivers
					inner.chan.0.pending.set(inner.chan.0.pending.get() - 1);

					// defer
					Poll::Pending
				}
				Err(TryRecvError::Disconnected) => {
					// receiving has failed and the channel has been closed
					// return with an error
					Poll::Ready(Err(RecvError))
				}
				Ok(msg) => {
					// message was successfully received
					// return with success
					Poll::Ready(Ok(msg))
				}
			},
			Some(link) => {
				// reactivation after subscribing to the chain

				// either the message has been received (success)
				// or the channel has been closed (failure)
				Poll::Ready(link.note.take().ok_or(RecvError))
			}
		}
	}
}

#[pin_project::pinned_drop]
impl<R: RingBuf + ?Sized> PinnedDrop for RecvFuture<'_, R> {
	#[inline]
	fn drop(self: Pin<&mut Self>) {
		let inner = self.project();

		// make sure to unsubscribe the soon-to-be-dangling link from the chain
		match inner.link.as_pin_mut() {
			Some(link) if link.is_linked() => {
				unsafe {
					inner.chan.0.chain.unsubscribe(link.as_ref());
				}
				inner.chan.0.pending.set(inner.chan.0.pending.get() + 1);
			}
			_ => {}
		}
	}
}

/// Inner state for a [SendFuture].
#[pin_project::pin_project(
	project = SendFutureStateProject,
	project_ref = SendFutureStateProjectRef
)]
enum SendFutureState<T> {
	/// The future contains a message that hasn't been sent yet.
	Primed(Option<T>),
	/// The future's message has been attempted to send but the message queue
	/// was full.
	Linked(#[pin] Link<OneShot<T, Waker>>),
}

impl<T> SendFutureState<T> {
	/// Creates a link into the provided chain carrying a specific message and
	/// waking a specific task.
	///
	/// # Safety
	/// The caller is responsible to call this method at most once and to
	/// [unlink](Self::unlink) from the chain before the created link goes out
	/// of scope.
	unsafe fn link(
		mut self: Pin<&mut Self>,
		chain: &Chain<OneShot<T, Waker>>,
		waker: Waker,
		msg: T,
	) {
		// create the new link
		self.set(Self::Linked(Link::new(OneShot::new(waker, Some(msg)))));

		// subscribe to it
		match self.project() {
			SendFutureStateProject::Linked(link) => chain.subscribe(link.as_ref()),
			_ => unreachable_unchecked(),
		}
	}

	/// Unlinks a link from the chain.
	///
	/// # Safety
	/// The caller is responsible to unlink the link from the same chain it may
	/// have previously been linked to. If the link was never established or
	/// unlinking has already happened, this method does nothing.
	unsafe fn unlink(self: Pin<&Self>, chain: &Chain<OneShot<T, Waker>>) -> bool {
		match self.project_ref() {
			SendFutureStateProjectRef::Linked(link) if link.is_linked() => {
				chain.unsubscribe(link);
				true
			}
			_ => false,
		}
	}
}

/* ******************************************************** Ring Buffer Trait */

/// The trait that qualifies a buffer to be used as a backing store for
/// [Channels](Channel).
pub trait RingBuf {
	/// The type of item stored inside the ring buffer.
	type Item;

	/// Places an item into the buffer, returning an `Ok` result if successful
	/// and the item wrapped in an `Err` result if the buffer is full.
	fn give(&mut self, item: Self::Item) -> Result<(), Self::Item>;

	/// Extracts an item from the buffer, returning `None` if the buffer is
	/// empty.
	fn take(&mut self) -> Option<Self::Item>;

	/// Removes all stored elements from the buffer.
	fn clear(&mut self);

	/// Returns the number of elements that are currently stored in the buffer.
	fn len(&self) -> usize;

	/// Returns whether the buffer is currently empty.
	#[inline]
	fn is_empty(&self) -> bool {
		self.len() == 0
	}
}

/* ******************************************************** Rendezvous Buffer */

/// A buffer that provides an empty backing store for a [Channel].
///
/// This leads to a direct synchronization between sender and receiver before
/// messages are exchanged.
pub struct Rendezvous<T>(PhantomData<T>);

impl<T> Rendezvous<T> {
	/// Creates a new rendezvouz buffer.
	pub const fn new() -> Self {
		Rendezvous(PhantomData)
	}
}

impl<T> Default for Rendezvous<T> {
	#[inline]
	fn default() -> Self {
		Rendezvous::new()
	}
}

impl<T> fmt::Debug for Rendezvous<T> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		write!(f, "Rendezvous")
	}
}

impl<T> RingBuf for Rendezvous<T> {
	type Item = T;

	#[inline]
	fn give(&mut self, item: T) -> Result<(), T> {
		Err(item)
	}

	#[inline]
	fn take(&mut self) -> Option<T> {
		None
	}

	#[inline]
	fn clear(&mut self) {}

	#[inline]
	fn len(&self) -> usize {
		0
	}
}

/* ************************************************** Fixed Size Array Buffer */

/// A fixed-size array buffer that can be allocated on the stack and used as a
/// backing store for [channels].
///
/// [channels]: Channel
pub struct ArrayBuf<const N: usize, T> {
	/// Index of the first item stored.
	off: usize,
	/// Number of elements currently held.
	len: usize,
	/// Pre-allocated space for the stored items.
	buf: [MaybeUninit<T>; N],
}

impl<const N: usize, T> ArrayBuf<N, T> {
	/// Creates a new, empty array buffer with a user-defined - but fixed -
	/// capacity.
	#[inline]
	pub const fn new() -> Self {
		ArrayBuf {
			off: 0,
			len: 0,
			buf: unsafe { MaybeUninit::uninit().assume_init() },
		}
	}
}

impl<const N: usize, T> Default for ArrayBuf<N, T> {
	#[inline]
	fn default() -> Self {
		ArrayBuf::new()
	}
}

impl<const N: usize, T> Drop for ArrayBuf<N, T> {
	fn drop(&mut self) {
		for i in self.off..self.off + self.len {
			unsafe { self.buf.get_unchecked_mut(i % N).assume_init_drop() }
		}
	}
}

impl<const N: usize, T> RingBuf for ArrayBuf<N, T> {
	type Item = T;

	#[inline]
	fn give(&mut self, item: T) -> Result<(), T> {
		// test if there is still space in the buffer
		if self.len < N {
			// calculate the prospective index
			let idx = (self.off + self.len) % N;

			// fill the slot with the item
			// SAFETY: safe, because there is an empty slot due to the fulfilled
			// condition above, and it has the calculated index due to the
			// values of `off` and `len`
			unsafe { self.buf.get_unchecked_mut(idx).write(item) };

			// increase the number of elements
			self.len += 1;

			// return success
			Ok(())
		} else {
			// not enough space, so return the item
			Err(item)
		}
	}

	#[inline]
	fn take(&mut self) -> Option<T> {
		// test if there are still elements in the buffer
		if self.len > 0 {
			// read the first element
			// SAFETY: the item exists due to the condition fulfilled above,
			// it has the calculated index due to the value of `off`
			let item = unsafe { self.buf.get_unchecked_mut(self.off).assume_init_read() };

			// increase the offset and wrap it around
			self.off = (self.off + 1) % N;

			// decrease the number of stored items
			self.len -= 1;

			// return the item
			Some(item)
		} else {
			// no elements: return `None`
			None
		}
	}

	#[inline]
	fn clear(&mut self) {
		while self.len > 0 {
			// read the first element
			// SAFETY: the item exists due to the condition fulfilled above,
			// it has the calculated index due to the value of `off`
			unsafe { self.buf.get_unchecked_mut(self.off).assume_init_read() };

			// increase the offset and wrap it around
			self.off = (self.off + 1) % N;

			// decrease the number of stored items
			self.len -= 1;
		}
	}

	#[inline]
	fn len(&self) -> usize {
		self.len
	}
}

/* ********************************************* Variably Sized Vector Buffer */

#[cfg(feature = "alloc")]
impl<T> RingBuf for alloc::collections::VecDeque<T> {
	type Item = T;

	#[inline]
	fn give(&mut self, item: T) -> Result<(), T> {
		// simply push the item into the deque
		self.push_back(item);

		// this operation cannot fail without a panic
		Ok(())
	}

	#[inline]
	fn take(&mut self) -> Option<T> {
		// attempt to extract the first element from the deque
		Self::pop_front(self)
	}

	#[inline]
	fn clear(&mut self) {
		Self::clear(self);
	}

	#[inline]
	fn len(&self) -> usize {
		self.len()
	}
}

#[cfg(feature = "alloc")]
impl<T: Ord> RingBuf for alloc::collections::BinaryHeap<T> {
	type Item = T;

	#[inline]
	fn give(&mut self, item: T) -> Result<(), T> {
		self.push(item);
		Ok(())
	}

	#[inline]
	fn take(&mut self) -> Option<T> {
		self.pop()
	}

	#[inline]
	fn clear(&mut self) {
		self.clear();
	}

	#[inline]
	fn len(&self) -> usize {
		self.len()
	}
}

/* ****************************************************** OneShot Signal Type */

/// Wrapper for [Subscriber] objects that allow transmission of a single
/// value as part of the notification.
pub struct OneShot<T, N> {
	/// The value being transmitted.
	payload: Cell<Option<T>>,
	/// The object to be notified when the value has been sent or received.
	notify: N,
}

impl<T, N> OneShot<T, N> {
	/// Creates a new signal with the payload either left empty or initialized
	/// with some value.
	pub const fn new(notify: N, payload: Option<T>) -> Self {
		OneShot {
			payload: Cell::new(payload),
			notify,
		}
	}

	/// Removes the payload and returns it.
	#[inline]
	pub fn take(&self) -> Option<T> {
		self.payload.replace(None)
	}

	/// Overwrites the current payload with an item.
	#[inline]
	pub fn give(&self, item: T) {
		self.payload.set(Some(item));
	}
}

impl<T, N: Subscriber> Subscriber for OneShot<T, N> {
	#[inline]
	fn notify(&self) {
		self.notify.notify();
	}
}

#[cfg(test)]
mod tests {
	use super::*;
	use crate::ForkExt;
	use odem_rs_core::simulator::{Sim, Simulator};
	use proptest::prelude::*;

	async fn test_insertion_order(
		sim: &Sim,
		buf: impl RingBuf<Item = usize> + 'static,
		elems: usize,
	) {
		let (sx, rx) = channel(buf);

		sim.fork(async move {
			for i in 0..elems {
				sx.send(i).await.unwrap();
				sim.advance(1.0).await;
			}
		})
		.and(async move {
			sim.advance(10.0).await;
			for exp in 0..elems {
				let val = rx.recv().await.unwrap();
				assert_eq!(val, exp);
			}
		})
		.await;
	}

	proptest! {
		#[test]
		fn rendezvous_channels_respect_insertion_order(elems in 1..100usize) {
			Simulator::default()
				.run(|sim| test_insertion_order(sim, Rendezvous::default(), elems))
				.expect("unexpected deadlock");
		}

		#[test]
		fn array_channels_respect_insertion_order(elems in 1..100usize) {
			Simulator::default()
				.run(|sim| test_insertion_order(sim, ArrayBuf::<2,_>::new(), elems))
				.expect("unexpected deadlock");
		}

		#[cfg(feature = "alloc")]
		#[test]
		fn deque_channels_respect_insertion_order(elems in 1..100usize) {
			use alloc::collections::VecDeque;
			Simulator::default()
				.run(|sim| test_insertion_order(sim, VecDeque::new(), elems))
				.expect("unexpected deadlock");
		}
	}
}
