//! Module for the owned and reference-counted version of the [Channel].

use alloc::rc::{Rc, Weak};

use std::{
	future::Future,
	pin::Pin,
	task::{Context, Poll},
};

use super::{
	Channel, RecvFuture, RingBuf, SendError, SendFutureState, SendFutureStateProject, TryRecvError,
	TrySendError,
};

/* ***************************************************** Sender/Receiver Pair */

/// The sending half of the [Channel] type. This half can be cloned and is
/// bound to the hosting channel through the reference counted [Rc].
#[derive(Default)]
pub struct Sender<R: RingBuf + ?Sized>(pub(super) Weak<Channel<R>>);

impl<'r, R: RingBuf + 'r> Sender<R> {
	/// Returns a `Sender` that is type-erased regarding the buffer underlying
	/// the [Channel].
	#[inline]
	pub fn erase(self) -> Sender<dyn RingBuf<Item = R::Item> + 'r> {
		Sender(self.0.clone())
	}
}

impl<R: RingBuf + ?Sized> Sender<R> {
	/// Attempt to send the value without waiting.
	///
	/// This can fail for two reasons:
	/// 1. the channel is closed, i.e. [`close`] has been called,
	/// 2. the message buffer is full.
	///
	/// The `Err`-result contains the initially provided item in either case.
	///
	/// [`close`]: Channel::close
	#[inline]
	pub fn try_send(&self, item: R::Item) -> Result<(), TrySendError<R::Item>> {
		if let Some(chan) = self.0.upgrade() {
			chan.try_send(item)
		} else {
			Err(TrySendError::Disconnected(item))
		}
	}

	/// Returns a future that gets fulfilled when the value has been written to
	/// the channel. If the channel gets closed while the send is in progress,
	/// sending the value will fail, and the future will deliver the value back.
	#[inline]
	pub const fn send(&self, item: R::Item) -> SendFuture<R> {
		SendFuture::new(self, item)
	}

	/// Returns the number of elements currently inside the channel of this
	/// sender.
	#[inline]
	pub fn count(&self) -> isize {
		self.0.upgrade().map(|inner| inner.count()).unwrap_or(0)
	}

	/// Closes the channel from the sending side. Already sent values may
	/// still be received.
	#[inline]
	pub fn close(&self) {
		if let Some(chan) = self.0.upgrade() {
			chan.close();
		}
	}
}

impl<R: RingBuf + ?Sized> Clone for Sender<R> {
	#[inline]
	fn clone(&self) -> Self {
		Sender(self.0.clone())
	}
}

impl<R: RingBuf + ?Sized> Drop for Sender<R> {
	#[inline]
	fn drop(&mut self) {
		if self.0.weak_count() == 1 {
			if let Some(chan) = self.0.upgrade() {
				chan.close();
			}
		}
	}
}

/// The receiving half of the [Channel] type. This half can be cloned and is
/// bound to the hosting channel through the reference counted [Rc].
#[derive(Default)]
pub struct Receiver<R: ?Sized + RingBuf>(pub(super) Rc<Channel<R>>);

impl<'r, R: RingBuf + 'r> Receiver<R> {
	/// Returns a `Receiver` that is type-erased regarding the buffer underlying
	/// the [Channel].
	#[inline]
	pub fn erase(self) -> Receiver<dyn RingBuf<Item = R::Item> + 'r> {
		Receiver(self.0.clone())
	}
}

impl<R: RingBuf + ?Sized> Receiver<R> {
	/// Attempts to receive a value of the channel without waiting.
	///
	/// This can fail for two reasons:
	/// 1. the channel is closed, i.e. [`close`] has been called,
	/// 2. the message buffer is empty.
	///
	/// [`close`]: Channel::close
	#[inline]
	pub fn try_recv(&self) -> Result<R::Item, TryRecvError> {
		self.0.try_recv()
	}

	/// Returns a future that gets fulfilled when a value is written to the
	/// channel. If the channels gets closed, the future will resolve to an
	/// `Err` variant.
	#[inline]
	pub fn recv(&self) -> RecvFuture<R> {
		self.0.recv()
	}

	/// Returns the number of elements currently inside the channel of this
	/// receiver.
	#[inline]
	pub fn count(&self) -> isize {
		self.0.count()
	}

	/// Closes the channel from the receiving side. Already sent values may
	/// still be received.
	#[inline]
	pub fn close(&self) {
		self.0.close();
	}
}

impl<R: RingBuf + ?Sized> Clone for Receiver<R> {
	#[inline]
	fn clone(&self) -> Self {
		Receiver(self.0.clone())
	}
}

impl<R: ?Sized + RingBuf> Drop for Receiver<R> {
	#[inline]
	fn drop(&mut self) {
		// notify all the senders of the implicit closing of the channel
		if Rc::strong_count(&self.0) == 1 {
			self.close();
		}
	}
}

/* *************************************************** Specialized SendFuture */

/// Future for the [send](Sender::send) operation of the shared channel.
///
/// We cannot re-use the SendFuture of the channel because our sender only
/// contains a weak reference to the channel.
#[pin_project::pin_project(PinnedDrop)]
pub struct SendFuture<'s, R: RingBuf + ?Sized> {
	/// The sender used to send the message.
	chan: &'s Sender<R>,
	/// Either the message or the link into the pending chain.
	#[pin]
	state: SendFutureState<R::Item>,
}

impl<'s, R: RingBuf + ?Sized> SendFuture<'s, R> {
	/// Creates a [send](Sender::send)-future and primes it with a message.
	///
	/// The send-operation is attempted once this future is resolved.
	#[inline]
	const fn new(chan: &'s Sender<R>, msg: R::Item) -> Self {
		SendFuture {
			chan,
			state: SendFutureState::Primed(Some(msg)),
		}
	}
}

impl<R: RingBuf + ?Sized> Future for SendFuture<'_, R> {
	type Output = Result<(), SendError<R::Item>>;

	fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
		let mut inner = self.project();

		match inner.state.as_mut().project() {
			SendFutureStateProject::Primed(msg) => {
				// extract the message
				let msg = msg.take().unwrap();

				// attempt to get an owning reference to the channel
				if let Some(chan) = inner.chan.0.upgrade() {
					// attempt to send the message
					match chan.try_send(msg) {
						Err(TrySendError::Full(msg)) => {
							// the message queue is already full
							// create a link and subscribe to the pending chain
							unsafe {
								inner.state.link(&chan.0.chain, cx.waker().clone(), msg);
							}
							// increase the number of pending senders
							chan.0.pending.set(chan.0.pending.get() + 1);
							// defer
							Poll::Pending
						}
						Err(TrySendError::Disconnected(msg)) => {
							// the channel has been closed
							// return an error
							Poll::Ready(Err(SendError(msg)))
						}
						Ok(()) => {
							// the message was sent successfully
							// return with success
							Poll::Ready(Ok(()))
						}
					}
				} else {
					// failed: the channel is closed and dropped
					// return with an error
					Poll::Ready(Err(SendError(msg)))
				}
			}
			SendFutureStateProject::Linked(link) => {
				// this future has been reactivated after subscribing
				// to the pending chain

				// either the message has been received (success)
				// or the channel has been closed (failure)
				Poll::Ready(if let Some(msg) = link.note.take() {
					Err(SendError(msg))
				} else {
					Ok(())
				})
			}
		}
	}
}

#[pin_project::pinned_drop]
impl<R: RingBuf + ?Sized> PinnedDrop for SendFuture<'_, R> {
	#[inline]
	fn drop(self: Pin<&mut Self>) {
		// make sure to unlink the soon-to-be-dangling link if the future has
		// been dropped after subscribing but before reactivation
		if let Some(chan) = self.chan.0.upgrade() {
			if unsafe { self.as_ref().project_ref().state.unlink(&chan.0.chain) } {
				chan.0.pending.set(chan.0.pending.get() - 1);
			}
		}
	}
}
