//! This module contains support for awaiting arbitrary state events in the form
//! of control expressions.
//!
//! Control expressions are types that implement the [Publisher] and [Expr]
//! trait, which allows them to [calculate] a value and [notify] dependent
//! tasks of changes in their value.
//! Once a task has subscribed, it is automatically activated by the control
//! expression whenever its internal state changes.
//!
//! This is used for implementing a mechanism that allows processes to await
//! specific conditions with regards to the system state, which is made
//! ergonomic through the [until!]-macro.
//!
//! The simplest control expressions are [Control]-variables that capture a
//! single, [copyable] value and automatically notify subscribed tasks when it
//! changes.
//!
//! [calculate]: Expr::get
//! [notify]: Subscriber::notify
//! [until!]: crate::until
//! [copyable]: Copy

use core::{
	cell::Cell,
	fmt,
	future::{Future, IntoFuture},
	pin::Pin,
	task::Waker,
	task::{Context, Poll},
};

use intrusive_collections::UnsafeRef;

use crate::{
	chain::{Chain, Link},
	Publisher, Subscriber,
};

pub mod imp;

/* ******************************************************* Control Expression */

/// An object-safe trait for repeatedly evaluating a complex expression.
pub trait Expr {
	/// The result type.
	type Output;

	/// The routine used for evaluation.
	fn get(&self) -> Self::Output;
}

/* ********************************************************* Control Variable */

/// Marks a variable as a state variable which can be used in conjunction with
/// the [until!]-macro.
///
/// Whenever a control variable changes its value, the runtime system checks
/// whether any other parts of the model are currently waiting for the variable
/// to attain a certain value.
///
/// [until!]: crate::until!
#[derive(Default)]
pub struct Control<T: ?Sized> {
	/// A [Chain] of waiting tasks.
	queue: Chain<UnsafeRef<Predicate>>,
	/// The inner value.
	value: Cell<T>,
}

impl<T> Control<T> {
	/// Creates a new control variable and initializes its value.
	#[inline]
	pub fn new(value: T) -> Self {
		Self {
			value: Cell::new(value),
			queue: Chain::new(),
		}
	}

	/// Assigns a new value to the control variable.
	///
	/// This can lead to potential activations of waiting processes.
	pub fn set(&self, val: T) {
		self.value.set(val);
		self.notify();
	}

	/// Returns a copy of the stored inner value.
	#[inline]
	pub fn get(&self) -> T
	where
		T: Copy,
	{
		self.value.get()
	}

	/// Swaps the values of two control variables. The difference to
	/// [`core::mem::swap`] is that this function doesn't require an exclusive
	/// reference.
	pub fn swap(&self, other: &Self) {
		self.value.swap(&other.value);
		self.notify();
	}

	/// Replaces the current value of the control variable with a new one and
	/// returns it.
	///
	/// This can lead to potential activations of waiting processes.
	pub fn replace(&self, val: T) -> T {
		let res = self.value.replace(val);
		self.notify();
		res
	}

	/// Updates the contained value using a function and returns the new value.
	#[inline]
	pub fn update<F>(&self, f: F)
	where
		F: FnOnce(T) -> T,
		T: Copy,
	{
		self.set(f(self.get()))
	}

	/// Takes the value of the control variable, leaving `Default::default()` in
	/// its place.
	#[inline]
	pub fn take(&self) -> T
	where
		T: Default,
	{
		self.replace(T::default())
	}

	/// Returns a mutable reference to the underlying data.
	///
	/// This call borrows the control variable mutably (at compile-time) which
	/// guarantees that we possess the only reference.
	#[inline]
	pub fn get_mut(&mut self) -> &mut T {
		self.value.get_mut()
	}

	/// Assigns a new value to the control variable **without notifying pending
	/// control conditions**.
	///
	/// Use this method if you want to change the inner value without scheduling
	/// tasks that depend on it. Once finished, you may call [notify] directly.
	///
	/// [notify]: Self::notify
	#[inline]
	pub fn unchecked_set(&self, val: T) {
		self.value.set(val);
	}
}

impl<T: ?Sized> Publisher for Control<T> {
	type Link = Link<UnsafeRef<Predicate>>;

	#[inline]
	unsafe fn subscribe(&self, link: Pin<&Self::Link>) {
		self.queue.subscribe(link);
	}

	#[inline]
	unsafe fn unsubscribe(&self, link: Pin<&Self::Link>) {
		self.queue.unsubscribe(link);
	}
}

impl<T: ?Sized> Subscriber for Control<T> {
	#[inline]
	fn notify(&self) {
		// notify the set of waiting tasks with fulfilled conditions
		self.queue.notify_all().filter(Expr::get).go();
	}
}

impl<T: Copy> Expr for Control<T> {
	type Output = T;

	#[inline]
	fn get(&self) -> T {
		self.value.get()
	}
}

impl<T: fmt::Debug> fmt::Debug for Control<T> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		f.debug_tuple("Control")
			.field(unsafe { &*self.value.as_ptr() })
			.finish()
	}
}

impl<T: fmt::Display> fmt::Display for Control<T> {
	#[inline]
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		unsafe { &*self.value.as_ptr() }.fmt(f)
	}
}

/* ******************************************************* Control Expression */

/// Container for complex control expressions that combines a set of [Control]
/// variables with a closure that depends on the values of the control set.
#[derive(Copy, Clone)]
pub struct ControlExpr<C, F> {
	/// The set of (potential) control variables that the closure depends on.
	depends: C,
	/// A function closure accepting the current values of the control variables
	/// and calculating the result.
	closure: F,
}

impl<C, F> ControlExpr<C, F>
where
	C: Publisher,
	F: Fn(&C) -> bool,
{
	/// Creates a new control expression given a set of [Control] variables and
	/// a closure that depends on them.
	#[inline]
	pub const fn new(ctrl: C, code: F) -> Self {
		ControlExpr {
			depends: ctrl,
			closure: code,
		}
	}
}

impl<C, F> Publisher for ControlExpr<C, F>
where
	C: Publisher,
{
	type Link = C::Link;

	#[inline]
	unsafe fn subscribe(&self, link: Pin<&Self::Link>) {
		self.depends.subscribe(link);
	}

	#[inline]
	unsafe fn unsubscribe(&self, link: Pin<&Self::Link>) {
		self.depends.unsubscribe(link);
	}
}

impl<C, F, R> Expr for ControlExpr<C, F>
where
	F: Fn(&C) -> R,
{
	type Output = R;

	fn get(&self) -> R {
		(self.closure)(&self.depends)
	}
}

impl<C, F> IntoFuture for ControlExpr<C, F>
where
	C: Publisher,
	F: Fn(&C) -> bool,
	C::Link: From<UnsafeRef<Predicate>>,
{
	type Output = ();
	type IntoFuture = Until<C, F>;

	#[inline]
	fn into_future(self) -> Self::IntoFuture {
		Until::new(self)
	}
}

/* ************************************************************* Until Future */

/// Notify dispatcher that includes the boolean condition being awaited.
///
/// The basic idea is to reduce the number of spurious reactivations of waiting
/// tasks by checking if the condition being awaited evaluated to `true` for the
/// specific state-change triggering the re-evaluation and not re-awakening the
/// task if it doesn't.
///
/// Even if it does, the re-awakened task still needs to check the condition
/// since its truth value may have changed between the scheduling and the
/// activation, but this technique still has the potential to cut down on a lot
/// of superfluous scheduling operations.
#[pin_project::pin_project(PinnedDrop)]
pub struct Until<C: Publisher, F> {
	/// A predicate of the boolean condition.
	#[pin]
	poll: Predicate<ControlExpr<C, F>>,
	/// Space for the link needed during polling.
	#[pin]
	link: Option<C::Link>,
}

impl<C: Publisher, F> Until<C, F> {
	/// Creates a new future that blocks on being awaited until the passed
	/// condition is fulfilled.
	#[inline]
	pub const fn new(cond: ControlExpr<C, F>) -> Self {
		Until {
			poll: Predicate {
				note: Cell::new(None),
				cond,
			},
			link: None,
		}
	}
}

impl<C, F> Future for Until<C, F>
where
	C: Publisher,
	F: Fn(&C) -> bool,
	C::Link: From<UnsafeRef<Predicate>>,
{
	type Output = ();

	fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
		// test if the condition evaluates to true and terminate if it does
		if self.poll.get() {
			return Poll::Ready(());
		}

		let mut projection = self.project();

		// clone the waker
		projection.poll.note.set(Some(cx.waker().clone()));

		// unsubscribe from the control set
		// SAFETY: we will only unsubscribe if we have been subscribed before
		// using the original link
		if let Some(link) = projection.link.as_ref().as_pin_ref() {
			unsafe {
				projection.poll.cond.unsubscribe(link);
			}
		}

		// fill the option with the link
		// SAFETY: the lifetime-extension is justified because we are sure to
		// unsubscribe during drop; drop happens because of pin
		projection.link.set(Some(
			unsafe {
				UnsafeRef::from_raw(&*core::mem::transmute::<
					Pin<&Predicate<dyn Expr<Output = bool> + '_>>,
					Pin<&Predicate<dyn Expr<Output = bool> + 'static>>,
				>(projection.poll.as_ref()))
			}
			.into(),
		));

		// subscribe to the control set
		// SAFETY: pinning guarantees that the link's address is stable
		// until this future drops, which will also unsubscribe the link
		unsafe {
			projection
				.poll
				.cond
				.subscribe(projection.link.as_ref().as_pin_ref().unwrap_unchecked());
		}

		// await the next reactivation
		Poll::Pending
	}
}

#[pin_project::pinned_drop]
impl<C: Publisher, F> PinnedDrop for Until<C, F> {
	#[inline]
	fn drop(self: Pin<&mut Self>) {
		// unsubscribe from control set
		if let Some(link) = self.as_ref().project_ref().link.as_pin_ref() {
			unsafe { self.poll.cond.unsubscribe(link) }
		}
	}
}

/// A boolean predicate with an attached [Waker] to notify.
pub struct Predicate<C: ?Sized = dyn Expr<Output = bool>> {
	/// The [Waker] to notify.
	note: Cell<Option<Waker>>,
	/// The boolean condition as a dynamically dispatched trait.
	cond: C,
}

impl<C: Expr<Output = bool> + ?Sized> Expr for Predicate<C> {
	type Output = bool;

	#[inline]
	fn get(&self) -> bool {
		self.cond.get()
	}
}

impl<C: ?Sized> Subscriber for Predicate<C> {
	#[inline]
	fn notify(&self) {
		// only wake the process once
		if let Some(waker) = self.note.take() {
			waker.wake();
		}
	}
}

/* *************************************************************** Expr Impls */

// implement Expr for primitive data types
macro_rules! impl_primitive_expr {
	($($L:ty),*) => {$(
		impl Expr for $L {
			type Output = $L;

			#[inline(always)]
			fn get(&self) -> $L { *self }
		}
	)*};
}

// implement Expr for smart pointer data types
macro_rules! impl_deref_expr {
	($($($L:ident)::+),*) => {$(
		impl<T: ?Sized + Expr> Expr for $($L)::*<T> {
			type Output = T::Output;

			#[inline(always)]
			fn get(&self) -> Self::Output { (**self).get() }
		}
	)*};
}

impl_primitive_expr!(
	(),
	bool,
	char,
	f32,
	f64,
	u8,
	u16,
	u32,
	u64,
	u128,
	usize,
	i8,
	i16,
	i32,
	i64,
	i128,
	isize
);

impl_deref_expr!(UnsafeRef);

impl<T: ?Sized + Expr> Expr for Pin<&T> {
	type Output = T::Output;

	#[inline(always)]
	fn get(&self) -> Self::Output {
		(**self).get()
	}
}

#[cfg(feature = "alloc")]
impl_deref_expr!(alloc::boxed::Box, alloc::rc::Rc, alloc::sync::Arc);

/* ******************************************************************* Macros */

/// A macro to simplify the formulation of state events.
///
/// The underlying implementation is procedural and defined in crate
/// `odem-rs-meta` which is automatically included as a dependency.
///
/// The macro contains the expression to wait for and uses the [Waker] to
/// identify and reactivate the current task.
///
/// Please note that [ControlExpr] implement [IntoFuture] which means that they
/// can be `await`ed directly.
///
/// # Semantics
///
/// The macro considers all combinations of paths, field accesses and index
/// operations as possible control variables, except inside of index
/// calculations. Control expressions are allowed to contain normal
/// (non-control) variables and constants - they will simply not be notified
/// when their value changes.
///
/// The reason why control variables in index calculations are not supported is
/// because allowing them would create dependent variables, resulting in
/// dynamic sets of control variables, e.g. by specifying a dependency to a
/// whole slice of control variables via a dependency to its index. Allowing
/// that would complicate the dispatch mechanism and prevent certain
/// optimizations that the implementation performs.
///
/// # Examples
///
/// Simple macro call:
/// ```
/// # use odem_rs_sync::{until, control::{Control, ControlExpr, Expr}};
/// # async fn foo(i: Control<i32>) {
/// // this macro invocation
/// until!(i != 0);
/// // expands roughly into this definition
/// ControlExpr::new(&i, |args| args.get() != 0);
/// # }
/// ```
/// Macro call with multiple control variables:
/// ```
/// # use odem_rs_sync::{until, control::{Control, ControlExpr, Expr}};
/// # async fn foo(i: Control<bool>, j: Control<bool>) {
/// // this macro invocation
/// until!(i && j);
/// // expands roughly into this definition
/// ControlExpr::new((&i,&j), |args| args.0.get() && args.1.get());
/// # }
/// ```
/// Macro call with multiple, complex control variables.
/// ```
/// # use odem_rs_sync::{until, control::{Control, ControlExpr, Expr}};
/// # struct V { inner: [Control<bool>; 2] }
/// # async fn foo(i: V) {
/// // this macro invocation
/// until!(i.inner[0] && i.inner[1]);
/// // expands roughly into this definition
/// ControlExpr::new((&i.inner[0],&i.inner[1]), |args| args.0.get() && args.1.get());
/// # }
/// ```
/// Local variables inside the macro are fully supported and will neither be
/// placed outside the closure nor replaced by references to the function
/// argument.
#[macro_export]
macro_rules! until {
	($E:expr) => {{
		#[allow(unused_imports)]
		use core::ops::Deref;
		#[allow(unused_imports)]
		use $crate::control::{
			imp::{DefaultTerm, ExprTerm, PubTerm},
			ControlExpr, Expr,
		};
		$crate::control_expr!($E)
	}};
	($($T:tt)*) => {
		compile_error!("only expressions are allowed to be used here")
	};
}

#[cfg(test)]
mod tests {
	use crate::{Control, Publisher};
	use core::pin::Pin;
	use odem_rs_core::{config::DefaultConfig, fiber::Fiber, simulator::Simulator, unique};

	struct Foo(Control<bool>);

	impl Foo {
		fn is_false(&self) -> bool {
			!self.0.get()
		}
	}

	impl Publisher for Foo {
		type Link = <Control<bool> as Publisher>::Link;

		unsafe fn subscribe(&self, link: Pin<&Self::Link>) {
			self.0.subscribe(link);
		}

		unsafe fn unsubscribe(&self, link: Pin<&Self::Link>) {
			self.0.unsubscribe(link);
		}
	}

	#[test]
	fn simple_control_expr() {
		let mut simulator = Simulator::new(DefaultConfig);

		simulator
			.run(|_| async move {
				until!(true).await;
			})
			.expect("no errors during execution");

		assert_eq!(simulator.model_time(), 0.0);
	}

	#[test]
	fn complex_control_expr() {
		let mut simulator = Simulator::new(DefaultConfig);

		simulator
			.run(|_| async move {
				let t1 = false;
				let t2 = Foo(Control::new(false));
				let t3 = Control::new(false);
				let t4 = Box::new(Control::new(false));
				let t5 = &t4;

				until!(t1 || t2.is_false() || t3 || t4 || t5).await;
			})
			.expect("no errors during execution");

		assert_eq!(simulator.model_time(), 0.0);
	}

	#[test]
	fn simple_scenario() {
		let mut simulator = Simulator::new(DefaultConfig);

		simulator
			.run(|sim| async move {
				let t1 = Control::new(false);
				let fiber = unique!(Fiber::new(async {
					sim.advance(1.0).await;
					t1.set(true);
				}));

				sim.activate(fiber);
				until!(t1).await;
			})
			.expect("no errors during execution");

		assert_eq!(simulator.model_time(), 1.0);
	}

	#[test]
	fn complex_scenario() {
		let mut simulator = Simulator::new(DefaultConfig);

		simulator
			.run(|sim| async move {
				let t1 = Control::new(0);
				let t2 = Control::new(false);

				let fiber = unique!(Fiber::new(async {
					loop {
						sim.advance(1.0).await;
						t1.update(|inner| inner + 1);
						t2.update(|inner| !inner);
					}
				}));

				sim.activate(fiber);
				until!(t1 > 5 && t2).await;
			})
			.expect("no errors during execution");

		assert_eq!(simulator.model_time(), 7.0);
	}
}
