//! Utility crate that exposes traits with identically named trait methods that
//! are used to detect whether expressions implement [Expr] or [Publisher] and
//! provide reasonable default behavior if they don't.

use super::Expr;
use crate::Publisher;

use core::pin::Pin;

/* ******************************************************** Control Terminals */

/// Fallback trait that is implemented for *all* reference-types to be imported
/// via `use` and selected when the instance a trait method is called on
/// *doesn't* implement the [Publisher]- or [Expr]-traits respectively.
///
/// The mechanism affords us a certain optionality when it comes to implementing
/// those two traits for custom types by automatically wrapping instances into
/// wrapper types ([DefaultPub] and [DefaultExpr]) that provide reasonable
/// default implementations for the `Publisher` and `Expr` traits.
///
/// If neither of the two traits are implemented, it is considered an error
/// and the compilation will fail during type-checking.
///
/// This implementation can be seen as an application of the *specialization
/// in stable Rust*-trick described [here].
///
/// [here]: https://github.com/dtolnay/case-studies/blob/master/autoref-specialization/README.md
pub trait DefaultTerm: Copy {
	/// Associated *inner* type obtained after stripping the reference.
	type Inner;

	/// Trait method that is selected if the instance in question *doesn't*
	/// implement the [Expr]-trait.
	fn expr_term(&self) -> &DefaultExpr<Self::Inner>;

	/// Trait method that is selected if the instance in question *doesn't*
	/// implement the [Publisher]-trait.
	fn pub_term(&self) -> &DefaultPub<Self::Inner>;

	/// Trait method that is selected if the instance in question implements
	/// the [Expr]- or the [Publisher]-trait.
	fn expr_or_publisher(&self) -> Self {
		*self
	}
}

/// Trait declaration that is competing with the [DefaultTerm]-trait and is
/// *only* implemented for types that implement the [Publisher]-trait.
///
/// Since the compiler prefers copying before auto-referencing, the
/// [Self::pub_term]-method will always have precedence if it is in scope.
pub trait PubTerm {
	/// Trait method that is selected if the instance in question implements
	/// the [Publisher]-trait.
	#[inline(always)]
	fn pub_term(&self) -> &Self {
		self
	}
}

/// Trait declaration that is competing with the [DefaultTerm]-trait and is
/// *only* implemented for types that implement the [Expr]-trait.
///
/// Since the compiler prefers copying before auto-referencing, the
/// [Self::expr_term]-method will always have precedence if it is in scope.
pub trait ExprTerm {
	/// Trait method that is selected if the instance in question implements
	/// the [Expr]-trait.
	#[inline(always)]
	fn expr_term(&self) -> &Self {
		self
	}
}

/// A wrapper for type-instances that don't implement [Expr] natively.
#[derive(Copy, Clone)]
#[repr(transparent)]
pub struct DefaultExpr<T>(pub T);

// implement Expr by dereferencing to the inner T
impl<'s, T> Expr for &'s DefaultExpr<T> {
	type Output = &'s T;

	#[inline]
	fn get(&self) -> Self::Output {
		&self.0
	}
}

// delegate to the inner Publisher when appropriate
impl<T: Publisher> Publisher for DefaultExpr<T> {
	type Link = T::Link;

	#[inline(always)]
	unsafe fn subscribe(&self, link: Pin<&Self::Link>) {
		self.0.subscribe(link);
	}

	#[inline(always)]
	unsafe fn unsubscribe(&self, link: Pin<&Self::Link>) {
		self.0.unsubscribe(link);
	}
}

/// A wrapper for type-instances that don't implement [Publisher] natively.
///
/// This structure provides an implementation of the `Publisher`-trait that
/// doesn't subscribe to anything and is therefore not alerted when the inner
/// value changes.
#[derive(Copy, Clone)]
#[repr(transparent)]
pub struct DefaultPub<T>(pub T);

impl<T> DefaultPub<DefaultExpr<T>> {
	/// Inherent method that is selected if the type-instance didn't implement
	/// neither [Publisher] nor [Expr].
	///
	/// The returned type doesn't implement `Expr` and cannot be used to access
	/// the inner instance, leading to compile-time type errors.
	#[inline(always)]
	pub const fn expr_or_publisher(&self) -> NeitherExprNorPublisherImplemented<T> {
		NeitherExprNorPublisherImplemented(core::marker::PhantomData)
	}
}

// implement Publisher trivially
impl<T> Publisher for DefaultPub<T> {
	type Link = crate::Tuple<()>;

	#[inline(always)]
	unsafe fn subscribe(&self, _link: Pin<&Self::Link>) {}

	#[inline(always)]
	unsafe fn unsubscribe(&self, _link: Pin<&Self::Link>) {}
}

// delegate to the inner Expr when appropriate
impl<T: Expr> Expr for DefaultPub<T> {
	type Output = T::Output;

	#[inline(always)]
	fn get(&self) -> Self::Output {
		self.0.get()
	}
}

/// Helper structure to diagnose cases when a referenced expression doesn't
/// implement [Expr] nor [Publisher].
///
/// This will cause the `until!`-macro to produce a type-error for the
/// expression in question.
pub struct NeitherExprNorPublisherImplemented<T>(core::marker::PhantomData<T>);

// implement Publisher in order to reduce the number of spurious errors
impl<T> Publisher for NeitherExprNorPublisherImplemented<T> {
	type Link = crate::Tuple<()>;

	#[inline(always)]
	unsafe fn subscribe(&self, _link: Pin<&Self::Link>) {}

	#[inline(always)]
	unsafe fn unsubscribe(&self, _link: Pin<&Self::Link>) {}
}

/* **************************************************** Trait Implementations */

// blanket implementation for all reference types to fall back to
impl<T> DefaultTerm for &T {
	type Inner = T;

	#[inline(always)]
	fn expr_term(&self) -> &DefaultExpr<T> {
		// SAFETY: validity of this cast is guaranteed by #[repr(transparent)]
		unsafe { core::mem::transmute::<&T, &DefaultExpr<T>>(*self) }
	}

	#[inline(always)]
	fn pub_term(&self) -> &DefaultPub<T> {
		// SAFETY: validity of this cast is guaranteed by #[repr(transparent)]
		unsafe { core::mem::transmute::<&T, &DefaultPub<T>>(*self) }
	}
}

// blanket implement PubTerm for all Publisher types
impl<T: Publisher> PubTerm for T {}

// blanket implement ExprTerm for all Expr types
impl<T: Expr> ExprTerm for T {}
