//! A module for sync-specific error types.

/* ************************************************************** Error Types */

/// Enumeration of errors for the functions and structures contained in this
/// module of the simulation library.
#[derive(Copy, Clone, Debug, thiserror::Error)]
pub enum Error {
	/// The operation needs to block in order to complete.
	#[error("operation would block")]
	WouldBlock,
	/// The connection between Sender and Receiver has been terminated.
	#[error("connection terminated")]
	Disconnected,
	/// The facility is already seized.
	#[error("facility already seized")]
	Occupied,
}

/* ******************************************************** Minor Error Types */

/// attempt to send a value through a disconnected channel
#[derive(Debug, thiserror::Error)]
#[error("no receiver left")]
pub struct SendError<T>(pub T);

impl<T> From<SendError<T>> for Error {
	#[inline]
	fn from(_: SendError<T>) -> Self {
		Error::Disconnected
	}
}

/// attempted to receive a value from a disconnected channel
#[derive(Debug, thiserror::Error)]
#[error("no sender left")]
pub struct RecvError;

impl From<RecvError> for Error {
	#[inline]
	fn from(_: RecvError) -> Self {
		Error::Disconnected
	}
}

/// Signals that a non-blocking attempt at sending a value through a channel has
/// failed. There are two possible reasons for this: (1) the message buffer of
/// the channel is full, or (2) the channel is closed.
#[derive(Debug, thiserror::Error)]
pub enum TrySendError<T> {
	/// attempt to send a value has failed for now because the channel is full
	#[error("channel buffer is full")]
	Full(T),
	/// attempt to send a value has failed because the channel is disconnected
	#[error("channel disconnected")]
	Disconnected(T),
}

impl<T> From<TrySendError<T>> for Error {
	#[inline]
	fn from(err: TrySendError<T>) -> Self {
		match err {
			TrySendError::Full(_) => Error::WouldBlock,
			TrySendError::Disconnected(_) => Error::Disconnected,
		}
	}
}

/// Signals that a non-blocking attempt at receiving a value through a channel
/// has failed. There are two possible reasons for this: (1) the message buffer
/// of the channel is empty, or (2) the channel is closed.
#[derive(Debug, thiserror::Error)]
pub enum TryRecvError {
	/// attempt to receive a value has failed because the channel is empty
	#[error("channel buffer is empty")]
	Empty,
	/// attempt to receive a value has failed because the channel is disconnected
	#[error("channel disconnected")]
	Disconnected,
}

impl From<TryRecvError> for Error {
	#[inline]
	fn from(err: TryRecvError) -> Self {
		match err {
			TryRecvError::Empty => Error::WouldBlock,
			TryRecvError::Disconnected => Error::Disconnected,
		}
	}
}

/// Signals that a facility is already seized and cannot be seized at this time.
#[derive(Debug, thiserror::Error)]
#[error("facility already seized")]
pub struct Occupied;

impl From<Occupied> for Error {
	#[inline]
	fn from(_: Occupied) -> Self {
		Error::Occupied
	}
}
