//! This module contains support for a GPSS-like facility.

use core::{
	cell::Cell,
	future::Future,
	pin::Pin,
	task::Waker,
	task::{Context, Poll},
};

use crate::{
	chain::{Chain, Link},
	error::Occupied,
	Publisher,
};

/* ***************************************************************** Facility */

/// GPSS-inspired facility that houses one exclusive resource to be used by
/// arbitrary many processes.
pub struct Facility {
	/// A queue of waiting-to-be-activated processes.
	queue: Chain<Waker>,
	/// A boolean flag indicating whether the facility is in use or not.
	in_use: Cell<bool>,
}

impl Facility {
	/// Creates a new facility.
	pub fn new() -> Self {
		Facility {
			queue: Chain::new(),
			in_use: Cell::new(false),
		}
	}

	/// Attempts to seize the facility without blocking, returning `Err` if
	/// it failed and `Ok` otherwise.
	#[inline]
	pub fn try_seize(&self) -> Result<SeizeGuard, Occupied> {
		if self.in_use.replace(true) {
			Err(Occupied)
		} else {
			Ok(SeizeGuard(self))
		}
	}

	/// Attempts to seize the facility, blocking until it's possible.
	#[inline]
	pub fn seize(&self) -> impl Future<Output = SeizeGuard> + '_ {
		SeizeFuture::new(self)
	}

	/// Releases the facility and activates the next waiting process.
	#[inline]
	fn release(&self) {
		self.in_use.set(false);
		self.queue.notify_one().go();
	}
}

impl Default for Facility {
	#[inline]
	fn default() -> Self {
		Facility::new()
	}
}

/* ************************************************************* Seize Future */

/// Future that blocks until the underlying [Facility] can be seized.
#[pin_project::pin_project(PinnedDrop)]
struct SeizeFuture<'f> {
	/// Shared reference to the facility.
	facility: &'f Facility,
	/// Intrusive [Link] into a [Chain]. Initially `None`.
	#[pin]
	link: Option<Link<Waker>>,
}

impl<'f> SeizeFuture<'f> {
	/// Creates the future from the [Facility].
	#[inline]
	const fn new(facility: &'f Facility) -> Self {
		SeizeFuture {
			facility,
			link: None,
		}
	}
}

impl<'f> Future for SeizeFuture<'f> {
	type Output = SeizeGuard<'f>;

	fn poll(self: Pin<&mut Self>, cx: &mut Context<'_>) -> Poll<Self::Output> {
		let mut inner = self.project();

		// test if the link has been initialized previously without moving it
		if let Some(_link) = inner.link.as_ref().as_pin_ref() {
			// we must have been reactivated because the facility is free again

			// note that it is not necessary to unsubscribe here, since the
			// reactivation unsubscribed us automatically
			debug_assert!(
				!_link.is_linked(),
				"reactivation should not be possible without being \
				 unsubscribed from the chain"
			);

			// drop the old link
			inner.link.set(None);

			debug_assert!(
				!inner.facility.in_use.get(),
				"reactivation should not be possible if the facility is \
				 still in use"
			);

			// seize the facility

			// note that we only get reactivated if the facility is free
			inner.facility.in_use.set(true);
		} else if inner.facility.in_use.replace(true) {
			// the facility is currently in use and cannot be seized

			// create the pinned link
			inner.link.set(Some(Link::new(cx.waker().clone())));

			// initialize the corresponding attribute
			let link = unsafe { inner.link.as_ref().as_pin_ref().unwrap_unchecked() };

			// register for reactivation
			unsafe {
				inner.facility.queue.subscribe(link);
			}

			// await reactivation
			return Poll::Pending;
		}

		// the facility has successfully been seized by us, so return the guard
		Poll::Ready(SeizeGuard(inner.facility))
	}
}

#[pin_project::pinned_drop]
impl PinnedDrop for SeizeFuture<'_> {
	#[inline]
	fn drop(self: Pin<&mut Self>) {
		// make sure to unsubscribe on drop
		if let Some(link) = self.as_ref().project_ref().link.as_pin_ref() {
			unsafe {
				self.facility.queue.unsubscribe(link);
			}
		}
	}
}

/* ************************************************************** Seize Guard */

/// Guards the referenced [Facility] to ensure that it is vacated at the end of
/// the Guard's scope.
#[must_use = "the lifetime of this guard restricts the time this facility is occupied for"]
pub struct SeizeGuard<'f>(&'f Facility);

impl SeizeGuard<'_> {
	/// Releases the guarded facility and activates the next waiting process.
	#[inline]
	pub fn release(self) {}
}

impl Drop for SeizeGuard<'_> {
	#[inline]
	fn drop(&mut self) {
		self.0.release();
	}
}
