//! Contains an extension trait for [simulation contexts] to support
//! process-internal parallelism through [Fork].
//!
//! [simulation contexts]: Sim

use core::{
	cell::Cell,
	future::{Future, IntoFuture},
	pin::Pin,
	ptr::NonNull,
	task::{Context, Poll},
};

use odem_rs_core::{
	config::Config,
	fiber::{self, Fiber, Settle},
	fsm::Brand,
	simulator::{Prec, Sim},
	task::{token, Puck as TaskPuck},
	ExitStatus, Puck, Unique,
};

/* ***************************************************** Fork Extension Trait */

/// Extension trait for [simulation contexts] to support the [fork]-operation.
///
/// [simulation contexts]: Sim
/// [fork]: ForkExt::fork
pub trait ForkExt {
	/// Associated [simulation configuration] type.
	///
	/// [simulation configuration]: Config
	type Config: Config;

	/// Returns a future combinator that allows concurrent waiting for
	/// **any** or **all** of arbitrary many futures.
	///
	/// Alternative futures may be combined using [`or`]-method-call-chains,
	/// which will suspend until the first future terminates and return its
	/// result.
	/// Joined futures may be combined with the [`and`]-method-call-chains,
	/// which suspends until all of the futures terminate, returning a tuple
	/// of the results.
	///
	/// It is not allowed to combine alternative and joined futures in a single
	/// method-call-chain, but they may appear nested.
	///
	/// # Alternative Futures
	/// Since it is unknown beforehand which future will complete first, all of
	/// them are required to have the same return type. If you need to return
	/// different result-types from different branches, consider using an
	/// enumeration.
	///
	/// If multiple branches would complete at the same model time, only the
	/// first one in order of declaration is allowed to terminate, i.e. the
	/// branches are prioritized according to their occurrence. This is a
	/// recursive property, should you use this combinator inside one of the
	/// branches.
	///
	/// # Joined Futures
	/// This future only returns after all of the sub-futures are complete and
	/// the result type is a tuple containing all of the results.
	///
	/// There is no guarantee that multiple branches that complete at the same
	/// model time are scheduled in the order of occurrence.
	///
	/// [`or`]: Fork::or
	/// [`and`]: Fork::and
	#[track_caller]
	fn fork<F: IntoFuture>(&self, future: F) -> Fork<Self::Config, F>;
}

// implement the trait for all valid simulation contexts
impl<C: Config> ForkExt for Sim<C> {
	type Config = C;

	#[inline]
	fn fork<F: IntoFuture>(&self, future: F) -> Fork<C, F> {
		Fork {
			sim: self,
			fiber: Fiber::build().with_actions(future),
		}
	}
}

/* ********************************************************************* Fork */

/// A builder-type structure that allows assembling arbitrary many futures into
/// a concurrently-executed choice or collection.
pub struct Fork<'s, C: ?Sized + Config, F: 's> {
	/// Associated simulation context.
	sim: &'s Sim<C>,
	/// A [Cons]-list of [fiber::Builder].
	fiber: fiber::Builder<F>,
}

impl<'s, C: ?Sized + Config, F: IntoFuture> Fork<'s, C, F> {
	/// Converts the fork into a [AllOf]-type of collection.
	#[track_caller]
	#[inline]
	pub fn and<G: IntoFuture + 's>(
		self,
		future: G,
	) -> AllOf<'s, C, impl Join<C, Output = ((F::Output,), G::Output)>> {
		AllOf {
			sim: self.sim,
			tasks: Cons(
				self.fiber.with_finalizer(NotifyQuorum::new()).finish(),
				Fiber::build()
					.with_actions(future)
					.with_finalizer(NotifyQuorum::new())
					.finish(),
			),
			state: JoinState {
				counter: Cell::new(usize::MAX),
				parent: Cell::new(None),
			},
		}
	}

	/// Converts the fork into a [AnyOf]-type of choice.
	#[track_caller]
	#[inline]
	pub fn or<G: IntoFuture<Output = F::Output> + 's>(
		self,
		future: G,
	) -> AnyOf<'s, C, impl Pick<C, Output = F::Output>> {
		AnyOf {
			sim: Cell::new(Some(self.sim)),
			tasks: Cons(
				self.fiber.with_finalizer(NotifyOnExit::new()).finish(),
				Fiber::build()
					.with_actions(future)
					.with_finalizer(NotifyOnExit::new())
					.finish(),
			),
		}
	}
}

/* ***************************************************** Pick and Join traits */

/// Trait capturing a set of tasks that may be run concurrently and yield the
/// same result-type on completion such that subsequent computation does not
/// specifially depend on which task returns first.
pub trait Pick<C: ?Sized + Config> {
	/// Number of contained tasks.
	const LEN: usize;
	/// The shared result-type of all these tasks.
	type Output;

	/// Method to activate all of the tasks.
	fn start(self: Pin<&mut Self>, sim: &Sim<C>, prec: &mut impl Iterator<Item = Prec>);

	/// Method to query whether any of them ran to completion.
	fn query(self: Pin<&mut Self>) -> Option<Self::Output>;
}

/// Trait capturing a set of tasks that may be run concurrently and whose return
/// values are all needed to continue the computation.
pub trait Join<C: ?Sized + Config> {
	/// Number of contained tasks.
	const LEN: usize;
	/// The result type containing all of the contained tasks result types.
	type Output;

	/// Method to activate all of the tasks.
	///
	/// Requires some [shared state] to limit the number of reactivations of
	/// the main task.
	///
	/// [shared state]: JoinState
	fn start(self: Pin<&mut Self>, sim: &Sim<C>, state: Pin<&JoinState<C>>);

	/// Unsafe method to return the result of all of the completed tasks. Can
	/// only be safely called if all tasks completed. This can be established
	/// by observing the [shared state].
	///
	/// # Safety
	/// It's the caller's responsibility to only call this method once all
	/// tasks have completed.
	///
	/// [shared state]: JoinState
	unsafe fn query(self: Pin<&mut Self>) -> Self::Output;
}

/// Helper trait that allows flattening of the recursively-nested tuples
/// produced in [Join::Output].
///
/// The trait is only implemented for the cases required by `Join` implemented
/// here, not in general for all kinds of recursive nesting.
pub trait Flatten {
	/// Result type after the flattening.
	type Flattened;

	/// Method to perform the flattening.
	fn flatten(self) -> Self::Flattened;
}

/* ***************************************************************** Pick Any */

/// Builder-structure for a set of branches containing different futures with
/// identical result-type.
///
/// Instances are used to construct concurrently executed tasks that are
/// automatically cancelled once the first task completes. The result of the
/// future corresponding with the first task is returned upon awaiting.
#[pin_project::pin_project]
pub struct AnyOf<'s, C: ?Sized + Config, T: 's> {
	/// Associated simulation context. Doubles as the internal state.
	sim: Cell<Option<&'s Sim<C>>>,
	/// Recursively defined set of tasks.
	#[pin]
	tasks: T,
}

impl<'s, C: ?Sized + Config, T: Pick<C>> AnyOf<'s, C, T> {
	/// Adds another branch to the builder.
	///
	/// The future has to produce results of identical type when compared to the
	/// previous and following branches.
	#[track_caller]
	#[inline]
	pub fn or<F: IntoFuture<Output = T::Output> + 's>(
		self,
		future: F,
	) -> AnyOf<'s, C, impl Pick<C, Output = T::Output>> {
		AnyOf {
			sim: self.sim,
			tasks: Cons(
				self.tasks,
				Fiber::build()
					.with_actions(future)
					.with_finalizer(NotifyOnExit::new())
					.finish(),
			),
		}
	}
}

impl<'s, C: ?Sized + Config, T: Pick<C>> Future for AnyOf<'s, C, T> {
	type Output = T::Output;

	fn poll(self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<Self::Output> {
		// copy the simulation context and the previous state
		let sim = self.sim.take();

		// project the pin onto the contained tasks
		let tasks = self.project().tasks;

		// match over the current state
		match sim {
			// not initialized: recursively activate all of the tasks
			Some(sim) => tasks.start(
				sim,
				&mut sim
					.active()
					.expect("no active task")
					.prec()
					.split(T::LEN as _)
					.expect("ran out of bits to support this level of nesting"),
			),
			// active: recursively query the tasks to assess their readiness
			None => {
				if let Some(res) = tasks.query() {
					// success: return the final value
					return Poll::Ready(res);
				}
			}
		}

		Poll::Pending
	}
}

/* ***************************************************************** Join All */

/// Builder-structure for a set of branches that are started and awaited
/// concurrently.
///
/// Instances of this structure may be awaited, returning a tuple containing
/// all of the results produced by all of the branches.
#[pin_project::pin_project]
pub struct AllOf<'s, C: ?Sized + Config, T: 's> {
	/// Associated simulation context and internal state.
	sim: &'s Sim<C>,
	/// Recursively defined set of tasks.
	#[pin]
	tasks: T,
	/// State shared with all of the branches.
	#[pin]
	state: JoinState<C>,
}

/// The state shared with the tasks generated by [AllOf].
pub struct JoinState<C: ?Sized + Config> {
	/// Number of tasks that have to be completed before the [AllOf]-future
	/// can return.
	counter: Cell<usize>,
	/// Puck of the calling [AllOf]-future. Is reactivated only by the task
	/// reducing the [counter] to 0.
	///
	/// [counter]: #structfield.counter
	parent: Cell<Option<TaskPuck<C>>>,
}

impl<'s, C: ?Sized + Config, T: Join<C>> AllOf<'s, C, T> {
	/// Adds another branch to the builder.
	///
	/// There are no restrictions on how many branches the type may have or
	/// on their return type.
	#[track_caller]
	#[inline]
	pub fn and<F: IntoFuture + 's>(
		self,
		future: F,
	) -> AllOf<'s, C, impl Join<C, Output = (T::Output, F::Output)>> {
		AllOf {
			sim: self.sim,
			tasks: Cons(
				self.tasks,
				Fiber::build()
					.with_actions(future)
					.with_finalizer(NotifyQuorum::new())
					.finish(),
			),
			state: self.state,
		}
	}
}

impl<'s, C: ?Sized + Config, T: Join<C>> Future for AllOf<'s, C, T>
where
	T::Output: Flatten,
{
	type Output = <T::Output as Flatten>::Flattened;

	fn poll(self: Pin<&mut Self>, _cx: &mut Context<'_>) -> Poll<Self::Output> {
		// we need multiple different pinning projections;
		// remove the outer pin to manually project those
		let inner = self.project();

		// project the pin onto the tasks
		let tasks = inner.tasks;

		// match over the state counter
		match inner.state.counter.get() {
			// has it been initialized already?
			usize::MAX => {
				let state = inner.state.as_ref();
				// initialize with the number of tasks to complete
				state.counter.set(T::LEN);
				// initialize the calling task pin
				state.parent.set(inner.sim.active());
				// recursively distribute the shared state to the sub-tasks
				tasks.start(inner.sim, state);
			}
			// have all the sub-tasks completed successfully?
			0 => return Poll::Ready(unsafe { tasks.query() }.flatten()),
			// have we been reawakened spuriously?
			_ => {}
		}

		Poll::Pending
	}
}

/* **************************************************************** Cons List */

impl<'s, C: ?Sized + Config, F: Future + 's> Pick<C> for Fiber<'static, C, F, NotifyOnExit<C>> {
	const LEN: usize = 1;
	type Output = F::Output;

	#[inline]
	fn start(self: Pin<&mut Self>, sim: &Sim<C>, prec: &mut impl Iterator<Item = Prec>) {
		// perform a sanity check that this method isn't called twice
		debug_assert!(self.state().is_born());

		// activate the pinned fiber
		sim.activate(unsafe {
			Unique::new_unchecked(self.map_unchecked_mut(move |inner| {
				// set its precedence
				inner.set_prec(prec.next().unwrap_unchecked());
				// register the currently active task for reactivation
				inner.brand(|mut inner, once| {
					let born = token::Born::new(once);
					inner
						.finalizer(&born)
						.bind(sim.active().expect("no active task"));
				});
				// return the inner fiber unchanged
				inner
			}))
		});
	}

	#[inline(always)]
	fn query(self: Pin<&mut Self>) -> Option<Self::Output> {
		// attempt to extract the result
		self.result().ok()
	}
}

impl<'s, C: ?Sized + Config, F: Future + 's> Join<C> for Fiber<'static, C, F, NotifyQuorum<C>> {
	const LEN: usize = 1;
	type Output = (F::Output,);

	#[inline]
	fn start(self: Pin<&mut Self>, sim: &Sim<C>, state: Pin<&JoinState<C>>) {
		// perform a sanity check that this method isn't called twice
		debug_assert!(self.state().is_born());

		// activate the pinned fiber
		sim.activate(unsafe {
			Unique::new_unchecked(self.map_unchecked_mut(|inner| {
				// register the currently active task for reactivation,
				// and the number of tasks to complete before meeting the quorum
				inner.brand(|mut inner, once| {
					let born = token::Born::new(once);
					inner.finalizer(&born).bind(state);
				});
				// return the inner fiber unchanged
				inner
			}))
		});
	}

	#[inline(always)]
	unsafe fn query(self: Pin<&mut Self>) -> Self::Output {
		// extract the result
		(self.result().unwrap_unchecked(),)
	}
}

/// Helper structure for recursively defined lists.
#[pin_project::pin_project]
struct Cons<L, R>(#[pin] L, #[pin] R);

impl<C, L, R> Pick<C> for Cons<L, R>
where
	C: ?Sized + Config,
	L: Pick<C>,
	R: Pick<C, Output = L::Output>,
{
	const LEN: usize = L::LEN + R::LEN;
	type Output = L::Output;

	#[inline(always)]
	fn start(self: Pin<&mut Self>, sim: &Sim<C>, prec: &mut impl Iterator<Item = Prec>) {
		let p = self.project();

		p.0.start(sim, prec);
		p.1.start(sim, prec);
	}

	#[inline(always)]
	fn query(self: Pin<&mut Self>) -> Option<Self::Output> {
		let p = self.project();
		p.0.query().or_else(|| p.1.query())
	}
}

impl<'s, C, L, R> Join<C> for Cons<L, Fiber<'static, C, R, NotifyQuorum<C>>>
where
	C: ?Sized + Config,
	L: Join<C>,
	R: Future + 's,
{
	const LEN: usize = L::LEN + 1;
	type Output = (L::Output, R::Output);

	#[inline(always)]
	fn start(self: Pin<&mut Self>, sim: &Sim<C>, state: Pin<&JoinState<C>>) {
		let p = self.project();

		p.0.start(sim, state);
		p.1.start(sim, state);
	}

	#[inline(always)]
	unsafe fn query(self: Pin<&mut Self>) -> Self::Output {
		let p = self.project();
		(p.0.query(), p.1.query().0)
	}
}

/* ****************************************************************** Flatten */

/// Helper macro to implement the Flatten-trait generically.
macro_rules! impl_flatten {
	(@name => $T:tt) => { $T };

	(@name $HID:ident $($TID:ident)* =>) => {
		impl_flatten!(@name $($TID)* => ($HID,))
	};

	(@name $HID:ident $($TID:ident)* => $T:tt) => {
		impl_flatten!(@name $($TID)* => ($T, $HID))
	};

	(@leaf $HID:ident) => {
		impl<$HID> Flatten for ($HID,) {
			type Flattened = $HID;

			#[inline(always)]
			fn flatten(self) -> Self::Flattened { self.0 }
		}
	};

	(@leaf $($FID:ident)+) => {
		impl<$($FID,)*> Flatten for impl_flatten!(@name $($FID)* =>) {
			type Flattened = ($($FID,)*);

			#[inline(always)]
			fn flatten(self) -> Self::Flattened {
				#[allow(non_snake_case)]
				let impl_flatten!(@name $($FID)* =>) = self;
				($($FID,)*)
			}
		}
	};

	(@node $($FID:ident)* .) => {
		impl<H: Flatten, $($FID),*> Flatten for impl_flatten!(@name $($FID)* => H) {
			type Flattened = (H::Flattened, $($FID),*);

			#[inline(always)]
			fn flatten(self) -> Self::Flattened {
				#[allow(non_snake_case)]
				let impl_flatten!(@name $($FID)* => H) = self;
				(H.flatten(), $($FID,)*)
			}
		}
	};

	(@node $($HID:ident)* . $MID:ident $($TID:ident)*) => {
		impl_flatten!(@leaf $($HID)* $MID);
		impl_flatten!(@node $($HID)* $MID . $($TID)*);
	};

	($($FID:ident),* $(,)?) => {
		impl_flatten!(@node . $($FID)*);
	};
}

// specialize tuple-flattening for up to 12-tuples before recursively returning
// the remaining tuples partially flattened; this should be sufficient for
// most cases and also allows debug printing the tuple's contents which is
// only implemented up to tuple-size 12
impl_flatten!(T1, T2, T3, T4, T5, T6, T7, T8, T9, TA, TB);

/* *************************************************** Notify Future Adapters */

/// Future adapter that reactivates the parent task when the child task
/// terminates. Doesn't reactivate the parent if the child is cancelled
/// prematurely.
struct NotifyOnExit<C: ?Sized + Config> {
	/// Handle to the parent task.
	parent: Option<TaskPuck<C>>,
}

impl<C: ?Sized + Config> NotifyOnExit<C> {
	/// Initializes the notifier from a future. The parent task is initialized
	/// later during binding to ensure the reactivation of the correct task,
	/// even if the fiber has been moved to a different task between creation
	/// and activation.
	#[inline]
	const fn new() -> Self {
		NotifyOnExit { parent: None }
	}

	/// Method initializing the parent task.
	#[inline]
	fn bind(&mut self, puck: TaskPuck<C>) {
		self.parent = Some(puck);
	}
}

impl<C: ?Sized + Config, R> Settle<R> for NotifyOnExit<C> {
	#[inline]
	fn settle(self, _result: &R) -> ExitStatus {
		// wake the parent; if it has already been woken up, ignore
		if let Some(mut puck) = self.parent {
			puck.wake().ok();
		}

		// report success
		Ok(odem_rs_core::Success)
	}
}

/// Future adapter that reactivates a given task when a quorum with regards to
/// the number of completed sub-tasks is reached.
struct NotifyQuorum<C: ?Sized + Config> {
	/// Pair of shared counter and handle to the parent task.
	parent: Option<NonNull<JoinState<C>>>,
}

impl<C: ?Sized + Config> NotifyQuorum<C> {
	/// Initializes the notifier from a future. Parent task and quorum are
	/// initialized later during binding to ensure the reactivation of the
	/// correct task, even if the fiber has been moved to a different task
	/// between creation and activation.
	#[inline]
	const fn new() -> Self {
		NotifyQuorum { parent: None }
	}

	/// Method initializing the parent task and the shared quorum countdown.
	#[inline]
	fn bind(&mut self, state: Pin<&JoinState<C>>) {
		self.parent = Some(NonNull::from(&*state));
	}
}

impl<C: ?Sized + Config, R> Settle<R> for NotifyQuorum<C> {
	#[inline]
	fn settle(self, _result: &R) -> ExitStatus {
		// since we do not expose this adapter, a debug assert should suffice
		debug_assert!(
			self.parent.is_some(),
			"shared data is expected to have been initialized beforehand"
		);

		// get a reference to the shared state
		// SAFETY: it should not be possible to get to this point without
		// calling the bind-method with a shared state, so this should be safe.
		let state = unsafe { self.parent.unwrap_unchecked().as_ref() };

		// decrease the number of expected terminations
		let n = state.counter.get() - 1;

		// if it is reduced to 0, activate the parent
		if n == 0 {
			if let Some(mut puck) = state.parent.take() {
				puck.wake().ok();
			}
		}

		// update the counter
		state.counter.set(n);

		// return success
		Ok(odem_rs_core::Success)
	}
}
