use core::{
	fmt,
	ops::RangeBounds
};

use alloc::vec::Vec;

use cursor::{BitCursor, Cursor};
pub use iter::{
	IntoOnes, IntoZeroes,
	Ones, Zeroes
};
use iter::RawOnes;
pub use update::BitSliceUpdate;

mod iter;
mod update;
pub mod cursor;

type BaseType = usize;
pub const BITS: usize = BaseType::BITS as usize;

/// A hierarchical bitset that allows very fast location of the smallest bit
/// index set.
/// 
/// The structure of the hierarchical bitset is a sequence of trees with
/// increasing levels of indirection and can be visualized as follows:
/// (`D`: direct entry, `I<n>`: entry for indirection level n)
/// ```(text)
/// D I1 D..D I2 I1 D..D.. I3 I2 I1 D..D.... I4 I3 I2 I1 D..D......
/// BITS×^^^^  BITS×^^^^       BITS×^^^^            BITS×^^^^
///         BITS×^^^^^^^^^  BITS×^^^^^^^^^       BITS×^^^^^^^^^
///                      BITS×^^^^^^^^^^^^^^  BITS×^^^^^^^^^^^^^^
///                                        BITS×^^^^^^^^^^^^^^^^^^^
/// ```
/// This means that every level of indirection has a prefix consisting of
/// the indirection level below, followed by an indirect entry for that
/// level and `BITS` repetitions of the level below it. Each indirect entry
/// of level `n` contains a one bit summary for each of the `BITS` many
/// entries of level `n-1`. Entries of indirection level `0` are direct.
#[derive(Clone)]
pub struct BitSet {
	/// Vector of bits containing space for direct and indirect entries.
	bits: Vec<BaseType>
}

impl BitSet {
	/// Creates a new, empty bitset without allocating memory.
	#[inline]
	pub const fn new() -> Self {
		Self {
			bits: Vec::new()
		}
	}
	
	/// Creates a new, empty bitset with a user-defined capacity pre-allocated.
	#[inline]
	pub fn with_capacity(capacity: usize) -> Self {
		Self {
			bits: Vec::with_capacity(
				Cursor::from_bit(capacity + BITS - 1).efa()
			)
		}
	}
	
	/// Removes all bits from the bitset.
	#[inline]
	pub fn clear(&mut self) {
		self.bits.clear();
	}
	
	/// Enables or disables bits in a range depending on the flag. Returns the
	/// previous value if the slice is a single index.
	#[inline]
	pub fn update<I>(&mut self, range: I, flag: bool) -> I::Output
	where I: BitSliceUpdate {
		range.update(self, flag)
	}
	
	/// Inserts bits into the bitset. Returns a boolean value if the slice is a
	/// single index indicating the previous state of the bit.
	#[inline]
	pub fn insert<I>(&mut self, range: I) -> I::Output
	where I: BitSliceUpdate {
		range.insert(self)
	}
	
	/// Removes bits from the bitset. Returns a boolean value if the slice is a
	/// single index indicating the previous state of the bit.
	#[inline]
	pub fn remove<I>(&mut self, range: I) -> I::Output
	where I: BitSliceUpdate {
		range.remove(self)
	}
	
	/// Tests if the set is empty.
	/// 
	/// This function is very fast because it can take advantage of the
	/// hierarchical structure of the bitset.
	pub fn is_empty(&self) -> bool {
		let mut cursor = Cursor::from_root(0);
		
		while let Some(&word) = self.bits.get(cursor.efa()) {
			// exit if the word is non-empty
			if word != 0 { return false; }
			
			// select the next root node
			cursor = cursor.successor();
		}
		
		// the set must have been empty for us to have made it here
		true
	}
	
	/// Returns whether the bit at the specified index is set or not.
	#[inline]
	pub fn contains(&self, bit: usize) -> bool {
		let cursor = Cursor::from_bit(bit);
		
		if let Some(word) = self.bits.get(cursor.efa()) {
			((*word >> cursor.bit()) & 1) != 0
		} else {
			false
		}
	}
	
	/// Counts the number of set bits in the set.
	#[inline]
	pub fn count(&self) -> usize {
		RawOnes::new(&self.bits).count(&self.bits)
	}
	
	/// Clears the smallest bit that is set. Returns `Some(index)` if that is
	/// possible and `None` otherwise.
	/// 
	/// This function is very fast because it can take advantage of the
	/// hierarchical structure of the bitset. It is also optimized for the task,
	/// only performing the steps necessary to locate and extract one bit.
	/// Prefer using this function over calling `next()` on [`drain`] once.
	/// 
	/// [`drain`]: #method.drain 
	#[inline]
	pub fn drain_one(&mut self) -> Option<usize> {
		RawOnes::new(&self.bits).drain_bit(&mut self.bits)
	}
	
	/// Constructs a draining iterator that removes every bit right before its
	/// index is returned. This iterator is lazy and will not clear bits that
	/// haven't been pulled.
	/// 
	/// The iterator is very fast because it can take advantage of the
	/// hierarchical structure of the bitset in addition to operating parallel
	/// on the word level.
	// #[inline]
	// pub fn drain(&mut self, range: impl RangeBounds<usize>) -> Drain {
	// 	Drain::in_range(&mut self.bits, range)
	// }
	
	/// Constructs a filling iterator that inserts every bit right before its
	/// index is returned. This iterator is lazy and will not fill bits that
	/// haven't been pulled.
	/// 
	/// This iterator is not so fast because it cannot take advantage of the
	/// hierarchical structure of the bitset. It still benefits from the
	/// standard parallelism on the word level.
	// #[inline]
	// pub fn fill(&mut self) -> Fill {
	// 	Fill::new(&mut self.bits)
	// }
	
	/// Constructs an iterator that returns the unset bit indices in the set
	/// within a range.
	/// 
	/// This iterator is not so fast because it cannot take advantage of the
	/// hierarchical structure of the bitset. It still operates in parallel on
	/// the word level.
	#[inline]
	pub fn zeroes(&self, range: impl RangeBounds<usize>) -> Zeroes {
		Zeroes::in_range(&self.bits, range)
	}
	
	/// Constructs an iterator that returns the set bit indices in the set
	/// within a range.
	/// 
	/// This iterator is very fast because it can take advantage of the
	/// hierarchical structure of the bitset in addition to operating in
	/// parallel on the word level.
	#[inline]
	pub fn ones(&self, range: impl RangeBounds<usize>) -> Ones {
		Ones::in_range(&self.bits, range)
	}
	
	/// Constructs an owning iterator that returns non-set bit indices within
	/// a range.
	/// 
	/// Identical restrictions as for the iterator constructed by [`zeroes`]
	/// apply.
	/// 
	/// [`zeroes`]: #method.zeroes
	#[inline]
	pub fn into_zeroes(self, range: impl RangeBounds<usize>) -> IntoZeroes {
		IntoZeroes::in_range(self.bits, range)
	}
	
	/// Constructs an owning iterator that returns set bit indices within a
	/// range.
	/// 
	/// Identical restrictions as for the iterator constructed by [`ones`]
	/// apply.
	/// 
	/// [`ones`]: #method.ones
	#[inline]
	pub fn into_ones(self, range: impl RangeBounds<usize>) -> IntoOnes {
		IntoOnes::in_range(self.bits, range)
	}
}

impl Default for BitSet {
	#[inline]
	fn default() -> Self {
		Self::new()
	}
}

impl IntoIterator for BitSet {
	type Item = usize;
	type IntoIter = IntoOnes;
	
	#[inline]
	fn into_iter(self) -> Self::IntoIter {
		IntoOnes::new(self.bits)
	}
}

impl<'a> IntoIterator for &'a BitSet {
	type Item = usize;
	type IntoIter = Ones<'a>;
	
	#[inline]
	fn into_iter(self) -> Self::IntoIter {
		Ones::new(&self.bits)
	}
}

impl fmt::Debug for BitSet {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		const WIDTH: usize = BaseType::BITS as usize / 4;
		let mut iter = self.bits.iter();
		
		write!(f, "BitSet(")?;
		
		if let Some(word) = iter.next() {
			let mut cursor = Cursor::from_word(0);
			let mut lvl = 0;
			
			// pretty print the forest in alternate mode
			if f.alternate() { write!(f, "\n    ")?; }
			
			// print the first direct entry
			write!(f, "{:0width$x}", word.reverse_bits(), width = WIDTH)?;
			
			while let Some(mut word) = iter.next() {
				// calculate the effective address of the next leaf node
				let successor = cursor.successor();
				let diff = successor.efa() - cursor.efa();
				cursor = successor;
				
				lvl = lvl.max(diff);
				
				// specifically mark the indirect entries for easier debugging
				for n in (1..diff).rev() {
					// indent the tree in the alternate mode
					if f.alternate() {
						write!(f, "\n{:1$}", "", (lvl - n)*4 - 1)?;
					}
					
					// apologies for this format string but it does the trick
					write!(
						f, " {1:[<n$}{:0width$x}{1:]<n$}",
						word.reverse_bits(), "", n = n, width = WIDTH
					)?;
					
					// advance (should succeed, but better safe than sorry)
					word = iter.next().unwrap_or(&0);
				}
				
				// print the leaf word
				write!(f, " {:0width$x}", word.reverse_bits(), width = WIDTH)?;
			}
			
			if f.alternate() { writeln!(f)?; }
		}
		
		write!(f, ")")
	}
}

/* *************************************************************** Unit Tests */

#[cfg(test)]
mod tests {
	use std::collections::BTreeSet;
	
	use prop::collection::{btree_set, vec};
	use proptest::prelude::*;
	
	use super::*;
	
	/// Checks that newly instantiated bitsets are registered as empty.
	#[test]
	fn new_bitset_is_empty() {
		assert!(BitSet::new().is_empty());
		assert!(BitSet::with_capacity(100).is_empty());
	}
	
	proptest! {
		/// Checks the invariants that (1) not-previously-inserted elements are
		/// absent when they're inserted and (2) that all of them are present
		/// after insertion.
		#[test]
		fn inserted_elems_contained(elems in btree_set(0..10000usize, 0..500)) {
			let elems: BTreeSet<usize> = elems;
			let capacity = elems.iter().next_back().map_or(0, |&e| e + 1);
			let mut set = BitSet::with_capacity(capacity);
			
			// insert the elements in the BTree set
			for &elem in elems.iter() {
				prop_assert!(
					!set.insert(elem),
					"freshly inserted element {} already present in bitset: {:?}",
					elem, set
				);
			}
			
			// test for the presence of the elements
			for &elem in elems.iter() {
				prop_assert!(
					set.contains(elem),
					"previously inserted element {} not present in bitset: {:?}",
					elem, set
				);
			}
		}
		
		/// Ensures that previously inserted elements are still reported as
		/// present when inserted a second time.
		#[test]
		fn doubly_inserted_elements_present(elems in btree_set(0..10000usize, 0..500)) {
			let elems: BTreeSet<usize> = elems;
			let capacity = elems.iter().next_back().map_or(0, |&e| e + 1);
			let mut set = BitSet::with_capacity(capacity);
			
			for &elem in elems.iter() {
				prop_assert!(
					!set.insert(elem),
					"freshly inserted element {} already present in bitset: {:?}",
					elem, set
				);
			}
			
			for &elem in elems.iter() {
				prop_assert!(
					set.insert(elem),
					"already inserted element {} not present in bitset: {:?}",
					elem, set
				);
			}
		}
		
		/// Ensures that previously present and then removed elements are still
		/// reported absent when removed a second time.
		#[test]
		fn doubly_removed_elements_absent(elems in btree_set(0..10000usize, 0..500)) {
			let elems: BTreeSet<usize> = elems;
			let capacity = elems.iter().next_back().map_or(0, |&e| e + 1);
			let mut set = BitSet::with_capacity(capacity);
			
			for &elem in elems.iter() {
				prop_assert!(
					!set.insert(elem),
					"freshly inserted element {} already present in bitset: {:?}",
					elem, set
				);
			}
			
			for &elem in elems.iter() {
				prop_assert!(
					set.remove(elem),
					"previously inserted element {} not present in bitset: {:?}",
					elem, set
				);
			}
			
			for &elem in elems.iter() {
				prop_assert!(
					!set.remove(elem),
					"previously removed element {} present in bitset: {:?}",
					elem, set
				);
			}
		}
		
		/// Checks if all of the previously inserted elements are covered
		/// by the draining iterator and that the set is left empty.
		// #[test]
		// fn inserted_elems_drained(elems in btree_set(0..10000usize, 0..500)) {
		// 	let mut elems: BTreeSet<usize> = elems;
		// 	let capacity = elems.iter().rev().next().map_or(0, |&e| e + 1);
		// 	let mut set = BitSet::with_capacity(capacity);
		// 	
		// 	for &elem in elems.iter() {
		// 		prop_assert!(
		// 			!set.insert(elem),
		// 			"freshly inserted element {} already present in bitset: {:?}",
		// 			elem, set
		// 		);
		// 	}
		// 	
		// 	for (e1, e2) in set.drain().zip(elems.clone()) {
		// 		prop_assert_eq!(e1, e2);
		// 		elems.remove(&e1);
		// 	}
		// 	
		// 	prop_assert!(
		// 		set.is_empty(),
		// 		"bitset still contains elements: {:?}", set
		// 	);
		// 	prop_assert!(
		// 		elems.is_empty(),
		// 		"hashset still contains elements: {:?}", elems
		// 	);
		// }
		
		/// Ensures that a partial drain leaves the bitset in a state that
		/// contains all of the non-drained bits and none of the drained ones.
		// #[test]
		// fn partial_drain_leaves_elems(
		// 	elems in btree_set(0..10000usize, 1..500),
		// 	split in any::<prop::sample::Index>()
		// ) {
		// 	let elems: BTreeSet<usize> = elems;
		// 	let capacity = elems.iter().rev().next().map_or(0, |&e| e + 1);
		// 	let mut set = BitSet::with_capacity(capacity);
		// 	
		// 	for &elem in elems.iter() {
		// 		prop_assert!(
		// 			!set.insert(elem),
		// 			"freshly inserted element {} already present in bitset: {:?}",
		// 			elem, set
		// 		);
		// 	}
		// 	
		// 	let split: usize = split.index(elems.len());
		// 	for (e1, &e2) in set.drain().zip(elems.iter()).take(split) {
		// 		prop_assert_eq!(
		// 			e1, e2,
		// 			"drained element is not one of the first {} elements in \
		// 			 the original set",
		// 			split
		// 		);
		// 	}
		// 	
		// 	prop_assert_eq!(
		// 		set.count(), elems.iter().skip(split).count(),
		// 		"draining {} element(s) has left differing remainders:\n{:?}\n{:?}",
		// 		split, set, elems
		// 	);
		// 	
		// 	for (e1, &e2) in set.drain().zip(elems.iter().skip(split)) {
		// 		prop_assert_eq!(
		// 			e1, e2,
		// 			"drained element is not an element after the first {} \
		// 			 elements in the original set",
		// 			split
		// 		);
		// 	}
		// 	
		// 	prop_assert!(
		// 		set.is_empty(),
		// 		"not all of the elements have been drained: {:?}",
		// 		set
		// 	);
		// }
		
		/// Ensures that the filling iterator only fills holes and updates the
		/// bitset correctly.
		// #[test]
		// fn filled_elements_inserted(elems in btree_set(0..10000usize, 2..500)) {
		// 	let elems: BTreeSet<usize> = elems;
		// 	let capacity = elems.iter().rev().next().map_or(0, |&e| e + 1);
		// 	let mut set = BitSet::with_capacity(capacity);
		// 	
		// 	for &elem in elems.iter() {
		// 		prop_assert!(
		// 			!set.insert(elem),
		// 			"freshly inserted element {} already present in bitset: {:?}",
		// 			elem, set
		// 		);
		// 	}
		// 	
		// 	let mut iter = set.fill().take(elems.len() / 2);
		// 	let mut split = iter.next().unwrap();
		// 	for elem in iter {
		// 		prop_assert!(
		// 			!elems.contains(&elem),
		// 			"non-inserted element {} present in bitset",
		// 			elem
		// 		);
		// 		split = elem;
		// 	}
		// 	
		// 	prop_assert_eq!(
		// 		set.ones(..=split).count(), split + 1,
		// 		"not all bits have been set in the filled bitset: {:?}",
		// 		set
		// 	);
		// 	
		// 	for (e1, &e2) in set.drain().skip(split + 1).zip(elems.iter().skip_while(|&&e| e <= split)) {
		// 		prop_assert_eq!(
		// 			e1, e2,
		// 			"element is not an element after the first {} \
		// 			 elements in the original set",
		// 			split
		// 		);
		// 	}
		// }
		
		/// Checks if the bitset doesn't include non-inserted elements.
		#[test]
		fn non_inserted_elems_vacant(elems in btree_set(0..10000usize, 0..500)) {
			let elems: BTreeSet<usize> = elems;
			let capacity = elems.iter().next_back().map_or(0, |&e| e + 1);
			let mut set = BitSet::with_capacity(capacity);
			
			for &elem in elems.iter() {
				prop_assert!(
					!set.insert(elem),
					"freshly inserted element {} already present in bitset: {:?}",
					elem, set
				);
			}
			
			for elem in set.zeroes(..capacity) {
				prop_assert!(
					!elems.contains(&elem),
					"inserted element {} returned when asking for non-existing \
					 elements in bitset: {:?}",
					elem, set
				);
			}
		}
		
		/// Checks if draining the bitset one-by-one returns all of the elements
		/// at some point.
		#[test]
		fn drain_one_returns_elems_eventually(elems in btree_set(0..10000usize, 0..500)) {
			let mut elems: BTreeSet<usize> = elems;
			let capacity = elems.iter().next_back().map_or(0, |&e| e + 1);
			let mut set = BitSet::with_capacity(capacity);
			
			for &elem in elems.iter() {
				prop_assert!(
					!set.insert(elem),
					"freshly inserted element {} already present in bitset: {:?}",
					elem, set
				);
			}
			
			while let Some(elem) = set.drain_one() {
				prop_assert!(
					elems.remove(&elem),
					"never inserted element {} drained from the bitset: {:?}",
					elem, set
				);
			}
			
			prop_assert!(
				set.is_empty(),
				"bitset still contains elements: {:?}", set
			);
			
			prop_assert!(
				elems.is_empty(),
				"bitset didn't return all elements: {:?}", elems
			);
		}
		
		/// Tests if counting set bits in arbitrary slices agrees with the
		/// number of inserted elements within that slice's range.
		#[test]
		fn counting_ones_yields_correct_results(
			elems in btree_set(0..10000usize, 1..500),
			indices in vec(any::<prop::sample::Index>(), 2..10)
		) {
			let elems: BTreeSet<usize> = elems;
			let capacity = elems.iter().next_back().map_or(0, |&e| e + 1);
			let mut set = BitSet::with_capacity(capacity);
			let mut indices = indices
				.into_iter()
				.map(|idx| idx.index(capacity))
				.collect::<Vec<usize>>();
			
			for &elem in elems.iter() {
				prop_assert!(
					!set.insert(elem),
					"freshly inserted element {} already present in bitset: {:?}",
					elem, set
				);
			}
			
			prop_assert_eq!(
				set.count(), elems.len(),
				"number of set bits different from length of collection"
			);
			
			// sort the indices first
			indices.sort();
			
			// convert them into ranges
			for range in indices.windows(2).map(|range| range[0]..range[1]) {
				// count elements in the slices of both structures
				prop_assert_eq!(
					elems.range(range.clone()).count(),
					set.ones(range.clone()).count(),
					"number of set bits different in range {:?} for bitset: {:#?}",
					range, set
				);
			}
		}
		
		/// Tests if counting unset bits in arbitrary slices agrees with the
		/// number of non-inserted elements within that slice's range.
		#[test]
		fn counting_zeroes_yields_correct_results(
			elems in btree_set(0..10000usize, 1..500),
			indices in vec(any::<prop::sample::Index>(), 2..10)
		) {
			use std::collections::BTreeSet;
			
			let elems: BTreeSet<usize> = elems;
			let capacity = elems.iter().next_back().map_or(0, |&e| e + 1);
			let mut set = BitSet::with_capacity(capacity);
			
			let mut indices: Vec<usize> = indices
				.into_iter()
				.map(|idx| idx.index(capacity))
				.collect();
			
			for &elem in elems.iter() {
				prop_assert!(
					!set.insert(elem),
					"freshly inserted element {} already present in bitset: {:?}",
					elem, set
				);
			}
			
			prop_assert_eq!(
				set.zeroes(..capacity).count(), capacity - elems.len(),
				"number of unset bits different from number of holes in collection"
			);
			
			// sort the indices first
			indices.sort();
			
			// convert them into ranges
			for range in indices.windows(2).map(|range| range[0]..range[1]) {
				let len = range.len();
				
				// count holes in both structures
				prop_assert_eq!(
					len - elems.range(range.clone()).count(),
					set.zeroes(range.clone()).count()
				);
			}
		}
	}
}
