use std::cmp::Ordering;
use super::BITS;

/// Maximal number of trees for a known fixed word length.
const MAX_LVL: usize = tree_count();

/// Effective addresses for all the root words for a known fixed word length.
const ROOT_TABLE: [usize; MAX_LVL] = root_table();

/// A helper to navigate the 'Ahnentafel' structure of the bit set.
/// 
/// Instances of this type can be manipulated during compile time and allow to
/// express the algorithms used for the implementation of the bit set at a
/// higher level of abstraction.
#[derive(Copy, Clone, Debug)]
pub struct Cursor {
	/// Field for the effective (actual) address of a word.
	efa: usize,
	
	/// Field for the path in the tree. Is equal to the index for leaf words and
	/// equal to 0 for tree roots.
	pth: usize,
	
	/// Number of the individual entries at this level of indirection. Is equal
	/// to 1 for leaf words.
	size: usize
}

/// Specialized cursor that operates on the root words of the bit set.
/// 
/// This cursor can benefit from the knowledge that the `pth` attribute is 0.
#[derive(Copy, Clone, Debug)]
pub struct RootCursor {
	/// Field for the effective (actual) address of a root word.
	efa: usize,
	
	/// Number of individual words in this tree.
	size: usize
}

/// Specialized cursor that operates on the leaf words of the bit set.
/// 
/// The leaf words store the actual contents of the bit set. This cursor
/// benefits from the knowledge that the `size` attribute is 1.
#[derive(Copy, Clone, Debug)]
pub struct LeafCursor {
	/// Field for the effective (actual) address of a direct entry.
	efa: usize,
	
	/// Index of the selected direct entry.
	pth: usize
}

/// Specialized cursor that addresses specific bits in leaf words of the bit
/// set.
/// 
/// Leaf words store the actual contents of the bit set. This cursor
/// additionally stores bit indices within the leaf words.
#[derive(Copy, Clone, Debug)]
pub struct BitCursor {
	/// Field for the effective (actual) address of a direct entry.
	efa: usize,
	
	/// Index of the selected direct entry, including the bit index.
	pth: usize
}

impl Cursor {
	/// Creates a new cursor pointing to the first word.
	#[inline]
	pub const fn new() -> Self {
		Self { efa: 0, pth: 0, size: 1 }
	}
	
	/// Given a root number, creates the [`RootCursor`] corresponding with the
	/// root number.
	/// 
	/// [`RootCursor`]: struct.RootCursor.html
	#[inline]
	pub const fn from_root(root_id: usize) -> RootCursor {
		RootCursor::new(root_id)
	}
	
	/// Given the index of a direct entry, creates the [`LeafCursor`]
	/// corresponding with the index.
	/// 
	/// [`LeafCursor`]: struct.LeafCursor.html
	#[inline]
	pub const fn from_word(word_id: usize) -> LeafCursor {
		LeafCursor::new(word_id)
	}
	
	/// Given a bit index, creates the [`BitCursor`] corresponding with the
	/// bit index.
	/// 
	/// [`BitCursor`]: struct.BitCursor.html
	#[inline]
	pub const fn from_bit(bit_id: usize) -> BitCursor {
		BitCursor::new(bit_id)
	}
	
	/// For a given `Cursor`, calculates the corresponding [`RootCursor`].
	/// 
	/// [`RootCursor`]: struct.RootCursor.html
	pub const fn to_root(mut self) -> RootCursor {
		while let Ok((parent, _)) = self.parent() { self = parent; }
		
		RootCursor {
			efa: self.efa,
			size: self.size
		}
	}
	
	/// For a given `Cursor`, calculates the first [`LeafCursor`] inside the
	/// interval it encompasses.
	/// 
	/// [`LeafCursor`]: struct.LeafCursor.html
	pub const fn to_leaf(self) -> LeafCursor {
		LeafCursor {
			efa: self.efa + size_to_lvl(self.size),
			pth: self.idx()
		}
	}
	
	/// For a given `Cursor` and a bit index, calculates a [`BitCursor`] to the
	/// first direct entry within its interval and sets its bit index
	/// accordingly.
	/// 
	/// [`BitCursor`]: struct.BitCursor.html
	#[inline]
	pub const fn to_bit(self, bit_no: usize) -> BitCursor {
		self.to_leaf().to_bit(bit_no)
	}
	
	/// Calculates the next entry of the same level within the bit set.
	/// 
	/// The `Cursor` cannot be moved past the final entry and will panic in
	/// debug mode but silently wrap around in release mode.
	pub const fn successor(self) -> Self {
		Self {
			efa: self.efa + next_effective_address(self.pth) + self.size,
			pth: self.pth + 1,
			size: self.size
		}
	}
	
	/// Calculates the previous entry of the same level within the bit set.
	/// 
	/// The `Cursor` cannot be moved before the first entry and will panic in
	/// debug mode but silently wrap around around in release mode.
	pub const fn predecessor(self) -> Self {
		Self {
			efa: self.efa - next_effective_address(self.pth - 1) - self.size,
			pth: self.pth - 1,
			size: self.size
		}
	}
	
	/// Calculates the direct parent of a cursor and returns it together with
	/// the branch index that was taken.
	/// 
	/// The `Cursor` cannot be moved above the local root and will return itself
	/// converted into a [`RootCursor`] in this case.
	/// 
	/// [`RootCursor`]: struct.RootCursor.html
	pub const fn parent(self) -> Result<(Self, usize), RootCursor> {
		if !self.is_root() {
			let pth = self.pth - 1;
			Ok((
				Self {
					efa: self.efa - (pth % BITS) * self.size - 1,
					pth: pth / BITS,
					size: self.size * BITS + 1
				},
				pth % BITS
			))
		} else {
			Err(RootCursor { efa: self.efa, size: self.size })
		}
	}
	
	/// Calculates a specified child of a cursor.
	/// 
	/// The `Cursor` cannot be moved below the direct word level and will return
	/// itself as a [`LeafCursor`] in this case. Consider converting it into a
	/// [`BitCursor`], if you want to address an individual bit in the bit set.
	/// 
	/// [`LeafCursor`]: struct.LeafCursor.html
	/// [`BitCursor`]: struct.BitCursor.html
	pub const fn child(self, child_id: usize) -> Result<Self, LeafCursor> {
		if !self.is_leaf() {
			Ok(Self {
				efa: self.efa + child_id * self.size / BITS + 1,
				pth: self.pth * BITS + child_id + 1,
				size: self.size / BITS
			})
		} else {
			Err(
				LeafCursor {
					efa: self.efa,
					pth: self.pth
				}
			)
		}
	}
	
	/// Returns the effective (actual) address of the current cursor position.
	#[inline]
	pub const fn efa(&self) -> usize {
		self.efa
	}
	
	/// Returns the direct word index of the first word in the interval
	/// encompassed by the current position of the `Cursor`.
	#[inline]
	pub const fn idx(&self) -> usize {
		self.pth * (self.size - self.size / BITS) + self.size / BITS
	}
	
	/// Returns whether the `Cursor` currently points at a root entry.
	#[inline]
	pub const fn is_root(&self) -> bool {
		self.pth == 0
	}
	
	/// Returns whether the `Cursor` currently points at a direct (leaf) entry.
	#[inline]
	pub const fn is_leaf(&self) -> bool {
		self.size == 1
	}
}

impl Default for Cursor {
	#[inline]
	fn default() -> Self {
		Self::new()
	}
}

impl PartialEq for Cursor {
	#[inline]
	fn eq(&self, other: &Self) -> bool {
		self.efa == other.efa
	}
}

impl Eq for Cursor {}

impl PartialOrd for Cursor {
	#[inline]
	fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
		Some(self.cmp(other))
	}
}

impl Ord for Cursor {
	#[inline]
	fn cmp(&self, other: &Self) -> Ordering {
		self.efa.cmp(&other.efa)
	}
}

impl RootCursor {
	/// Creates a `RootCursor` to the specified root index.
	/// 
	/// This function returns the last root cursor for root indices outside
	/// of the scope.
	#[inline]
	pub const fn new(root_id: usize) -> Self {
		// test if the root index is within the range of possible root nodes
		if root_id < MAX_LVL - 1 {
			// use the precomputed table to initialize the root cursor
			Self {
				efa: ROOT_TABLE[root_id],
				size: ROOT_TABLE[root_id + 1] - ROOT_TABLE[root_id]
			}
		} else {
			// initialize the root cursor with the final table entry
			Self {
				efa: ROOT_TABLE[MAX_LVL - 1],
				size: (ROOT_TABLE[MAX_LVL - 1] - ROOT_TABLE[MAX_LVL - 2]) * BITS + 1
			}
		}
	}
	
	/// Constructs a `RootCursor` for an effective address.
	/// 
	/// This can be used to find the local root for a given [`Cursor`].
	/// 
	/// [`Cursor`]: struct.Cursor.html
	pub const fn from_efa(mut efa: usize) -> Self {
		let mut size = 1;
		let mut root = 0;
		
		while efa > size {
			efa -= size;
			root += size;
			size = size * BITS + 1;
		}
		
		Self { efa: root, size }
	}
	
	/// Returns the given `RootCursor`.
	#[inline]
	pub const fn to_root(self) -> Self {
		self
	}
	
	/// For a given `RootCursor`, calculates the first [`LeafCursor`] inside the
	/// interval it encompasses.
	///
	/// [`LeafCursor`]: struct.LeafCursor.html
	#[inline]
	pub const fn to_leaf(self) -> LeafCursor {
		LeafCursor {
			efa: self.efa + size_to_lvl(self.size),
			pth: self.idx()
		}
	}
	
	/// For a given `RootCursor` and a bit index, calculates a [`BitCursor`] to
	/// the first direct entry within its interval and sets its bit index
	/// accordingly.
	///
	/// [`BitCursor`]: struct.BitCursor.html
	#[inline]
	pub const fn to_bit(self, bit_no: usize) -> BitCursor {
		self.to_leaf().to_bit(bit_no)
	}
	
	/// Selects the following `RootCursor`.
	/// 
	/// Note that the result is equivalent to converting to [`Cursor`], calling
	/// [`successor()`] and [`parent()`] and converting back into a
	/// `RootCursor`.
	///
	/// The `RootCursor` cannot be moved past the final entry and will panic in
	/// debug mode but silently wrap around in release mode.
	/// 
	/// [`Cursor`]: struct.Cursor.html
	/// [`successor()`]: struct.Cursor.html#method.successor
	/// [`parent()`]: struct.Cursor.html#method.parent
	#[inline]
	pub const fn successor(self) -> Self {
		Self { efa: self.efa + self.size, size: self.size * BITS + 1 }
	}
	
	/// Selects the previous `RootCursor`.
	/// 
	/// Note that the result is equivalent to converting to [`Cursor`], calling
	/// [`child()`] and [`predecessor()`] and converting back into a
	/// `RootCursor`.
	/// 
	/// The `RootCursor` cannot be moved before the first root and will panic in
	/// debug mode but silently wrap around in release mode.
	/// 
	/// [`Cursor`]: struct.Cursor.html
	/// [`predecessor()`]: struct.Cursor.html#method.predecessor
	/// [`child()`]: struct.Cursor.html#method.child
	#[inline]
	pub const fn predecessor(self) -> Self {
		// uncomment once panic in const_fn is stable
		// debug_assert!(self.efa != 0, "no root node before the first one exists");
		Self { efa: self.efa - self.size / BITS, size: self.size / BITS }
	}
	
	/// Implicitly converts the `RootCursor` into a [`Cursor`] to the given
	/// child index.
	/// 
	/// Will return itself as a [`LeafCursor`] for the first root node, since
	/// that one is also a leaf node.
	/// 
	/// [`Cursor`]: struct.Cursor.html
	/// [`LeafCursor`]: struct.LeafCursor.html
	#[inline]
	pub const fn child(self, child_id: usize) -> Result<Cursor, LeafCursor> {
		Cursor {
			efa: self.efa,
			pth: 0,
			size: self.size
		}.child(child_id)
	}
	
	/// Returns the effective (actual) address of the current cursor position.
	#[inline]
	pub const fn efa(&self) -> usize {
		self.efa
	}
	
	/// Returns the direct word index of the first word in the interval
	/// encompassed by the current position of the `RootCursor`.
	#[inline]
	pub const fn idx(&self) -> usize {
		self.size / BITS
	}
	
	/// Returns whether the `RootCursor` currently points at a root entry. By
	/// definition, this is always the case.
	#[inline]
	pub const fn is_root(&self) -> bool {
		true
	}
	
	/// Returns whether the RootCursor` currently points at a direct word.
	/// This is only true for the first tree.
	#[inline]
	pub const fn is_leaf(&self) -> bool {
		self.size == 1
	}
}

impl Default for RootCursor {
	#[inline]
	fn default() -> Self {
		Self::new(0)
	}
}

impl From<RootCursor> for Cursor {
	#[inline]
	fn from(cur: RootCursor) -> Self {
		Self {
			efa: cur.efa(),
			pth: 0,
			size: cur.size
		}
	}
}

impl PartialEq for RootCursor {
	#[inline]
	fn eq(&self, other: &Self) -> bool {
		self.efa == other.efa
	}
}

impl PartialEq<Cursor> for RootCursor {
	#[inline]
	fn eq(&self, other: &Cursor) -> bool {
		self.efa == other.efa
	}
}

impl Eq for RootCursor {}

impl PartialOrd for RootCursor {
	#[inline]
	fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
		Some(self.cmp(other))
	}
}

impl Ord for RootCursor {
	#[inline]
	fn cmp(&self, other: &Self) -> Ordering {
		self.efa.cmp(&other.efa)
	}
}

impl LeafCursor {
	/// Given a direct word index, construct the corresponding `LeafCursor`.
	pub const fn new(word_id: usize) -> Self {
		Self {
			efa: effective_word(word_id),
			pth: word_id
		}
	}
	
	/// For a given `LeafCursor`, calculates the corresponding [`RootCursor`].
	///
	/// [`RootCursor`]: struct.RootCursor.html
	#[inline]
	pub const fn to_root(self) -> RootCursor {
		RootCursor::from_efa(self.efa)
	}
	
	/// Returns the given `LeafCursor`.
	#[inline]
	pub const fn to_leaf(self) -> Self {
		self
	}
	
	/// For a given `LeafCursor` and a bit index, constructs a [`BitCursor`]
	/// and sets its bit index accordingly.
	///
	/// [`BitCursor`]: struct.BitCursor.html
	#[inline]
	pub const fn to_bit(self, bit_no: usize) -> BitCursor {
		BitCursor {
			efa: self.efa,
			pth: self.pth * BITS + bit_no
		}
	}
	
	/// Constructs the next `LeafCursor`.
	///
	/// The `LeafCursor` cannot be moved past the final entry and will panic in
	/// debug mode but silently wrap around in release mode.
	pub const fn successor(self) -> Self {
		Self {
			efa: self.efa + next_effective_address(self.pth) + 1,
			pth: self.pth + 1
		}
	}
	
	/// Constructs the previous `LeafCursor`.
	/// 
	/// The `LeafCursor` cannot be moved before the first entry and will panic
	/// in debug mode but silently wrap around in release mode.
	pub const fn predecessor(self) -> Self {
		Self {
			efa: self.efa - next_effective_address(self.pth - 1) - 1,
			pth: self.pth - 1
		}
	}
	
	/// Implicitly converts the `LeafCursor` into a [`Cursor`] to the parent
	/// entry.
	/// 
	/// Will return itself as a [`RootCursor`] for the first leaf node, since
	/// that is also a root node.
	/// 
	/// [`Cursor`]: struct.Cursor.html
	/// [`RootCursor`]: struct.RootCursor.html
	#[inline]
	pub const fn parent(self) -> Result<(Cursor, usize), RootCursor> {
		Cursor {
			efa: self.efa,
			pth: self.pth,
			size: 1
		}.parent()
	}
	
	/// Returns the effective (actual) address of the current cursor position.
	#[inline]
	pub const fn efa(&self) -> usize {
		self.efa
	}
	
	/// Returns the direct word index of the selected word.
	#[inline]
	pub const fn idx(&self) -> usize {
		self.pth
	}
	
	/// Returns whether the `LeafCursor` currently points at a root entry.
	/// This is only the case for the first tree.
	#[inline]
	pub const fn is_root(&self) -> bool {
		self.pth == 0
	}
	
	/// Returns whether the `LeafCursor` currently points at a direct entry.
	/// By definition, this is always the case.
	#[inline]
	pub const fn is_leaf(&self) -> bool {
		true
	}
}

impl Default for LeafCursor {
	#[inline]
	fn default() -> Self {
		Self::new(0)
	}
}

impl From<LeafCursor> for Cursor {
	#[inline]
	fn from(cur: LeafCursor) -> Self {
		Self {
			efa: cur.efa,
			pth: cur.pth,
			size: 1
		}
	}
}

impl PartialEq for LeafCursor {
	#[inline]
	fn eq(&self, other: &Self) -> bool {
		self.efa == other.efa
	}
}

impl PartialEq<Cursor> for LeafCursor {
	#[inline]
	fn eq(&self, other: &Cursor) -> bool {
		self.efa == other.efa
	}
}

impl Eq for LeafCursor {}

impl PartialOrd for LeafCursor {
	#[inline]
	fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
		Some(self.cmp(other))
	}
}

impl Ord for LeafCursor {
	#[inline]
	fn cmp(&self, other: &Self) -> Ordering {
		self.efa.cmp(&other.efa)
	}
}

impl BitCursor {
	/// Given a bit index, constructs the corresponding `BitCursor`.
	pub const fn new(bit: usize) -> Self {
		Self {
			efa: effective_bit(bit),
			pth: bit
		}
	}
	
	/// Given a non-empty inclusive range of bit indices, constructs a
	/// corresponding pair of `BitCursor` instances.
	pub const fn from_range(range: (usize, usize)) -> (BitCursor, BitCursor) {
		let efa = effective_interval(range);
		(Self { efa: efa.0, pth: range.0 }, Self { efa: efa.1, pth: range.1 })
	}
	
	/// For a given `BitCursor`, calculates the corresponding [`RootCursor`].
	///
	/// [`RootCursor`]: struct.RootCursor.html
	#[inline]
	pub const fn to_root(self) -> RootCursor {
		RootCursor::from_efa(self.efa)
	}
	
	/// For a given `BitCursor`, constructs the corresponding [`LeafCursor`].
	///
	/// [`LeafCursor`]: struct.LeafCursor.html
	#[inline]
	pub const fn to_leaf(self) -> LeafCursor {
		LeafCursor {
			efa: self.efa,
			pth: self.pth / BITS
		}
	}
	
	/// Constructs a new `BitCursor` with the bit index set according to the
	/// function argument.
	#[inline]
	pub const fn to_bit(self, bit_no: usize) -> Self {
		Self {
			efa: self.efa,
			pth: self.pth / BITS * BITS + bit_no
		}
	}
	
	/// Selects the next bit index.
	///
	/// The `BitCursor` cannot be moved past the final entry and will panic in
	/// debug mode but silently wrap around in release mode.
	pub const fn successor(self) -> Self {
		Self {
			efa: self.efa + next_effective_address(self.pth + 1),
			pth: self.pth + 1
		}
	}
	
	/// Selects the previous bit index.
	///
	/// The `BitCursor` cannot be moved before the first bit and will panic in
	/// debug mode but silently wrap around in release mode.
	pub const fn predecessor(self) -> Self {
		Self {
			efa: self.efa - next_effective_address(self.pth),
			pth: self.pth - 1
		}
	}
	
	/// Returns the effective (actual) address of the current cursor position.
	#[inline]
	pub const fn efa(&self) -> usize {
		self.efa
	}
	
	/// Returns the path of the `BitCursor`.
	/// 
	/// This is equivalent to returning the bit index.
	#[inline]
	pub const fn pth(&self) -> usize {
		self.pth
	}
	
	/// Returns the direct word index of the selected word.
	#[inline]
	pub const fn idx(&self) -> usize {
		self.pth / BITS
	}
	
	/// Returns the currently selected bit index.
	#[inline]
	pub const fn bit(&self) -> usize {
		self.pth % BITS
	}
	
	/// Returns whether the `BitCursor` currently points at a root entry. By
	/// definition, this can never be the case.
	#[inline]
	pub const fn is_root(&self) -> bool {
		false
	}
	
	/// Returns whether the `BitCursor` currently points at a direct entry.
	/// By definition, this can never be the case.
	#[inline]
	pub const fn is_leaf(&self) -> bool {
		false
	}
}

impl Default for BitCursor {
	#[inline]
	fn default() -> Self {
		Self::new(0)
	}
}

impl PartialEq for BitCursor {
	#[inline]
	fn eq(&self, other: &Self) -> bool {
		self.pth == other.pth
	}
}

impl Eq for BitCursor {}

impl PartialOrd for BitCursor {
	#[inline]
	fn partial_cmp(&self, other: &Self) -> Option<Ordering> {
		Some(self.cmp(other))
	}
}

impl Ord for BitCursor {
	#[inline]
	fn cmp(&self, other: &Self) -> Ordering {
		self.pth.cmp(&other.pth)
	}
}

/// Given a *word* index, calculates the effective address of the corresponding
/// direct entry in the bit set.
pub const fn effective_word(mut idx: usize) -> usize {
	let mut efa = idx;
	
	// loop until all of the bits have been covered
	while idx > 0 {
		idx = (idx - 1) / BITS;
		efa += idx + 1;
	}
	
	efa
}

/// Given a *bit* index, calculates the effective address of the corresponding
/// direct entry in the bit set.
pub const fn effective_bit(mut bit: usize) -> usize {
	let mut efa = 0;
	
	// loop until all of the bits have been covered
	while bit >= BITS {
		bit /= BITS;
		efa += bit + 1;
		bit -= 1;
	}
	
	efa
}

/// Given a non-empty bit index interval, calculates a pair of effective
/// addresses for these indices.
pub const fn effective_interval(mut bit: (usize, usize)) -> (usize, usize) {
	let mut efa = (0, 0);
	
	// loop until all of the bits for the start of the interval are covered
	while bit.0 >= BITS {
		bit.0 /= BITS;
		efa.0 += bit.0 + 1;
		bit.0 -= 1;
		
		bit.1 /= BITS;
		efa.1 += bit.1 + 1;
		bit.1 -= 1;
	}
	
	// cover the remaining segment
	while bit.1 >= BITS {
		bit.1 /= BITS;
		efa.1 += bit.1 + 1;
		bit.1 -= 1;
	}
	
	efa
}

/// Given the path to a root entry, calculates the number of indirect entries
/// between the current and the next effective address.
pub const fn next_effective_address(mut pth: usize) -> usize {
	let mut step = 0;
	
	// as long as the path indicates an overflow to the next branch
	while (pth % BITS) == 0 {
		// increase the distance to the next entry for the additional entry
		step += 1;
		
		// select the path segment of the next higher layer
		// note that the wrapping subtract automatically fails the test
		// from the top of the loop for pth == 0, which is intended
		pth = pth.wrapping_sub(1) / BITS;
	}
	
	// return the distance
	step
}

/// Given the size of a level in the bit set, calculates the corresponding
/// level.
const fn size_to_lvl(mut size: usize) -> usize {
	// check if BITS is a power of two for optimized code generation
	if BITS.is_power_of_two() {
		// BITS is a power of two: every step in the construction of `size`
		// sets an additional bit and the first bit is always set;
		// we disable the first bit to reduce the result by one
		(size ^ 1).count_ones() as usize
	} else {
		// just count how often size can be divided by BITS;
		// the +1 can safely be ignored because of integer rounding rules
		let mut lvl = 0;
		
		while size > 1 {
			size /= BITS;
			lvl += 1;
		}
		
		lvl
	}
}

/// Calculates the total number of trees within a bit set, given a constant
/// number of bits in the base type and a constant number of bits in the index
/// type.
const fn tree_count() -> usize {
	let mut idx = usize::MAX;
	let mut n = 0;
	
	while idx > 0 {
		n += 1;
		idx /= BITS;
	}
	
	n
}

/// Fills and returns a table with root effective addresses given the constant
/// number of entries.
const fn root_table() -> [usize; MAX_LVL] {
	let mut lut = [0; MAX_LVL];
	let mut i = 1;
	let mut size = 1;
	
	while i < MAX_LVL {
		lut[i] = lut[i-1] + size;
		size = size * BITS + 1;
		i += 1;
	}
	
	lut
}

#[cfg(test)]
mod tests {
	use super::*;
	
	/// Checks that the incremental algorithm for calculating effective
	/// addresses from direct entries does not deviate from the direct one.
	#[test]
	fn iterate_effective_word() {
		let mut efa = 0;
		for i in 0..100000 {
			let step = next_effective_address(i);
			efa += step + 1;
			assert_eq!(efa, effective_word(i + 1));
		}
	}
	
	/// Checks that the algorithms calculating effective addresses from word and
	/// bit indices produce equivalent results.
	#[test]
	fn effective_word_equivalent_effective_bit() {
		for i in 0..1000000*BITS {
			assert_eq!(effective_word(i/BITS), effective_bit(i));
		}
	}
	
	/// Validates the equality between the `effective_range()` function and the
	/// `effective_address()` function for start and end of the interval.
	#[test]
	fn effective_range_equals_effective_bits() {
		for i in 0..100000 {
			assert_eq!(
				(effective_bit(i), effective_bit(i+1)),
				effective_interval((i, i+1))
			);
		}
	}
}
