use std::iter::FusedIterator;
use std::ops::RangeBounds;

use super::{
	BaseType,
	Cursor, BitCursor,
	BITS
};

/* *************************************************** Iterators for set bits */

/// Owning iterator that returns the indices of the set bits.
#[derive(Clone)]
pub struct IntoOnes {
	/// The bit vector.
	bits: Vec<BaseType>,
	/// The iterator helper for finding the locations of set bits.
	iter: RawOnes,
	/// The first bit index outside of the range.
	end: usize
}

impl IntoOnes {
	/// Constructs an owning iterator that returns the indices of the set bits.
	#[inline]
	pub(super) fn new(set: Vec<BaseType>) -> Self {
		Self {
			iter: RawOnes::new(&set),
			bits: set,
			end: !0
		}
	}
	
	/// Constructs an owning iterator that returns a range of indices for the
	/// set bits.
	#[inline]
	pub(super) fn in_range<R>(bits: Vec<BaseType>, range: R) -> Self
	where R: RangeBounds<usize> {
		use std::ops::{Range, Bound::*};
		
		// normalize the range argument
		let range = Range {
			start: match range.start_bound() {
				Included(&idx) => idx,
				Excluded(&idx) => idx.saturating_add(1),
				Unbounded => 0
			},
			end: match range.end_bound() {
				Included(&idx) => idx.saturating_add(1),
				Excluded(&idx) => idx,
				Unbounded => !0
			}
		};
		
		Self {
			iter: RawOnes::new_at(&bits, range.start),
			bits,
			end: range.end
		}
	}
}

impl Iterator for IntoOnes {
	type Item = usize;
	
	#[inline]
	fn next(&mut self) -> Option<Self::Item> {
		// calculate the next bit index
		let bit = self.iter.next(&self.bits)?;
		
		// don't return indices past the threshold
		if bit >= self.end { return None; }
		
		// return the bit index
		Some(bit)
	}
	
	#[inline]
	fn size_hint(&self) -> (usize, Option<usize>) {
		let len = self.len();
		(len, Some(len))
	}
	
	#[inline]
	fn count(self) -> usize {
		self.len()
	}
}

impl ExactSizeIterator for IntoOnes {
	#[inline]
	fn len(&self) -> usize {
		self.iter.count_to(&self.bits, self.end)
	}
}

impl FusedIterator for IntoOnes {}

/// Borrowing iterator that returns the indices of all set bits of a bitset.
#[derive(Copy, Clone)]
pub struct Ones<'a> {
	/// Slice into the bitset.
	bits: &'a Vec<BaseType>,
	/// Internal iterator over set bits.
	iter: RawOnes,
	/// The first bit index outside of the range.
	end: usize
}

impl<'a> Ones<'a> {
	/// Creates a borrowing iterator that returns the indices of the set bits.
	#[inline]
	pub(super) fn new(bits: &'a Vec<BaseType>) -> Self {
		Self {
			bits,
			iter: RawOnes::new(bits),
			end: !0
		}
	}
	
	/// Creates a borrowing iterator that returns the indices in a range of the
	/// set bits.
	#[inline]
	pub(super) fn in_range<R>(bits: &'a Vec<BaseType>, range: R) -> Self
	where R: RangeBounds<usize> {
		use std::ops::{Range, Bound::*};
		
		// normalize the range argument
		let range = Range {
			start: match range.start_bound() {
				Included(&idx) => idx,
				Excluded(&idx) => idx.saturating_add(1),
				Unbounded => 0
			},
			end: match range.end_bound() {
				Included(&idx) => idx.saturating_add(1),
				Excluded(&idx) => idx,
				Unbounded => !0
			}
		};
		
		Self {
			bits,
			iter: RawOnes::new_at(bits, range.start),
			end: range.end
		}
	}
}

impl Iterator for Ones<'_> {
	type Item = usize;
	
	#[inline]
	fn next(&mut self) -> Option<Self::Item> {
		// calculate the next bit index
		let bit = self.iter.next(self.bits)?;
		
		// don't return indices past the threshold
		if bit >= self.end { return None; }
		
		// return the index
		Some(bit)
	}
	
	#[inline]
	fn size_hint(&self) -> (usize, Option<usize>) {
		let len = self.len();
		(len, Some(len))
	}
	
	#[inline]
	fn count(self) -> usize {
		self.len()
	}
}

impl ExactSizeIterator for Ones<'_> {
	#[inline]
	fn len(&self) -> usize {
		self.iter.count_to(self.bits, self.end)
	}
}

impl FusedIterator for Ones<'_> {}

/* ************************************************* Iterators for clear bits */

/// Owning iterator that yields the non-set bits of a bitset in a specific range.
#[derive(Clone)]
pub struct IntoZeroes {
	/// Vector containing all of the bits.
	bits: Vec<BaseType>,
	/// Helper iterator that yields the bit positions.
	iter: RawZeroes,
	/// The bit index one past the last index to return.
	end: usize
}

impl IntoZeroes {
	/// Creates an owning iterator that returns the bit indices of the non-set
	/// bits in a range.
	#[inline]
	pub(super) fn in_range<R>(bits: Vec<BaseType>, range: R) -> Self
	where R: RangeBounds<usize> {
		use std::ops::{Range, Bound::*};
		
		// normalize the range argument
		let range = Range {
			start: match range.start_bound() {
				Included(&idx) => idx,
				Excluded(&idx) => idx.saturating_add(1),
				Unbounded => 0
			},
			end: match range.end_bound() {
				Included(&idx) => idx.saturating_add(1),
				Excluded(&idx) => idx,
				Unbounded => !0
			}
		};
		
		Self {
			iter: RawZeroes::new_at(&bits, range.start),
			bits,
			end: range.end
		}
	}
}

impl Iterator for IntoZeroes {
	type Item = usize;
	
	#[inline]
	fn next(&mut self) -> Option<Self::Item> {
		// calculate the next bit index
		let bit = self.iter.next(&self.bits);
		
		// don't return indices past the threshold
		if bit >= self.end { return None; }
		
		// return the bit index
		Some(bit)
	}
	
	#[inline]
	fn size_hint(&self) -> (usize, Option<usize>) {
		let len = self.len();
		(len, Some(len))
	}
	
	#[inline]
	fn count(self) -> usize {
		self.len()
	}
}

impl ExactSizeIterator for IntoZeroes {
	#[inline]
	fn len(&self) -> usize {
		self.iter.count_to(&self.bits, self.end)
	}
}

impl FusedIterator for IntoZeroes {}

/// Iterator that returns the indices of all cleared bits of a bitset.
#[derive(Copy, Clone)]
pub struct Zeroes<'a> {
	/// Slice of bits.
	bits: &'a Vec<BaseType>,
	/// Raw internal iterator over unset bits.
	iter: RawZeroes,
	/// Last index to return (exclusive).
	end: usize
}

impl<'a> Zeroes<'a> {
	/// Creates a borrowing iterator that returns the bit indices of the non-set
	/// bits in a range.
	#[inline]
	pub(super) fn in_range<R>(bits: &'a Vec<BaseType>, range: R) -> Self
	where R: RangeBounds<usize> {
		use std::ops::{Range, Bound::*};
		
		// normalize the range argument
		let range = Range {
			start: match range.start_bound() {
				Included(&idx) => idx,
				Excluded(&idx) => idx.saturating_add(1),
				Unbounded => 0
			},
			end: match range.end_bound() {
				Included(&idx) => idx.saturating_add(1),
				Excluded(&idx) => idx,
				Unbounded => !0
			}
		};
		
		Self {
			bits,
			iter: RawZeroes::new_at(bits, range.start),
			end: range.end
		}
	}
}

impl Iterator for Zeroes<'_> {
	type Item = usize;
	
	#[inline]
	fn next(&mut self) -> Option<Self::Item> {
		// calculate the index
		let bit = self.iter.next(self.bits);
		
		// don't return indices past the threshold
		if bit >= self.end { return None; }
		
		// return the index
		Some(bit)
	}
	
	#[inline]
	fn size_hint(&self) -> (usize, Option<usize>) {
		let len = self.len();
		(len, Some(len))
	}
	
	#[inline]
	fn count(self) -> usize {
		self.len()
	}
}

impl ExactSizeIterator for Zeroes<'_> {
	#[inline]
	fn len(&self) -> usize {
		self.iter.count_to(self.bits, self.end)
	}
}

impl FusedIterator for Zeroes<'_> {}

/* ******************************************************* Internal Iterators */

/// Iterator helper that returns all of the non-cleared bit-indices in the set.
#[derive(Copy, Clone)]
pub(super) struct RawOnes {
	/// Currently cached word.
	cache: BaseType,
	/// Cursor to the currently selected word.
	cursor: BitCursor
}

impl RawOnes {
	/// Creates a new raw iterator over set bits that starts at the beginning.
	#[inline]
	pub(super) fn new(bits: &[BaseType]) -> Self {
		Self {
			cache: bits.first().cloned().unwrap_or_default(),
			cursor: BitCursor::new(0)
		}
	}
	
	/// Creates a new raw iterator over set bits that starts at a given bit
	/// index.
	#[inline]
	pub(super) fn new_at(bits: &[BaseType], start: usize) -> Self {
		let cursor = BitCursor::new(start);
		let mask = !0 << cursor.bit();
		
		Self {
			cache: bits.get(cursor.efa()).map_or(0, |&w| w & mask),
			cursor: cursor.to_bit(0)
		}
	}
	
	/// Finds the next bit index with a set bit and returns it if it exists.
	#[inline]
	fn next(&mut self, bits: &[BaseType]) -> Option<usize> {
		// short-circuit if the cached word is exhausted and we're outside the
		// valid range or the step moved us there
		if self.cache == 0 && (self.cursor.efa() >= bits.len() || !self.step(bits)) {
			return None;
		}
		
		let bit = self.cursor.pth() + self.cache.trailing_zeros() as usize;
		self.cache &= self.cache - 1;
		Some(bit)
	}
	
	/// Performs the heavy lifting in advancing the iterator, returning `true`
	/// if the next non-empty word has been located and `false` if there is no
	/// non-empty word remaining.
	fn step(&mut self, bits: &[BaseType]) -> bool {
		// convert the bit cursor into a leaf cursor
		let mut cursor: Cursor = self.cursor.to_leaf().into();
		let mut cache;
		
		// locate an entry with a not-yet-traversed path
		'select_branch: loop {
			match cursor.parent() {
				Ok((parent, branch)) => {
					// update the cursor
					cursor = parent;
					
					// select the next word
					cache = *unsafe { bits.get_unchecked(cursor.efa()) };
					// disregard branches already visited
					cache &= !1 << branch;
					
					// stop ascending if we found a non-clear bit
					if cache != 0 { break 'select_branch; }
				}
				
				Err(mut root) => loop {
					// select the next root node
					root = root.successor();
					
					// read the next root entry
					match bits.get(root.efa()) {
						Some(&w) if w != 0 => {
							// found a non-empty root entry
							cache = w;
							cursor = root.into();
							break 'select_branch;
						}
						Some(_) => { /* continue with the next root */ }
						None => {
							// no additional roots within range
							self.cache = 0;
							self.cursor = Cursor::from_bit(usize::MAX);
							return false;
						}
					}
				}
			}
		}
		
		// descend until we arrive at the leaf entry
		loop {
			match cursor.child(cache.trailing_zeros() as usize) {
				Ok(child) => {
					cursor = child;
					cache = *unsafe { bits.get_unchecked(cursor.efa()) };
				}
				Err(leaf) => {
					// convert the leaf cursor back into a bit cursor
					self.cache = cache;
					self.cursor = leaf.to_bit(0);
					break true;
				}
			}
		}
	}
	
	/// Drains a single bit from the set and returns it, if it exists.
	pub(super) fn drain_bit(&mut self, bits: &mut [BaseType]) -> Option<usize> {
		let bit = self.next(bits)?;
		let mut cursor: Cursor = self.cursor.to_leaf().into();
		
		// update the bits in the higher layers
		loop {
			// SAFETY: first iteration: next() only returns `Some` if there is
			// a valid effective address with a non-drained word;
			// every other iteration: since we're only navigating up the tree,
			// the parent nodes have to exist by construction
			let adr = unsafe {
				bits.get_unchecked_mut(cursor.efa())
			};
			
			// disable the least significant bit that was just extracted
			*adr &= *adr - 1;
			
			// if the word is not 0, exit
			if *adr != 0 { break; }
			
			// select the next path segment
			match cursor.parent() {
				Ok((parent, _)) => { cursor = parent; }
				Err(_) => break
			}
		}
		
		// return the originally extracted bit
		Some(bit)
	}
	
	/// Counts the number of remaining set bits.
	pub(super) fn count(mut self, bits: &[BaseType]) -> usize {
		let mut cnt = 0;
		
		loop {
			cnt += self.cache.count_ones() as usize;
			if bits.len() <= self.cursor.efa() || !self.step(bits) { break cnt; }
		}
	}
	
	/// Counts the number of remaining set bits up to a bit index.
	pub(super) fn count_to(mut self, bits: &[BaseType], end: usize) -> usize {
		let mut cnt = 0;
		
		// count all of the complete words
		while end > self.cursor.pth() + BITS - 1 {
			cnt += self.cache.count_ones() as usize;
			if bits.len() <= self.cursor.efa() || !self.step(bits) { return cnt; }
		}
		
		// pay special attention to the final, incomplete word
		let diff = end.saturating_sub(self.cursor.pth());
		cnt + (self.cache & ((1 << diff) - 1)).count_ones() as usize
	}
}

/// Iterator helper that returns all of the non-set bits in a bitset.
#[derive(Copy, Clone)]
struct RawZeroes {
	/// Cached and inverted word currently being drained.
	cache: BaseType,
	/// Cursor to the currently selected word.
	cursor: BitCursor
}

impl RawZeroes {
	/// Given a slice of bits, creates an iterator that traverses the non-set
	/// bits.
	// #[inline]
	// fn new(bits: &[BaseType]) -> Self {
	// 	Self {
	// 		cache: bits.first().map_or(!0, |&w| !w),
	// 		cursor: Cursor::from_bit(0)
	// 	}
	// }
	
	/// Given a slice of bits and a start bit, creates an iterator that
	/// traverses the non-set bits after (and including) the start bit.
	#[inline]
	fn new_at(bits: &[BaseType], start: usize) -> Self {
		let cursor = Cursor::from_bit(start);
		let mask = !0 << cursor.bit();
		
		Self {
			cache: bits.get(cursor.efa()).map_or(mask, |&w| !w & mask),
			cursor: cursor.to_bit(0)
		}
	}
	
	/// Returns the next unset bit-index, which is always successful but
	/// undefined if the value overflows `usize` word boundaries.
	#[inline]
	fn next(&mut self, bits: &[BaseType]) -> usize {
		// check if we need to advance the position in the bit array
		if self.cache == 0 { self.step(bits); }
		
		// extract the next bit from the set and clear it from the stored word
		let bit = self.cursor.pth() + self.cache.trailing_zeros() as usize;
		self.cache &= self.cache - 1;
		
		// return the bit index
		bit
	}
	
	/// Advances the iterator to the next word with non-set bits.
	fn step(&mut self, bits: &[BaseType]) {
		// convert into a leaf cursor
		let mut cursor = self.cursor.to_leaf();
		
		// loop until we find a non-full word
		self.cache = loop {
			// select the next word
			cursor = cursor.successor();
			
			// check if the word contains 0-bits
			match bits.get(cursor.efa()) {
				None => break !0,
				Some(&adr) if adr != !0 => break !adr,
				Some(_) => { /* continue searching for a non-full word */ }
			}
		};
		
		// convert back into a bit cursor
		self.cursor = cursor.to_bit(0);
	}
	
	/// Counts the number of unset bits from the current bit position to some
	/// bit index.
	fn count_to(mut self, bits: &[BaseType], end: usize) -> usize {
		let mut cnt = 0;
		
		while bits.len() > self.cursor.efa() && end > self.cursor.pth() + BITS - 1 {
			cnt += self.cache.count_ones() as usize;
			self.step(bits);
		}
		
		let diff = end.saturating_sub(self.cursor.pth());
		
		cnt
		+ diff / BITS * BITS
		+ (self.cache & ((1 << (diff % BITS)) - 1)).count_ones() as usize
	}
}
