use super::{BitSet, BaseType, Cursor, BitCursor};
use std::ops::{
	Bound, Range, RangeInclusive, RangeToInclusive, RangeFrom, RangeTo,
	RangeFull, RangeBounds
};

mod private_bit_slice_update {
	use std::ops;
	
	pub trait Sealed: Sized {}
	
	impl Sealed for usize {}
	impl Sealed for ops::Range<usize> {}
	impl Sealed for ops::RangeTo<usize> {}
	impl Sealed for ops::RangeFrom<usize> {}
	impl Sealed for ops::RangeFull {}
	impl Sealed for ops::RangeInclusive<usize> {}
	impl Sealed for ops::RangeToInclusive<usize> {}
	impl Sealed for (ops::Bound<usize>, ops::Bound<usize>) {}
}

/// A helper trait used to unify insertion and removal of bits from [`BitSet`]
/// instances.
///
/// The mechanism of this is very similar to [`SliceIndex`] from the standard
/// library.
///
/// [`BitSet`]: ../struct.BitSet.html
/// [`SliceIndex`]: https://doc.rust-lang.org/std/slice/trait.SliceIndex.html
pub trait BitSliceUpdate: private_bit_slice_update::Sealed {
	/// The result-type of the update method.
	type Output;
	
	/// Sets or clears a single bit or a range of bits.
	fn update(self, set: &mut BitSet, flag: bool) -> Self::Output;
	
	/// Inserts a single bit index or a range of bits.
	#[inline(always)]
	fn insert(self, set: &mut BitSet) -> Self::Output {
		self.update(set, true)
	}
	
	/// Removes a single bit index or a range of bits.
	#[inline(always)]
	fn remove(self, set: &mut BitSet) -> Self::Output {
		self.update(set, false)
	}
}

impl BitSliceUpdate for usize {
	type Output = bool;
	
	#[inline]
	fn update(self, set: &mut BitSet, flag: bool) -> bool {
		#[cold]
		unsafe fn consolidate(bits: &mut [BaseType], mut cursor: Cursor, f_bits: BaseType) {
			// ascend to the root node
			while let Ok((parent, branch)) = cursor.parent() {
				// advance the cursor position
				cursor = parent;
				
				// access the word
				let word = bits.get_unchecked_mut(cursor.efa());
				let mask = 1 << branch;
				
				// set or clear the bit based on `f_bits`
				// credits to https://graphics.stanford.edu/~seander/bithacks.html
				*word = (*word & !mask) | (f_bits & mask);
				
				// update and exit if higher layers cannot have been affected
				if (*word & !mask) != 0 { break; }
			}
		}
		
		// construct a bit cursor for the bit index
		let cursor = Cursor::from_bit(self);
		
		// prepare the mask to isolate the selected bit
		let mask = 1 << cursor.bit();
		
		// convert the flag into a bitmask
		// (compiles down to a branchless negation of `flag`)
		let f_bits = if flag { !0 } else { 0 };
		
		// try to access the word containing our bit
		if let Some(word) = set.bits.get_mut(cursor.efa()) {
			// save the value
			let w = *word;
			
			// set or clear the bit based on `f_bits`
			// credits to https://graphics.stanford.edu/~seander/bithacks.html
			*word = (*word & !mask) | (f_bits & mask);
			
			// exit early if no higher layers could have been affected
			// Here is how it works in principle: we're testing whether the
			// current word either had bits other than the one from `mask` set,
			// or if setting/clearing of that bit made no difference for the
			// value of `word`. Either case allows us to shortcut.
			
			// If `flag` is true (we're inserting) then f_bits will be all ones.
			// In this case, the following condition checks whether w != 0, i.e.
			// whether the previous value of `word` was different from 0.
			// This would mean that the higher layers must have already been set
			// for this word and no update is necessary.
			
			// If `flag` is false (we're clearing) then `f_bits` will be zero.
			// In this case, the following condition checks whether w != mask,
			// i.e. whether the previous value of `word` had bits other than
			// the one from `mask` set.
			// This would mean that the higher layers are either justified
			// to having been set for this word (because there are bits *other*
			// than the one in `mask` set) or that they are justified to be
			// cleared (because the bit in mask was clear already).
			// Either case allows the shortcut.
			
			if w != (!f_bits & mask) { return (w & mask) != 0; }
			
			// we need to update the higher layers
		} else if flag {
			// we weren't able to access the word, so let's extend the vector
			set.bits.resize(cursor.efa() + 1, 0);
			
			// SAFETY: since we extended the vector to len == efa+1,
			// accessing the vector at position `efa` is safe
			*unsafe { set.bits.get_unchecked_mut(cursor.efa()) } = mask;
			
			// we need to update the higher layers
		} else {
			// if we wanted to clear the bit, there is nothing more to be done
			return false;
		}
		
		// update the higher layers
		unsafe {
			consolidate(&mut set.bits, cursor.to_leaf().into(), f_bits);
		}
		
		// we must have flipped the bit to have made it here
		!flag
	}
}

impl BitSliceUpdate for (Bound<usize>, Bound<usize>) {
	type Output = ();
	
	#[inline]
	fn update(self, set: &mut BitSet, flag: bool) {
		// convert the bounds into an inclusive interval
		let bit = (match self.0 {
			Bound::Included(bound) => bound,
			Bound::Excluded(usize::MAX) => return,
			Bound::Excluded(bound) => bound + 1,
			Bound::Unbounded => 0
		}, match self.1 {
			Bound::Included(bound) => bound,
			Bound::Excluded(0) => return,
			Bound::Excluded(bound) => bound - 1,
			Bound::Unbounded => !0
		});
		
		// short-circuit empty intervals
		if bit.0 > bit.1 { return; }
		
		// perform the heavy lifting
		if flag {
			insert_interval(set, bit);
		} else {
			remove_interval(set, bit);
		}
	}
}

struct Updater<'a, const F: bool>(&'a mut BitSet);

impl<const F: bool> Updater<'_,F> {
	#[inline]
	unsafe fn update(&mut self, efa: usize, bits: BaseType) -> bool {
		// access the word
		let word = self.bits().get_unchecked_mut(efa);
		
		// set the bits
		if F { *word |= bits; } else { *word &= !bits; }
		
		// returns if the bit in the higher layer should be set already
		(*word & !bits) != 0
	}
	
	#[inline]
	unsafe fn fill(&mut self, rng: Range<usize>) {
		self.bits().get_unchecked_mut(rng).fill(if F { !0 } else { 0 });
	}
	
	#[inline]
	fn bits(&mut self) -> &mut Vec<BaseType> {
		&mut self.0.bits
	}
}

fn insert_interval(set: &mut BitSet, range: (usize, usize)) {
	let (start, end) = BitCursor::from_range(range);
	let mut set = Updater::<true>(set);
	
	// resize the bit set if necessary
	if set.bits().len() <= end.efa() {
		set.bits().resize(end.efa() + 1, 0);
	}
	
	// extract the bit indices
	let (mut bs, mut be) = (start.bit(), end.bit());
	
	// upgrade the bit cursors to fully fledged cursors
	let (mut start, mut end): (Cursor, Cursor) =
		(start.to_leaf().into(), end.to_leaf().into());
	
	// test for single-word-inserts
	if start != end {
		let mut sup;
		
		// handle the back end of the set
		loop {
			sup = end.efa();
			unsafe { set.update(sup, !(!1 << be)); }
			
			match end.parent() {
				Ok((parent, branch)) => {
					end = parent;
					be = branch;
					
					if start.efa() > end.efa() { break; }
					unsafe { set.fill(end.efa() + 1 .. sup); }
				}
				Err(_) => break
			}
		}
		
		// fill the main body
		unsafe { set.fill(start.efa() + 1 .. sup); }
		
		// handle the front end of the set
		while start != end {
			unsafe { set.update(start.efa(), !0 << bs); }
			
			match start.parent() {
				Ok((parent, branch)) => {
					start = parent;
					bs = branch;
				}
				Err(_) => return
			}
		}
	}
	
	// update the intersection
	if unsafe {
		!set.update(start.efa(), (!0 << bs) ^ (!1 << be))
	} {
		// consolidate the shared layers
		while let Ok((parent, branch)) = start.parent() {
			// update and exit if higher layers cannot have been affected
			if unsafe { set.update(parent.efa(), 1 << branch) } { break; }
			
			// advance the cursor position
			start = parent;
		}
	}
}

fn remove_interval(set: &mut BitSet, range: (usize, usize)) {
	let (start, end) = BitCursor::from_range(range);
	let mut set = Updater::<false>(set);
	
	// test if the back end is inside of the allocated range
	if set.bits().len() <= end.efa() {
		// test if the front end is outside as well
		if set.bits().len() > start.efa() {
			// `end` is outside of the allocated range but `start` isn't
			set.bits().truncate(start.efa() + 1);
			
			// clear the bits after `start` in the selected leaf word
			let mut flag = BaseType::from(unsafe {
				set.update(start.efa(), !0 << start.bit())
			});
			
			// propagate the clearing upwards to the root
			let mut start: Cursor = start.to_leaf().into();
			while let Ok((parent, branch)) = start.parent() {
				start = parent;
				flag = BaseType::from(unsafe {
					set.update(start.efa(), !flag << branch)
				});
			}
		}
		
		// EITHER: front and back ends are both outside the allocated range
		//     OR: the situation has been dealt with through truncation + update
		return;
	}
	
	// extract the bit indices
	let (mut bs, mut be) = (start.bit(), end.bit());
	
	// upgrade the bit cursors into fully fledged cursors
	let (mut start, mut end): (Cursor, Cursor) =
		(start.to_leaf().into(), end.to_leaf().into());
	
	// store the state of the updated words
	let mut flags = (!0, !1);
	
	// test for single-word-removes
	if start != end {
		let mut sup;
		
		// handle the back end of the set
		loop {
			// store the effective address at the end
			sup = end.efa();
			
			// convert the mask from inclusive to exclusive with regards
			// to the previous branch decision; this prevents disabling
			// the bit corresponding to the branch in higher layers if
			// the word in the layer below it wasn't 0
			flags.1 |= BaseType::from(unsafe {
				set.update(sup, !(flags.1 << be))
			});
			
			match end.parent() {
				Ok((parent, branch)) => {
					end = parent;
					be = branch;
					
					if start.efa() > end.efa() { break; }
					unsafe { set.fill(end.efa() + 1 .. sup); }
				}
				Err(_) => break
			}
		}
		
		// clear the gap between the start cursor and the last separated word
		unsafe { set.fill(start.efa() + 1 .. sup); }
		
		// handle the front end of the set
		while start != end {
			// convert the mask from inclusive to exclusive with regards
			// to the previous branch decision; this prevents disabling
			// the bit corresponding to the branch in higher layers if
			// the word in the layer below it wasn't empty
			flags.0 &= !BaseType::from(unsafe {
				set.update(start.efa(), flags.0 << bs)
			});
			
			match start.parent() {
				Ok((parent, branch)) => {
					start = parent;
					bs = branch;
				}
				Err(_) => return
			}
		}
	}
	
	// update and test if higher layers are affected by this change
	if unsafe {
		!set.update(start.efa(), (flags.0 << bs) ^ (flags.1 << be))
	} {
		// consolidate the remaining layers
		while let Ok((parent, branch)) = start.parent() {
			// update and exit if higher layers cannot have been affected
			if unsafe { set.update(parent.efa(), 1 << branch) } {
				break;
			}
			
			// advance the cursor position
			start = parent;
		}
	}
}

macro_rules! bit_slice_range_update_impl {
	($($T:ident),*) => {$(
		impl BitSliceUpdate for $T<usize> {
			type Output = ();
			
			#[inline(always)]
			fn update(self, set: &mut BitSet, flag: bool) {
				(self.start_bound().cloned(), self.end_bound().cloned())
					.update(set, flag)
			}
			
			#[inline(always)]
			fn insert(self, set: &mut BitSet) {
				(self.start_bound().cloned(), self.end_bound().cloned())
					.insert(set)
			}
			
			#[inline(always)]
			fn remove(self, set: &mut BitSet) {
				(self.start_bound().cloned(), self.end_bound().cloned())
					.remove(set)
			}
		}
	)*};
}

bit_slice_range_update_impl!(
	Range, RangeInclusive, RangeTo, RangeFrom, RangeToInclusive
);

impl BitSliceUpdate for RangeFull {
	type Output = ();
	
	#[inline(always)]
	fn update(self, set: &mut BitSet, flag: bool) {
		if flag {
			self.insert(set);
		} else {
			self.remove(set);
		}
	}
	
	#[inline(always)]
	fn insert(self, _set: &mut BitSet) {
		panic!("this cannot work, you'd need more memory than can reasonably \
		        be assumed to be available to set every single bit")
	}
	
	#[inline(always)]
	fn remove(self, set: &mut BitSet) {
		set.clear();
	}
}
