//! This module defines an arena allocator for [Active] simulation objects that
//! performs automatic garbage collection on terminated and unreferenced
//! instances.
//!
//! # Arena Allocator
//!
//! The [`Pool`] struct is the arena allocator that manages the allocation and
//! recycling of [Active] simulation objects. It includes methods for allocating
//! new objects, checking the number of allocated objects, and handling object
//! recycling.
//!
//! # Backing Store Types
//!
//! The [`Storage`] trait unifies the backing store of the `Pool`, providing
//! methods for allocating and managing storage. Two implementations, [`Fixed`]
//! and [`Arena`](typed_arena::Arena), serve as backing store types for the
//! `Pool`.
//!
//! # Garbage Collecting Type
//!
//! The [`Gc`] struct is a wrapper type that accepts [Active] types and
//! automatically returns them to the `Pool` once they are no longer
//! referenced. It includes methods for creating a new garbage-collected instance,
//! polling the active object, and reclaiming the object to the `Pool`.
//!
//! # Example Usage
//!
//! ```
//! # use odem_rs_util::pool;
//! # use odem_rs_core::{Fiber, Sim};
//!
//! async fn sim_main(sim: &Sim) {
//!     // Create a dynamically growing pool with default initial capacity.
//!     let pool = pool!();
//!     
//!     // Allocate a new Fiber and activate it in the simulation.
//!     sim.activate(pool.alloc(Fiber::new(async {})));
//! }
//! ```
//!
//! For more specific details, refer to the individual documentation of each type
//! and method within this module.

use core::{
	cell::{Cell, RefCell},
	mem::MaybeUninit,
	pin::Pin,
	ptr::NonNull,
	task::{Context, Poll},
};

use intrusive_collections::{intrusive_adapter, SinglyLinkedList, SinglyLinkedListLink, UnsafeRef};

use odem_rs_core::{error, Active, Config, Dispatch, ExitStatus, Share, Unique};

/* ********************************************************** Arena Allocator */

/// An arena allocator for [Active] simulation objects that performs automatic
/// garbage collection on terminated and unreferenced instances.
pub struct Pool<A, S> {
	/// A linked list of terminated instances ready to be recycled.
	#[doc(hidden)]
	pub reclaim: RefCell<SinglyLinkedList<GcAdapter<A>>>,
	/// A backing store used to allocate more instances.
	#[doc(hidden)]
	pub storage: S,
}

impl<A, S: Storage<Gc<A>>> Pool<A, S> {
	/// Allocates a new, [Active] object, ready to be activated.
	///
	/// It is possible that the same memory location is returned multiple
	/// times, but only if the object is no longer in use.
	#[inline]
	pub fn try_alloc(&self, value: A) -> Result<Unique<Gc<A>>, A> {
		self.try_recycle(value)
			.or_else(|value| self.storage.try_alloc(Gc::new(value, &self.reclaim)))
			.map_err(|reclaim| reclaim.active)
	}

	/// Allocates a new, [Active] object, ready to be activated.
	///
	/// It is possible that the same memory location is returned multiple
	/// times, but only if the object is no longer in use.
	#[inline]
	pub fn alloc(&self, value: A) -> Unique<Gc<A>> {
		self.try_recycle(value)
			.unwrap_or_else(|value| self.storage.alloc(Gc::new(value, &self.reclaim)))
	}

	/// Returns the number of simultaneously allocated objects.
	#[inline]
	pub fn len(&self) -> usize {
		self.storage.len()
	}

	/// Returns whether there are no allocated objects in the pool.
	#[inline]
	pub fn is_empty(&self) -> bool {
		self.storage.is_empty()
	}

	/// Attempts to initialize a recycled object, returning a pinned exclusive
	/// reference to it if successful or the value itself otherwise.
	fn try_recycle(&self, value: A) -> Result<Unique<Gc<A>>, A> {
		if let Some(item) = self.reclaim.borrow_mut().pop_front() {
			let mut old = unsafe { Pin::new_unchecked(&mut *UnsafeRef::into_raw(item)) };
			old.set(Gc::new(value, &self.reclaim));
			Ok(unsafe { Unique::new_unchecked(old) })
		} else {
			Err(value)
		}
	}
}

impl<'a, A, const N: usize> Pool<A, Pin<&'a mut Fixed<Gc<A>, N>>> {
	/// Creates an arena with a fixed backing store given a maximal size.
	///
	/// This arena will panic if too many instances exist simultaneously
	/// during a simulation run.
	#[inline]
	pub fn fixed(store: Pin<&'a mut Fixed<Gc<A>, N>>) -> Self {
		Pool::from(store)
	}
}

#[cfg(feature = "alloc")]
impl<A> Pool<A, typed_arena::Arena<Gc<A>>> {
	/// Creates an automatically growing arena.
	#[inline]
	pub fn dynamic() -> Self {
		Pool::from(typed_arena::Arena::new())
	}

	/// Creates an automatically growing arena with an initial minimal capacity.
	#[inline]
	pub fn with_capacity(n: usize) -> Self {
		Pool::from(typed_arena::Arena::with_capacity(n))
	}
}

impl<A, S> From<S> for Pool<A, S> {
	#[inline]
	fn from(store: S) -> Self {
		Pool {
			reclaim: RefCell::new(SinglyLinkedList::new(GcAdapter::new())),
			storage: store,
		}
	}
}

impl<A, S: Default> Default for Pool<A, S> {
	#[inline]
	fn default() -> Self {
		Pool::from(S::default())
	}
}

/* ****************************************************** Backing Store Types */

/// A trait to unify the backing store of a [`Pool`].
///
/// This trait defines methods for allocating and managing storage for a `Pool`
/// of [Active] simulation objects. It provides flexibility for different types
/// of backing stores, such as fixed-size arrays or dynamically growing
/// allocators.
///
/// # Implementations
///
/// Implementations of this trait are expected to define how storage is
/// allocated, whether allocation is attempted, and how the number of allocated
/// items is tracked. Two primary implementations, [`Fixed`] and [`Arena`],
/// serve as backing store types for the `Pool`.
///
/// [`Arena`]: typed_arena::Arena
pub trait Storage<T> {
	/// Allocates a [pinned](Pin) mutable pointer to the object type, causing a
	/// panic if the backing store is full.
	#[inline]
	fn alloc(&self, value: T) -> Unique<T> {
		self.try_alloc(value)
			.ok()
			.expect("sufficient storage space")
	}

	/// Returns whether there have not been allocation attempts yet.
	#[inline]
	fn is_empty(&self) -> bool {
		self.len() == 0
	}

	/// Attempts to allocate a [pinned](Pin) mutable pointer to the object type,
	/// returning the value if the backing store is full.
	fn try_alloc(&self, value: T) -> Result<Unique<T>, T>;

	/// Returns the number of allocated items in the backing store.
	fn len(&self) -> usize;
}

/// A constant-sized array, usable as [Storage] for a [Pool].
#[pin_project::pin_project(PinnedDrop)]
pub struct Fixed<A, const N: usize> {
	/// The number of allocated elements.
	len: Cell<usize>,
	/// The storage array.
	#[pin]
	mem: [Cell<MaybeUninit<A>>; N],
}

impl<A, const N: usize> Fixed<A, N> {
	#[allow(clippy::declare_interior_mutable_const)]
	const FRESH: Cell<MaybeUninit<A>> = Cell::new(MaybeUninit::uninit());

	/// Creates a new, empty storage array.
	#[inline]
	pub const fn new() -> Self {
		Fixed {
			len: Cell::new(0),
			mem: [Self::FRESH; N],
		}
	}
}

impl<A, const N: usize> Default for Fixed<A, N> {
	#[inline]
	fn default() -> Self {
		Fixed::new()
	}
}

impl<A, const N: usize> Storage<A> for Pin<&mut Fixed<A, N>> {
	#[inline]
	fn try_alloc(&self, value: A) -> Result<Unique<A>, A> {
		// take the next available slot or return an error
		let res = match self.mem.get(self.len.get()) {
			Some(res) => res,
			None => return Err(value),
		};

		// increase the length by one
		self.len.set(self.len.get() + 1);

		// return the initialized slot
		// SAFETY: pinning the exclusive reference is safe because the whole
		// array is pinned, and we won't move the value before dropping it
		Ok(unsafe { Unique::new_unchecked(Pin::new_unchecked((*res.as_ptr()).write(value))) })
	}

	#[inline]
	fn len(&self) -> usize {
		self.len.get()
	}
}

// take care of dropping the allocated values
#[pin_project::pinned_drop]
impl<A, const N: usize> PinnedDrop for Fixed<A, N> {
	#[inline]
	fn drop(self: Pin<&mut Self>) {
		let inner = self.project();

		for item in inner.mem.iter().take(inner.len.get()).rev() {
			unsafe {
				item.replace(MaybeUninit::uninit()).assume_init_drop();
			}
		}
	}
}

// make the typed-arena allocator compatible with our pool
#[cfg(feature = "alloc")]
impl<A> Storage<A> for typed_arena::Arena<A> {
	#[inline]
	fn alloc(&self, value: A) -> Unique<A> {
		// SAFETY: the typed-arena cannot move any values internally before
		// they are all dropped, since they all have the lifetime of the arena
		unsafe { Unique::new_unchecked(Pin::new_unchecked(self.alloc(value))) }
	}

	#[inline]
	fn try_alloc(&self, value: A) -> Result<Unique<A>, A> {
		Ok(Storage::alloc(self, value))
	}

	#[inline]
	fn len(&self) -> usize {
		self.len()
	}
}

/// Safely creates and initializes a [`Pool`] given a specific initial capacity
/// or storage size.
///
/// The `pool!` macro provides a convenient way to create instances of the
/// [`Pool`] structure. It allows users to specify the initial capacity or
/// storage size of the [`Pool`] at compile time. If the capacity is prefixed by
/// `const`, the initialized pool has a fixed size and doesn't require feature
/// `alloc` to be enabled. Otherwise, it grows dynamically in capacity.
/// The initial capacity is optional in the latter case.
///
/// # Examples
///
/// A dynamically growing pool with default initial capacity:
///
/// ```
/// # use odem_rs_util::pool;
/// # use odem_rs_core::{Fiber, Sim};
/// # async fn test(sim: &Sim) {
/// let pool = pool!();
/// # sim.activate(pool.alloc(Fiber::new(async {})));
/// # }
/// ```
///
/// A dynamically growing pool with specific initial capacity:
///
/// ```
/// # use odem_rs_util::pool;
/// # use odem_rs_core::{Fiber, Sim};
/// # async fn test(sim: &Sim, n: usize) {
/// let pool = pool!(n + 2);
/// # sim.activate(pool.alloc(Fiber::new(async {})));
/// # }
/// ```
///
/// A fixed-size pool with capacity:
///
/// ```
/// # use odem_rs_util::pool;
/// # use odem_rs_core::{Fiber, Sim};
/// # async fn test(sim: &Sim) {
/// let pool = pool!(const 8 + 4);
/// # sim.activate(pool.alloc(Fiber::new(async {})));
/// # }
/// ```
///
/// # Notes
///
/// - The macro internally creates a [`Fixed`] or [`Arena`](typed_arena::Arena)
///   based on the specified capacity, providing flexibility for different use
///   cases.
/// - [`Pool`] instances manage terminated instances for automatic recycling.
#[macro_export]
macro_rules! pool {
	(const $N:expr) => {
		$crate::pool::Pool {
			storage: ::core::pin::pin!($crate::pool::Fixed::<_, { $N }>::new()),
			reclaim: ::core::cell::RefCell::new(::intrusive_collections::SinglyLinkedList::new(
				$crate::pool::GcAdapter::NEW,
			)),
		}
	};

	() => {
		$crate::pool::Pool::dynamic()
	};
	($N:expr) => {
		$crate::pool::Pool::with_capacity($N)
	};
}

/* ************************************************** Garbage Collecting Type */

/// Wrapper-type that accepts [Active] types and automatically returns them
/// to the [Pool] that spawned it once they are no longer referenced.
#[pin_project::pin_project]
pub struct Gc<A> {
	/// An intrusive hook.
	#[pin]
	hook: SinglyLinkedListLink,
	/// A raw pointer to the buffer inside the arena.
	buffer: NonNull<RefCell<SinglyLinkedList<GcAdapter<A>>>>,
	/// The active object.
	#[pin]
	active: A,
}

intrusive_adapter! {
	/// The [Adapter] type for [Gc].
	///
	/// [Adapter]: intrusive_collections::Adapter
	pub GcAdapter<A> = UnsafeRef<Gc<A>>: Gc<A> {
		hook: SinglyLinkedListLink
	}
}

impl<A> Gc<A> {
	/// Creates a new garbage-collected instance linked to an intrusive list.
	#[inline]
	fn new(active: A, buffer: &RefCell<SinglyLinkedList<GcAdapter<A>>>) -> Self {
		Self {
			hook: SinglyLinkedListLink::new(),
			buffer: NonNull::from(buffer),
			active,
		}
	}
}

impl<A: Dispatch> Dispatch for Gc<A> {
	#[inline]
	fn poll(self: Pin<&Self>, cx: &mut Context<'_>) -> Poll<ExitStatus> {
		self.project_ref().active.poll(cx)
	}

	#[inline]
	fn reclaim(self: Pin<&mut Self>) {
		// add the garbage collected item to the stored list
		unsafe { self.buffer.as_ref().borrow_mut() }
			.push_front(unsafe { UnsafeRef::from_raw(self.get_unchecked_mut()) });
	}
}

impl<C: Config, A: Active<C>> Active<C> for Gc<A> {
	type Puck<'p> = <A as Active<C>>::Puck<'p>
	where
		A: 'p;

	#[inline]
	unsafe fn bind<'p>(
		this: Unique<'p, Self>,
		vtab: NonNull<dyn Dispatch + 'p>,
		sctx: NonNull<Share<C>>,
	) -> Result<Self::Puck<'p>, error::NotBorn> {
		Active::bind(
			this.map_unchecked(|inner| inner.project().active),
			vtab,
			sctx,
		)
	}
}
