//! Contains facilities to ease the creation, distribution and use of random
//! number generators based on the [`rand`]-crate.
//!
//! The module offers [RngStream] as a utility structure to deal with
//! pseudo-randomness. It provides a stream of independent random-number
//! generators for use in a single simulator.

use core::cell::RefCell;

use rand::{rngs::mock::StepRng, SeedableRng};

#[cfg(feature = "rand_pcg")]
use rand_pcg::*;

#[cfg(feature = "rand_xoshiro")]
use rand_xoshiro::*;

/// Type-alias for the default random-number generator that is used in absense
/// of another user-selected one.
#[cfg(feature = "rand_xoshiro")]
pub type DefaultRng = Xoshiro256PlusPlus;

#[cfg(all(not(feature = "rand_xoshiro"), feature = "rand_pcg"))]
pub type DefaultRng = Pcg64Mcg;

/// Helper-trait providing a way to create new random-number generators based
/// on internal state.
///
/// This trait provides glue code between [RngStream] and PRNGs.
pub trait RngGen {
	/// The internal state necessary to generate random-number generators.
	type State;

	/// Initializes the state using a default value.
	fn default_state() -> Self::State;

	/// Generates new state from a seed value.
	fn from_seed(seed: u64) -> Self::State;

	/// Generates a new, independent random-number generator from the state.
	fn jump(state: &mut Self::State) -> Self;

	/// Generates new state suitable for the initialization of a [RngStream]
	/// producing independent random-number generators and updates the previous
	/// state.
	fn fork(state: &mut Self::State) -> Option<Self::State>;
}

/// Represents a stream of independent random-number generators, suitable for
/// one or more sequential simulation runs.
///
/// The underlying random-number generators is selectable by the user. The
/// default choice is using the [Xoshiro256PlusPlus] generator, allowing for
/// up to `2^128` independent generators with a sequence length of `2^128`.
///
/// It is important to note that the creation of another random-number generator
/// is not achieved through re-seeding but by jumping ahead in the random-number
/// stream in a constant operation, which efficiently creates provably
/// independent random-number streams.
pub struct RngStream<R: RngGen = DefaultRng>(RefCell<R::State>);

impl<R: RngGen> RngStream<R> {
	/// Creates a new RngStream from inner state.
	#[inline]
	pub const fn new(inner: R::State) -> Self {
		RngStream(RefCell::new(inner))
	}

	/// Creates a new RngStream from a seed value.
	///
	/// The function can safely be called with small low-entropy seed values
	/// and may be relied upon to generate uncorrelated RngStreams with high
	/// probability. This makes it suitable for reproducably running simulations
	/// in parallel with slowly incrementing seed values.
	#[inline]
	pub fn from_seed(seed: u64) -> Self {
		RngStream(RefCell::new(R::from_seed(seed)))
	}

	/// Creates a new random number generator of type `R` and returns it.
	///
	/// The random sequence generated from the returned generator is independent
	/// from all of the previously generated ones.
	#[inline]
	pub fn rng(&self) -> R {
		R::jump(&mut self.0.borrow_mut())
	}

	/// Creates an iterator that returns independent random number generators
	/// of type `R`.
	#[inline]
	pub fn rng_iter(&self) -> RngIter<R> {
		RngIter(self)
	}
}

impl<R: RngGen> IntoIterator for RngStream<R> {
	type Item = Self;
	type IntoIter = IntoIter<R>;

	#[inline]
	fn into_iter(self) -> Self::IntoIter {
		IntoIter(self)
	}
}

impl<R: RngGen> Default for RngStream<R> {
	#[inline]
	fn default() -> Self {
		RngStream::new(R::default_state())
	}
}

/* ************************************************************ Rng Iterators */

/// An iterator over a set of independent Random Number generators.
pub struct RngIter<'r, R: RngGen>(&'r RngStream<R>);

impl<R: RngGen> Iterator for RngIter<'_, R> {
	type Item = R;

	#[inline]
	fn next(&mut self) -> Option<Self::Item> {
		Some(self.0.rng())
	}
}

/// An owning iterator over a set of independent Random Number Stream
/// generators.
///
/// This iterator can be used to seed a set of independent simulators each with
/// an [RngStream] producing independent random-number generators.
pub struct IntoIter<R: RngGen>(RngStream<R>);

impl<R: RngGen> Iterator for IntoIter<R> {
	type Item = RngStream<R>;

	#[inline]
	fn next(&mut self) -> Option<Self::Item> {
		Some(RngStream::new(R::fork(self.0 .0.get_mut())?))
	}
}

/* ************************************************* Built-In Implementations */

// Xoshiro-family of PRNGs
macro_rules! impl_stream_xoshiro {
	($($UID:ident),* $(,)?) => {$(
		impl RngGen for $UID {
			type State = $UID;

			#[inline]
			fn default_state() -> $UID {
				$UID::seed_from_u64(0)
			}

			#[inline]
			fn from_seed(seed: u64) -> Self::State {
				// the seed is hashed by SplitMix64 with excellent avalanche
				// effects before its used to initialize the state, so passing
				// it here directly should be fine
				$UID::seed_from_u64(seed)
			}

			#[inline]
			fn jump(state: &mut $UID) -> $UID {
				let res = state.clone();
				state.jump();
				res
			}

			#[inline]
			fn fork(state: &mut $UID) -> Option<$UID> {
				let res = state.clone();
				state.long_jump();
				Some(res)
			}
		}
	)*};
}

#[cfg(feature = "rand_xoshiro")]
impl_stream_xoshiro!(
	Xoroshiro128Plus,
	Xoroshiro128PlusPlus,
	Xoroshiro128StarStar,
	// Xoshiro128Plus, // doesn't support long_jump
	Xoshiro128PlusPlus,
	Xoshiro128StarStar,
	Xoshiro256Plus,
	Xoshiro256PlusPlus,
	Xoshiro256StarStar,
	Xoshiro512Plus,
	Xoshiro512PlusPlus,
	Xoshiro512StarStar,
);

// PCG-family of PRNGs
#[cfg(feature = "rand_pcg")]
impl RngGen for Pcg32 {
	type State = (u64, u64);

	#[inline]
	fn default_state() -> Self::State {
		(0xcafef00dd15ea5e5, 0xa02bdbf7bb3c0a7)
	}

	#[inline]
	fn from_seed(seed: u64) -> Self::State {
		let mut res = Self::default_state();

		// increase the stream sequence by seed * 2^32 in order to prevent
		// overlaps between the RNGs of different simulations
		res.1 += seed << 32;

		res
	}

	#[inline]
	fn jump((state, seq): &mut Self::State) -> Self {
		use rand::RngCore;
		let mut res = Pcg32::new(*state, *seq);

		// increase the sequence number
		*seq = seq.wrapping_add(1);

		// call the generated PRNG in order to vary the initial state;
		// RNGs may be closely correlated if only the sequence number differs
		*state = state.wrapping_add(res.next_u64());
		res
	}

	#[inline]
	fn fork(state: &mut Self::State) -> Option<Self::State> {
		let res = *state;
		state.1 = state.1.checked_add(1u64 << 32)?;
		Some(res)
	}
}

#[cfg(feature = "rand_pcg")]
impl RngGen for Pcg64 {
	type State = (u128, u128);

	#[inline]
	fn default_state() -> Self::State {
		(0xcafef00dd15ea5e5, 0xa02bdbf7bb3c0a7ac28fa16a64abf96)
	}

	#[inline]
	fn from_seed(seed: u64) -> Self::State {
		let mut res = Self::default_state();

		// increase the stream sequence by seed * 2^64 in order to prevent
		// overlaps between the RNGs of different simulations
		res.1 += (seed as u128) << 64;

		res
	}

	#[inline]
	fn jump((state, seq): &mut Self::State) -> Self {
		use rand::RngCore;
		let mut res = Pcg64::new(*state, *seq);

		// increase the sequence number
		*seq = seq.wrapping_add(1);

		// call the generated PRNG in order to vary the initial state;
		// RNGs may be closely correlated if only the sequence number differs
		*state = state.wrapping_add((res.next_u64() as u128) | ((res.next_u64() as u128) << 64));

		res
	}

	#[inline]
	fn fork(state: &mut Self::State) -> Option<Self::State> {
		let res = *state;
		state.1 = state.1.checked_add(1u128 << 64)?;
		Some(res)
	}
}

#[cfg(feature = "rand_pcg")]
impl RngGen for Pcg64Mcg {
	type State = Pcg64Mcg;

	#[inline]
	fn default_state() -> Self::State {
		Pcg64Mcg::new(0xcafef00dd15ea5e5)
	}

	#[inline]
	fn from_seed(seed: u64) -> Self::State {
		let mut res = Self::default_state();
		res.advance((seed as u128) << 96);
		res
	}

	#[inline]
	fn jump(state: &mut Self::State) -> Self {
		let res = state.clone();
		state.advance(1u128 << 64);
		res
	}

	#[inline]
	fn fork(state: &mut Self::State) -> Option<Self::State> {
		let res = state.clone();
		state.advance(1u128 << 96);
		Some(res)
	}
}

impl RngGen for StepRng {
	type State = (u64, u64);

	#[inline]
	fn default_state() -> Self::State {
		(0, 1)
	}

	#[inline]
	fn from_seed(seed: u64) -> Self::State {
		let mut res = Self::default_state();
		res.1 += seed;
		res
	}

	#[inline]
	fn jump(state: &mut Self::State) -> Self {
		let res = StepRng::new(state.0, state.1);
		state.0 += 1u64 << 32;
		res
	}

	#[inline]
	fn fork(state: &mut Self::State) -> Option<Self::State> {
		let res = *state;
		state.0 = state.0.checked_add(1u64 << 48)?;
		Some(res)
	}
}
