#include <odemx/odemx.h>
#include <vector>
#include <memory>
#include <iostream>
#include <utility>

using odemx::base::SimTime;
using odemx::base::Process;
using odemx::base::Simulation;
using odemx::statistics::Tally;

const SimTime SIM_DURATION = 24.0 * 60.0 * 7.0;
const SimTime HARBOR_DISTANCE = 10.0;
const SimTime FERRY_TIMEOUT = 5.0;
const unsigned int FERRY_CAPACITY = 5;
const unsigned int FERRY_COUNT = 2;
const unsigned int HARBOR_COUNT = 4;

///@brief A passive structure for cars.
struct Car {
	SimTime arrival_time;
	SimTime load_duration;
};

// These are just ODEMx-flavored sender/receiver-pairs for the same channel.
typedef odemx::synchronization::PortTailT<Car> PortTail;
typedef odemx::synchronization::PortHeadT<Car> PortHead;

/**
 * @brief The simulation context for a car ferry simulation.
 * 
 * This class simulates the operation of a car ferry system, including harbors
 * and ferries. It also collects statistical data on ferry cargo length, ferry
 * load time, and car wait time.
 */
struct CarFerry: Simulation {
	///@brief A unique pointer to the main process of the simulation.
	std::unique_ptr<Process> main;
	
	///@brief Constructor setting up the harbors and ferries.
	CarFerry(SimTime duration, unsigned int ferries, unsigned int harbors);

	///@brief A tally for aggregating the length of the ferry cargo.
	Tally ferry_cargo_len;
	///@brief A tally for aggregating the time taken to load the ferry.
	Tally ferry_load_time;
	///@brief A tally for aggregating the wait time of cars.
	Tally car_wait_time;

protected:
	///@brief Activates the main process.
	void initSimulation() override {
		main->activate();
	}
};

///@brief The process modeling a harbor with arriving cars.
struct Pier: Process {
	///@brief Writing end for the queue with arriving cars.
	PortTail::Ptr landing_site;

	///@brief The constructor initializing the car queue.
	Pier(CarFerry& sim)
	:	Process(sim, "Harbor"),
		landing_site(PortTail::create(sim, "Harbor Port")) {}
	
	///@brief Models the lifecycle of the harbor.
	int main() override {
		using namespace odemx::random;
		NegativeExponential arrival_delay(getSimulation(), "Arrival Delay", 0.1);
		Normal loading_delay(getSimulation(), "Loading Delay", 0.5, 0.2);

		while (true) {
			double load_duration;

			holdFor(arrival_delay.sample());

			do load_duration = loading_delay.sample();
			while (load_duration < 0.0);

			landing_site->put(Car { getTime(), load_duration });
		}
	}
};

///@brief The process modeling an individual ferry.
class Ferry: public Process {
	///@brief Space for the cars.
	std::vector<Car> cargo;
	///@brief Maximum car capacity.
	unsigned int capacity;
	///@brief Time to wait for another car before leaving before filling all
	/// of the available cargo space.
	SimTime timeout;
	///@brief Travel time between (any) two harbors.
	SimTime travel_time;
	///@brief Vector of reading ends for all of the harbors.
	std::vector<PortHead::Ptr> piers;

public:
	///@brief Constructor.
	Ferry(CarFerry& sim, unsigned int capacity, SimTime timeout,
	      SimTime travel_time, std::vector<PortHead::Ptr>&& piers)
	:	Process(sim, "Ferry"),
		capacity(capacity),
		timeout(timeout),
		travel_time(travel_time),
		piers(piers) {}

	///@brief Models the lifecycle of the ferry.
	int main() override {
		using namespace odemx::synchronization;
		auto& sim = static_cast<CarFerry&>(getSimulation());
		Timer timer(getSimulation(), "FerryTimer");
		IMemory* notifier;
		Car car;

		while (true) {
			for (auto& pier: piers) {
				// unload the cars
				for (const auto& car: cargo) {
					holdFor(car.load_duration);
				}
				cargo.clear();

				SimTime begin_loading = getTime();

				// wait until new cars arrive or a timeout occurs
				while (cargo.size() < capacity) {
					timer.setIn(timeout);
					notifier = wait(&timer, pier.get());

					if (notifier != &timer) {
						// a car arrived in time
						car = *pier->get();
						timer.stop();
						sim.car_wait_time.update(getTime() - car.arrival_time);
						holdFor(car.load_duration);
						cargo.push_back(car);
					} else {
						// the timeout has been triggered
						break;
					}
				}

				sim.ferry_load_time.update(getTime() - begin_loading);
				sim.ferry_cargo_len.update(cargo.size());

				// travel to the next harbor
				holdFor(travel_time);
			}
		}
	}
};

class SimMain: public Process {
	///@brief A vector for all of the harbors.
	std::vector<std::unique_ptr<Pier>> pierVec;
	///@brief A vector for all of the ferries.
	std::vector<std::unique_ptr<Ferry>> ferryVec;
	
	SimTime duration;
	
public:
	SimMain(CarFerry& sim, SimTime duration, unsigned int ferries, unsigned int harbors)
	: Process(sim, "Main Process"), duration(duration) {
		// create the harbors
		for (unsigned int i = 0; i < harbors; ++i) {
			pierVec.push_back(std::make_unique<Pier>(sim));
		}

		// create all of the ferries
		for (unsigned int i = 0; i < ferries; ++i) {
			std::vector<PortHead::Ptr> ports;

			for (unsigned int j = i; j < harbors; ++j) {
				ports.push_back(pierVec[j]->landing_site->getHead());
			}

			for (unsigned int j = 0; j < i; ++j) {
				ports.push_back(pierVec[j]->landing_site->getHead());
			}

			ferryVec.push_back(std::make_unique<Ferry>(
				sim, FERRY_CAPACITY, FERRY_TIMEOUT, HARBOR_DISTANCE,
				std::move(ports)
			));
		}
	}
	
	int main() override {
		// activate the harbors
		for (auto& pier: pierVec) {
			pier->activate();
		}
		
		// activate the ferries
		for (auto& ferry: ferryVec) {
			ferry->activate();
		}
		
		// wait until the simulation ends
		holdFor(duration);
		
		// finalize statistics collection
		auto& sim = static_cast<CarFerry&>(getSimulation());
		for (auto& pier: pierVec) {
			unsigned int count = pier->landing_site->count();
			PortHead::Ptr head = pier->landing_site->getHead();

			// take cars into account that weren't picked up by a ferry
			for (unsigned int i = 0; i < count; ++i) {
				sim.car_wait_time.update(sim.getTime() - head->get()->arrival_time);
			}
			
			// stop running the harbor
			pier->cancel();
		}
		
		// cancel the ferries
		for (auto& ferry: ferryVec) {
			ferry->cancel();
		}
		
		return 0;
	}
};

// Implementation of the simulation class.

CarFerry::CarFerry(SimTime duration, unsigned int ferries, unsigned int harbors)
:	Simulation("Ferry"),
	ferry_cargo_len(*this, "Ferry Cargo Len"),
	ferry_load_time(*this, "Ferry Load Time"),
	car_wait_time(*this, "Car Wait Time"),
	main(std::make_unique<SimMain>(*this, duration, ferries, harbors)) {}


#ifdef RUST_FFI
///@brief Entry point for the function used to benchmark this scenario.
extern "C" void ferry(SimTime duration, unsigned int ferries, unsigned int harbors) {
	// set up and run the simulation model
	CarFerry(duration, ferries, harbors).run();
}
#else
int main() {
	using odemx::data::output::OStreamReport;
	
	// set up the simulation model
	CarFerry sim(SIM_DURATION, FERRY_COUNT, HARBOR_COUNT);
	sim.run();

	// generate the report
	OStreamReport report(std::cout);
	report.addReportProducer(sim.ferry_cargo_len);
	report.addReportProducer(sim.ferry_load_time);
	report.addReportProducer(sim.car_wait_time);
	report.generateReport();

	return 0;
}
#endif
