//! This example simulates a simple barbershop scenario, demonstrating the use
//! of facilities, random number generation, and statistics collection in the
//! ODΞM-rs library.
//!
//! # Overview
//!
//! The simulation models a barbershop with a single barber (`joe`). Customers
//! arrive at random intervals, and each goes through a sequence of actions:
//! arrival, seizing the barber for a haircut, and leaving after the haircut.
//! The simulation also collects statistics on customer waiting times.
//!
//! # Key Elements
//! - **Facility Usage (Barber):** The barber (`joe`) is represented as a `Facility`.
//! - **Random Number Generation:** Used for simulating random customer arrivals and haircut durations.
//! - **Statistics Collection:** Collects and calculates average waiting times for performance analysis.

use odem_rs::{prelude::*, sync::Facility, util::random::RngStream};

/// Configuration type.
#[derive(Config, Default)]
struct Barbershop {
	duration: f64,
	/// Stream for generating independent random number generators.
	rng_stream: RngStream,
	/// The barber facility for haircuts.
	joe_barber: Facility,
	/// Random variable for recording customer wait times.
	wait_time: RandomVariable,
}

/// Customer process with access to the barber and a random processing delay.
struct Customer;

impl Customer {
	/// Creates a new customer process.
	fn process() -> impl Active<Barbershop> {
		Process::new(Customer, Customer::actions)
	}

	/// Defines the actions a customer goes through in the barbershop.
	async fn actions(&self, sim: &Sim<Barbershop>) {
		let mut rng = sim.global().rng_stream.rng();

		// access the barber and record the time for the report
		let arrival_time = sim.now();
		let barber = sim.global().joe_barber.seize().await;
		sim.global().wait_time.tabulate(sim.now() - arrival_time);

		// spend time
		sim.advance(rng.gen_range(12.0..18.0)).await;

		// release the barber
		barber.release();
	}
}

/// Orchestrates the barbershop simulation for a given duration.
async fn sim_main(sim: &Sim<Barbershop>) {
	let pool = pool::Pool::dynamic();

	// specify a customer generating fiber
	let customers = unique!(Fiber::new(async {
		let mut rng = sim.global().rng_stream.rng();

		loop {
			sim.advance(rng.gen_range(12.0..24.0)).await;
			sim.activate(pool.alloc(Customer::process()));
		}
	}));

	// activate the fiber and wait for the duration to pass
	let puck = sim.activate(customers);
	sim.advance(sim.global().duration).await;
	puck.abort();

	// finish processing the queue (no more customers arrive)
	let _ = sim.global().joe_barber.seize().await;
}

#[cfg(not(test))]
fn main() {
	let sim = Simulator::new(Barbershop {
		duration: 3.0 * 7.0 * 24.0 * 60.0,
		..Default::default()
	});
	
	sim.run(sim_main).unwrap();
	
	// print statistics
	println!("wait_time: {:#.3?}", sim.into_inner().wait_time);
}

#[cfg(test)]
criterion::criterion_main!(bench::benches);

#[cfg(all(test, windows))]
mod slx;

#[cfg(test)]
mod bench {
	use super::*;
	use criterion::{
		criterion_group, AxisScale, BatchSize, BenchmarkId, Criterion, PlotConfiguration,
	};

	#[cfg(feature = "odemx")]
	mod odemx {
		use std::os::raw::c_double;

		#[link(name = "odemx", kind = "static")]
		extern "C" {
			pub fn barbershop(duration: c_double);
		}
	}
	#[cfg(windows)]
	const SLX_PATH: &'static str = "C:\\Wolverine\\SLX";
	const RANGE: u32 = 10;
	const STEP: f64 = 1000.0;

	fn barbershop_bench(c: &mut Criterion) {
		let mut group = c.benchmark_group("Barbershop");

		// set up the benchmark parameters
		group
			.confidence_level(0.99)
			.plot_config(PlotConfiguration::default().summary_scale(AxisScale::Logarithmic))
			.measurement_time(std::time::Duration::from_secs(30));

		#[cfg(windows)]
		let slx_path = {
			let path = slx::slx_version(SLX_PATH);

			if path.is_none() {
				println!("SLX not found, skipping SLX benchmarks!");
			} else {
				println!("Using SLX program at {:?}", path.as_ref().unwrap());
			}

			path
		};

		// vary in the length of the simulation run
		for duration in (0..RANGE).map(|c| f64::from(1 << c) * STEP) {
			#[cfg(windows)]
			if let Some(path) = slx_path.as_ref() {
				let dt = duration.to_string();
				let args = [
					"/silent",
					"/stdout",
					"/noicon",
					"/nowarn",
					"/noxwarn",
					"/#BENCH",
					"slx\\barbershop.slx",
					dt.as_str(),
				];

				// benchmark the SLX implementation
				group.bench_function(BenchmarkId::new("SLX", duration), |b| {
					b.iter_custom(|iters| {
						slx::slx_bench(path.as_os_str(), &args, iters as usize)
							.expect("couldn't benchmark the SLX program")
					})
				});
			}

			// benchmark the C++ implementation
			#[cfg(feature = "odemx")]
			group.bench_function(BenchmarkId::new("ODEMx", duration), |b| {
				b.iter(|| unsafe { odemx::barbershop(duration as _) })
			});

			// benchmark the Rust implementation
			group.bench_function(BenchmarkId::new("Rust", duration), |b| {
				b.iter(|| simulation(&Barbershop {
					duration, ..Default::default()
				}, sim_main))
			});
		}

		group.finish();
	}

	criterion_group!(benches, barbershop_bench);
}
