use std::{collections::VecDeque, rc::Rc};

use rand::Rng;
use rand_distr::{num_traits::Float, Distribution, Exp, Normal, Poisson};
use tracing::debug;

use odem_rs::{
	prelude::*,
	stat::random_variable::Utilized,
	sync::{
		channel::shared::{Receiver, Sender},
		Facility,
	},
	util::{random::DefaultRng, uom::*},
};

#[derive(Config, Default)]
struct GroceryStore {
	#[time]
	time: Time<f64>,
	rng_stream: RngStream,
	parameters: Parameters,
	statistics: Statistics,
}

struct Parameters {
	conveyor_belt_capacity: u32,
	active_customers: Control<u32>,

	customer_arrival_distribution: Exp<f64>,
	item_amount_distribution: Poisson<f32>,
	move_item_distribution: Normal<f64>,
	price_distribution: Exp<f64>,
	billing_distribution: Normal<f64>,
}

impl Default for Parameters {
	fn default() -> Self {
		Self {
			conveyor_belt_capacity: 20,
			active_customers: Control::new(0),

			customer_arrival_distribution: Exp::new(1.0 / 30.0).unwrap(),
			item_amount_distribution: Poisson::new(3.0).unwrap(),
			move_item_distribution: Normal::new(3.0, 2.0).unwrap(),
			price_distribution: Exp::new(1.0 / 5.0).unwrap(),
			billing_distribution: Normal::new(10.0, 5.0).unwrap(),
		}
	}
}

#[derive(Default)]
struct Statistics {
	revenue: [RandomVariable; CASH_REGISTER],
	cashier_idle_time: [RandomVariable<Utilized<Time<f64>>>; CASH_REGISTER],
	customer_processing_time: RandomVariable<Time<f64>>,
}

impl Statistics {
	#[cfg(not(test))]
	fn print(&self) {
		let mut total_revenue = RandomVariable::new();
		let mut total_idle_time = RandomVariable::new();

		for i in 0..CASH_REGISTER {
			total_revenue = total_revenue.join(self.revenue[i].clone());
			total_idle_time = total_idle_time.join(self.cashier_idle_time[i].clone());
		}

		println!(
			"Customer processing time: {:#.2}",
			self.customer_processing_time.display(time::second)
		);
		println!("Revenue: {:.2} €", total_revenue.sum());

		for (i, revenue) in self.revenue.iter().enumerate() {
			println!("\tCheckout [{}] = {:.2} €", i + 1, revenue.sum());
		}

		println!("Cashier idle time: {:#}", total_idle_time.sum().display());
		for (i, idle_time) in self.cashier_idle_time.iter().enumerate() {
			println!("\tCashier [{}] = {:#}", i + 1, idle_time.sum().display());
		}
	}
}

pub fn sample_positive<T>(dist: impl Distribution<T>, rng: &mut DefaultRng) -> T
where
	T: Float,
{
	dist.sample_iter(rng)
		.find(|&val| val.is_sign_positive())
		.unwrap()
}

struct CashRegister {
	id: usize,
	customer_count: Control<u32>,
	conveyor_belt: Facility,
	conveyor_belt_head: Sender<VecDeque<Option<Item>>>,
	conveyor_belt_count: Control<u32>,
	conveyor_belt_capacity: u32,
	item_storage: Facility,
	item_storage_tail: Receiver<VecDeque<Option<Item>>>,
	item_storage_count: Control<u32>,
	bill_paid: Control<bool>,
}

impl CashRegister {
	fn new(
		id: usize,
		conveyor_belt_head: Sender<VecDeque<Option<Item>>>,
		item_storage_tail: Receiver<VecDeque<Option<Item>>>,
		conveyor_belt_capacity: u32,
	) -> CashRegister {
		CashRegister {
			id,
			customer_count: Control::new(0),
			conveyor_belt: Facility::new(),
			conveyor_belt_head,
			conveyor_belt_count: Control::new(0),
			conveyor_belt_capacity,
			item_storage: Facility::new(),
			item_storage_tail,
			item_storage_count: Control::new(0),
			bill_paid: Control::new(false),
		}
	}
}

struct Cashier {
	conveyor_belt_tail: Receiver<VecDeque<Option<Item>>>,
	item_storage_head: Sender<VecDeque<Option<Item>>>,
	checkout: Rc<CashRegister>,
}

impl Cashier {
	fn process(
		conveyor_belt_tail: Receiver<VecDeque<Option<Item>>>,
		item_storage_head: Sender<VecDeque<Option<Item>>>,
		checkout: Rc<CashRegister>,
	) -> impl Active<GroceryStore> {
		Process::new(
			Cashier {
				conveyor_belt_tail,
				item_storage_head,
				checkout,
			},
			Cashier::actions,
		)
	}

	async fn scan_items_from_belt(&self, sim: &Sim<GroceryStore>, rng: &mut DefaultRng) {
		let Parameters {
			move_item_distribution,
			price_distribution,
			..
		} = &sim.global().parameters;
		let Statistics {
			revenue,
			cashier_idle_time,
			..
		} = &sim.global().statistics;
		let CashRegister {
			id,
			conveyor_belt_count,
			item_storage_count,
			bill_paid,
			..
		} = &*self.checkout;

		// cashier starts out idle
		cashier_idle_time[*id].tabulate(Utilized(sim.now(), true));

		while let Ok(option_item) = self.conveyor_belt_tail.recv().await {
			// cashier is not idle anymore
			cashier_idle_time[*id].tabulate(Utilized(sim.now(), false));

			match option_item {
				Some(item) => {
					// take the item from the belt
					sim.advance(second(sample_positive(move_item_distribution, rng)))
						.await;
					conveyor_belt_count.update(|count| count - 1);

					// add product price to total revenue
					let price = rng.sample(price_distribution);
					revenue[*id].tabulate(price);

					// move the item to the item storage
					sim.advance(second(sample_positive(move_item_distribution, rng)))
						.await;
					self.item_storage_head.try_send(Some(item)).unwrap();
					item_storage_count.update(|count| count + 1);
				}
				None => {
					// encountered item separator
					self.item_storage_head.try_send(None).unwrap();
					bill_paid.set(false);
					break;
				}
			}

			// cashier is idle again
			cashier_idle_time[*id].tabulate(Utilized(sim.now(), true));
		}
	}

	async fn actions(&self, sim: &Sim<GroceryStore>) {
		let mut rng = sim.global().rng_stream.rng();

		debug!("Cashier opened checkout!");

		loop {
			self.scan_items_from_belt(sim, &mut rng).await;

			// wait until the customer collected his items
			until!(self.checkout.item_storage_count == 0).await;

			// produce the bill and charge them the amount
			let billing_duration = sim
				.global()
				.parameters
				.billing_distribution
				.sample_iter(&mut rng)
				.find_map(|v| (v >= 0.0).then_some(second(v)))
				.unwrap();

			sim.advance(billing_duration).await;

			self.checkout.bill_paid.set(true);
		}
	}
}

// Would be possible to add addtional useful data (e.g. price, manufacturer, ...)
#[derive(Debug, Clone, Copy, Default)]
struct Item;

struct Customer {
	basket: Vec<Item>,
	checkout: Rc<CashRegister>,
}

impl Customer {
	fn process(number_of_items: usize, checkout: Rc<CashRegister>) -> impl Active<GroceryStore> {
		Process::new(
			Customer {
				basket: vec![Item; number_of_items],
				checkout,
			},
			Customer::actions,
		)
	}

	async fn place_items_on_belt(&self, sim: &Sim<GroceryStore>, rng: &mut DefaultRng) {
		let Parameters {
			move_item_distribution,
			..
		} = sim.global().parameters;
		let CashRegister {
			conveyor_belt,
			conveyor_belt_count,
			conveyor_belt_capacity,
			conveyor_belt_head,
			..
		} = &self.checkout.as_ref();

		// wait until first customer in line
		let conveyor_belt = conveyor_belt.seize().await;

		for item in &self.basket {
			// wait until there is enough space for another item
			until!(conveyor_belt_count < conveyor_belt_capacity).await;

			// advance simulation by the time it takes to move an item
			sim.advance(second(sample_positive(move_item_distribution, rng)))
				.await;

			// place item on the belt
			conveyor_belt_head.try_send(Some(*item)).unwrap();
			conveyor_belt_count.update(|count| count + 1);
		}

		// place the item separator
		conveyor_belt_head.try_send(None).unwrap();

		conveyor_belt.release();
	}

	async fn collect_items_from_storage(&self, sim: &Sim<GroceryStore>, rng: &mut DefaultRng) {
		let Parameters {
			move_item_distribution,
			..
		} = sim.global().parameters;
		let CashRegister {
			item_storage,
			item_storage_tail,
			item_storage_count,
			bill_paid,
			..
		} = &self.checkout.as_ref();

		// wait until the previous customer has collected his items
		let item_storage = item_storage.seize().await;

		while let Ok(option_item) = item_storage_tail.recv().await {
			match option_item {
				Some(_item) => {
					// pick up the item from the storage
					sim.advance(second(sample_positive(move_item_distribution, rng)))
						.await;

					item_storage_count.update(|count| count - 1)
				}
				None => {
					until!(bill_paid).await;
					break;
				}
			}
		}

		// leave the checkout
		item_storage.release();
	}

	async fn actions(&self, sim: &Sim<GroceryStore>) {
		let Parameters {
			active_customers, ..
		} = &sim.global().parameters;
		let Statistics {
			customer_processing_time,
			..
		} = &sim.global().statistics;
		let CashRegister { customer_count, .. } = &self.checkout.as_ref();
		let mut rng = sim.global().rng_stream.rng();
		let arrival_time = sim.now();

		debug!(
			"Arrived after {:#} with {} items",
			arrival_time.display(),
			self.basket.len()
		);
		customer_count.update(|count| count + 1);
		active_customers.update(|total| total + 1);

		self.place_items_on_belt(sim, &mut rng).await;
		self.collect_items_from_storage(sim, &mut rng).await;

		customer_count.update(|count| count - 1);
		active_customers.update(|total| total - 1);
		customer_processing_time.tabulate(sim.now() - arrival_time);
	}
}

async fn sim_main(sim: &Sim<GroceryStore>, duration: Time<f64>) {
	let Parameters {
		active_customers,
		conveyor_belt_capacity,
		..
	} = &sim.global().parameters;

	let mut checkouts = Vec::with_capacity(CASH_REGISTER);
	let cashiers = pool!(const CASH_REGISTER);

	for id in 0..CASH_REGISTER {
		let (band_sender, band_receiver) = channel(VecDeque::new());
		let (item_storage_sender, item_storage_receiver) = channel(VecDeque::new());

		let checkout = Rc::new(CashRegister::new(
			id,
			band_sender,
			item_storage_receiver,
			*conveyor_belt_capacity,
		));
		sim.activate(cashiers.alloc(Cashier::process(
			band_receiver,
			item_storage_sender,
			checkout.clone(),
		)));
		checkouts.push(checkout.clone());
	}

	let customer_pool = pool!();

	sim.fork(sim.advance(duration))
		.or(async {
			let Parameters {
				customer_arrival_distribution,
				item_amount_distribution,
				..
			} = &sim.global().parameters;
			let mut rng = sim.global().rng_stream.rng();

			loop {
				// advance simulation until customer arrives
				let arrival_delay = second(customer_arrival_distribution.sample(&mut rng));
				sim.advance(arrival_delay).await;

				// set amount of items and select shortest lane
				let item_count = item_amount_distribution.sample(&mut rng).ceil() as usize;
				let shortest_lane = checkouts.iter().min_by(|c1, c2| {
					c1.conveyor_belt_count
						.get()
						.cmp(&c2.conveyor_belt_count.get())
						.then_with(|| c1.customer_count.get().cmp(&c2.customer_count.get()))
				});

				// spawn new customer
				sim.activate(customer_pool.alloc(Customer::process(
					item_count,
					shortest_lane.unwrap().clone(),
				)));
			}
		})
		.await;

	// serve remaining customers that are still waiting in line
	until!(active_customers == 0).await;
}

const CASH_REGISTER: usize = 2;

#[cfg(not(test))]
fn main() {
	tracing_subscriber::fmt()
		.with_max_level(tracing::Level::DEBUG)
		.with_target(false)
		.with_timer(model_time!("[{time:#2.3}]"))
		.init();

	let sim_duration = hour(0.5);

	let mut sim = Simulator::default();

	sim.run(|sim| sim_main(sim, sim_duration)).unwrap();

	println!("Simulation duration: {:#}", sim.model_time().display());
	sim.global().statistics.print();
}

#[cfg(test)]
criterion::criterion_main!(bench::benches);

#[cfg(all(test, windows))]
mod slx;

#[cfg(test)]
mod bench {
	use criterion::{
		criterion_group, AxisScale, BatchSize, BenchmarkId, Criterion, PlotConfiguration,
	};

	use super::*;

	#[cfg(feature = "odemx")]
	mod odemx {
		use std::os::raw::c_int;

		#[link(name = "odemx", kind = "static")]
		extern "C" {
			pub fn checkout(duration: c_int);
		}
	}

	// #[cfg(windows)]
	// const SLX_PATH: &'static str = "C:\\Wolverine\\SLX";
	const RANGE: u32 = 10;
	const STEP: f64 = 1000.0;

	fn checkout_bench(c: &mut Criterion) {
		let mut group = c.benchmark_group("Checkout");

		// set up the benchmark parameters
		group
			.confidence_level(0.99)
			.plot_config(PlotConfiguration::default().summary_scale(AxisScale::Logarithmic))
			.measurement_time(std::time::Duration::from_secs(30));

		// #[cfg(windows)]
		// 	let slx_path = {
		// 	let path = slx::slx_version(SLX_PATH);
		//
		// 	if path.is_none() {
		// 		println!("SLX not found, skipping SLX benchmarks!");
		// 	} else {
		// 		println!("Using SLX program at {:?}", path.as_ref().unwrap());
		// 	}
		//
		// 	path
		// };

		// vary in the length of the simulation run
		for sim_duration in (0..RANGE).map(|c| f64::from(1 << c) * STEP) {
			// #[cfg(windows)]
			// if let Some(path) = slx_path.as_ref() {
			// 	let duration = sim_duration.to_string();
			// 	let args = [
			// 		"/silent",
			// 		"/stdout",
			// 		"/noicon",
			// 		"/nowarn",
			// 		"/noxwarn",
			// 		"/#BENCH",
			// 		"slx\\checkout.slx",
			// 		duration.as_str()
			// 	];
			//
			// 	// benchmark the SLX implementation
			// 	group.bench_function(
			// 		BenchmarkId::new("SLX", sim_duration),
			// 		|b| b.iter_custom(|iters|
			// 			slx::slx_bench(
			// 				path.as_os_str(),
			// 				&args,
			// 				iters as usize
			// 			).expect("couldn't benchmark the SLX program")
			// 		)
			// 	);
			// }

			// benchmark the C++ implementation
			#[cfg(feature = "odemx")]
			group.bench_function(BenchmarkId::new("ODEMx", sim_duration), |b| {
				b.iter(|| unsafe { odemx::checkout(sim_duration as _) })
			});

			// benchmark the Rust implementation
			group.bench_function(BenchmarkId::new("Rust", sim_duration), |b| {
				b.iter_batched(
					Simulator::default,
					|mut sim| sim.run(|sim| sim_main(sim, second(sim_duration))),
					BatchSize::SmallInput,
				)
			});
		}

		group.finish();
	}

	criterion_group!(benches, checkout_bench);
}
