//! This crate contains a simulation model of a ferry system with multiple
//! harbors, ferries, and cars. The simulation is built using the ODΞM-rs
//! crate, incorporating asynchronous processes and random variable
//! distributions.

use std::{collections::VecDeque, fmt};

use rand_distr::{Distribution, Exp, Normal};

use tracing::debug;

use odem_rs::{
	prelude::*,
	sync::{channel::shared::{Receiver, Sender}, error::RecvError},
};

/// Configuration structure for the ferry simulation model.
///
/// Implements the [`CarFerry`] trait, specifying various parameters such as the
/// type and initial value of the model time, a random number generation stream,
/// and random variables for ferry-related statistics.
#[derive(Config, Default)]
struct CarFerry {
	/// The simulation time type.
	#[time]
	time: Time<f64>,
	
	/// The number of harbors.
	harbors: usize,
	
	/// The number of ferries serving these harbors.
	ferries: usize,
	
	/// The simulation duration.
	duration: Time<f64>,
	
	/// A stream of independent random number generators (RNGs)
	rng_stream: RngStream,
	
	/// Statistics for cargo lengths.
	ferry_cargo_len: RandomVariable<usize>,
	
	/// Statistics for load times.
	ferry_load_time: RandomVariable<Time<f64>>,
	
	/// Statistics for individual wait times of the cars.
	car_wait_time: RandomVariable<Time<f64>>,
}

impl fmt::Debug for CarFerry {
	fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
		f.debug_struct("CarFerry")
			.field("ferry_cargo_len", &self.ferry_cargo_len)
			.field("ferry_load_time", &self.ferry_load_time.display(time::minute))
			.field("car_wait_time", &self.car_wait_time.display(time::minute))
			.finish()
	}
}

// Constants for simulation parameters.
const FERRY_COUNT: usize = 1;
const HARBOR_COUNT: usize = 2;
const HARBOR_DISTANCE: Time<f64> = second(10.0 * 60.0);
const FERRY_TIMEOUT: Time<f64> = second(5.0 * 60.0);
const FERRY_CAPACITY: usize = 5;

/// Structure representing a car in the simulation.
///
/// Contains information about the arrival time and loading duration of the car.
#[derive(Debug)]
struct Car {
	arrival_time: Time<f64>,
	load_duration: Time<f64>,
}

/// Structure representing a harbor pier.
///
/// It includes a landing site for cars and an asynchronous `actions` method
/// that simulates the arrival of cars to the pier.
struct Pier {
	/// Channel to send cars to the ferry.
	landing_site: Sender<VecDeque<Car>>,
}

impl Pier {
	/// Instantiates a new Pier process that generates new [Car]s to be picked
	/// up by [Ferry]-processes.
	fn process(landing_site: Sender<VecDeque<Car>>) -> impl Active<CarFerry> {
		Process::new(Pier { landing_site }, Self::actions)
	}
	
	/// Asynchronous method simulating actions at the pier, including the
	/// arrival of cars.
	async fn actions(&self, sim: &Sim<CarFerry>) {
		let mut rng = sim.global().rng_stream.rng();
		let mut cars = 0;
		
		// Arrival and loading time distributions.
		let arrival_delay = Exp::new(0.1).unwrap();
		let loading_delay = Normal::new(0.5, 0.2).unwrap();
		
		loop {
			// Wait for the next car to arrive.
			sim.advance(minute(arrival_delay.sample(&mut rng))).await;
			
			cars = cars + 1;
			debug!(%cars, "cars generated");
			
			// Create a new car with arrival and load times.
			let car = Car {
				arrival_time: sim.now(),
				load_duration: minute(
					loading_delay
						.sample_iter(&mut rng)
						.find(|&val| val >= 0.0)
						.unwrap(),
				),
			};
			self.landing_site
				.try_send(car)
				.expect("No ferries in the simulation");
		}
	}
}

/// Structure representing a ferry in the simulation.
struct Ferry {
	/// Receivers from piers to accept arriving cars.
	piers: Vec<Receiver<VecDeque<Car>>>,
}

impl Ferry {
	/// Instantiates a new ferry process that begins its journey at the
	/// harbor indexed by `id` and visits the neighboring harbours in circular
	/// order.
	fn process(p: &[Receiver<VecDeque<Car>>], id: usize) -> impl Active<CarFerry> {
		Process::new(
			Ferry {
				piers: p
					.iter()
					.skip(id)
					.chain(p.iter().take(id))
					.cloned()
					.collect(),
			},
			Self::actions,
		)
	}
	
	/// Asynchronous method simulating ferry operations, including loading,
	/// unloading, and traveling between piers.
	async fn actions(&self, sim: &Sim<CarFerry>) {
		let mut cargo = Vec::<Car>::with_capacity(FERRY_CAPACITY);
		
		loop {
			for pier in self.piers.iter() {
				debug!("unloading {} cars", cargo.len());
				
				// Unload the cars at the current pier.
				for car in cargo.drain(..) {
					sim.advance(car.load_duration).await;
				}
				
				let begin_loading = sim.now();
				
				// Load cars until capacity is reached or timeout occurs.
				for _ in 0..cargo.capacity() {
					match sim.fork(pier.recv()).or(async {
						sim.advance(FERRY_TIMEOUT).await;
						Err(RecvError)
					}).await {
						// A car arrived before timeout.
						Ok(car) => {
							sim.global()
								.car_wait_time
								.tabulate(sim.now() - car.arrival_time);
							sim.advance(car.load_duration).await;
							cargo.push(car);
						}
						// Timeout occurred; depart to next harbor.
						Err(_) => break
					}
				}
				
				// Record ferry loading statistics.
				sim.global()
					.ferry_load_time
					.tabulate(sim.now() - begin_loading);
				sim.global().ferry_cargo_len.tabulate(cargo.len());
				
				// Travel to the next harbor.
				sim.advance(HARBOR_DISTANCE).await;
			}
		}
	}
}

/// Asynchronous function representing the main ferry simulation.
///
/// It creates and activates harbors, ferries, and their corresponding
/// processes. The simulation duration is specified, and the simulation is
/// advanced until the specified duration is reached.
async fn sim_main(sim: &Sim<CarFerry>) {
	let harbors = sim.global().harbors;
	let ferries = sim.global().ferries;
	let duration = sim.global().duration;
	
	let mut ports = Vec::with_capacity(harbors);
	
	let harbor_pool = pool!(harbors);
	let ferry_pool = pool!(ferries);
	
	// Create all the harbors (piers).
	for _ in 0..harbors {
		let (sx, rx) = channel(VecDeque::new());
		sim.activate(harbor_pool.alloc(Pier::process(sx)));
		ports.push(rx);
	}
	
	// Create all the ferries.
	for id in 0..ferries {
		sim.activate(ferry_pool.alloc(Ferry::process(&ports, id)));
	}
	
	// Run the simulation for the specified duration.
	sim.advance(duration).await;
	
	// Handle any cars that weren't picked up by a ferry.
	for port in ports {
		while let Ok(car) = port.try_recv() {
			sim.global()
				.car_wait_time
				.tabulate(sim.now() - car.arrival_time);
		}
	}
}

/// Main function for running the ferry simulation.
///
/// Initializes the simulator, runs the ferry simulation using the `ferry`
/// function, and displays relevant statistics at the end.
#[cfg(not(test))]
fn main() {
	use tracing::Level;
	tracing_subscriber::fmt()
		.with_max_level(Level::DEBUG)
		.with_target(false)
		.with_timer(model_time!("[{time:#2}]"))
		.init();
	
	let sim = Simulator::new(CarFerry {
		harbors: HARBOR_COUNT,
		ferries: FERRY_COUNT,
		duration: day(1.0),
		..Default::default()
	});
	sim.run(sim_main).unwrap();
	let config = sim.into_inner();
	
	println!("Ferry cargo len: {:#.3?}", config.ferry_cargo_len);
	println!(
		"Ferry load time: {:#.3?}",
		config.ferry_load_time.display(time::minute)
	);
	println!(
		"Car wait time: {:#.3?}",
		config.car_wait_time.display(time::minute)
	);
}

#[cfg(test)]
criterion::criterion_main!(bench::benches);

#[cfg(all(test, windows))]
mod slx;

#[cfg(test)]
mod bench {
	use criterion::{
		criterion_group, AxisScale, BatchSize, BenchmarkId, Criterion, PlotConfiguration,
	};

	use super::*;

	#[cfg(feature = "odemx")]
	mod odemx {
		use std::os::raw::{c_double, c_uint};

		#[link(name = "odemx", kind = "static")]
		extern "C" {
			pub fn ferry(duration: c_double, ferries: c_uint, harbors: c_uint);
		}
	}
	#[cfg(windows)]
	const SLX_PATH: &'static str = "C:\\Wolverine\\SLX";
	const RANGE: u32 = 10;
	const STEP: f64 = 1000.0;

	fn ferry_bench(c: &mut Criterion) {
		let mut group = c.benchmark_group("Ferry");

		// set up the benchmark parameters
		group
			.confidence_level(0.99)
			.plot_config(PlotConfiguration::default().summary_scale(AxisScale::Logarithmic))
			.measurement_time(std::time::Duration::from_secs(30));

		#[cfg(windows)]
		let slx_path = {
			let path = slx::slx_version(SLX_PATH);

			if path.is_none() {
				println!("SLX not found, skipping SLX benchmarks!");
			} else {
				println!("Using SLX program at {:?}", path.as_ref().unwrap());
			}

			path
		};

		// vary in the length of the simulation run
		for sim_duration in (0..RANGE).map(|c| f64::from(1 << c) * STEP) {
			#[cfg(windows)]
			if let Some(path) = slx_path.as_ref() {
				let duration = sim_duration.to_string();
				let args = [
					"/silent",
					"/stdout",
					"/noicon",
					"/nowarn",
					"/noxwarn",
					"/#BENCH",
					"slx\\ferry.slx",
					duration.as_str(),
				];

				// benchmark the SLX implementation
				group.bench_function(BenchmarkId::new("SLX", sim_duration), |b| {
					b.iter_custom(|iters| {
						slx::slx_bench(path.as_os_str(), &args, iters as usize)
							.expect("couldn't benchmark the SLX program")
					})
				});
			}

			// benchmark the C++ implementation
			#[cfg(feature = "odemx")]
			group.bench_function(BenchmarkId::new("ODEMx", sim_duration), |b| {
				b.iter(|| unsafe {
					odemx::ferry(sim_duration as _, FERRY_COUNT as _, HARBOR_COUNT as _)
				})
			});

			// benchmark the Rust implementation
			group.bench_function(BenchmarkId::new("Rust", sim_duration), |b| {
				b.iter_batched(
					Simulator::default,
					|mut sim| {
						sim.run(|sim| {
							sim_main(sim, minute(sim_duration), FERRY_COUNT, HARBOR_COUNT)
						})
					},
					BatchSize::SmallInput,
				)
			});
		}

		group.finish();
	}

	criterion_group!(benches, ferry_bench);
}
