//! This crate provides a simulation of the classic Dining Philosophers problem
//! using the ODΞM-rs simulation library.
//!
//! The Dining Philosophers problem is a classic synchronization and concurrency
//! problem where a number of philosophers sit at a dining table, and each
//! philosopher alternates between thinking and eating. The challenge is usually
//! to design a solution that avoids deadlock and ensures proper resource
//! sharing (forks in this case). This simulation model attempts to determine
//! statistics about the amount of time it would take for a deadlock to occur if
//! the deadlock is not prevented.

use std::rc::Rc;

use rand_distr::{Distribution, Exp, Normal, Uniform};
use rayon::prelude::*;

use odem_rs::prelude::*;

/// Configuration type for the Dining Philosophers simulation.
#[derive(Config, Default)]
struct DiningPhilosophers {
	/// The type and initial value of the model time.
	#[time]
	time: Time<f64>,
	/// The number of dining philosophers.
	count: usize,
	/// A stream of independent random-number generators.
	rng_stream: RngStream,
}

/// Represents the dining table with forks in the Dining Philosophers scenario.
struct Table {
	/// Forks on the table, represented as binary [Control] variables
	/// (true if taken).
	forks: Vec<Control<bool>>,
	/// Number of forks currently held by philosophers.
	forks_held: Control<usize>,
	/// Number of philosophers waiting for forks.
	forks_awaited: Control<usize>,
}

impl Table {
	/// Creates a new dining table with a specified number of forks.
	fn new(forks: usize) -> Self {
		Table {
			forks: (0..forks).map(|_| Control::new(false)).collect(),
			forks_held: Control::new(0),
			forks_awaited: Control::new(0),
		}
	}

	/// Blocks until the requested fork is available and acquires it.
	async fn acquire_fork(&self, i: usize) {
		let num = i % self.forks.len();
		self.forks_awaited.set(self.forks_awaited.get() + 1);

		until!(!self.forks[num]).await;

		self.forks[num].set(true);
		self.forks_awaited.set(self.forks_awaited.get() - 1);
		self.forks_held.set(self.forks_held.get() + 1);
	}

	/// Releases the requested fork back to the table.
	fn release_fork(&self, i: usize) {
		self.forks[i % self.forks.len()].set(false);
		self.forks_held.set(self.forks_held.get() - 1);
	}
}

/// Represents a philosopher in the Dining Philosophers scenario.
struct Philosopher {
	table: Rc<Table>,
	seat: usize,
}

impl Philosopher {
	/// Creates a new philosopher process.
	fn process(table: Rc<Table>, seat: usize) -> impl Active<DiningPhilosophers> {
		Process::build()
			.with_subject(Philosopher { table, seat })
			.with_actions(Self::actions)
			.finish()
	}

	/// Defines the actions a philosopher goes through in the Dining
	/// Philosophers scenario.
	async fn actions(&self, sim: &Sim<DiningPhilosophers>) {
		let mut rng = sim.global().rng_stream.rng();
		let thinking_duration = Exp::new(1.0).unwrap();
		let artificial_delay = Uniform::new(0.1, 0.2);
		let eating_duration = Normal::new(0.5, 0.2).unwrap();

		loop {
			// spend some time pondering the nature of things
			sim.advance(minute(thinking_duration.sample(&mut rng))).await;

			// acquire the first fork
			self.table.acquire_fork(self.seat).await;

			// introduce an artificial delay to leave room for deadlocks
			sim.advance(minute(artificial_delay.sample(&mut rng))).await;

			// acquire the second fork
			self.table.acquire_fork(self.seat + 1).await;

			// spend some time eating
			sim.advance(
				minute(eating_duration
					.sample_iter(&mut rng)
					.find(|&val| val >= 0.0)
					.unwrap()
				),
			)
			.await;

			// release the forks
			self.table.release_fork(self.seat + 1);
			self.table.release_fork(self.seat);
		}
	}
}

/// Runs the Dining Philosophers simulation once and returns the model time
/// until the deadlock.
async fn sim_main(sim: &Sim<DiningPhilosophers>) -> Time<f64> {
	let count = sim.global().count;
	let table = Rc::new(Table::new(count));
	let arena = pool!(count);

	// create and activate the philosopher-processes
	for i in 0..count {
		sim.activate(arena.alloc(Philosopher::process(table.clone(), i)));
	}

	// wait for the precise configuration indicating a deadlock
	// (we technically don't have to do this, because deadlocks imply that no
	//  processes can advance, causing the simulation to end;
	//  SLX doesn't allow us to record the time of that event, so we can't use
	//  it here either)
	until!(table.forks_held == count && table.forks_awaited == count).await;

	// tabulate the current system time
	sim.now()
}

/// Executes the Dining Philosophers simulation with multithreading and returns
/// the average time until the deadlock.
#[allow(dead_code)]
fn philosophers_mt(count: usize, reruns: usize) -> RandomVariable<Time<f64>> {
	// use thread-based parallelism to concurrently run simulation models
	(0..reruns)
		.into_par_iter()
		.map(|seed| {
			Simulator::new(DiningPhilosophers {
				time: second(0.0),
				count,
				rng_stream: RngStream::from_seed(seed as u64),
			})
			.run(sim_main)
			.unwrap()
		})
		.fold(RandomVariable::default, |var, duration| {
			var.tabulate(duration);
			var
		})
		.reduce(RandomVariable::default, RandomVariable::join)
}

/// Executes the Dining Philosophers simulation with single-threading and
/// returns the average time until the deadlock.
#[allow(dead_code)]
fn philosophers_st(count: usize, reruns: usize) -> RandomVariable<Time<f64>> {
	// run the simulations sequentially
	let random_var = RandomVariable::default();
	let mut simulator = Simulator::new(DiningPhilosophers {
		count, ..DiningPhilosophers::default()
	});

	for seed in 0..reruns {
		simulator.inner_mut().rng_stream = RngStream::from_seed(seed as u64);
		random_var.tabulate(simulator.run(sim_main).unwrap());
	}

	random_var
}

// /// Executes the Dining Philosophers simulation with multithreading and returns
// /// the total steps until the deadlock.
// #[allow(dead_code)]
// fn philosopher_steps(count: usize, reruns: usize) -> usize {
// 	// use thread-based parallelism to concurrently run simulation models
// 	(0..reruns)
// 		.into_par_iter()
// 		.map(|seed| {
// 			pin!(Simulator::new(Config {
// 				rng_stream: RngStream::from_seed(seed as u64)
// 			})
// 			.steps(|sim| run_once(sim, count)))
// 			.iter()
// 			.count()
// 		})
// 		.fold(|| 0, |var, steps| var + steps)
// 		.reduce(|| 0, |var_a, var_b| var_a + var_b)
// }

#[cfg(not(test))]
fn main() {
	const EXPERIMENT_COUNT: usize = 100;
	const PHILOSOPHER_COUNT: usize = 5;

	let res = philosophers_mt(PHILOSOPHER_COUNT, EXPERIMENT_COUNT);
	println!("result: {:#.2}", res.display(time::minute));
	// let res = philosophers_st(PHILOSOPHER_COUNT, EXPERIMENT_COUNT);
	// println!("result: {:#.2}", res.display(time::minute));
}

#[cfg(test)]
criterion::criterion_main!(bench::benches);

#[cfg(all(test, windows))]
mod slx;

#[cfg(test)]
mod bench {
	use criterion::{criterion_group, AxisScale, BenchmarkId, Criterion, PlotConfiguration};

	use super::*;

	#[cfg(feature = "odemx")]
	mod odemx {
		use std::os::raw::c_uint;

		#[link(name = "odemx", kind = "static")]
		extern "C" {
			pub fn philosophers(count: c_uint, reruns: c_uint);
		}
	}
	#[cfg(windows)]
	const SLX_PATH: &'static str = "C:\\Wolverine\\SLX";
	const RANGE: u32 = 10;
	const STEP: usize = 3;

	fn philosopher_bench(c: &mut Criterion) {
		let mut group = c.benchmark_group("Philosophers");

		// set up the benchmark parameters
		group
			.confidence_level(0.99)
			.plot_config(PlotConfiguration::default().summary_scale(AxisScale::Logarithmic))
			.measurement_time(std::time::Duration::from_secs(300));

		#[cfg(windows)]
		let slx_path = {
			let path = slx::slx_version(SLX_PATH);

			if path.is_none() {
				println!("SLX not found, skipping SLX benchmarks!");
			} else {
				println!("Using SLX program at {:?}", path.as_ref().unwrap());
			}

			path
		};

		// vary the number of performed reruns in each experiment
		for experiment_count in (0..RANGE).map(|c| (1 << c) * STEP) {
			#[cfg(windows)]
			if let Some(path) = slx_path.as_ref() {
				let count = experiment_count.to_string();
				let args = [
					"/silent",
					"/stdout",
					"/noicon",
					"/nowarn",
					"/noxwarn",
					"/#BENCH",
					"slx\\philosophers.slx",
					count.as_str(),
				];

				// benchmark the SLX implementation
				group.bench_function(BenchmarkId::new("SLX", experiment_count), |b| {
					b.iter_custom(|iters| {
						slx::slx_bench(path.as_os_str(), &args, iters as usize)
							.expect("couldn't benchmark the SLX program")
					})
				});
			}

			// benchmark the C++ implementation
			#[cfg(feature = "odemx")]
			group.bench_function(BenchmarkId::new("ODEMx", experiment_count), |b| {
				b.iter(|| unsafe {
					odemx::philosophers(PHILOSOPHER_COUNT as _, experiment_count as _)
				})
			});

			// benchmark the (parallel) Rust implementation
			group.bench_function(BenchmarkId::new("Rust", experiment_count), |b| {
				b.iter(|| philosophers_mt(PHILOSOPHER_COUNT, experiment_count))
			});

			// benchmark the (sequential) Rust implementation
			group.bench_function(BenchmarkId::new("Rust_ST", experiment_count), |b| {
				b.iter(|| philosophers_st(PHILOSOPHER_COUNT, experiment_count))
			});
		}

		group.finish();
	}

	criterion_group!(benches, philosopher_bench);
}
