use std::{
	cell::Cell,
	rc::{Rc, Weak},
};

use tracing::{debug, Level};

use odem_rs::{core::config::DefaultConfig, prelude::*};

#[derive(Copy, Clone)]
struct Signal(usize);

#[derive(Default)]
struct PingPong {
	message: Control<Option<Signal>>,
	sender: Cell<Option<Weak<PingPong>>>,
}

impl PingPong {
	fn process() -> Process<impl Actions<Subject = Rc<Self>, Config = DefaultConfig>> {
		Process::build()
			.with_subject(Rc::default())
			.with_actions(Self::actions)
			.with_name("PingPong")
			.finish()
	}

	fn send(&self, msg: Signal, source: &Rc<Self>) {
		self.message.set(Some(msg));
		self.sender.set(Some(Rc::downgrade(source)));
	}

	async fn actions(self: &Rc<Self>, sim: &Sim) {
		let mut no = 1;
		
		loop {
			until!(self.message.is_some()).await;
			debug!("Message '{}' received", self.message.take().unwrap().0);
			
			sim.advance(30.0).await;
			match self.sender.replace(None).unwrap().upgrade() {
				Some(sender) => sender.send(Signal(no), self),
				None => break,
			}
			no += 1;
			debug!("Message sent");
		}
	}
}

async fn sim_main(sim: &Sim) {
	let p1 = unique!(PingPong::process());
	let p2 = unique!(PingPong::process());

	let h1 = sim.activate(p1);
	let h2 = sim.activate(p2);

	h1.subject().send(Signal(0), h2.subject());
	sim.advance(100.0).await;
}

fn main() {
	tracing_subscriber::fmt()
		.with_max_level(Level::TRACE)
		.with_target(false)
		.with_timer(model_time!("[{time:2}]"))
		.init();

	simulation(sim_main).expect("unexpected deadlock");
}
