//! A simulation model of a warehouse distribution system with forklifts and
//! stations.
//!
//! Forklifts retrieve packages from stations, deliver them to the distribution
//! center, and return engines for packing.
//! Stations pack packages when they have enough engines available, waiting
//! until they're unblocked by a forklift.
//! An abstract storage queue manages a pool of available engines, emptied by
//! forklifts taking engines and replenished by the simulation itself.

use intrusive_collections::{intrusive_adapter, KeyAdapter, PointerOps, RBTree, RBTreeLink};
use odem_rs::prelude::*;
use std::{
	cell::{Cell, RefCell},
	rc::Rc,
};
use tracing::{debug, Level};

/// Represents the simulation configuration.
#[derive(Config, Default)]
struct Transport {
	#[time]
	time: usize,
	/// The total number of workers.
	workers: usize,
	/// The total number of stations.
	stations: usize,
	/// Set of stations in the warehouse.
	station_set: StationSet,
	/// Cumulative number of packed packages.
	packed: Cell<usize>,
}

/// A custom data structure to manage the set of stations.
#[derive(Default)]
struct StationSet {
	/// Red-black tree to store stations based on their distance from the
	/// distribution center.
	set: RefCell<RBTree<StationAdapter>>,
	/// Counter to track the number of stations in the set.
	size: Control<usize>,
}

impl StationSet {
	/// Inserts a new station into the set.
	pub fn insert(&self, elem: Rc<Station>) {
		self.set.borrow_mut().insert(elem);
		self.size.set(self.size.get() + 1);
	}

	/// Removes a specified station from the set.
	pub fn remove(&self, elem: &Station) -> Rc<Station> {
		assert!(elem.hook.is_linked());

		self.size.set(self.size.get() - 1);

		unsafe { self.set.borrow_mut().cursor_mut_from_ptr(elem) }
			.remove()
			.unwrap()
	}

	/// Pops the nearest station from the set.
	pub async fn pop(&self) -> Rc<Station> {
		until!(self.size > 0).await;
		self.size.set(self.size.get() - 1);
		self.set.borrow_mut().front_mut().remove().unwrap()
	}
}

intrusive_adapter!(
	/// Intrusive adapter trait for embedding `Station` into the `RBTree`.
	StationAdapter = Rc<Station>: Station { hook: RBTreeLink }
);

impl<'a> KeyAdapter<'a> for StationAdapter {
	type Key = usize;

	fn get_key(&self, value: &'a <Self::PointerOps as PointerOps>::Value) -> Self::Key {
		value.distance
	}
}

/// A placeholder for an engine.
struct Engine;

/// A pool of available engines.
struct Queue {
	engines: Control<usize>,
}

impl Queue {
	/// Creates a new engine pool.
	fn new() -> Self {
		Queue {
			engines: Control::new(0),
		}
	}

	/// Gives an engine to the pool.
	fn give(&self, _engine: Engine) {
		self.engines.set(self.engines.get() + 1);
	}

	/// Takes an engine from the pool.
	async fn take(&self) -> Engine {
		until!(self.engines > 0).await;
		self.engines.set(self.engines.get() - 1);
		Engine
	}
}

/// Represents a warehouse station.
struct Station {
	/// Link in the `RBTree`.
	hook: RBTreeLink,
	/// Flag to indicate whether the station is blocked.
	blocked: Control<bool>,
	/// Number of engines available for packing.
	engines: Control<usize>,
	/// Distance from the distribution center.
	distance: usize,
}

impl Station {
	/// Creates a new station.
	fn new(distance: usize) -> Process<impl Actions<Subject = Rc<Self>, Config =Transport>> {
		Process::build()
			.with_subject(Rc::new(Station {
				hook: RBTreeLink::new(),
				blocked: Control::new(false),
				engines: Control::new(0),
				distance,
			}))
			.with_actions(Self::actions)
			.with_name("Station")
			// .with_rank(1)
			.finish()
	}

	/// Acquires the station lock before packing.
	async fn acquire(&self) {
		until!(!self.blocked).await;
		self.blocked.set(true);
	}

	/// Releases the station lock after packing.
	fn release(&self) {
		self.blocked.set(false);
	}

	/// Processes the station's actions, including packing and release.
	async fn actions(self: &Rc<Self>, sim: &Sim<Transport>) {
		let Transport {
			station_set,
			packed,
			..
		} = sim.global();
		loop {
			until!(self.engines >= 3).await;

			self.acquire().await;
			station_set.remove(self);

			debug!("starting packing");

			sim.advance(60usize).await;
			packed.set(packed.get() + 3);
			self.engines.set(self.engines.get() - 3);

			debug!("finished packing (total: {:2})", packed.get());

			self.release();
			station_set.insert(self.clone());
		}
	}
}

/// Represents a forklift in the warehouse.
struct Forklift {
	queue: Rc<Queue>,
}

impl Forklift {
	/// Creates a new forklift.
	fn new(queue: Rc<Queue>) -> Process<impl Actions<Subject = Self, Config =Transport>> {
		Process::new(Forklift { queue }, Self::actions)
	}

	/// Processes the forklift's actions, including picking up packages,
	/// delivering, and returning engines.
	async fn actions(&self, sim: &Sim<Transport>) {
		let station_set = &sim.global().station_set;

		loop {
			self.queue.take().await;
			let station = station_set.pop().await;
			station.acquire().await;

			debug!("selected station {}", station.distance / 5);

			let distance = station.distance;
			sim.advance(distance).await;

			station.release();
			station_set.insert(station.clone());
			station.engines.set(station.engines.get() + 1);

			sim.advance(distance).await;
		}
	}
}

/// Performs the simulation setup and execution.
async fn sim_main(sim: &Sim<Transport>) {
	let workers = sim.global().workers;
	let stations = sim.global().stations;
	let queue = Rc::new(Queue::new());
	
	let station_arena = pool!(stations);
	let worker_arena = pool!(workers);

	for i in 1..=stations {
		let station = Station::new(i * 5);
		sim.global().station_set.insert(station.clone());
		sim.activate(station_arena.alloc(station));
	}

	for _ in 0..workers {
		sim.activate(worker_arena.alloc(Forklift::new(queue.clone())));
	}

	sim.fork(sim.advance(60 * 60usize))
		.or(async {
			loop {
				sim.advance(6usize).await;
				queue.give(Engine);
			}
		})
		.await;

	debug!(
		%workers,
		%stations,
		packed = %sim.global().packed.get(),
		engines = %queue.engines.get()
	);
}

/// Initializes tracing and runs the simulation.
fn main() {
	tracing_subscriber::fmt()
		.with_max_level(Level::DEBUG)
		.with_target(false)
		.with_timer(model_time!("[{time:#.2}]"))
		.init();

	Simulator::new(Transport {
		workers: 8, stations: 6, ..Default::default()
	}).run(sim_main).expect("unexpected deadlock");
}
